package quest.client;

import quest.client.net.*;
import quest.client.chat.*;
import quest.global.chat.*;
import quest.global.net.*;
import quest.global.login.*;

import javax.swing.*;
import java.awt.*;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.net.InetAddress;
import java.net.URL;
import java.net.MalformedURLException;

/**
 * <p>Represents the frame that provides the chat, search, help, etc functionality after the user has succesfully logged in.
 * The codebase is passed through in the constructor in order to get the coherence as low as possible with the applet. This way 
 * the implementation of ClientFrame in an application is facilitated.</p>
 *
 * This class is an example of a <a href="../../../overview-summary.html#patternSingleton">Singleton design pattern</a>.
 * @author Sander Brandenburg
 */
public class ClientFrame extends JFrame implements MessageClientListener {

    private static ClientFrame uniqueInstance = null;
    private static GUITab_GlobalChat	global;	
    private static GUITab_Chat			chat;
    private static GUITab_Channel		channels;
    private GUITab_Search		search;
    private GUITab_Ranking		ranking;
    private GUITab_Help			help;
    private static JTabbedPane			jtp;
	
    private static URL codebase = null;
	
    private static String username;
    private static String password;
	
    private static boolean enableglobal = false;
    private static boolean enablechat = false;

    private static boolean finishing = false;

    /**
     * Toggles whether to enable the chatwindow in the GUI.
     * @param flag Whether to enable the chat tab in the GUI overview.
     */
    public static void setEnabledChat(boolean flag) {
	channels.setEnabledCreateChannel(!flag);
	channels.setEnabledJoinChannel(!flag);
	global.setEnabledCreateChannel(!flag);
	jtp.setEnabledAt(1, flag);
	enablechat = flag;
		
	if (flag)
	    jtp.setSelectedIndex(1);
	else
	    jtp.setSelectedIndex(0);
    }
	
    private static void setEnabledGlobalChat(boolean flag) {
	jtp.setEnabledAt(0, flag);
	enablechat = flag;
    }
	
    /**
     * Returns the username which the client is using.
     * @exception NullPointerException When ClientFrame hasn't been initialized yet.
     */
    public static String getUsername() {
	if (uniqueInstance == null) {
	    throw new RuntimeException("ClientFrame hasn't been initialized!");
	}
	return username;
    }
	
    /**
     * Returns password which the client is using.
     * @exception NullPointerException When ClientFrame hasn't been initialized yet.
     */
    public static String getPassword() {
	if (uniqueInstance == null) {
	    throw new RuntimeException("ClientFrame hasn't been initialized!");
	}
	return password;
    }
	
    // returned null als onbekend (voor debug-standalone runs van de game)
    /**
     * Returns the codebase the amazing quest game is started from (set in constructor).
     * @return URL The codebase AQ is started from.
     */
    public static URL getCodeBase() {
	return codebase;
    }
     
    /**
     * Tries to gracefully quit (log out properly) Amazing quest. If the lower layers (read: chat and game layer) 
     * time out in 3 secs on their log out procedure, it will quit forcefully (kills the threads).
     */
    public static void quit() {
      	// a hack to restrict calling this method only once
	if (finishing)
	    return;
	
	username = null;
	password = null;
		
	finishing = true;
	new Thread(new Runnable() {
		public void run() {
		    System.out.println("exiter thread started");
				
		    try {
			Thread.currentThread().sleep(3000);
		    } catch (InterruptedException ie) {
			System.out.println("exiter interrupted");
			return;
		    }
				
		    System.out.println("exiter exits");
				//check if clientapplet is still in loggedin state
		    ClientApplet ca = ClientApplet.getInstance();
		    if (ca.isLoggedIn()) 
			ca.exit("exit timeout");
				
				//TODO possible race condition: in 3000ms the client could have reinitialized 
				//on the server again, so clientapplet should somehow kill or synchronize on 
				//this thread, but who cares
				//update: rudimentary fixed in ClientApplet#incomingMessage, but not tested or untestable
		}
	    }, "exiter").start();
		
	if (enablechat) {
	    chat.quit();
	    enablechat = false;
	}
	if (enableglobal) {
	    global.quit();
	    enableglobal = false;
	}
		
	//TODO exit from game somehow
	cd.sendMessage(new LoginMessage.C2S_Logout());
    }
	
    /**
     * Sole constructor that creates the supporting Chat GUI that is used to spawn games.
     * @exception RuntimeException When class is initialized more than once in the same context.
     */
    public ClientFrame(String username, String password, ClientDaemon cd, URL codebase) {
	if (uniqueInstance != null)
	    throw new RuntimeException("can only be initialized once in a context!");
	
	uniqueInstance = this;

	this.codebase = codebase;
	this.username = username;
	this.password = password;
	this.cd = cd;
		
	chat = new GUITab_Chat();
	global = new GUITab_GlobalChat();
	channels = new GUITab_Channel();
	search = new GUITab_Search();
	ranking = new GUITab_Ranking();
	help = new GUITab_Help();

	cd.registerProces(ChatMessage.class, this);
	cd.registerProces(ChannelMessage.class, this);

	try {
	    ChatMessage cm = new ChatMessage.C2S_Login(ChatMessage.GLOBALCHANNELNAME);
	    cd.sendMessage(cm);
	} catch (Exception e) {
	    e.printStackTrace();
	}

	Container cp = getContentPane();
		
	jtp = new JTabbedPane();		
		
	jtp.addTab(ChatMessage.GLOBALCHANNELNAME,  global);
	jtp.addTab("Chat",    chat);
	jtp.addTab("Channel", channels);
	jtp.addTab("Search",  search);
	jtp.addTab("Ranking", ranking);
	jtp.addTab("Help",    help);
		
	setEnabledChat(false);
	setEnabledGlobalChat(true);
		
	cp.setLayout(new GridLayout(1, 1)); 
	cp.add(jtp);
		
	//TODO change title
	setTitle("Amazing Quest - ALL YOUR QUEST ARE BELONG TO US!");
	try {
	    setIconImage(getToolkit().getImage(new URL(getCodeBase(), "quest/client/chat/images/aqicon.gif")));
	} catch(MalformedURLException mue) {
	    mue.printStackTrace();
	}
		
	addWindowListener(new WindowAdapter() {
		public void windowClosing(WindowEvent e) {
		    quit();
		}
			
		public void windowClosed(WindowEvent e) {
		    quit();
		}
	    });
    }
		
    private static ClientDaemon cd; 
	
    /**
     * Gets the ClientDaemon.
     * @return ClientDaemon The communication layer.
     * @see ClientDaemon
     */
    public static ClientDaemon getCD() {
	return cd;
    }
	
    /**
     * Messagehandler called by ClientDaemon.
     * @param msg Chatmessage that has to be routed to the globalchat channel or other channels.
     */
    public void incomingMessage(Message msg) {
	if (msg instanceof ChatMessage) {
	    if (msg instanceof ChatMessage.S2C_Joinself) {
		ChatMessage.S2C_Joinself js = (ChatMessage.S2C_Joinself) msg;
				//TODO HACK FIXME
		if (js.channel.equals(ChatMessage.GLOBALCHANNELNAME) || js.channel.equals("null")) {
		    if (enableglobal) {
			System.out.println("warning: received duplicate joinself on global channel");
		    }
		    enableglobal = true;
		} else {
		    // to avoid race conditions and server checking this check is omitted
		    // though ClientFrame.setEnabledChat(true) should conceptually only be done here
		    /*if (enablechat == true) {
		      if (js.channel.equals(chat.getChannelname())) {
		      System.out.println("warning: received duplicate joinself on the same channel");
		      } else {
		      System.out.println("warning: received duplicate joinself on unknown channel" + js.channel);
		      }
		      }*/
		    enablechat = true;
		}
	    } else {
		ChatMessage cm = (ChatMessage) msg;
				// it's not an initialization packet, so check if the chatchannels are enabled;
		if (cm.channel.equals(ChatMessage.GLOBALCHANNELNAME)) {
		    if (!enableglobal) {
			System.out.println("expected joinself packet instead of " + cm);
			return;
		    }
		} else {
		    if (enablechat) {
			if (!cm.channel.equals(chat.getChannelname())) {
			    System.out.println("packet received from unexpected channel");
			    return;
			}
		    } else {
			System.out.println("expected joinself packet instead of " + cm);
			return;
		    }
		}
	    }

	    // now handle the packets
	    ChatMessage cm = (ChatMessage) msg;
	    if (cm.channel.equals(ChatMessage.GLOBALCHANNELNAME)) {
		global.packetHandler(cm);
	    } else {
		chat.packetHandler(cm);
	    }
	} else {
	    System.out.println("unsupported message received: " + msg.toString());
	}
    }

}
