package quest.client.chat;

import quest.client.*;
import quest.client.net.*;
import quest.global.net.*;
import quest.global.chat.*;
import quest.client.game.*;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import javax.swing.table.*;
import javax.swing.event.*;
import javax.swing.text.*;

//
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Vector;
import java.util.Hashtable;
import java.net.URL;
import java.net.MalformedURLException;

/**
 * Represents the channel/game chat tab in the Chat GUI. The admin can start the game in this channel.
 *
 * @author Sander Brandenburg
 * @see quest.global.chat.ChatMessage
 */
public class GUITab_Chat extends JPanel implements ActionListener {
  /**
   * The PopupMenu that has to be displayed over the Player field.
   */
  protected JPopupMenu playerPopup;
  protected MouseAdapter playerPopupAdapter;
  /**
   * The PopupMenu that has to be displayed over the Spectator field.
   */
  protected JPopupMenu spectatorPopup;
  protected MouseAdapter spectatorPopupAdapter;
	
  /**
   * The RankingTable that represents the Player Field.
   */
  protected RankingTable playerTable;
  /**
   * The data vector of the Player Field RankingTable.
   */
  protected Vector players;

  /**
   * The RankingTable that represents the Spectator Field.
   */
  protected RankingTable spectatorTable;
  /**
   * The data vector of the Spectator Field RankingTable.
   */
  protected Vector spectators;
  
  /**
   * JTextField used for user chat or command input.
   */
  protected JTextField	textField;
  /**
   * Here the channel output is printed on.
   */
  protected JTextPane	textPane;
  /**
   * The internal Document representation in the textPane. Used for displaying colors, different fonts and the like.
   */
  protected Document	textDoc;

  /**
   * Label shownon top of screen to display channel/game status.
   */
  protected JLabel		chatLabel;
  
  /**
   * Label shown when entering whisper mode.
   */
  protected JLabel whisperLabel;
  /**
   * Panel used to include whisperlabel in in whisper mode.
   */
  protected JPanel  textFieldPanel;
	
  /**
   * The 'start game' button.
   */
  protected JButton		startGame;
  
  /**
   * The 'kick' button.
   */
  protected JButton		kick;
  /**
   * The 'green up arrow' button.
   */
  protected JButton		uparrow;
  /**
   * The 'red down arrow' button.
   */
  protected JButton		downarrow;
  /**
   * The 'whisper' button.
   */
  protected JButton		whisperb;
  /**
   * The lockout checkbox.
   */
  protected JCheckBox lockoutCheckbox;

  /**
   * ButtonModel for the lockoutCheckbox to only be selected until the server has granted the Selection.
   * Will send a request for lockout on firing the button. The button will be selected through the method packetHandler.
   * @see GUITab_Chat#packetHandler
   */
  protected final class DelayedCheckboxModel extends DefaultButtonModel {
	/**
	 * Will actually send a request to the server to change the lockout status according to flag.
	 */
	public void setSelected(boolean flag) {
	  cd.sendMessage(new ChatMessage.C2S_Lockout(channelname, flag));
	}
	/**
	 * Toggles the checkbox according to flag.
	 */
	public void setSelectedConfirmed(boolean flag) {
	  super.setSelected(flag);
	}
  }

  /**
   * Contains the usermode of the active player.
   */
  protected int usermode;
  /**
   * Boolean whether player is logged in to the channel or not.
   */
  protected boolean loggedin = false;
	
  /**
   * The ClientDaemon Object communicator with the server.
   */
  protected final ClientDaemon cd = ClientDaemon.getInstance();
	
  /**
   * Contains the name of the current Channel.
   */
  protected String channelname;
  /**
   * Contains the mode of the current channel.
   */
  protected int channelmode;
  /**
   * Contains the lockout status of the current channel.
   */
  protected boolean lockout;
  
  /**
   * Contains the boolean whether whispermode is enabled.
   */
  protected boolean whispermode = false;

  /**
   * Contains the name of the target where the popupmenu is issued at.
   */
  protected String tableTarget = null;
  private Object tableOrigin = null;
	
  private final static String STYLE_NORMAL	= "normal";
  private final static String STYLE_TEXT		= "text";
  private final static String STYLE_DATE		= "date";
  private final static String STYLE_NAME		= "name";
  private final static String STYLE_SPECIAL	= "special";
  private final static String STYLE_BROADCAST	= "broadcast";
	
  protected final static double leftwindow_weightx = 1;
  protected final static double rightwindow_weightx = 0.0;
	
  protected final static Insets insets_small = new Insets(2, 2, 2, 2);
  protected final static Insets insets_none = new Insets(0, 0, 0, 0);
  protected final static int rightwindowwidth = 167;


  /**
   * ActionListener that sends a startgame request to the server.
   */
  protected final ActionListener startGameActionListener = new ActionListener() {
	public void actionPerformed(ActionEvent ae) {
	  if (players.size() <= 1) {
              	appendSpecial("Can't start a game with only 1 player!");
	                return;
       		 }
	  cd.sendMessage(new ChatMessage.C2S_Startgame(channelname));
	}
  };
  /**
   * Actionlistener that sends a stopgame request to the server.
   */
  protected final ActionListener stopGameActionListener = new ActionListener() {
	public void actionPerformed(ActionEvent ae) {
	  cd.sendMessage(new ChatMessage.C2S_Stopgame(channelname, ChatMessage.C2S_Stopgame.STOP_STOP));
	}
  };

  /**
   * Constructor that calls initForm.
   */
  public GUITab_Chat() {  
	players = null;
	spectators = null;
	channelname = null;
	lockout = false;
		
	initForm();
  }
	
  /**
   * Toggles WhisperMode.
   */
  protected void setEnableWhisperMode(boolean flag) {
	textFieldPanel.removeAll();
	if (flag) {
	  textFieldPanel.add(whisperLabel);
	  textField.requestFocus();
	}
	textFieldPanel.add(textField);
	textFieldPanel.revalidate();
	whispermode = flag;
  }
  
  /**
   * Sets the whisper target.
   * @param target The name of the selected player to be whispered to.
   */
  protected void setWhisperTarget(String target) {
	tableTarget = target;
	whisperLabel.setText("whisper to " + target + ":");
  }
   
  /**
   * Sets the current user mode to mode. Will enable or disable the lockoutcheckbox, start game, uparrow, downarow and kick button according to the player mode policy.
   * @param mode The mode the current user mode has to be set to.
   * See the ChatMessage.USERMODE_* fields for the arguments.
   * @exception RuntimeException When an invalid mode is requested.
   */
  protected void setUserMode(int mode) {
	usermode = mode;
	JButton[] buts = new JButton[] {uparrow, downarrow, kick, startGame};
	boolean[] bools;
	if (usermode == ChatMessage.USERMODE_PLAYER) {
	  bools = new boolean[] {false, true, false, false};
	} else if (usermode == ChatMessage.USERMODE_PLAYERADMIN) {
	  bools = new boolean[] {false, false, true, true};
	} else if (usermode == ChatMessage.USERMODE_SPEC_PLAYER) {
	  bools = new boolean[] {false, true, false, false};
	} else if (usermode == ChatMessage.USERMODE_SPEC_SPEC) {
	  bools = new boolean[] {true, false, false, false};
	} else {
	  throw new RuntimeException("invalid mode");
	}
		
	for (int i = 0; i <  buts.length; i++) {
	  if (buts[i] != null)
		buts[i].setEnabled(bools[i]);
	}

	if (lockoutCheckbox != null) {
	  if (usermode == ChatMessage.USERMODE_PLAYERADMIN) {
		System.out.println("lockoutcheckbox enabled!"); //DEBUG
		lockoutCheckbox.setEnabled(true);
	  } else {
		System.out.println("lockoutcheckbox disabled!"); //DEBUG
		lockoutCheckbox.setEnabled(false);
	  }
	}

	initPopup();
  }
	
  /**
   * Retrieves channel name of the current channel.
   * @return channel name of the current channel.
   */
  public String getChannelname() {
	return channelname;
  }
	
  /**
   * Initializes the form.
   */
  protected void initForm() {
	ImageIcon up, down;
	try {
	  up = new ImageIcon(new URL(ClientFrame.getCodeBase(), "quest/client/chat/images/up.gif"));
	  down = new ImageIcon(new URL(ClientFrame.getCodeBase(), "quest/client/chat/images/down.gif"));
	} catch (MalformedURLException mue) {
	  mue.printStackTrace();
	  return;
	}
		
	GridBagLayout gbl = new GridBagLayout();
	setLayout(gbl);
		
	GridBagConstraints c = new GridBagConstraints();

	// 'chat' label
	JLabel lab1 = chatLabel = new JLabel();
	lab1.setDisplayedMnemonic(KeyEvent.VK_Y);
	c.gridx = 0;
	c.gridy = 0;
	c.fill = GridBagConstraints.NONE;
	c.anchor = GridBagConstraints.SOUTHWEST;
	c.weightx = leftwindow_weightx;
	c.weighty = 0.0;
	c.insets = insets_small;
	gbl.setConstraints(lab1, c);
	add(lab1);
		
	// 'players' label
	JLabel lab2 = new JLabel("Players");
	lab2.setDisplayedMnemonic(KeyEvent.VK_P);
	c.gridx = 1;
	c.gridy = 0;
	c.fill = GridBagConstraints.NONE;
	c.anchor = GridBagConstraints.SOUTHWEST;
	c.weightx = rightwindow_weightx;
	c.weighty = 0.0;
	c.insets = insets_small;
	gbl.setConstraints(lab2, c);
	add(lab2);
		
	// the chat textarea for chattext
	JTextPane tp1 = textPane = new JTextPane();
	tp1.setEditable(false);
	tp1.setAutoscrolls(true);
	textDoc = tp1.getDocument();
	initStyles();
	JScrollPane sp1 = new JScrollPane(tp1, JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
	sp1.setAutoscrolls(true);
		
	c.gridx = 0;
	c.gridy = 1;
	c.fill = GridBagConstraints.BOTH;
	c.anchor = GridBagConstraints.CENTER;
	c.weightx = leftwindow_weightx;
	c.weighty = 1.0;
	c.gridheight = 4;
	c.insets = insets_small;
	gbl.setConstraints(sp1, c);
	c.gridheight = 1;
	add(sp1);
		
	// the line for entering your own chat
	final JTextField tf1 = textField = new JTextField();
	lab1.setLabelFor(tf1);
	tf1.addActionListener(new ActionListener() {
	  public void actionPerformed(ActionEvent ae) {
		if (tf1.getText().length() <= 0)
		  return;
				
		if (whispermode)
		  sendWhisper();
		else
		  sendChat();
	  }
	});
	whisperLabel = new JLabel();
	textFieldPanel = new JPanel();
	textFieldPanel.setLayout(new BoxLayout(textFieldPanel, BoxLayout.X_AXIS));
	//textFieldPanel.add(whisperLabel);
	textFieldPanel.add(textField);
	c.gridx = 0;
	c.gridy = 5;
	c.fill = GridBagConstraints.HORIZONTAL;
	c.anchor = GridBagConstraints.EAST;
	c.weightx = leftwindow_weightx;
	c.weighty = 0.0;
	c.gridwidth = 2;
	c.insets = insets_small;
	gbl.setConstraints(textFieldPanel, c);
	c.gridwidth	= 1;
	add(textFieldPanel);
		
	RankingTable rt = playerTable = new RankingTable("");
	final String[] columnNames = {"Admin", "Players"};
	final Class[] columnTypes = {Boolean.class, String.class};
	players = new Vector();
	VectorTableModel tm = new VectorTableModel(columnNames, columnTypes, players);
		
	rt.setTableMode(tm, true);
	rt.table.getTableHeader().setReorderingAllowed(false);
	rt.table.getTableHeader().setResizingAllowed(false);
	rt.table.setShowHorizontalLines(false);
	rt.table.setShowVerticalLines(false);
	rt.table.setShowGrid(false);
	lab2.setLabelFor(rt);
	c.gridx = 1;
	c.gridy = 1;
	c.fill = GridBagConstraints.BOTH;
	c.anchor = GridBagConstraints.CENTER;
	c.weightx = rightwindow_weightx;
	c.weighty = 1.0;
	c.insets = insets_small;
	gbl.setConstraints(rt, c);
	add(rt);
		
	//the panel that contains 'spectators' and both arrows for moving spectators		
	JPanel jp1 = new JPanel();
	jp1.setLayout(new BoxLayout(jp1, BoxLayout.X_AXIS));
	JLabel lab3 = new JLabel("Spectators");
	lab3.setDisplayedMnemonic(KeyEvent.VK_S);
	jp1.add(lab3);
	jp1.add(Box.createRigidArea(new Dimension(5, 0)));		
	JButton b1 = uparrow = new JButton(up);
	b1.setMargin(insets_none);
	b1.setActionCommand("toplayers");
	b1.addActionListener(this);
	jp1.add(b1);
	JButton b2 = downarrow = new JButton(down);
	b2.setMargin(insets_none);
	b2.setActionCommand("tospecs");
	b2.addActionListener(this);
	jp1.add(b2);
	jp1.add(Box.createHorizontalGlue());
	c.gridx = 1;
	c.gridy = 2;
	c.fill = GridBagConstraints.BOTH;
	c.anchor = GridBagConstraints.CENTER;
	c.weightx = rightwindow_weightx;
	c.weighty = 0.0;
	c.insets = insets_small;
	gbl.setConstraints(jp1, c);
	add(jp1);
		
	RankingTable rt2 = spectatorTable = new RankingTable("");
	final String[] columnNames2 = {"Spectators", "X"};
	final Class[] columnTypes2 = {String.class, Boolean.class};
	spectators = new Vector();
	VectorTableModel tm2 = new VectorTableModel(columnNames2, columnTypes2, spectators);
		
	rt2.setTableMode(tm2, true);
	rt2.table.getTableHeader().setResizingAllowed(false);
	rt2.table.setShowHorizontalLines(false);
	rt2.table.setShowVerticalLines(false);
	rt2.table.setShowGrid(false);
	rt2.table.getTableHeader().setReorderingAllowed(false);
	c.gridx = 1;
	c.gridy = 3;
	c.fill = GridBagConstraints.BOTH;
	c.anchor = GridBagConstraints.CENTER;
	c.weightx = rightwindow_weightx;
	c.weighty = 1.0;
	c.insets = insets_small;
	gbl.setConstraints(rt2, c);
	add(rt2);
		
	// the administrator buttons and whisper mode
	JPanel jph = new JPanel();
	jph.setLayout(new BoxLayout(jph, BoxLayout.Y_AXIS));
	JPanel jp4 = new JPanel();
	jp4.setLayout(new BoxLayout(jp4, BoxLayout.X_AXIS));
	//			JLabel jlab = new JLabel("lockout channel");
	//			jp4.add(jlab);
	jp4.add(Box.createHorizontalGlue());
	lockoutCheckbox = new JCheckBox("Lockout channel", false);
	lockoutCheckbox.setEnabled(false);
	System.out.println("lockoutcheckbox disabled!"); //DEBUG
	lockoutCheckbox.setModel(new DelayedCheckboxModel());
	jp4.add(lockoutCheckbox);
	jph.add(jp4);
	JPanel jp2 = new JPanel();
	jp2.setLayout(new BoxLayout(jp2, BoxLayout.X_AXIS));
	JButton b3 = startGame = new JButton("Start game");
	b3.addActionListener(startGameActionListener);
	b3.setMargin(insets_small);
	b3.setMnemonic(KeyEvent.VK_S);
	b3.setEnabled(false);
	jp2.add(b3);
	JButton b5 = new JButton("Leave Channel");
	b5.setMargin(insets_small);
	b5.setMnemonic(KeyEvent.VK_L);
	b5.addActionListener(new ActionListener() {
	  public void actionPerformed(ActionEvent ae) {
		cd.sendMessage(new ChatMessage.C2S_Leave(channelname));
	  }
	});
	jp2.add(b5);
	JButton b4 = whisperb = new JButton("Whisper");
	b4.setMargin(insets_small);
	b4.setMnemonic(KeyEvent.VK_W);
	b4.addActionListener(new ActionListener() {
	  public void actionPerformed(ActionEvent ae) {
		setEnableWhisperMode(!whispermode);
		//tableTarget = null;
	  }
	});
	jp2.add(b4);
	jp2.setMinimumSize(new Dimension(rightwindowwidth, jp2.getPreferredSize().height));
	jph.add(jp2);
	c.gridx = 1;
	c.gridy = 4;
	c.fill = GridBagConstraints.VERTICAL;
	c.anchor = GridBagConstraints.CENTER;
	c.weightx = rightwindow_weightx;
	c.weighty = 0.0;
	c.insets = insets_small;
	gbl.setConstraints(jph, c);
	add(jph);
  }
	
  void initStyles() {
	Style def = StyleContext.getDefaultStyleContext().getStyle(StyleContext.DEFAULT_STYLE);
	StyleConstants.setFontFamily(def, "Tahoma");
	StyleConstants.setForeground(def, Color.blue);
	StyleConstants.setFontSize(def, 12);
	Style normal = textPane.addStyle(STYLE_NORMAL, def);

	Style s = textPane.addStyle(STYLE_DATE, normal);
	StyleConstants.setForeground(s, new Color(152, 30, 36));

	s = textPane.addStyle(STYLE_NAME, normal);
	StyleConstants.setBold(s, true);
	StyleConstants.setForeground(s, Color.black);

	s = textPane.addStyle(STYLE_SPECIAL, normal);
	StyleConstants.setBold(s, true);
	StyleConstants.setForeground(s, new Color(24, 174, 59));

	s = textPane.addStyle(STYLE_TEXT, normal);
	StyleConstants.setForeground(s, Color.black);

	s = textPane.addStyle(STYLE_BROADCAST, normal);
	StyleConstants.setForeground(s, new Color(223, 17, 27));
  }

  /** 
   * Adds a player to the player field.
   * @param name The name of the player.
   * @param modus Mode according to ChatMessage.USERMODE_*.
   */
  protected void addPlayer(String name, int modus) {
	Vector vec = new Vector(2);
	vec.addElement(new Boolean(modus == ChatMessage.USERMODE_PLAYERADMIN));
	vec.addElement(name);
		
	players.addElement(vec);
	int size = players.size() - 1;
	playerTable.tm.fireTableRowsInserted(size, size);
  }

  /** 
   * Returns the index of the player in the players vector.
   * @param name The name of the player to be found in the player vector.
   * @return The index of the player. -1 if not found.
   */
  protected int findPlayer(String name) {
	for (int i = 0; i < players.size(); i++) {
	  Vector vec = (Vector) players.elementAt(i);
	  String name2 = (String) vec.elementAt(1);
	  if (name.compareTo(name2) == 0) {
		return i;
	  }
	}
	return -1;
  }

  /**
   * Updates the mode of player.
   * @exception RuntimeException If name couldn't be found in the player vector.
   */
  protected void updatePlayer(String name, int modus) {
	int nr = findPlayer(name);
		
	if (nr == -1) {
	  System.out.println("couldn't find player " + name + " for update");
	  throw new RuntimeException("couldn't find player " + name + " for update");
	}

	Vector vec = (Vector) players.elementAt(nr);
	vec.setElementAt(new Boolean(modus == ChatMessage.USERMODE_PLAYERADMIN), 0);
	playerTable.tm.fireTableRowsUpdated(nr, nr);
  }
  
  /**
   * Removes the player from the player vector.
   */
  protected void removePlayer(String name) {
	int nr = findPlayer(name);
	if (nr == -1) {
	  System.out.println("couldn't remove " + name + " from playertable");
	  (new RuntimeException()).printStackTrace();
	  return;
	}

	players.removeElementAt(nr);
	playerTable.tm.fireTableRowsDeleted(nr, nr);		
  }
 
  /**
   * Adds a spectator to the Spectator field.
   */
  protected void addSpectator(String name, int modus) {
	Vector vec = new Vector(2);
	vec.addElement(name);
	vec.addElement(new Boolean(modus == ChatMessage.USERMODE_SPEC_PLAYER));
	spectators.addElement(vec);
	int size = spectators.size() - 1;
	spectatorTable.tm.fireTableRowsInserted(size, size);
  }

  /**
   * Finds a spectator in the spectator Vector.
   * @return The index of the spectator in the spectator Vector.
   */
  protected int findSpectator(String name) {
	for (int i = 0; i < spectators.size(); i++) {
	  Vector vec = (Vector) spectators.elementAt(i);
	  String name2 = (String) vec.elementAt(0);
	  if (name.compareTo(name2) == 0) {
		return i;
	  }
	}
	return -1;
  }
 
  /**
   * Updates the mode of a spectator in the spectatorfield.
   * @exception RuntimeException If the spectator couldn't be found.
   */
  protected void updateSpectator(String name, int modus) {
	int nr = findSpectator(name);
		
	if (nr == -1) {
	  System.out.println("couldn't find Spectator " + name + " for update");
	  (new RuntimeException()).printStackTrace();
	  return;
	}
		
	Vector vec = (Vector) spectators.elementAt(nr);
	vec.setElementAt(new Boolean(modus == ChatMessage.USERMODE_SPEC_PLAYER), 1);
	spectatorTable.tm.fireTableRowsUpdated(nr, nr);
  }
	
  /**
   * Removes a spectator from the spectatorfield.
   * @exception RuntimeException If the spectator couldn't be found.
   */
  protected void removeSpectator(String name) {
	int nr = findSpectator(name);
	if (nr == -1) {
	  (new RuntimeException()).printStackTrace();
	  System.out.println("couldn't remove " + name + " from spectatortable");
	  return;
	}
		
	spectators.removeElementAt(nr);
	spectatorTable.tm.fireTableRowsDeleted(nr, nr);
  }
	
  /**
   * Toggles whether this window is enabled.
   */
  public void setEnableWindow(boolean flag) {
	ClientFrame.setEnabledChat(flag);
  }

  // the clientframe class guarantees we receive the correct packet for this channel
  // and the correct order (first joinself, and then other packets
  /**
   * Used by ClientFrame to process the ChatMessage message Objects.
   * @param cm The ChatMessage to be processed.
   */
  public void packetHandler(ChatMessage cm) {		
	String name;
	int mode;
	Class c = cm.getClass();
		
	if (c == ChatMessage.S2C_Broadcast.class) {
	  ChatMessage.S2C_Broadcast cmc = (ChatMessage.S2C_Broadcast) cm;
	  appendSpecial(cmc.text, STYLE_BROADCAST);
	} else if (c == ChatMessage.S2C_Msg.class) {
	  ChatMessage.S2C_Msg cmc = (ChatMessage.S2C_Msg) cm;
	  //name, text
	  appendChat(cmc.username, cmc.text);
	} else if (c == ChatMessage.S2C_Joinself.class) {
	  handleJoinChannel((ChatMessage.S2C_Joinself) cm);
	} else if (c == ChatMessage.S2C_Leaveself.class) {
	  appendSpecial("*** You left channel " + cm.channel);
	  chatLabel.setText("");
	  setEnableWindow(false);
	  loggedin = false;
	} else if (c == ChatMessage.S2C_Userjoin.class) {
	  ChatMessage.S2C_Userjoin cmc = (ChatMessage.S2C_Userjoin) cm;
	  name = cmc.username;
	  appendSpecial("*** Joins: " + name);
	  mode = cmc.umode;
	  if (isPlayerMode(mode))
		addPlayer(name, mode);
	  else
		addSpectator(name, mode);
	} else if (c == ChatMessage.S2C_Signoff.class) {
	  ChatMessage.S2C_Signoff cmc = (ChatMessage.S2C_Signoff) cm;
	  name = cmc.username;
	  mode = cmc.mode;
	  if (mode == ChatMessage.SIGNOFF_LEAVE)
		appendSpecial("*** Parts: " + name);
	  else if (mode == ChatMessage.SIGNOFF_KICK)
		appendSpecial("*** Quits: " + name + " (Kicked)");
	  else if (mode == ChatMessage.SIGNOFF_TIMEOUT)
		appendSpecial("*** Quits: " + name + " (Timeout)");
	  else 
		appendSpecial("*** Quits: " + name + " (unknown reason)");
			
	  if (findPlayer(name) >= 0)
		removePlayer(name);
	  else
		removeSpectator(name);
	} else if (c == ChatMessage.S2C_Chusermode.class) { //TODO bail out if usermode is undef
	  ChatMessage.S2C_Chusermode cmc = (ChatMessage.S2C_Chusermode) cm;
	  int index;
	  boolean isnowplayer = true;
	  name = cmc.username;
	  mode = cmc.umode;
	  boolean wantstobecomeplayer = isPlayerMode(mode);
			
	  if (name.compareTo(ClientFrame.getUsername()) == 0) {
		//hey, it's you!
		setUserMode(mode);
	  }
			
	  index = findPlayer(name);
	  if (index == -1) {
		index = findSpectator(name);
		if (index == -1) {
		  System.out.println("nonexistant name " + name + " to change usermode");
		  return;
		}	
		isnowplayer = false;
	  }
	  // if he's player and wants to become spectator, or other way around - crossover
	  if (isnowplayer != wantstobecomeplayer) {
		if (isnowplayer) {
		  // player becomes spectator
		  removePlayer(name);
		  addSpectator(name, mode);
		} else {
		  // spectator becomes player
		  removeSpectator(name);
		  addPlayer(name, mode);
		}
	  } else {
		// player doesn't cross-over to spectator or spectator doesn't cross over to player
		// so only mode update is needed
		if (isnowplayer) {
		  updatePlayer(name, mode);
		} else {
		  updateSpectator(name, mode);
		}
	  }
	} else if (c == ChatMessage.S2C_Lockout.class) {
	  ChatMessage.S2C_Lockout cmc = (ChatMessage.S2C_Lockout) cm;
	  ((DelayedCheckboxModel) lockoutCheckbox.getModel()).setSelectedConfirmed(cmc.lockout);
	} else if (c == ChatMessage.S2C_Whisper.class) {
	  ChatMessage.S2C_Whisper cmw = (ChatMessage.S2C_Whisper) cm;
	  appendWhisper(cmw.username, cmw.text);
	} else if (c == ChatMessage.S2C_Whisperself.class) {
	  ChatMessage.S2C_Whisperself cmws = (ChatMessage.S2C_Whisperself) cm;
	  appendOwnWhisper(cmws.username, cmws.text);
	} else if (c == ChatMessage.S2C_Chchanmode.class) {
	  ChatMessage.S2C_Chchanmode ccm = (ChatMessage.S2C_Chchanmode) cm;
	  setChannelMode(ccm.mode);
	} else {
	  System.out.println("uncaught packet received: " + cm.toString());
	}
  }

  /**
   * Sets the Channelmode to ChatMessage.CHANMODE_BEGIN, ChatMessage.CHANMODE_END or ChatMessage.CHANMODE_GAME. 
   * Updates the chatLabel to the channel state and respectively does nothing, ends or starts the game using 
   * endGame and startGame.
   * @param mode According to ChatMessage.CHANMODE_*.
   */
  private void setChannelMode(int mode) {
	switch (mode) {
	case ChatMessage.CHANMODE_BEGIN:
	  chatLabel.setText("Game hasn't started yet - get ready!");
	  if (usermode == ChatMessage.USERMODE_PLAYERADMIN) {
		startGame.setText("Start Game");
		startGame.removeActionListener(stopGameActionListener);
		startGame.removeActionListener(startGameActionListener);
		startGame.addActionListener(startGameActionListener);
	  }
	  break;
	case ChatMessage.CHANMODE_END:
	  chatLabel.setText("Game is finished");
	  if (usermode == ChatMessage.USERMODE_PLAYERADMIN) {
		startGame.setText("Stop Game");
		startGame.removeActionListener(startGameActionListener);
		startGame.addActionListener(stopGameActionListener);
	  }
	  endGame();
	  break;
	case ChatMessage.CHANMODE_GAME:
	  chatLabel.setText("Game is started!");
	  if (usermode == ChatMessage.USERMODE_PLAYERADMIN) {
		startGame.setText("Stop Game");
		startGame.removeActionListener(startGameActionListener);
		startGame.addActionListener(stopGameActionListener);
	  }
	  startGame();
	  break;
	default:
	  throw new RuntimeException("Unexpected mode in ChatMessage.S2C_Setstatus");
	}
  }
  
  /**
   * Tries to log out gracefully sending a C2SLogout ChatMessage. Blocks until a reply has been received.
   */
  public void quit() {
	// assumed ClientApplet will kill the frame if logging out takes to long
	cd.sendMessage(new ChatMessage.C2S_Logout(channelname));

	while (loggedin) {
	  try {
		Thread.currentThread().sleep(10);
	  } catch (InterruptedException e) {} // ignore
	}
  }

  /**
   * Handles a S2C_Joinself ChatMessage packet.
   */
  private void handleJoinChannel(ChatMessage.S2C_Joinself js) {
	// TODO handle undefined shite
	if (loggedin == true) {
	  System.out.println("already logged in a channel and received duplicate joinself. ignored.");
	  System.out.println("packet: " + js);
	  return;
	}
	loggedin = true;
		
	setChannelMode(js.channelmode);
	boolean receivedSelf = false;
	try {
	  textDoc.remove(0, textDoc.getLength());
	} catch (BadLocationException ble) {} // ignore
	appendSpecial("*** Now talking in channel " + js.channel);
	channelname = js.channel;
		
	players.setSize(0);
	spectators.setSize(0);
	for (int i = 0; i < js.usernames.length; i++) {
	  int nr = js.umodes[i];
	  if (ClientFrame.getUsername().compareTo(js.usernames[i]) == 0) {
		receivedSelf = true;
		setUserMode(nr);
	  }
	  switch (nr) {
	  case ChatMessage.USERMODE_PLAYER: //FALL-THROUGH
	  case ChatMessage.USERMODE_PLAYERADMIN:
		addPlayer(js.usernames[i], nr);
		break;
	  case ChatMessage.USERMODE_SPEC_PLAYER: //FALL-THROUGH
	  case ChatMessage.USERMODE_SPEC_SPEC:
		addSpectator(js.usernames[i], nr);
		break;
	  case ChatMessage.USERMODE_UNDEFINED:
		addPlayer(js.usernames[i], 0);
		break;
	  default:
		if (ClientFrame.getUsername().compareTo(js.usernames[i]) == 0)
		  System.out.println("usermode corrupted");
		System.out.println("encountered undefined usermode: " + nr);
	  }	
	}
		
	if (receivedSelf == false) {
	  System.out.println("expected to receive own playername and state, but didn't");
	}
  }
	
  private final static SimpleDateFormat sdf = new SimpleDateFormat("HH:mm");
	
  /**
   * Appends an end of line character to the styled text document and scrolls to the end of the document.
   */
  private void appendEndOfLine() {
	try {
	  int l = textDoc.getLength();
	  textDoc.insertString(l, "\r\n", textPane.getStyle(STYLE_TEXT));
	  textPane.setCaretPosition(l);
	} catch (BadLocationException ble) {
	  System.err.println("Couldn't insert text.");
	}
  }
	
 /**
  * Appends current time to the styled text document according to the predefined styles.
  */
  private void appendTime() {
	try {
	  textDoc.insertString(textDoc.getLength(), "[" + sdf.format(new Date()) + "] ", textPane.getStyle(STYLE_DATE));
	} catch (BadLocationException ble) {
	  System.err.println("Couldn't insert text.");
	}
  }
  /**
   * Appends a chat message from a user to the styled text document according to the predefined styles.
   * @param user The username.
   * @param text The chat message the user has transmitted.
   */
  protected void appendChat(String user, String text) {
	try {
	  appendTime();
	  textDoc.insertString(textDoc.getLength(), "<" + user + "> ", textPane.getStyle(STYLE_NAME));
	  textDoc.insertString(textDoc.getLength(), text, textPane.getStyle(STYLE_TEXT));
	  appendEndOfLine();
	} catch (BadLocationException ble) {
	  System.err.println("Couldn't insert text.");
	}
  }
 /**
  * Appends a whisper message, that has been acknowledged to be sent to another user, to the styled text document according to the predefined styles.
  * @param user The destination user the whisper message has been sent to.
  * @param text The chat message the client has whispered to user.
  */
  protected void appendOwnWhisper(String username, String text) {
	try {
	  appendTime();
	  textDoc.insertString(textDoc.getLength(), "[whispered to " + username + "] ", textPane.getStyle(STYLE_NAME));
	  textDoc.insertString(textDoc.getLength(), text, textPane.getStyle(STYLE_TEXT));
	  appendEndOfLine();
	} catch (BadLocationException ble) {
	  System.err.println("Couldn't insert text.");
	}
  }
 /**
  * Appends a whisper message from a certain user to the styled text document according to the predefined styles.
  * @param user The user the whisper message has been received from.
  * @param text The chat message the user has whispered to.
  */
  protected void appendWhisper(String username, String text) {
	try {
	  appendTime();
	  textDoc.insertString(textDoc.getLength(), "[whisper from " + username + "] ", textPane.getStyle(STYLE_NAME));
	  textDoc.insertString(textDoc.getLength(), text, textPane.getStyle(STYLE_TEXT));
	  appendEndOfLine();
	} catch (BadLocationException ble) {
	  System.err.println("Couldn't insert text.");
	}
  }

  /** 
   * Appends a special message to the styled text document according a special style.
   * f.i. joining, leaving a channel.
   * @param text the text 
   */
  protected void appendSpecial(String text) {
	appendTime();
	appendText(text, STYLE_SPECIAL);
  }
	
   /** 
   * Appends a special message to the styled text document according a special style determined by style.
   * f.i. joining, leaving a channel.
   */
  protected void appendSpecial(String text, String style) {
	appendTime();
	appendText(text, style);
  }
	
  private void appendText(String text, String style) {
	try {
	  textDoc.insertString(textDoc.getLength(), text, textPane.getStyle(style));
	  appendEndOfLine();
	} catch (BadLocationException ble) {
	  System.err.println("Couldn't insert text.");
	}
  }
	
  /**
   * @return true if the client is a player (or playeradmin), false otherwise.
   */
  final boolean isPlayerMode(int mode) {
	switch(mode) {
	case ChatMessage.USERMODE_PLAYER:
	case ChatMessage.USERMODE_PLAYERADMIN:
	  return true;
	case ChatMessage.USERMODE_SPEC_PLAYER:
	case ChatMessage.USERMODE_SPEC_SPEC:
	  return false;
	default:
	  throw new RuntimeException("isPlayerMode: mode invalid");
	}
  }

  /**
   * Initializers a GameController that starts the game.
   * @see quest.client.game.GameController
   */
  protected void startGame() {
	GameController gc = new GameController(this.channelname, ClientFrame.getUsername());
  }

  /**
   * Sends a request to the server to abort the game.
   */
  protected void endGame() {
	cd.sendMessage(new ChatMessage.C2S_Abortgame(this.channelname));
  }

  // should only be called when tables are initialized
  /**
   * Initializes the specator and player popup menus.
   */
  protected void initPopup() {
	if (spectatorPopup != null) {
	  spectatorTable.table.removeMouseListener(spectatorPopupAdapter);
	}
	if (playerPopup != null) {
	  playerTable.table.removeMouseListener(playerPopupAdapter);
	}

	if (usermode == ChatMessage.USERMODE_UNDEFINED)
	  return;
		
	playerPopup = createPopupMenu("Player action");
	registerMenuItem(playerPopup, "Info", "info", KeyEvent.VK_I);
	registerMenuItem(playerPopup, "Whisper", "whisper", KeyEvent.VK_W);
	if (usermode == ChatMessage.USERMODE_PLAYERADMIN) {
	  registerMenuItem(playerPopup, "Kick", "kick", KeyEvent.VK_K);
	  registerMenuItem(playerPopup, "Move to Spectators", "tospecs", KeyEvent.VK_F);
	}
	playerPopupAdapter = new PopupMenuMouseListener(playerPopup, playerTable);
	playerTable.table.addMouseListener(playerPopupAdapter);

	spectatorPopup = createPopupMenu("Spectator Action");
	registerMenuItem(spectatorPopup, "Info", "info", KeyEvent.VK_I);
	registerMenuItem(spectatorPopup, "Whisper", "whisper", KeyEvent.VK_W);
	if (usermode == ChatMessage.USERMODE_PLAYERADMIN) {
	  registerMenuItem(spectatorPopup, "Kick", "kick", KeyEvent.VK_K);
	  registerMenuItem(spectatorPopup, "Move to Players", "toplayers", KeyEvent.VK_M);
	}
		
	spectatorPopupAdapter = new PopupMenuMouseListener(spectatorPopup, spectatorTable);
	spectatorTable.table.addMouseListener(spectatorPopupAdapter);
  }
	
  private JPopupMenu createPopupMenu(String label) {
	JPopupMenu jpum = new JPopupMenu();

	JPanel jpa = new JPanel();
	jpa.setLayout(new BoxLayout(jpa, BoxLayout.X_AXIS));
	jpa.add(Box.createHorizontalGlue());
	jpa.add(new JLabel(label));
	jpa.add(Box.createHorizontalGlue());
	jpum.add(jpa);
	jpum.addSeparator();

	return jpum;
  }
	

  /**
   * Adds a MenuItem to the menu using the actionlistener provided by this class.
   * @param menu The menu the menuitem has to be added to.
   * @param str The String to be represented as MenuItem.
   * @param actioncmd The actioncommand to be bound to.
   * @param key The shortcut key the action has to be bound to.
   */
  protected void registerMenuItem(JPopupMenu menu, String str, String actioncmd, int key) {
	JMenuItem mi = new JMenuItem(str, key);
	mi.setActionCommand(actioncmd);
	mi.addActionListener(this);
	menu.add(mi);
  }

  /**
   * Sends a chat message to the current channel, with text retrieved from the textField.
   */
  protected void sendChat() {
	ChatMessage.C2S_Chat chat  = new ChatMessage.C2S_Chat(channelname, textField.getText());
	cd.sendMessage(chat);
	textField.setText("");
  }

  /**
   * Sends a whisper message to the tableTarget, with text retrieved from the textField.
   */
  protected void sendWhisper() {
	ChatMessage.C2S_Whisper whisper = new ChatMessage.C2S_Whisper(channelname, textField.getText(), tableTarget);
	cd.sendMessage(whisper);
	textField.setText("");
	setEnableWhisperMode(false);
  }
	
  private void fireKick() {
	if (tableTarget == null) {
	  System.out.println("tabletarget null on firekick");
	  return;
	}
		
	ChatMessage.C2S_Kick kick = new ChatMessage.C2S_Kick(channelname, tableTarget);
	cd.sendMessage(kick);
  }
	
  private void enablePlayerAction() {
	whisperb.setEnabled(true);

	if (usermode == ChatMessage.USERMODE_PLAYERADMIN) {
	  kick.setEnabled(true);
	  if (tableOrigin == playerTable) {
		downarrow.setEnabled(true);
		uparrow.setEnabled(false);
	  } else {
		downarrow.setEnabled(false);
		uparrow.setEnabled(true);				
	  }
	}
  }
	
  private void disablePlayerAction() {
	whisperb.setEnabled(false);
	kick.setEnabled(false);
		
	if (usermode == ChatMessage.USERMODE_PLAYERADMIN) {
	  downarrow.setEnabled(false);
	  uparrow.setEnabled(false);
	}
  }

    /**
     * Catches the popupmenu actions.
     */
  public void actionPerformed(ActionEvent e) {
	String act = e.getActionCommand();
		
	if (act.equals("whisper")) {
	  setEnableWhisperMode(true);
	  setWhisperTarget(tableTarget);
	} else if (act.equals("kick")) {
	  fireKick();
	} else if (act.equals("tospecs")) {
	  if (usermode == ChatMessage.USERMODE_SPEC_SPEC) {
		System.out.println("shouldn't have happened: tospecs in spec_spec mode");
	  } else if (usermode == ChatMessage.USERMODE_SPEC_PLAYER) {
		ChatMessage.C2S_Chusermode chum = new ChatMessage.C2S_Chusermode(channelname, ClientFrame.getUsername(), ChatMessage.USERMODE_SPEC_SPEC);
		cd.sendMessage(chum);
	  } else if (usermode == ChatMessage.USERMODE_PLAYER) {
		ChatMessage.C2S_Chusermode chum = new ChatMessage.C2S_Chusermode(channelname, ClientFrame.getUsername(), ChatMessage.USERMODE_SPEC_SPEC);
		cd.sendMessage(chum);
	  } else if (usermode == ChatMessage.USERMODE_PLAYERADMIN) {
		if (tableTarget == null) {
		  System.out.println("tabletarget null on tospecs");
		  return;
		}
		ChatMessage.C2S_Chusermode chum = new ChatMessage.C2S_Chusermode(channelname, tableTarget, ChatMessage.USERMODE_SPEC_SPEC);
		cd.sendMessage(chum);
	  }
	} else if (act.equals("toplayers")) {
	  if (usermode == ChatMessage.USERMODE_SPEC_SPEC) {
		ChatMessage.C2S_Chusermode chum = new ChatMessage.C2S_Chusermode(channelname, ClientFrame.getUsername(), ChatMessage.USERMODE_SPEC_PLAYER);
		cd.sendMessage(chum);
	  } else if (usermode == ChatMessage.USERMODE_SPEC_PLAYER) {
		System.out.println("shouldn't have happened: toplayers in spec_player mode");
	  } else if (usermode == ChatMessage.USERMODE_PLAYER) {
		System.out.println("shouldn't have happened: toplayers in player mode");
	  } else if (usermode == ChatMessage.USERMODE_PLAYERADMIN) {
		if (tableTarget == null) {
		  System.out.println("tabletarget null on toplayers");
		  return;
		}
		ChatMessage.C2S_Chusermode chum = new ChatMessage.C2S_Chusermode(channelname, tableTarget, ChatMessage.USERMODE_PLAYER);
		cd.sendMessage(chum);
	  }
	} else {
	  JMenuItem source = (JMenuItem)(e.getSource());
	  String s = "Action event detected."
		+ "\n"
		+ "    Event source: " + source.getText()
		+ " (an instance of " + source.getClass().getName() + ")";
	  System.out.println(s);
	}
  }
	
  private void setChangeTarget(String target, RankingTable origin) {
	tableOrigin = origin;											 
	tableTarget = target;
		
	if (usermode == ChatMessage.USERMODE_PLAYERADMIN) {
	  if (origin == playerTable) {
		if (target.equals(ClientFrame.getUsername())) {
		  uparrow.setEnabled(false);
		  downarrow.setEnabled(false);
		} else {
		  // target is player
		  uparrow.setEnabled(false);
		  downarrow.setEnabled(true);
		}
	  } else {
		// target is spectator
		uparrow.setEnabled(true);
		downarrow.setEnabled(false);
	  }
	}
  }

  private class PopupMenuMouseListener extends MouseAdapter {
	private JPopupMenu	pum;
	private RankingTable		rt;
		
	public PopupMenuMouseListener(JPopupMenu pum, RankingTable rt) {
	  this.pum = pum;
	  this.rt = rt;
	}
		
	public void mousePressed(MouseEvent e) {
	  if (e.getSource() == playerTable.table) {
		spectatorTable.table.clearSelection();
	  } else if (e.getSource() == spectatorTable.table) {
		playerTable.table.clearSelection();
	  }
			
	  maybeShowPopup(e);
	}

	public void mouseReleased(MouseEvent e) {			
	  if (e.getSource() == playerTable.table) {
		spectatorTable.table.clearSelection();
	  } else if (e.getSource() == spectatorTable.table) {
		playerTable.table.clearSelection();
	  }
			
	  maybeShowPopup(e);
	}
		
	private void maybeShowPopup(MouseEvent e) {
	  int x = rt.table.rowAtPoint(e.getPoint());
	  if (rt == playerTable) 
		setChangeTarget((String) rt.tm.getValueAt(x, 1), rt);
	  else 
		setChangeTarget((String) rt.tm.getValueAt(x, 0), rt);
			
	  if (rt.table != null && e.isPopupTrigger() && x != -1) {
		//rt.table.clearSelection();
		//rt.table.setRowSelectionInterval(x, x);
		//rt.table.requestFocus();

		//tableTarget = (String) rt.tm.getValueAt(x, 1);
		//tableOrigin = rt;
				
		pum.show(e.getComponent(), e.getX(), e.getY());
	  }
	}
  }
}

