package quest.client.chat;

import javax.swing.*;
import javax.swing.table.*;
import java.util.Vector;
import java.util.Date;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.awt.*;
import java.awt.event.*;
import javax.swing.event.*;

import quest.client.net.ClientDaemon;
import quest.client.ClientFrame;
import quest.global.chat.RankingMessage;
import quest.global.chat.UserRanking;
import quest.global.net.Message;
import quest.global.net.MessageClientListener;

/**
 * Represents the ranking tab in the chat GUI. Here the rankings of all players, 
 * the top n players or n nearest neighbours can be retrieved and shown.
 *
 * @author Sander Brandenburg
 */
public class GUITab_Ranking extends JPanel implements MessageClientListener {
	
    /**
     * The ClientDaemon instance to provide communication to the server.
     */  
    protected final ClientDaemon cd = ClientDaemon.getInstance();

    /**
     * Registers RankingMessage to the ClientDaemon and initializes the Ranking form.
     */
    public GUITab_Ranking() {
	cd.registerProces(RankingMessage.class, this);
	initForm();
    }

    /**
     * The RankingTable visual representation of the three ranking overviews.
     */
    protected RankingTable RTAll, RTTop, RTRivals;
    /**
     * The Strings that contain the time a ranking update was received for any of the RankingTables.
     */
    protected String timelabels[] = new String[3];
	
    private final Insets insets_none = new Insets(0, 0, 0, 0);
    private final Insets insets_small = new Insets(2, 2, 2, 2);		
	
    /**
     * The tabbed pane that contains the RankingTables.
     */
    protected JTabbedPane tp;
    /**
     * Displays the current ranking status when retrieving ranking.
     */
    protected JLabel rankingLabel;
    /**
     * The actual label that will show the time of timelabels
     * @see GUITab_Ranking#timelabels
     */ 
    protected JLabel updateTimeLabel;
	
    /**
     * Initializes a RankingTable to support the default columntypes and show the default columnnames.
     * @param rt The RankingTable to be initialized
     */
    private void initRanking(RankingTable rt) {
	final String[] columnNames = {
	    "Position", "Ranking", "Username", "#Games", "#Wins", "#Lost"
	};
	final Class[] classTypes = {
	    Integer.class, Integer.class, String.class, Integer.class, Integer.class, Integer.class
	};

	Vector data = new Vector(50);
	VectorTableModel vtm = new VectorTableModel(columnNames, classTypes, data);
	rt.setTableMode(vtm, true);
    }
	
  /**
   * Messagehandler called by ClientDaemon.
   * @param m The RankingMessage.
   * @see quest.global.chat.RankingMessage
   * @exception RuntimeException when msg is not a RankingMessage.
   */
    public void incomingMessage(Message msg) {
	  if (!(msg instanceof RankingMessage)) {
		throw new RuntimeException("expected RankingMessage");
	  }
	RankingMessage rm = (RankingMessage) msg;
		
	if (rm instanceof RankingMessage.S2C_SendRanking) {
	    RankingMessage.S2C_SendRanking sr = (RankingMessage.S2C_SendRanking) rm;
	    RankingTable rt;
	    if (sr.mode == RankingMessage.RANKING_WHOLE) {
		rt = RTAll;
	    } else if (sr.mode == RankingMessage.RANKING_TOP) {
		rt = RTTop;
	    } else if (sr.mode == RankingMessage.RANKING_NEIGHBOUR) {
		rt = RTRivals;
	    } else {
		System.out.println("undefined mode received: " + sr);
		return;
	    }
	    if (sr.rankings == null) {
		System.out.println("blah");
		rt.setLabelMode("error retrieving scores");
		return;
	    }

	    initRanking(rt);
			
	    rt.getTableModel().getData().removeAllElements();
			
	    rt.getTableModel().getData().ensureCapacity(sr.rankings.length);
	    updateRankingLabel(sr.ownpos, sr.ranking, sr.maxpos);

	    for (int i = 0; i < sr.rankings.length; i++) {
		Vector vec = new Vector(6);
		UserRanking ur = sr.rankings[i];
		vec.addElement(new Integer(ur.position));
		vec.addElement(new Integer(ur.ranking));
		vec.addElement(ur.username);
		vec.addElement(new Integer(ur.nrgames));
		vec.addElement(new Integer(ur.nrwins));
		vec.addElement(new Integer(ur.nrlost));
		rt.getTableModel().getData().addElement(vec);
	    }
	    updateTimeLabel(rt);
	    rt.getTableModel().fireTableDataChanged();
	} else {
	    System.out.println("invalid packet received: " + msg);
	}
    }

    /**
     * Initializes the form.
     */
    private void initForm() {
	GridBagLayout gbl = new GridBagLayout();
	setLayout(gbl);
		
	GridBagConstraints c = new GridBagConstraints();
		
	JLabel lab1 = rankingLabel = new JLabel("");
	//lab1.setHorizontalAlignment(JLabel.CENTER);
	//lab1.setVerticalAlignment(JLabel.CENTER);
	//updateRankingLabel();
	c.gridx = 0;
	c.gridy = 0;
	c.fill = GridBagConstraints.NONE;
	c.anchor = GridBagConstraints.CENTER;
	c.weightx = 1.0;
	c.weighty = 0.0;
	c.insets = new Insets(5, 5, 5, 5);
	gbl.setConstraints(lab1, c);
	add(lab1);
		
		
	final JTabbedPane tp1 = new JTabbedPane(JTabbedPane.LEFT);
	tp = tp1;
		
	RTTop = new RankingTable("Press the update button to retrieve the scores");
	addShowTimeListener(RTTop);
	RTRivals = new RankingTable("Press the update button to retrieve the scores");
	addShowTimeListener(RTRivals);
	RTAll = new RankingTable("Press the update button to retrieve the scores");
	addShowTimeListener(RTAll);
		
		
	tp1.addTab("Top 50",      RTTop);
	tp1.addTab("Your Rivals", RTRivals);
	tp1.addTab("Whole List",  RTAll);
	c.gridx = 0;
	c.gridy = 1;
	c.fill = GridBagConstraints.BOTH;
	c.anchor = GridBagConstraints.CENTER;
	c.weightx = 1.0;
	c.weighty = 1.0;
	c.insets = new Insets(20, 2, 20, 30);
	gbl.setConstraints(tp1, c);
	add(tp1);
		
		
	JLabel lab2 = updateTimeLabel = new JLabel();
	c.gridx = 0;
	c.gridy = 2;
	c.fill = GridBagConstraints.NONE;
	c.anchor = GridBagConstraints.SOUTH;
	c.weightx = 1.0;
	c.weighty = 0.0;
	c.insets = insets_small;
	gbl.setConstraints(lab2, c);
	add(lab2);
		
	JButton b1 = new JButton("Update");
	b1.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent ae) {
		    final Runnable run = new Runnable() {
			    public void run() {
				int mode;
				switch(tp1.getSelectedIndex()) {
				case 0:
				    mode = RankingMessage.RANKING_TOP;
				    break;
				case 1:
				    mode = RankingMessage.RANKING_NEIGHBOUR;
				    break;
				case 2:
				    mode = RankingMessage.RANKING_WHOLE;
				    break;
				default:
				    System.out.println("invalid index selected in rankingtable");
				    return;
				}
				RankingMessage.C2S_RequestRanking rr = new RankingMessage.C2S_RequestRanking(mode);
				cd.sendMessage(rr);
			    }
			};
		    (new Thread(run)).start();
		}
	    });
	b1.setMnemonic(KeyEvent.VK_U);
	c.gridx = 0;
	c.gridy = 3;
	c.fill = GridBagConstraints.NONE;
	c.anchor = GridBagConstraints.CENTER;
	c.weightx = 1.0;
	c.weighty = 0.0;
	c.insets = insets_small;
	gbl.setConstraints(b1, c);
	add(b1);
		
    }
	
    /**
     * Updates the rankinglabel that shows current rank nr, position among other etc.
     * @param pos The position in rank of the user between all other users.
     * @param ranking The ranking score of this user.
     * @param maxpple The number of users in the database.
     */
    private void updateRankingLabel(int pos, int ranking, int maxpple) {
	rankingLabel.setText("Your current rank is " + ranking + " with position " + pos + " out of " + maxpple + " users");
    }
	
    private final static DateFormat df = new SimpleDateFormat("dd-MM-yy HH:mm:ss"); 
    /**
     * Updates the timelabel that shows when a ranking score reply has been received for the certain table.
     * @param rt The RankingTable RTAll, RTTop or RTRivals
     */
    private void updateTimeLabel(RankingTable rt) {
	int idx = tp.indexOfComponent(rt);
	if (idx == -1) {
	    System.out.println("couldn't retrieve index of rankingtable");
	    return;
	}
		
	timelabels[idx] = "List retrieved at " + df.format(new Date());
	updateTimeLabel.setText(timelabels[idx]);
    }

    /**
     * Updates the updateTimeLabel shown on screen out of field timelabels according to the given RankingTable.
     * @param rt The RankingTable: RTAll, RTTop or RTRivals.
     */
    private void showTimeLabel(RankingTable rt) {
	int idx = tp.indexOfComponent(rt);
	if (idx == -1) {
	    System.out.println("couldn't retrieve index of rankingtable");
	    return;
	}
	if (timelabels[idx] == null)
	    updateTimeLabel.setText("");
	else
	    updateTimeLabel.setText(timelabels[idx]);
    }

    /**
     * Adds ComponentAdapter to the specified RankingTable. This ComponentAdaoter will call showTimeLabel(rt) when the component is shown.
     * @param rt The RankingTable the ComponentAdapter has to be attached to.
     */
    private void addShowTimeListener(final RankingTable rt) {
	rt.addComponentListener(new ComponentAdapter() {
		public void componentShown(ComponentEvent e) {
		    showTimeLabel(rt);
		}
	    });
    }
}
