package quest.client.chat;

import java.util.Vector;
import javax.swing.table.TableModel;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.JTable;
import javax.swing.JPanel;
import javax.swing.JLabel;
import javax.swing.JScrollPane;
import javax.swing.ListSelectionModel;
import javax.swing.ScrollPaneLayout;
import java.awt.GridLayout;
import java.awt.BorderLayout;
import java.awt.event.MouseListener;
import java.awt.Component;							   
import java.util.EventListener;

/**
 * Written to represent a table based on a VectorTableModel.
 * RankingTable has two modes. It's initialized in Labelmode, where a message
 * will be displayed until Tablemode is entered using setTableMode.
 * 
 * @see RankingTable#setTableMode
 * @author Sander Brandenburg
 */
public class RankingTable extends JPanel {	
    private TableSorter tableSorter;
    /**
     * The actual table visualisation that's contained within RankingTable.
     */
	protected JTable table;
    /**
     * VectorTableModel that represents the table data.
     */
	protected VectorTableModel tm;
    /**
     * ScrollPane that's enabled in Tablemode.
     */
	protected JScrollPane sp;
    /**
     * Flag that contains the mode.
     */
	protected boolean labelMode = true;
		
    /**
     * Initializes the RankingTable in Labelmode using str as text.
     */
    RankingTable(String str) {
	setLayout(new GridLayout(1,1));
	setLabelMode(str);
    }
	
    /**
     * Sets RankingTable to Labelmode, with a centered JLabel containing the text.
     */
    public void setLabelMode(String text) {
	labelMode = true;
	
	removeAll();
	JLabel label = new JLabel(text);
	label.setHorizontalAlignment(JLabel.CENTER);
	label.setVerticalAlignment(JLabel.CENTER);
	
	add(label);
	revalidate();
    }
	
    /**
     * Sets RankingTable to Tablemode. 
     * @param tm The VectorTableModel to be represented.
     * @param showHeader Flag whether to show the header of the table.
     */
    public void setTableMode(VectorTableModel tm, boolean showHeader) {
	labelMode = false; 
	tableSorter = new TableSorter(tm);
	table = new JTable(tableSorter);
	tableSorter.addMouseListenerToHeaderInTable(table);
	table.getTableHeader().setReorderingAllowed(false);
	table.getTableHeader().setResizingAllowed(true);
	table.getTableHeader().setEnabled(showHeader);
	
	table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
	tableSorter.addTableModelListener(table);
	
	RankingTableCellRenderer rtcr = new RankingTableCellRenderer();
	table.setDefaultRenderer(String.class, rtcr);
	table.setDefaultRenderer(Integer.class, rtcr);
	
	this.tm = tm;
	
	removeAll();
	
	sp = new JScrollPane(table);
	add(sp);
	
	revalidate();
    }
	
    /**
     * @return the tablemodel when in TableMode
     * @exception RuntimeException when in Label mode.
     */
    public VectorTableModel getTableModel() {
	if (labelMode)
	    throw new RuntimeException("RankingTable is in Labelmode!");

	return tm;
    }
    
    /**
     * @return the table when in TableMode
     * @exception RuntimeException when in Label mode.
     */
    public JTable getTable() {
	if (labelMode)
	     throw new RuntimeException("RankingTable is in Labelmode!");

	return table;
    }
    
    /**
     * Simple TableCellRenderer that disregards focus of a cell. It aligns to the right.
     */
    private static class RankingTableCellRenderer extends DefaultTableCellRenderer {
	public RankingTableCellRenderer() {
	    setHorizontalAlignment(JLabel.RIGHT);
	}
	
	public Component getTableCellRendererComponent(JTable table,
						       Object value,
						       boolean isSelected,
						       boolean hasFocus,
						       int row,
						       int column) {
	    if (value == null) {
		System.err.println("NULL at " + row + ", " + column);
		return super.getTableCellRendererComponent(table, "NULL", isSelected, false, row, column);
	    }
	    
	    return super.getTableCellRendererComponent(table, value.toString(), isSelected, false, row, column);
	}
    }
}

