package quest.client.game;

import quest.global.game.*;
import javax.media.j3d.*;
import javax.vecmath.*;
import com.sun.j3d.utils.picking.PickTool;

/**
 * Represents a 3D arrow. 
 * An arrow has a position and an orientation, which are set at construction
 * time and can't be changed afterwards.
 *
 * @author Mathijs den Burger
 */

class Arrow extends PickableQuestObject {

	private static final float ARROW_HEAD_WIDTH = CorridorBlock.BLOCK_SIZE, // * (3f / 7f),
		ARROW_HEAD_HEIGHT = CorridorBlock.BLOCK_SIZE * (4f / 7f),
		ARROW_BODY_WIDTH = CorridorBlock.BLOCK_SIZE * (3f / 7f),
		ARROW_BODY_HEIGHT = CorridorBlock.BLOCK_SIZE * (3f / 7f);

	private MazePosition3D position;
	private Orientation2D orientation;
	private Shape3D model;

	// a hack, since an arrow hasn't an associated Selectable logical object.
	private static final Selectable selectable = new MazeCorridor3D(new MazePosition3D(0, 0, 0));

	/**
	 * Construct an arrow.
	 * @param position The position of the arrow in the games' board.
	 * @param orientation The 2D orientation of the arrow in the same plane as 
	 * the games' board levels.
	 * @see quest.client.game.GamesBoard
	 */
	public Arrow(MazePosition3D position, Orientation2D orientation) {
		super(selectable);
		this.position = position;
		this.orientation = orientation;
	}

	/**
	 * Gives the associated selectable object of this arrow.
	 * @return The associated selectable object
	 */
	public Selectable getLogicalArrow() {
		return selectable;
	}

	/**
	 * Gives the position of this arrow. Changing this position won't effect
	 * the position of the arrow in 3D space, so it can only be used for reading.
	 * @return The position of this arrow
	 */
	public MazePosition3D getPosition() {
		return position;
	}

	/**
	 * Gives the orientation of this arrow. This is the orientation in the same
	 * plane as the levels of the games' board. Changing the orientation won't
	 * effect the orientation of the arrow in 3D space, so it can only be used
	 * for reading.
	 * @return The orientation of this arrow
	 */
	public Orientation2D getOrientation() {
		return orientation;
	}

	/**
	 * Gives the transform object that can be used to rotate a scene graph below 
	 * it to the correct rotation, as specified at construction time of this arrow.
	 * @return The transform object
	 */
	protected Transform3D getRotationMatrix() {
		Transform3D moveToOrigin = new Transform3D();
		Transform3D rotation = new Transform3D();
		Transform3D result = new Transform3D();

		moveToOrigin.setTranslation(new Vector3f(-CorridorBlock.BLOCK_SIZE / 2, 0, -CorridorBlock.BLOCK_SIZE / 2));
		result.mulInverse(moveToOrigin);

		rotation.rotY(Math.toRadians(orientation.getRotation()));

		result.mul(rotation);
		result.mul(moveToOrigin);
		return result;
	}

	/**
	 * Gives the transform object that can be used to position a scene graph below 
	 * it to the correct position in 3D space, as specified at construction 
	 * time of this arrow. The actual translation is specified by the method
	 * {@link quest.client.game.GamesBoard#getGridTranslation(int, int, int)}.
	 * @return The transform object
	 */
	protected Transform3D getPositionMatrix() {
		Transform3D result = new Transform3D();
		result.set(GamesBoard.getGridTranslation(position.getX(), position.getY(), position.getZ()));
		return result;
	}

	/**
	 * Builds the 3D model that represent an arrow in 3D space. The model is
	 * oriented and positioned in 3D space as specified at construction time.
	 * The model can also be picked.
	 * @return The 3D model
	 */ 
	public Node buildSelectableModel() {
		NodeComponentFactory ncf = NodeComponentFactory.instance();
		ModelTool mt = ModelTool.instance();
		int vertexFormat = QuadArray.COORDINATES | QuadArray.NORMALS;
		model = new Shape3D();

		// build geometry
		float[] arrowHeadCoords = {
			0f, 0f, 0f,										// top
			0f, 0f, 0f,										// top
			-ARROW_HEAD_WIDTH / 2, 0f, ARROW_HEAD_HEIGHT,	// left corner
			ARROW_HEAD_WIDTH / 2, 0f, ARROW_HEAD_HEIGHT		// right corner
		};
		mt.translate(arrowHeadCoords, new Vector3f(CorridorBlock.BLOCK_SIZE / 2f, Floor.HEIGHT / 2f, (CorridorBlock.BLOCK_SIZE - ARROW_HEAD_HEIGHT - ARROW_BODY_HEIGHT) / 2f));
		float[] arrowBodyCoords = mt.planeSquareXZCoordinates(ARROW_BODY_WIDTH, ARROW_BODY_HEIGHT);
		mt.translate(arrowBodyCoords, new Vector3f((CorridorBlock.BLOCK_SIZE - ARROW_BODY_WIDTH) / 2f, Floor.HEIGHT / 2f, ((CorridorBlock.BLOCK_SIZE - ARROW_HEAD_HEIGHT - ARROW_BODY_HEIGHT) / 2f) + ARROW_HEAD_HEIGHT));

		// add geometry
		model.addGeometry(ncf.getQuadArray(mt.NR_PLANE_SQUARE_VERTICES, vertexFormat, arrowHeadCoords, mt.planeSquareXZNormals()));
		model.addGeometry(ncf.getQuadArray(mt.NR_PLANE_SQUARE_VERTICES, vertexFormat, arrowBodyCoords, mt.planeSquareXZNormals()));

		// build appearance
		model.setAppearance(ncf.getAppearance(ncf.getMaterial(new Color3f(0.4f, 0.4f, 0.4f), new Color3f(0.4f, 0.4f, 0.4f), new Color3f(0.4f, 0.4f, 0.4f), new Color3f(0.1f, 0.1f, 0.1f), 128f), null, ncf.getPolygonAttributes(PolygonAttributes.CULL_NONE, true)));

		// set capabilities
		PickTool.setCapabilities(model, PickTool.INTERSECT_TEST);

		// position and rotate arrow in the maze
		TransformGroup rotationTg = new TransformGroup(getRotationMatrix());
		rotationTg.addChild(model);
		TransformGroup positionTg = new TransformGroup(getPositionMatrix());
		positionTg.addChild(rotationTg);

		return positionTg;
	}

}