package quest.client.game;

import quest.global.game.*;
import java.util.*;
import java.net.URL;
import java.awt.event.*;
import javax.media.j3d.*;
import javax.vecmath.*;
import javax.media.j3d.PositionPathInterpolator;
import com.sun.j3d.utils.geometry.Box;
import com.sun.j3d.utils.picking.PickTool;
import java.security.InvalidParameterException;

/**
 * <P>
 * Represents a 3D corridor in the maze. This is the building block of a maze.
 * It is the 3D representation of the logical {@link quest.global.game.MazeCorridor3D}
 * object.</P> 
 * <P>
 * A corridor block's model depends on the type of corridor that is defined in
 * the associated {@link quest.global.game.MazeCorridor3D} object. A corridor block
 * has an position and orientation in 3D space, each of which can be altered at
 * runtime (by changing the {@link quest.global.game.MazePosition3D} and
 * {@link quest.global.game.Orientation2D} parts of the associated 
 * {@link quest.global.game.MazeCorridor3D} object).
 * The corridor block itself observes these parts and will adapt it's
 * 3D model when these parts change. These adaptations can be implemented as
 * animations of which the corridor block takes care itself. This means that if
 * you change e.g. the orientation of a corridor block, the 3D model will 
 * immediately execute an animation that rotates the corridor block to it's new
 * orientation. The code that does this looks something like:
 * <PRE>
 * CorridorBlock cb = new CorridorBlock(new MazePosition3D(0, 0, 0), new Orientation2D());
 * cb.getLogicalCorridor().getOrientation().rotate(90)
 * </PRE>
 * <P>
 * A change of the selection state is reflected by adapting the appearance of
 * the 3D model. This is also done automatically.</P>
 * <P>
 * When this corridor block contains a treasure, this is also included in the 3D model.
 * The corridor block removes this 3D treasure automatically from the model when
 * it is logically removed from the associated {@link quest.global.game.MazeCorridor3D}
 * object.</P>
 * <P>
 * A corridor block can be picked. When it contains a {@link quest.client.game.Teleport},
 * picks on the teleport will be passed to the pick listeners of the corridor block.</P>
 *
 * @author Mathijs den Burger
 */

class CorridorBlock extends PickableQuestObject implements PickListener {

    /**
     * The size of a corridor block (in metres). Corridor blocks are always
     * squares, and this is the size of an edge of the square that forms the
     * floor of a corridor block.
     */
    public static final float BLOCK_SIZE = (2 * Walls.WIDTH) + Walls.CORRIDOR_WIDTH;

    // animation constants
    private static final long SLIDE_TIME = 2000,	// time a slide to next grid position takes (in milliseconds)
		FREED_TIME = 1,
		POSITION_FREE_BLOCK_TIME = 1,
		ROTATION_TIME = 1000;

    private static final float[] SLIDE_KNOTS = {0f, 1f},
		FREED_KNOTS = {0f, SLIDE_TIME / (SLIDE_TIME + FREED_TIME), 1f},
		POSITION_FREE_BLOCK_KNOTS = {0f, 1f};

    // appearance modes
    private static final int WALLS = 0, FLOOR = 1, NR_GEOMETRIES = 2;

	/**
	 * Appearance constant, meaning the corridor block is illuminated. The corridor
	 * is lit, and all the walls and everything in it can be seen. The only
	 * thing that is different from the {@link quest.client.game.CorridorBlock#DAYLIGHT}
	 * mode are the colors used. This mode looks like a cave that is lit.
	 */
    private static final int ILLUMINATED = 0;
    
    /**
     * Appearance constant, meaning you can only see the contours of the corridor 
     * block. This means the form of the corridor can be seen, but you can't see
     * players or treasures that might be inside the corridor block.
     */
    private static final int CONTOURS = 1;
    
    /**
     * Appearance constant, meaning you can see nothing about the corridor block.
     * You can't tell it's shape nor if there are any players or treasures inside
     * it.
     */
    private static final int PITCH_DARK = 2;
    
	/**
	 * Appearance constant, meaning the corridor block is in clear daylight.
	 * The corridor is lit by the sun, and all the walls and everything in it can be seen. The only
	 * thing that is different from the {@link quest.client.game.CorridorBlock#ILLUMINATED}
	 * mode are the colors used. This mode looks like walls where the sun shines on.
	 */
    private static final int DAYLIGHT = 3;
    
    /**
     * The number of different appearance modes that are defined.
     */
    private static final int NR_APPEARANCE_MODES = 4;
    
    private static final int DEFAULT_APPEARANCE_MODE = DAYLIGHT;
    private static final float NOT_SHINY = 1f, MEDIUM_SHINY = 64, VERY_SHINY = 128f;
    private static Appearance[][][] appearances = new Appearance[NR_GEOMETRIES][NR_APPEARANCE_MODES][SelectionState.NR_STATES];
    static {
		NodeComponentFactory ncf = NodeComponentFactory.instance();

		for (int i = 0; i < NR_GEOMETRIES; i++)
		    for (int j = 0; j < NR_APPEARANCE_MODES; j++)
			for (int k = 0; k < SelectionState.NR_STATES; k++)
			    appearances[i][j][k] = null;

		appearances[WALLS][DAYLIGHT][SelectionState.UNSELECTABLE] = ncf.getAppearance(ncf.getMaterial(new Color3f(0.25f, 0.05f, 0f), new Color3f(0.25f, 0.05f, 0f), new Color3f(0.25f, 0.05f, 0f), new Color3f(0f, 0f, 0f), NOT_SHINY), null);
		appearances[WALLS][DAYLIGHT][SelectionState.UNSELECTED] = ncf.getAppearance(ncf.getMaterial(new Color3f(0.35f, 0.15f, 0f), new Color3f(0.35f, 0.15f, 0f), new Color3f(0.35f, 0.15f, 0f), new Color3f(0f, 0f, 0f), NOT_SHINY), null);
		appearances[WALLS][DAYLIGHT][SelectionState.SELECTED] = ncf.getAppearance(ncf.getMaterial(new Color3f(0.5f, 0.1f, 0f), new Color3f(0.5f, 0.5f, 0.5f), new Color3f(0.5f, 0.1f, 0f), new Color3f(0.5f, 0.0f, 0f), VERY_SHINY), null);
		appearances[WALLS][DAYLIGHT][SelectionState.TARGET] = ncf.getAppearance(ncf.getMaterial(new Color3f(0.5f, 0.0f, 0f), new Color3f(0.5f, 0.3f, 0.5f), new Color3f(0.5f, 0.0f, 0f), new Color3f(0.5f, 0.0f, 0f), VERY_SHINY), null);

		appearances[WALLS][ILLUMINATED][SelectionState.UNSELECTED] = ncf.getAppearance(ncf.getMaterial(new Color3f(0.35f, 0.15f, 0f), new Color3f(0.35f, 0.15f, 0f), new Color3f(0.35f, 0.15f, 0f), new Color3f(0f, 0f, 0f), NOT_SHINY), null);
		appearances[WALLS][ILLUMINATED][SelectionState.SELECTED] = ncf.getAppearance(ncf.getMaterial(new Color3f(0.5f, 0.1f, 0f), new Color3f(0.5f, 0.5f, 0.5f), new Color3f(0.5f, 0.1f, 0f), new Color3f(0.5f, 0.0f, 0f), VERY_SHINY), null);
		appearances[WALLS][ILLUMINATED][SelectionState.TARGET] = ncf.getAppearance(ncf.getMaterial(new Color3f(0.5f, 0.0f, 0f), new Color3f(0.5f, 0.3f, 0.5f), new Color3f(0.5f, 0.0f, 0f), new Color3f(0.5f, 0.0f, 0f), VERY_SHINY), null);

		appearances[FLOOR][DAYLIGHT][SelectionState.UNSELECTABLE] = ncf.getAppearance(ncf.getMaterial(new Color3f(0.49f, 0.29f, 0f), new Color3f(0.49f, 0.29f, 0f), new Color3f(0.49f, 0.29f, 0f), new Color3f(0f, 0f, 0f), NOT_SHINY), null);
		appearances[FLOOR][DAYLIGHT][SelectionState.UNSELECTED] = ncf.getAppearance(ncf.getMaterial(new Color3f(0.5f, 0.3f, 0f), new Color3f(0.5f, 0.3f, 0f), new Color3f(0.5f, 0.3f, 0f), new Color3f(0f, 0f, 0f), NOT_SHINY), null);
		appearances[FLOOR][DAYLIGHT][SelectionState.SELECTED] = ncf.getAppearance(ncf.getMaterial(new Color3f(1.0f, 0f, 0f), new Color3f(0.7f, 0.7f, 0.7f), new Color3f(0.8f, 0.3f, 0f), new Color3f(1.0f, 0f, 0f), NOT_SHINY), null);
		appearances[FLOOR][DAYLIGHT][SelectionState.TARGET] = ncf.getAppearance(ncf.getMaterial(new Color3f(1.0f, 0.0f, 0f), new Color3f(0.5f, 0f, 0.0f), new Color3f(0.5f, 0f, 0f), new Color3f(0.5f, 0f, 0f), VERY_SHINY), null);

		appearances[FLOOR][ILLUMINATED][SelectionState.UNSELECTED] = ncf.getAppearance(ncf.getMaterial(new Color3f(0.24f, 0.14f, 0f), new Color3f(0.24f, 0.14f, 0f), new Color3f(0.24f, 0.14f, 0f), new Color3f(0f, 0f, 0f), NOT_SHINY), null);
		appearances[FLOOR][ILLUMINATED][SelectionState.SELECTED] = ncf.getAppearance(ncf.getMaterial(new Color3f(0.25f, 0.15f, 0f), new Color3f(0.35f, 0.35f, 0.35f), new Color3f(0.4f, 0.15f, 0f), new Color3f(0f, 0f, 0f), NOT_SHINY), null);
		appearances[FLOOR][ILLUMINATED][SelectionState.TARGET] = ncf.getAppearance(ncf.getMaterial(new Color3f(1.0f, 0.0f, 0f), new Color3f(0.25f, 0f, 0f), new Color3f(0.5f, 0f, 0f), new Color3f(0.5f, 0f, 0f), VERY_SHINY), null);

		appearances[FLOOR][CONTOURS][SelectionState.UNSELECTABLE] = ncf.getAppearance(ncf.getMaterial(new Color3f(0.5f, 0f, 0f), new Color3f(0.1f, 0.1f, 0.1f), new Color3f(0.3f, 0.3f, 0.3f), new Color3f(0.0f, 0.0f, 0.0f), NOT_SHINY), null);
		appearances[FLOOR][CONTOURS][SelectionState.UNSELECTED] = ncf.getAppearance(ncf.getMaterial(new Color3f(0.5f, 0f, 0f), new Color3f(0.1f, 0.1f, 0.1f), new Color3f(0.3f, 0.3f, 0.3f), new Color3f(0.0f, 0.0f, 0.0f), NOT_SHINY), null);

		appearances[FLOOR][PITCH_DARK][SelectionState.UNSELECTABLE] = ncf.getAppearance(ncf.getMaterial(new Color3f(0.5f, 0f, 0f), new Color3f(0.1f, 0.1f, 0.1f), new Color3f(0.3f, 0.3f, 0.3f), new Color3f(0.0f, 0.0f, 0.0f), NOT_SHINY), null);
		appearances[FLOOR][PITCH_DARK][SelectionState.UNSELECTED] = ncf.getAppearance(ncf.getMaterial(new Color3f(0.5f, 0f, 0f), new Color3f(0.1f, 0.1f, 0.1f), new Color3f(0.3f, 0.3f, 0.3f), new Color3f(0.0f, 0.0f, 0.0f), NOT_SHINY), null);
    }

    // switch values
    private static final int SWITCH_VISIBLE = 0, SWITCH_INVISIBLE = 1,
		SWITCH_CONTOURS_ON = 0, SWITCH_CONTOURS_OFF = 1;

    // static model geometry
    private static QuadArray floorGeometry = null;
    private static QuadArray pitchDarkGeometry = null;

    // non-static model parameters and parts
    private Floor floor;
    private Walls visibleWalls;
    private Treasure3D treasure;
    private Switch wallSwitch, treasureSwitch;
    private TransformGroup positionTg, orientationTg;

    // abstract block contents
    private MazeCorridor3D logicalCorridor;

	/**
	 * Constructs a corridor block that represents the given logical corridor.
	 * @param logicalCorridor The logical representation of this corridor block.
	 */
    public CorridorBlock(MazeCorridor3D logicalCorridor) {
		super(logicalCorridor);
		this.logicalCorridor = logicalCorridor;
		this.logicalCorridor.addObserver(this);
		this.logicalCorridor.getPosition().addObserver(this);
		this.logicalCorridor.getOrientation().addObserver(this);
		this.logicalCorridor.getSelectionState().addObserver(this);
		getModel();		// build model instantly, since a change of appearance needs to change the wallswitch
    }

	/**
	 * Gives the logical representation of this corridor block. When you change
	 * parts of this representation, this corridor block will immediately reflect
	 * the changes in the 3D model.
	 */
    MazeCorridor3D getLogicalCorridor() {
		return logicalCorridor;
    }

	/**
	 * Gives the 3D treasure that lies in this corridor.
	 * @return The 3D treasure that lies in this corridor, of null if there
	 * never was a treasure in this corridor. Note that this doesn't have anything to do
	 * with the logical presence of the treasure in this corridor: when there's
	 * was a treasure in this corridor but it is logically removed, you can still
	 * get the 3D model of the treasure.
	 * @return The 3D treasure, or null if this corridor never contained a treasure.
	 */
    Treasure3D getTreasure() {
    	return treasure;
    }

	/**
	 * This method is called whenever an observable part of this corridor block
	 * is changed. Applications shouldn't call this method directly.
	 */
    protected void updateSelectable(Observable o, Object arg) {
		if (o instanceof MazeCorridor3D) {
			treasurePresenceChanged();
		} else if (o instanceof Orientation2D) {
		    orientationChanged();
		} else if (o instanceof MazePosition3D) {
		    positionChanged();
		} else if (o instanceof SelectionState) {
		    selectionStateChanged();
	    }
	}

    private void treasurePresenceChanged() {
    	if (logicalCorridor.containsTreasure()) {
    		treasureSwitch.setWhichChild(Switch.CHILD_ALL);
    	} else {
    		treasureSwitch.setWhichChild(Switch.CHILD_NONE);
    	}
    }

    private void orientationChanged() {
		//orientationTg.setTransform(getRotationMatrix());
		Behavior animation = buildOrientationAnimation(logicalCorridor.getOrientation());
		BranchGroup animator = new BranchGroup();
		animator.addChild(animation);
		animator.compile();
		orientationTg.addChild(animator);
    }

    private void positionChanged() {
		Behavior animation = buildPositionAnimation(logicalCorridor.getPosition(), positionTg);
		BranchGroup animator = new BranchGroup();
		animator.addChild(animation);
		animator.compile();
		positionTg.addChild(animator);
    }

    private void selectionStateChanged() {
		updateAppearance();
    }

    private Behavior buildOrientationAnimation(Orientation2D orientation) {
    	// build alpha value
		Alpha alpha = new Alpha(1, ROTATION_TIME);
		alpha.setStartTime(System.currentTimeMillis());

		// coordinate system
		Transform3D transform = new Transform3D();
		transform.setTranslation(new Vector3f(BLOCK_SIZE / 2, 0, BLOCK_SIZE / 2));

		// build behavior
    	RotationInterpolator result = new RotationInterpolator(
							       alpha, orientationTg, transform,
							       (float)Math.toRadians(orientation.getOldRotation()),
							       (float)Math.toRadians(orientation.getRotation())
							       //(float)Math.toRadians(0),
							       //(float)Math.toRadians(Math.PI)
							       );
    	result.setSchedulingBounds(Game.EVERYTHING);

    	return result;
    }

	/**
	 * Builds an animation to reflect the change from the old coordinates to
	 * the new ones in the given {@link quest.global.game.MazePosition3D} object. 
	 * The actual animation depends on the new and old 'free' status of 
	 * the position:
	 * <UL>
	 * <LI><B>Is free and was free?</B><BR>
	 *     Slide the (still free) block into position
	 * <LI><B>Is free but wasn't free?</B><BR>
	 *     Slide the block outside the maze and then hide it by sliding it 
	 *     very quickly outside the visible range
	 * <LI><B>Else?</B><BR>
	 *     Slide the block from the old to the new position
	 * </UL>
	 * @param position The position to build the animation for
	 * @param positionTg The transformgroup to modify during the animation
	 * @return The animation
	 */
	static PositionPathInterpolator buildPositionAnimation(MazePosition3D position, TransformGroup positionTg) {
		if (position.isFree() && position.wasFree()) {
			//System.out.println("buildPositionFreeBlockAnim");
		    return buildPositionFreeBlockAnimation(position, positionTg);
		} else if (position.isFree() && !position.wasFree()) {
			//System.out.println("buildFreedAnim");
		    return buildFreedAnimation(position, positionTg);
		} else {	// !position.isFree() && position.wasFree() doesn't matter
			//System.out.println("buildSlideAnim from (" + position.getOldX() + ", " + position.getOldY() + ", " + position.getOldZ() + ") -> (" + position.getX() + ", " + position.getY() + ", " + position.getZ() + ")");
		    return buildSlideAnimation(position, positionTg);
		}
    }

    private static PositionPathInterpolator buildPositionFreeBlockAnimation(MazePosition3D pos, TransformGroup positionTg) {
		// build slide positions
		Point3f[] positions = new Point3f[2];
		positions[0] = new Point3f(GamesBoard.getGridTranslation(pos.getOldX(), pos.getOldY(), pos.getOldZ()));
		positions[1] = new Point3f(GamesBoard.getGridTranslation(pos.getX(), pos.getY(), pos.getZ()));

		// build alpha value
		Alpha alpha = new Alpha(1, POSITION_FREE_BLOCK_TIME);
		alpha.setStartTime(System.currentTimeMillis());

		// build behavior
		PositionPathInterpolator slide = new PositionPathInterpolator(
									      alpha, positionTg, new Transform3D(),
									      POSITION_FREE_BLOCK_KNOTS, positions);
		slide.setSchedulingBounds(Game.EVERYTHING);

		return slide;
    }

    private static PositionPathInterpolator buildFreedAnimation(MazePosition3D pos, TransformGroup positionTg) {
		// build animation positions
		Point3f[] positions = new Point3f[3];
		positions[0] = new Point3f(GamesBoard.getGridTranslation(pos.getOldX(), pos.getOldY(), pos.getOldZ()));
		positions[1] = new Point3f(GamesBoard.getGridTranslation(pos.getX(), pos.getY(), pos.getZ()));
		positions[2] = new Point3f(GamesBoard.getFreeTranslation(pos.getZ()));

		// build alpha
		Alpha alpha = new Alpha(1, SLIDE_TIME + FREED_TIME);
		alpha.setStartTime(System.currentTimeMillis());

		// build behavior
		PositionPathInterpolator slide = new PositionPathInterpolator(
									      alpha, positionTg, new Transform3D(),
									      FREED_KNOTS, positions);
		slide.setSchedulingBounds(Game.EVERYTHING);
		return slide;
    }

    private static PositionPathInterpolator buildSlideAnimation(MazePosition3D pos, TransformGroup positionTg) {
		// build slide positions
		Point3f[] positions = new Point3f[2];
		positions[0] = new Point3f(GamesBoard.getGridTranslation(pos.getOldX(), pos.getOldY(), pos.getOldZ()));
		positions[1] = new Point3f(GamesBoard.getGridTranslation(pos.getX(), pos.getY(), pos.getZ()));

		// alpha value
		Alpha alpha = new Alpha(1, SLIDE_TIME);
		alpha.setStartTime(System.currentTimeMillis());

		// build behavior
		PositionPathInterpolator slide = new PositionPathInterpolator(
									      alpha, positionTg, new Transform3D(),
									      SLIDE_KNOTS, positions);
		slide.setSchedulingBounds(Game.EVERYTHING);

		return slide;
    }

    private void updateAppearance() {
		int appearanceMode = calculateAppearanceMode();
		int selectionState = logicalCorridor.getSelectionState().getState();
		switch(appearanceMode) {
		    case DAYLIGHT:
				visibleWalls.setAppearance(appearances[WALLS][DAYLIGHT][selectionState]);
				break;
		    case ILLUMINATED: case CONTOURS: case PITCH_DARK:
		    	if (selectionState != SelectionState.UNSELECTABLE) {
			    visibleWalls.setAppearance(appearances[WALLS][ILLUMINATED][selectionState]);
			    wallSwitch.setWhichChild(ILLUMINATED);
		    	} else {
			    wallSwitch.setWhichChild(appearanceMode);
			}
			//wallSwitch.setWhichChild(ILLUMINATED); // USE THIS ONE TO ALWAYS SHOW THE LOWER LEVEL
			break;
		    default:
			throw new RuntimeException("Unknown appearanceMode in CorridorBlock.updateAppearanceMode: " + appearanceMode);
		}
		floor.setAppearance(appearances[FLOOR][appearanceMode][selectionState]);
    }

    private int calculateAppearanceMode() {
		MazePosition3D pos = logicalCorridor.getPosition();
		if (pos.getZ() == GameLogic.DARK_LEVEL) {		// limited visibility here
		    if (Maze3D.isSteadyX(pos.getX()) && Maze3D.isSteadyY(pos.getY())) {	// of these the contours are shown
			return CONTOURS;
		    } else {				// these are so dark you don't see anything
			return PITCH_DARK;
		    }
		} else {	// pos.getZ() == GameLogic.LIGHT_LEVEL || block is free
		    return DAYLIGHT;
		}
    }

	/**
	 * Gives the rotation transform that reflects the current logical orientation.
	 * @return The current rotation transform
	 */
    protected Transform3D getRotationMatrix() {
		Transform3D moveToOrigin = new Transform3D();
		Transform3D rotation = new Transform3D();
		Transform3D result = new Transform3D();

		moveToOrigin.setTranslation(new Vector3f(-BLOCK_SIZE / 2, 0, -BLOCK_SIZE / 2));
		result.mulInverse(moveToOrigin);

		rotation.rotY(Math.toRadians(logicalCorridor.getOrientation().getRotation()));

		result.mul(rotation);
		result.mul(moveToOrigin);
		return result;
    }

	/**
	 * Gives the position transform that reflects the current logical position.
	 * @return The current position transform
	 */
    protected Transform3D getPositionMatrix() {
		Transform3D result = new Transform3D();
		MazePosition3D pos = logicalCorridor.getPosition();
		if (pos.isFree()) {
		    result.set(GamesBoard.getFreeTranslation(pos.getZ()));
		} else {
		    result.set(GamesBoard.getGridTranslation(pos.getX(), pos.getY(), pos.getZ()));
		}
		return result;
    }

	/**
	 * Handles the picks performed on the teleport of this corridor block
	 * (if there is one). These pick events are passed to the pick listeners
	 * of this corridor block.
	 */
	public void pickPerformed(PickEvent e) {
		// this corridor block contains a teleport, and that has been picked
		// forward the pick event to all the listeners of this corridor block
		// so it looks like this block has been picked
		dispatchEvent(e);
	}

	/**
	 * Builds the 3D model of the corridor block. The contents of the model
	 * depends on the logical contents in the {@link quest.global.game.MazeCorridor3D}
	 * object that was set at construction time.
	 * @return The 3D model
	 */
    protected Node buildSelectableModel() {
		int appearanceMode = calculateAppearanceMode();
		int selectionState = logicalCorridor.getSelectionState().getState();

		// build rotation
		orientationTg = new TransformGroup(getRotationMatrix());
		orientationTg.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		orientationTg.setCapability(TransformGroup.ALLOW_CHILDREN_EXTEND);

		// build position
		positionTg = new TransformGroup(getPositionMatrix());
		positionTg.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		positionTg.setCapability(TransformGroup.ALLOW_CHILDREN_EXTEND);
		positionTg.addChild(orientationTg);

		// floor
		floor = new Floor(BLOCK_SIZE, false);
		floor.setCapability(Shape3D.ALLOW_APPEARANCE_WRITE);
		PickTool.setCapabilities(floor, PickTool.INTERSECT_TEST);
		//floor.setAppearance(appearances[FLOOR][appearanceMode][selectionState]);
		orientationTg.addChild(floor);

		// teleport
		if (logicalCorridor.passageOpen(MazeCorridor3D.FLOOR) ||
		    logicalCorridor.passageOpen(MazeCorridor3D.CEILING)) {
			Teleport teleport = new Teleport(logicalCorridor);
			teleport.addPickListener(this);
		    orientationTg.addChild(teleport.getModel());
		}

		// walls and treasure
		visibleWalls = new Walls(logicalCorridor.getShape2D(), Floor.HEIGHT, Walls.SOLID);
		visibleWalls.setCapability(Shape3D.ALLOW_APPEARANCE_WRITE);
		PickTool.setCapabilities(visibleWalls, PickTool.INTERSECT_TEST);
		//visibleWalls.setAppearance(appearances[WALLS][appearanceMode][selectionState]);
		BranchGroup visibleWallsAndTreasure = new BranchGroup();
	    visibleWallsAndTreasure.addChild(visibleWalls);
		if (logicalCorridor.containsTreasure()) {
			treasure = new Treasure3D(logicalCorridor.getTreasure());
			treasureSwitch = new Switch(Switch.CHILD_ALL);
			treasureSwitch.setCapability(Switch.ALLOW_SWITCH_WRITE);
			treasureSwitch.addChild(treasure.getModel());
		    visibleWallsAndTreasure.addChild(treasureSwitch);
		} else {
			treasure = null;
		}

		if (appearanceMode == DAYLIGHT) {
		    // this block will always be in the upper level, so don't build
		    // the switch node to save rendering time
		    orientationTg.addChild(visibleWallsAndTreasure);
		} else {
		    // this block will always be in the lower level, so build the
		    // switch node for the other geometry of CONTOURS and PITCH_DARK blocks
		    wallSwitch = new Switch(ILLUMINATED);
		    wallSwitch.setCapability(Switch.ALLOW_SWITCH_WRITE);
		    wallSwitch.setCapability(Switch.ALLOW_CHILDREN_READ);

		    // add illuminated walls and treasure
		    wallSwitch.addChild(visibleWallsAndTreasure);

		    // add the dark contours
		    BranchGroup darkContours = new BranchGroup();
		    Shape3D dark = new Walls(logicalCorridor.getShape2D(), Floor.HEIGHT, Walls.PITCH_DARK);
		    PickTool.setCapabilities(dark, PickTool.INTERSECT_TEST);
		    dark.setAppearance(appearances[WALLS][CONTOURS][SelectionState.UNSELECTABLE]);
		    darkContours.addChild(dark);
		    Shape3D contours = new Walls(logicalCorridor.getShape2D(), Floor.HEIGHT, Walls.CONTOURS);
		    PickTool.setCapabilities(contours, PickTool.INTERSECT_TEST);
		    Appearance contoursApp = new Appearance();
		    contoursApp.setLineAttributes(new LineAttributes(2, LineAttributes.PATTERN_SOLID, false));
		    contours.setAppearance(contoursApp);
		    darkContours.addChild(contours);
		    wallSwitch.addChild(darkContours);

		    // add the pitch dark walls
		    Shape3D pitchDark = new Walls(logicalCorridor.getShape2D(), Floor.HEIGHT, Walls.PITCH_DARK);
		    PickTool.setCapabilities(pitchDark, PickTool.INTERSECT_TEST);
		    pitchDark.setAppearance(appearances[WALLS][PITCH_DARK][SelectionState.UNSELECTABLE]);
		    wallSwitch.addChild(pitchDark);

		    orientationTg.addChild(wallSwitch);
		}

		updateAppearance();

		return positionTg;
    }

}
