package quest.client.game;

import javax.media.j3d.*;
import javax.vecmath.*;

/**
 * A 3D floor. A floor is a square platform of an adaptable size and a fixed
 * height. A floor can contain a square fixed-size hole in the middle.
 *
 * @author Mathijs den Burger
 */

class Floor extends Shape3D {

	/**
	 * The height of the floor (in metres). You can also call it the thickness 
	 * of the floor if you want.
	 */
	public static final float HEIGHT = 1f;		// height (or thickness) of the floor
	
	/**
	 * The size of the hole in the middle (in metres)
	 */
	public static final float HOLE_SIZE = Walls.CORRIDOR_WIDTH * 0.666f;

	/**
	 * Construct a floor of the given size and with an optional hole in the middle.
	 * @param size The size of the floor (in metres)
	 * @param hasHole true if the floor should contain a hole, false if not
	 */
	public Floor(float size, boolean hasHole) {
		if (hasHole) {
			buildFloorWithHole(size);
		} else {	// floor without a hole
			buildFloorSolid(size);
		}
	}

	private void buildFloorWithHole(float size) {
		NodeComponentFactory ncf = NodeComponentFactory.instance();
		ModelTool mt = ModelTool.instance();
		int vertexFormat = QuadArray.COORDINATES | QuadArray.NORMALS;
		float side2hole = (size - HOLE_SIZE) / 2;

		// make coordinates
		float[] floorNorthCoords = mt.blockCoordinates(HOLE_SIZE + 1f, HEIGHT, side2hole);
		float[] floorSouthCoords = mt.blockCoordinates(HOLE_SIZE + 1f, HEIGHT, side2hole);
		float[] floorEastCoords = mt.blockCoordinates(side2hole, HEIGHT, size);
		float[] floorWestCoords = mt.blockCoordinates(side2hole, HEIGHT, size);
		
		// translate to the right place
		mt.translate(floorNorthCoords, new Vector3f(side2hole - 0.5f, 0, 0));
		mt.translate(floorSouthCoords, new Vector3f(side2hole - 0.5f, 0, side2hole + HOLE_SIZE));
		mt.translate(floorEastCoords, new Vector3f(side2hole + HOLE_SIZE , 0, 0));
		
		// add geometry
		addGeometry(ncf.getQuadArray(mt.NR_BLOCK_VERTICES, vertexFormat, floorNorthCoords, mt.blockNormals()));
		addGeometry(ncf.getQuadArray(mt.NR_BLOCK_VERTICES, vertexFormat, floorSouthCoords, mt.blockNormals()));
		addGeometry(ncf.getQuadArray(mt.NR_BLOCK_VERTICES, vertexFormat, floorEastCoords, mt.blockNormals()));
		addGeometry(ncf.getQuadArray(mt.NR_BLOCK_VERTICES, vertexFormat, floorWestCoords, mt.blockNormals()));
	}

	private void buildFloorSolid(float size) {
		NodeComponentFactory ncf = NodeComponentFactory.instance();
		ModelTool mt = ModelTool.instance();
		int vertexFormat = QuadArray.COORDINATES | QuadArray.NORMALS;

		// make coordinates
		float[] floorCoords = mt.blockCoordinates(size, HEIGHT, size);
		
		// add geometry
		addGeometry(ncf.getQuadArray(mt.NR_BLOCK_VERTICES, vertexFormat, floorCoords, mt.blockNormals()));
	}

}