package quest.client.game;

import quest.global.util.FiniteStateMachine;
import quest.global.game.*;
import java.applet.Applet;
import com.sun.j3d.utils.applet.MainFrame;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.net.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.JPopupMenu;
import javax.swing.JMenuItem;
import javax.media.j3d.*;
import com.sun.j3d.utils.universe.SimpleUniverse;
import com.sun.j3d.utils.behaviors.keyboard.KeyNavigatorBehavior;
import com.sun.j3d.utils.behaviors.mouse.MouseRotate;
import javax.vecmath.*;

/**
 * The constructor and controller of the interface of an Amazing Quest game.
 * After a Game is constructed the complete interface is build and ready to use.
 * The game GUI consists of:
 * <UL>
 * <LI>The 3D games board (of which the 3D world is controlled by a {@link quest.client.game.GamesBoard} object)
 * <LI>The list of players (implemented by a {@link quest.client.game.PlayerList} object)
 * <LI>The list of spectators (implemented by a {@link quest.client.game.SpectatorList} object)
 * <LI>The history of the game (implemented by a {@link quest.client.game.History} object
 * <LI>The status bar
 * <LI>The control buttons.
 * </UL>
 * The various parts are initialized with the data available in the {@link quest.global.game.GameLogic} object
 * the Game receives when it is constructed. Various parts of this GameLogic object
 * are from that moment on observed by the Game, so it can update it's GUI components
 * when that's nescessary. The observed parts are:
 * <UL>
 * <LI>The game states of all game users (called 'players' on the client side)
 * <LI>The {@link quest.global.game.GameUserList}
 * <LI>The {@link quest.global.game.SpectatorSet}
 * </UL>
 *
 * @author Mathijs den Burger
 */

class Game extends JFrame implements WindowListener, ActionListener, Observer {

    private static final int GUI_WIDTH = 700,
	GUI_HEIGHT = 500,
	BUTTON_BAR_HEIGHT = 30,
	CANVAS3D_WIDTH = GUI_HEIGHT - BUTTON_BAR_HEIGHT,
	CANVAS3D_HEIGHT = GUI_HEIGHT - BUTTON_BAR_HEIGHT,
	INFO_PANEL_WIDTH = 250,
	MISSION_INFO_HEIGHT = 60;

    private static final Color GUI_BACKGROUND = Color.gray;

    private static final Vector3f INIT_VIEWER_POSITION;
    static {
    	if (GameLogic.X_DIMENSION == 7 && GameLogic.Y_DIMENSION == 7) {
	    INIT_VIEWER_POSITION = new Vector3f(0f, 0f, 93f);
    	} else {
	    INIT_VIEWER_POSITION = new Vector3f(0f, 0f, 110f);
    	}
    }

    /**
     * The backing clip distance (in 3D world metres). This means
     * the viewer in the 3D world can't see any object that is further
     * away from him than this distance.
     */
    public static final double BACK_CLIP_DISTANCE = 300.0;

    private static final double GREATEST_DISTANCE = 10000.0;

    /**
     * The bounding sphere that includes everything in the 3D world.
     */
    public static final BoundingSphere EVERYTHING = new BoundingSphere(new Point3d(0.0, 0.0, 0.0), GREATEST_DISTANCE);

    private static final String LABEL_END_WALKING_BUTTON = "End walking",
	LABEL_ULV_BUTTON_HIDE = "Hide upper level  ",
	LABEL_ULV_BUTTON_SHOW = "Show upper level",
	LABEL_QUIT_BUTTON = "Quit",
	CONFIRM_QUIT_MESSAGE = "Do you want to quit this game?",
	CONFIRM_QUIT_TITLE = "Quit game?";

    private static final String STATUS_TEXT_WALK = "It's your turn. Click in the maze to walk around. Click on a treasure to pick it up. Drag to rotate the maze. Click '" + LABEL_END_WALKING_BUTTON + "' when you're done.",
	STATUS_CHANGE_MAZE = "Now you must change the maze. Click on an arrow to position the free corridors there.",
	STATUS_BLOCK_TO_INSERT_POSITIONED = "Click on an arrow to position the free corridors there. Right-click on the free corridors to rotate them, left-click on them to insert them.",
	STATUS_OBSERVE_PLAYER = "You can now only observe the moves of the other players. Wait for your turn...",
	STATUS_OBSERVE_SPECTATOR = "You are a spectator, and can only observe the moves of the players. Drag to rotate the maze.";

    private static final Font STATUS_TEXT_FONT = new Font("SansSerif", Font.BOLD, 12);
    private static final Color STATUS_TEXT_COLOR = new Color(0x33, 0x33, 0x44);

    private static final Insets BUTTON_INSETS = new Insets(5, 11, 5, 11);
    private static final int SPACE_BETWEEN_BUTTON_GROUPS = 3;	// pixels

    private Canvas3D canvas;
    private SimpleUniverse universe;
    private GamesBoard gamesBoard;
    private TransformGroup gamesBoardTg;
    private MouseRotate mouseRotator;
    private GameLogic logicalGame;
    private Uplink uplink;
    private String gameName;
    private PlayerList playerList;
    private MissionInfo missionInfo;
    private History history;
    private SpectatorList spectatorList;
    private JButton ulvButton, resetViewButton, endWalkingButton, quitButton;
    private JLabel statusText;
    private StatusBar statusBar;
    private JPopupMenu popupMenu;
    private JMenuItem popupResetView;

    /**
     * Constructs the GUI of an Amazing Quest game. After the construction the
     * GUI is ready to use.
     * @param logicalGame The logical representation of the game to construct
     *                    the interface for
     * @param uplink The link to the outside world
     * @param gameName The name of the game
     */
    public Game(GameLogic logicalGame, Uplink uplink, String gameName) {
	this.logicalGame = logicalGame;
	this.uplink = uplink;
	this.gameName = gameName;

	// observe the game state of all the players
	for (int i = 0; i < logicalGame.getNrUsers(); i++) {
	    this.logicalGame.getUserGameState(i).addObserver(this);
	}

	System.out.println("Creating canvas...");
	canvas = new Canvas3D(getGraphicsConfiguration3D());
	canvas.setSize(CANVAS3D_WIDTH, CANVAS3D_HEIGHT);

	System.out.println("Creating universe...");
	universe = new SimpleUniverse(canvas);

	System.out.println("Creating content...");
	gamesBoard = new GamesBoard(logicalGame, uplink);

	System.out.println("Creating scene...");
	BranchGroup scene = buildScene(canvas, universe, gamesBoard);

	System.out.println("Compiling scene...");
	scene.compile();

	System.out.println("Creating GUI...");
	buildGUI(logicalGame, canvas);

	universe.addBranchGraph(scene);
    }

    private BranchGroup buildScene(Canvas3D canvas, SimpleUniverse universe, GamesBoard gamesBoard) {
	BranchGroup scene = new BranchGroup();

	// initialize view
	universe.getViewingPlatform().setNominalViewingTransform();
	universe.getViewingPlatform().getViewPlatform().setActivationRadius((float)GREATEST_DISTANCE);
	universe.getViewer().getView().setBackClipDistance(BACK_CLIP_DISTANCE);

	// move viewer to initial position
	TransformGroup viewerTg = universe.getViewingPlatform().getViewPlatformTransform();
	Transform3D viewerTrans = new Transform3D();
	viewerTrans.setTranslation(INIT_VIEWER_POSITION);
	viewerTg.setTransform(viewerTrans);

	// add keyboard navigation to the viewer
	KeyNavigatorBehavior keyNavigator = new KeyNavigatorBehavior(canvas, viewerTg);
	keyNavigator.setSchedulingBounds(EVERYTHING);
	scene.addChild(keyNavigator);

	// add GamesBoard under a transformgroup (for behaviours)
	Transform3D gamesBoardTrans = new Transform3D();
	gamesBoardTg = new TransformGroup(gamesBoardTrans);
	gamesBoardTg.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
	gamesBoardTg.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
	gamesBoardTg.addChild(gamesBoard.getModel());

	// enable rotation of the gamesboard
	mouseRotator = new MouseRotate(canvas, gamesBoardTg);
	mouseRotator.setSchedulingBounds(EVERYTHING);
	scene.addChild(gamesBoardTg);
	scene.addChild(mouseRotator);

	// add pick behavior to the games board
	PickBehavior behavior = new PickBehavior(canvas, (BranchGroup)gamesBoard.getModel(), EVERYTHING);
	scene.addChild(behavior);

	// add light
    	DirectionalLight sun = new DirectionalLight(true, new Color3f(1, 1, 1), new Vector3f(0, 0, -1));
    	sun.setInfluencingBounds(EVERYTHING);
    	scene.addChild(sun);

    	return scene;
    }

    private GraphicsConfiguration getGraphicsConfiguration3D() {
	GraphicsConfigTemplate3D tmpl = new GraphicsConfigTemplate3D();
	GraphicsEnvironment env = GraphicsEnvironment.getLocalGraphicsEnvironment();
	GraphicsDevice device = env.getDefaultScreenDevice();
	return device.getBestConfiguration(tmpl);
    }

    private void buildGUI(GameLogic logicalGame, Canvas3D canvas) {
	setBackground(GUI_BACKGROUND);
	Box contentBox = new Box(BoxLayout.Y_AXIS);

	Box box = new Box(BoxLayout.X_AXIS);	//Box.createHorizontalBox();
	box.setBackground(GUI_BACKGROUND);

        box.add(canvas);
        updateCanvasCursor();

        JPanel infoPanel = buildInfoPanel(logicalGame);
        infoPanel.setPreferredSize(new Dimension(INFO_PANEL_WIDTH, GUI_HEIGHT));
        box.add(infoPanel);

        contentBox.add(box);
        contentBox.add(buildBottomBar());

        getContentPane().add(contentBox);

	addWindowListener(this);
    }

    private JPanel buildInfoPanel(GameLogic logicalGame) {
	JPanel result = new JPanel();
	result.setLayout(new BoxLayout(result, BoxLayout.Y_AXIS));

        // add player list
        GameUserList players = logicalGame.getUsers();
        playerList = new PlayerList(players.elements(), logicalGame.getActiveUserRank());
      	players.addObserver(this);
      	result.add(playerList);

	// add mission info
	missionInfo = new MissionInfo();
	missionInfo.setMaximumSize(new Dimension(INFO_PANEL_WIDTH, MISSION_INFO_HEIGHT));
	missionInfo.setPreferredSize(new Dimension(INFO_PANEL_WIDTH, MISSION_INFO_HEIGHT));
	result.add(missionInfo);

	// add history
	history = new History();
	result.add(history);

        // add spectator list
        SpectatorSet spectators = logicalGame.getSpectators();
        spectatorList = new SpectatorList(spectators.elements());
        spectators.addObserver(this);
        result.add(spectatorList);

	// put border around the control panel
	result.setBorder(BorderFactory.createEtchedBorder(EtchedBorder.LOWERED));

	// set colors
	result.setBackground(GUI_BACKGROUND);

	return result;
    }

    private JPanel buildBottomBar() {
	JPanel result = new JPanel();
	result.setLayout(new BoxLayout(result, BoxLayout.X_AXIS));

	// add status bar
	statusBar = new StatusBar();
	result.add(statusBar);

	// add "hide/show upper level" button
	ulvButton = new JButton(LABEL_ULV_BUTTON_HIDE);
	ulvButton.setEnabled(true);
	ulvButton.setMargin(BUTTON_INSETS);
	ulvButton.setMnemonic(KeyEvent.VK_U);
	ulvButton.addActionListener(this);
	result.add(ulvButton);

	// add "reset view" button
	resetViewButton = new JButton("Reset view");
	resetViewButton.setEnabled(true);
	resetViewButton.setMargin(BUTTON_INSETS);
	resetViewButton.setMnemonic(KeyEvent.VK_R);
	resetViewButton.addActionListener(this);
	result.add(resetViewButton);

	// add space
	result.add(Box.createHorizontalStrut(SPACE_BETWEEN_BUTTON_GROUPS));

	// add "end walk" button
	endWalkingButton = new JButton(LABEL_END_WALKING_BUTTON);
	endWalkingButton.setEnabled(false);
	endWalkingButton.setMargin(BUTTON_INSETS);
	endWalkingButton.setMnemonic(KeyEvent.VK_E);
	endWalkingButton.addActionListener(this);
	result.add(endWalkingButton);

	// add space
	result.add(Box.createHorizontalStrut(SPACE_BETWEEN_BUTTON_GROUPS));

	// add "quit" button
	quitButton = new JButton(LABEL_QUIT_BUTTON);
	quitButton.setEnabled(true);
	quitButton.setMargin(BUTTON_INSETS);
	quitButton.setMnemonic(KeyEvent.VK_Q);
	quitButton.addActionListener(this);
	result.add(quitButton);

	// set colors
	result.setBackground(GUI_BACKGROUND);
	result.setBorder(BorderFactory.createEtchedBorder(EtchedBorder.LOWERED));

	return result;
    }

    private void updateStatusText() {
    	switch(logicalGame.getMyGameState().getCurrentState()) {
	    case GameLogic.WALK:
		statusBar.setText(STATUS_TEXT_WALK);
		break;
	    case GameLogic.CHANGE_MAZE:
		statusBar.setText(STATUS_CHANGE_MAZE);
		break;
	    case GameLogic.BLOCK_TO_INSERT_POSITIONED: case GameLogic.ROTATION_CHANGED:
		statusBar.setText(STATUS_BLOCK_TO_INSERT_POSITIONED);
		break;
	    case GameLogic.OBSERVE:
		if (logicalGame.getMyRank() == GameLogic.NOBODY) {
		    // I'm a spectator
		    statusBar.setText(STATUS_OBSERVE_SPECTATOR);
		} else {
		    // I'm a player
		    statusBar.setText(STATUS_OBSERVE_PLAYER);
		}
		break;
	    case GameLogic.GAME_WON:
		String winnerName = logicalGame.getUser(logicalGame.getWinnerRank()).getName();
		String message = winnerName + " has won '" + gameName + "'. Press '" + LABEL_QUIT_BUTTON + "' to return to the chat.";
		statusBar.setText(message);
		break;
	    default:
		statusBar.setText("");
		break;
    	}
    }

    private void updateCanvasCursor() {
	if (logicalGame.getMyGameState().getCurrentState() == GameLogic.GAME_WON) {
	    canvas.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
	} else if (logicalGame.getMyRank() == logicalGame.getActiveUserRank()) {
	    canvas.setCursor(Cursor.getPredefinedCursor(Cursor.CROSSHAIR_CURSOR));
	} else {
	    canvas.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
	}
    }

    private boolean gameStarted = false;
    private void addLastGameMoveToHistory() {
    	GameMove lastMove = logicalGame.getLastGameMove();
    	if (lastMove == null) {
	    if (!gameStarted) {
		// the game has just been started
		history.addMetaGameStateChange("Welcome " + logicalGame.getUser(logicalGame.getMyRank()).getName());
		gameStarted = true;
	    }
    	} else if (logicalGame.getMyGameState().getCurrentState() == GameLogic.GAME_WON) {
	    // game finished
	    history.addMetaGameStateChange("Game finished, " + logicalGame.getUser(logicalGame.getWinnerRank()).getName() + " wins.");
	    history.repaint();
	    if (logicalGame.getWinnerRank() == logicalGame.getMyRank()) {
		history.addMetaGameStateChange("Congratulations!");
	    }
    	} else {
	    // someone performed a game move
	    String initiatorName = logicalGame.getUser(lastMove.getInitiatorRank()).getName();
	    if (lastMove.getInitiatorRank() == logicalGame.getMyRank()) {
		history.addOwnGameMove(lastMove, initiatorName);
	    } else {
		history.addGameMove(lastMove, initiatorName);
	    }
    	}
	// check if it's now the next player's turn
	if (lastMove instanceof GameMove.InsertFreeCorridorBlock &&
	    logicalGame.getActiveUserRank() != lastMove.getInitiatorRank()) {
	    // next player's turn
	    String nextName = logicalGame.getUser(logicalGame.getActiveUserRank()).getName();
	    String suffix = nextName.toLowerCase().endsWith("s") ? "'" : "'s";
	    history.addMetaGameStateChange("It's " +  nextName + suffix + " turn now");
    	}
    }

    /**
     * The update routine that is called when some Observable part that is
     * observed by this Game changes. You should not call this method directly.
     */
    public void update(Observable o, Object arg) {
	if (o instanceof FiniteStateMachine) {		// game state changed
	    // check if your mission has been set
	    if (!missionInfo.hasMission() || GameController.LOCAL_TEST_MODE) {
		// mission not set, which means the game has just been started
		// set your mission now
		if (logicalGame.getMyRank() >= 0) {
		    // I'm a player and have a mission, show it
		    Mission myMission = logicalGame.getUser(logicalGame.getMyRank()).getMission();
		    missionInfo.setMission(myMission);
		} else {
		    // I'm not a player, so remove the mission info component
		    missionInfo = null;
		}
	    }
	
	    // set the active player in the player list
	    playerList.setActivePlayerRank(logicalGame.getActiveUserRank());
	
	    // update the GUI according to your game state
	    if (logicalGame.getMyGameState().getCurrentState() == GameLogic.WALK) {
		endWalkingButton.setEnabled(true);
	    } else {
		endWalkingButton.setEnabled(false);
	    }
	    updateStatusText();
	    updateCanvasCursor();
	    addLastGameMoveToHistory();
	
	    if (logicalGame.getMyGameState().getCurrentState() == GameLogic.GAME_WON) {
		// the game has ended, no need to observe the game states anymore
		for (int i = 0; i < logicalGame.getNrUsers(); i++) {
		    logicalGame.getUserGameState(i).deleteObserver(this);
		}
	    }
	} else if (o instanceof SpectatorSet) {		// spectator joined or left
	    spectatorList.setData(logicalGame.getSpectators().elements());
	    history.addSpectatorPresenceChange(
					       logicalGame.getSpectators().getLastPresenceChangingSpectatorName(),
					       logicalGame.getSpectators().lastPresenceChangeWasJoining()
					       );
	} else if (o instanceof GameUserList) {		// game user left
	    GameUserList gameUserList = (GameUserList)o;
	    GameUser leftUser = gameUserList.getLastChangedPresentStatusUser();
				
	    // update the interface information components
	    playerList.repaint();		
	    history.addUserPresenceChange(leftUser.getName());
	}
    }

    /**
     * Handles the events generated by the interface. You should not call
     * this method directly.
     */
    public void actionPerformed(ActionEvent e) {
	if (e.getSource() == endWalkingButton) {
	    // end the walking of the active user
	    uplink.executeGameMove(new GameMove.ChangeMaze(logicalGame.getMyRank()));
	} else if (e.getSource() == ulvButton) {
	    // show or hide the upper level of the games' board
	    gamesBoard.setUpperLevelVisible(!gamesBoard.getUpperLevelVisible());
	    if (gamesBoard.getUpperLevelVisible()) {
		ulvButton.setText(LABEL_ULV_BUTTON_HIDE);
	    } else {
		ulvButton.setText(LABEL_ULV_BUTTON_SHOW);
	    }
	} else if (e.getSource() == resetViewButton) {
	    // reset the games board to it's (default) starting orientation
	    gamesBoardTg.setTransform(new Transform3D());
	    mouseRotator.setTransformGroup(gamesBoardTg);
	} else if (e.getSource() == quitButton) {
	    // quit the game, if you confirm this action
	    if (logicalGame.getMyGameState().getCurrentState() == GameLogic.GAME_WON ||
		quitConfirmed()) {
		uplink.logout();
	    }
	} else if (e.getSource() == popupResetView) {
	    System.out.println("View reset");
	}
    }

    private boolean quitConfirmed() {
	Object[] options = { "Yes", "No" };
	int chosenOption = JOptionPane.showOptionDialog(
							this,
							CONFIRM_QUIT_MESSAGE,
							CONFIRM_QUIT_TITLE,
							JOptionPane.YES_NO_OPTION,
							JOptionPane.QUESTION_MESSAGE,
							null,
							options,
							options[1]
							);
	return (chosenOption == JOptionPane.YES_OPTION);
    }

    private void popupGameFinished() {
	String winnerName = logicalGame.getUser(logicalGame.getWinnerRank()).getName();
	String message = winnerName + " has won " + gameName + ". Press OK to return to the chat";
	JOptionPane.showMessageDialog(this, message, "Game finished", JOptionPane.INFORMATION_MESSAGE);
    }

    /**
     * No special action is taken when the window is activated.
     * @param e The triggering window event
     */
    public void windowActivated(WindowEvent e) {}

    /**
     * Forces the Java3D renderer to stop when the window is closed.
     * @param e The triggering window event
     */
    public void windowClosed(WindowEvent e) {
	if (canvas.isRendererRunning()) canvas.stopRenderer();
    }

    /**
     * Logs gracefully out of the game server when the window is closing.
     * @param e The triggering window event
     */
    public void windowClosing(WindowEvent e) {
	if (quitConfirmed()) uplink.logout();
    }

    /**
     * No special action is taken when the window is deactivated.
     * @param e The triggering window event
     */
    public void windowDeactivated(WindowEvent e) {}

    /**
     * Starts the Java3D renderer again when the window is deiconified
     * @param e The triggering window event
     */
    public void windowDeiconified(WindowEvent e) {
	if (!canvas.isRendererRunning()) canvas.startRenderer();
    }

    /**
     * Stops the Java3D renderer to save CPU cycles when the window is iconified
     * @param e The triggering window event
     */
    public void windowIconified(WindowEvent e) {
	if (canvas.isRendererRunning()) canvas.stopRenderer();
    }

    /**
     * No special action is taken when the window is opened
     * @param e The triggering window event
     */
    public void windowOpened(WindowEvent e) {}

}
