package quest.client.game;

import quest.global.util.FiniteStateMachine;
import quest.client.net.*;
import quest.global.game.*;
import quest.global.net.Message;
import quest.global.net.MessageClientListener;
import quest.global.game.*;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.Observer;
import java.util.Observable;
import javax.swing.JFrame;
import java.awt.Dimension;

/**
 * <P>
 * The controller on the client side of a single Amazing Quest game.
 * By constructing an game controller object, a client is ready to interact with
 * the game server and play a game.</P>
 * <P>
 * The control consists of:
 * <UL>
 * <LI>Constructing the interface
 * <LI>Sending and receiving game moves
 * <LI>Handling joining and leaving spectators
 * </UL>
 * The main parts a game controller consists of are a {@link quest.global.game.GameLogic}
 * object for the logical representation of the game, a  {@link quest.client.net.ClientDaemon}
 * object for the communication with the server and a {@link quest.global.util.FiniteStateMachine}
 * object to keep track of the meta game state (called the 'control state').</P>
 *
 * @author Mathijs den Burger
 */

public class GameController implements MessageClientListener {

    /**
     * Debug variable. If true, the game doens't use the game server
     * at all, but emulates multiple users by managing one game logic
     * object and excuting the moves on that object directly.
     * In the online version, the variable will always be false, of course.
     */
    public static boolean LOCAL_TEST_MODE = false;
    static {
	if (LOCAL_TEST_MODE) {
	    try {
		//codeBase = new URL("file:///C:/MyFiles/Code/AmazingQuest/");
		//codeBase = new URL("file:///home/mms/src/current/");
		//Settings.instance().setCodeBase(new URL("file:///E:/MyFiles/Code/AmazingQuest/"));
		Settings.instance().setCodeBase(new URL("file:///H:/mms/src/current/"));
	    } catch (MalformedURLException e) {
		System.out.println("Wrong manual codebase in GameController");
	    }
	}
    }

    private Game gameFrame;

    /**
     * Control state, indicating the game is initializing. This means you're
     * waiting for other players and spectators to join and the GameLogic object
     * of this game.
     */
    public static final int INITIALIZING = 0;

    /**
     * Control state, indicating the game is running. This means the active players
     * can perform game moves and everyone else can observe them.
     */
    public static final int GAME_RUNNING = 1;

    /**
     * Control state, indicating the game is finished. This means someone has won the
     * game (check the GameLogic object to find out who).
     */
    public static final int GAME_FINISHED = 2;

    /**
     * The total number of control states.
     */
    public static final int NR_STATES = 3;

    private static final int[][] CONTROL_STATE_TRANSITIONS = {
	{ INITIALIZING, GAME_RUNNING },
	{ GAME_RUNNING, GAME_FINISHED }
    };

    private FiniteStateMachine controlState;
    private ClientDaemon postman;
    private String gameName, myName;
    private GameLogic logicalGame;

    /**
     * Constructs a game controller. This initializes the control state to
     * INITIALIZING and logs in to the game server. After the contruction
     * it's waiting for other players and spectators to join and the GameLogic
     * object the server will construct. The actual interface will not be
     * constructed and displayed until all players have joined and the game
     * controller has received the GameLogic object from the server.
     * @param gameName The name of the game this game controller will be controlling
     * @param myName The name of the user (player or spectator) whose game this
     * 				 game controller will be controlling.
     * @param codeBase The code base of the applet. This is needed by some parts
     *				   of the interface to load e.g. texture images.
     * @see quest.global.game.GameLogic
     */
    public GameController(String gameName, String myName) {
    	System.out.println("Initializing \"" + gameName + "\"...");
	if (LOCAL_TEST_MODE) System.out.println("WARNING: you're in LOCAL TEST MODE");
	this.gameName = gameName;
	this.myName = myName;
	logicalGame = null;

	// initialize the control state
	controlState = new FiniteStateMachine(NR_STATES, CONTROL_STATE_TRANSITIONS);
	controlState.start(INITIALIZING);

	if (!LOCAL_TEST_MODE) {
	    // get the message controller
	    postman = ClientDaemon.getInstance();
	    postman.registerProces(GameMessage.class, this);

	    // login to the game server
	    login(gameName);

	    System.out.println(gameName + " initialized, waiting for other players...");
	} else {
	    startLocalGame(gameName, myName);
	}
    }

    private void login(String gameName) {
	GameMessage.C2S_Login msg = new GameMessage.C2S_Login(gameName);
	postman.sendMessage(msg);
    }

    private void startLocalGame(String gameName, String myName) {
    	//String[] userNames = { "Rhythmic", "Dessignator", "Blizzard", "Player 4", "Player 5", "Player 6", "Player 7", "Player 8" };
    	String[] userNames = { "Rhythmic", "Dessignator" }; //, "Blizzard", "Player 4" };
    	String[] spectatorNames = { "Jopie", "Joepie", "Japie", "Jip", "Janneke", "Jaap", "Joris", "Jan", "Job" };
	GameLogic logicalGameOriginal = new GameLogic(userNames, spectatorNames);
	//logicalGame = new GameLogic(userNames, spectatorNames);

	// DEBUG: clone test
	System.out.println("CLONE TEST ENABLED");
	logicalGame = (GameLogic)logicalGameOriginal.clone();
	startGame(logicalGame, myName, new Uplink(this, gameName));
    }

    private void startGame(GameLogic logicalGame, String myName, Uplink uplink) {
    	// build game frame
	gameFrame = new Game(logicalGame, uplink, gameName);
	gameFrame.setTitle("Amazing Quest - " + gameName);
	gameFrame.pack();

	// position the frame in the middle of the screen
	Dimension screenSize = gameFrame.getToolkit().getScreenSize();
	Dimension frameSize = gameFrame.getSize();
	if (frameSize.height > screenSize.height)
	    frameSize.height = screenSize.height;
	if (frameSize.width > screenSize.width)
	    frameSize.width = screenSize.width;

	gameFrame.setLocation((screenSize.width - frameSize.width) / 2,
			      (screenSize.height - frameSize.height) / 2);

	// start the game
	logicalGame.start(logicalGame.getRank(myName));
	controlState.setState(GAME_RUNNING);
	gameFrame.show();

	System.out.println(gameName + " started, have fun.");
    }

    /**
     * Gives the finite state machine that contains the control state of this
     * game controller. These state can be INITIALIZING, RUNNING or FINISHED.
     * @return The finite state machine that contains the control state of this game controller.
     */
    FiniteStateMachine getControlState() {
	return controlState;
    }

    /**
     * Gives the game logic object that contains the logical representation of
     * the game this game controller is controlling.
     * @return The game logic object
     */
    GameLogic getLogicalGame() {
    	return logicalGame;
    }

    /**
     * <P>
     * Handles incoming messages that have arrived at the client deamon.
     * The contents of the message is checked, and only if it's a legal message
     * (that is: it doesn't conflict with the current control state and the
     * contents makes sense) some action is taken. What kind of action depends
     * on the type of the message. See {@link quest.global.game.GameMessage} for
     * more details.</P>
     * <P>
     * All messsages are assumed to be all of the type {@link quest.global.game.GameMessage}.</P>
     * @param message The incoming message
     */
    public void incomingMessage(Message message) {
	GameMessage msg = (GameMessage)message;
	if (legalMessage(msg)) {
	    if (msg instanceof GameMessage.S2C_InitGame) {
		doInitGame((GameMessage.S2C_InitGame)msg);
	    } else if (msg instanceof GameMessage.S2C_Gamemove) {
		doGameMove((GameMessage.S2C_Gamemove)msg);
	    } else if (msg instanceof GameMessage.S2C_Userjoin) {
		doUserJoin((GameMessage.S2C_Userjoin)msg);
	    } else if (msg instanceof GameMessage.S2C_Userleave) {
		doUserLeave((GameMessage.S2C_Userleave)msg);
	    } else if (msg instanceof GameMessage.S2C_Joinself) {
		doJoinSelf((GameMessage.S2C_Joinself)msg);
	    } else if (msg instanceof GameMessage.S2C_Leaveself) {
		doLeaveSelf((GameMessage.S2C_Leaveself)msg);
	    } else if (msg instanceof GameMessage.S2C_AbortGame) {
		doAbortGame((GameMessage.S2C_AbortGame)msg);
	    }
	} else {
	    System.out.println("GameController.incomingMessage: illegal message received: " + msg.toString());
	}
    }

    private boolean legalMessage(GameMessage msg) {
	if (!msg.gamename.equals(gameName)) {
	    System.err.println("Warning: received a stray game message from game " + msg.gamename);
	    return false;
	} else if (msg instanceof GameMessage.S2C_InitGame) {
	    return legalInitGame((GameMessage.S2C_InitGame)msg);
	} else if (msg instanceof GameMessage.S2C_Gamemove) {
	    return legalGameMove((GameMessage.S2C_Gamemove)msg);
	} else if (msg instanceof GameMessage.S2C_Userjoin) {
	    return legalUserJoin((GameMessage.S2C_Userjoin)msg);
	} else if (msg instanceof GameMessage.S2C_Userleave) {
	    return legalUserLeave((GameMessage.S2C_Userleave)msg);
	} else if (msg instanceof GameMessage.S2C_Joinself) {
	    return legalJoinSelf((GameMessage.S2C_Joinself)msg);
	} else if (msg instanceof GameMessage.S2C_Leaveself) {
	    return legalLeaveSelf((GameMessage.S2C_Leaveself)msg);
	} else {
	    return false;
	}
    }

    private boolean legalInitGame(GameMessage.S2C_InitGame msg) {
	System.out.println("Gamelogic heeft " + msg.gamelogic.getNrUsers() + "users:");
	for (int i = 0; i < msg.gamelogic.getNrUsers(); i++) {
	    System.out.println("rank " + i + ": " + msg.gamelogic.getUser(i).getName());
	}
	System.out.println("Ik ben " + myName);
	if (controlState.getCurrentState() != INITIALIZING &&
	    controlState.getCurrentState() != GAME_RUNNING) {
	    return false;
	} else if (msg.gamelogic.getRank(myName) == GameLogic.NOBODY) {
	    System.err.println("GameController.legalInitGame: a spectator cannot start a game.");
	    return false;
	} else if (msg.gamelogic.getRank(myName) == GameLogic.UNKNOWN) {
	    System.err.println("GameController.legalInitGame: an unknown user cannot start a game.");
	    return false;
	} else {
	    return true;
	}
    }

    private boolean legalGameMove(GameMessage.S2C_Gamemove msg) {
	return (controlState.getCurrentState() == GAME_RUNNING);
    }

    private boolean legalUserJoin(GameMessage.S2C_Userjoin msg) {
	return (controlState.getCurrentState() == INITIALIZING ||
		controlState.getCurrentState() == GAME_RUNNING);
    }

    private boolean legalUserLeave(GameMessage.S2C_Userleave msg) {
	return (controlState.getCurrentState() == INITIALIZING ||
		controlState.getCurrentState() == GAME_RUNNING);
    }

    private boolean legalJoinSelf(GameMessage.S2C_Joinself msg) {
	return (controlState.getCurrentState() == INITIALIZING);
    }

    private boolean legalLeaveSelf(GameMessage.S2C_Leaveself msg) {
	return (controlState.getCurrentState() == INITIALIZING ||
		controlState.getCurrentState() == GAME_RUNNING ||
		controlState.getCurrentState() == GAME_FINISHED);
    }

    private void doInitGame(GameMessage.S2C_InitGame msg) {
	logicalGame = msg.gamelogic;
	startGame(logicalGame, myName, new Uplink(postman, gameName));
    }

    private void doGameMove(GameMessage.S2C_Gamemove msg) {
	try {
	    if (logicalGame.doMove(msg.gamemove)) {
	    	System.out.println("DEBUG: game is finished");
	    }
	} catch (IllegalMoveException e) {
	    System.err.println("GameController.doGameMove: user " + msg.gamemove.getInitiatorRank() + " in game \"" + gameName + "\" tried to make an illegal game move.");
	}
    }

    private void doUserJoin(GameMessage.S2C_Userjoin msg) {
	switch(controlState.getCurrentState()) {
	    case INITIALIZING:
		// ik ontvang nu info over een player die "online" is.
		// TODO: laat hiermee een lijstje vollopen ofzo
		System.out.println("Player " + msg.uname + " joins");
		break;
	    case GAME_RUNNING:   // new spectator joins
		logicalGame.getSpectators().add(new GameSpectator(msg.uname));
	}
    }

    private void doUserLeave(GameMessage.S2C_Userleave msg) {
	switch(controlState.getCurrentState()) {
	    case INITIALIZING:        // player that came online now leaves
		// TODO: haal player uit het vollooplijstje
		System.out.println("Player " + msg.uname + " leaves");
		break;
	    case GAME_RUNNING: case GAME_FINISHED:
		switch (logicalGame.getRank(msg.uname)) {
		    case GameLogic.NOBODY:
			// spectator left
			logicalGame.getSpectators().remove(new GameSpectator(msg.uname));
			break;
		    case GameLogic.UNKNOWN:
			// name is unknown
			System.err.println("Warning: unknown user '" + msg.uname + "' left the game");
			break;
		    default:
			// user left
			int rank = logicalGame.getRank(msg.uname);
			if (rank >= 0 && rank < logicalGame.getNrUsers()) {		// it's a known user
			    logicalGame.getUser(rank).setPresent(false);
			} else {												// it's an unknown user!?
			    System.err.println("Warning: an unknown user '" + msg.uname + "' just left the game");
			}
			break;
		}
	}
    }

    private void doJoinSelf(GameMessage.S2C_Joinself msg) {
	// you just logged in, and get a list of players that are already online
	// TODO: maak een lijstje en plemp daar die namen in.
	System.out.print("You logged in. Players online: ");
	for (int i = 0; i < msg.unames.length; i++) {
	    System.out.print(msg.unames[i] + " ");
	}
	System.out.println();
    }

    private void doLeaveSelf(GameMessage.S2C_Leaveself msg) {
	switch(controlState.getCurrentState()) {
	    case INITIALIZING:  // while waiting you choose to quit, so close your "waiting list"
		// TODO: kill venster met vollooplijstje
		System.out.println("You have left the building");
		break;
	    case GAME_RUNNING: case GAME_FINISHED:
		// you have quit the game
		gameFrame.dispose();
		System.out.println("You have quit the game, see you next time.");
		if (LOCAL_TEST_MODE) System.exit(0);
		break;
	}
    }

    private void doAbortGame(GameMessage.S2C_AbortGame msg) {
	// TODO
    }

}
