/**
 * Manages all the data about the games' board.
 *
 * @author Mathijs den Burger
 *
 * @version 0.1
 *
 */

package quest.client.game;

import quest.global.game.*;
import quest.global.util.FiniteStateMachine;
import quest.client.net.ClientDaemon;
import java.net.URL;
import java.util.*;
import java.awt.event.*;
import javax.media.j3d.*;
import javax.vecmath.*;
import javax.swing.event.*;

/**
 * <P>
 * The games' board of Amazing Quest, which includes all the visual objects
 * the game is composed of.</P>
 */

class GamesBoard extends VisibleObject implements PickListener, Observer {

    public static final float LEVEL_HEIGHT = Walls.HEIGHT * 6;

    private GameLogic logicalGame;
    private Uplink uplink;
    private Vector corridorBlocks, arrows, treasures;
    private Player[] players;
    private Switch switchArrows, switchCorridors;
    private boolean upperLevelVisible;

    GamesBoard(GameLogic logicalGame, Uplink uplink) {
		this.logicalGame = logicalGame;
		for (int i = 0; i < logicalGame.getNrUsers(); i++) {
		    this.logicalGame.getUserGameState(i).addObserver(this);
		}
		this.uplink = uplink;
		initCorridorBlocks(logicalGame.getMaze());
		initPlayers(logicalGame);
		initArrows(logicalGame.getMaze());
		registerPickHandler();
		upperLevelVisible = true;

	}

    private void initCorridorBlocks(Maze3D maze) {
		corridorBlocks = new Vector(maze.getXDimension() * maze.getYDimension() * maze.getZDimension());
		for (int x = 0; x < maze.getXDimension(); x++) {
		    for (int y = 0; y < maze.getYDimension(); y++) {
			for (int z = 0; z < maze.getZDimension(); z++) {
			    CorridorBlock corridorBlock = new CorridorBlock(maze.getCorridor(x, y, z));
			    corridorBlocks.add(corridorBlock);
			}
		    }
		}
		corridorBlocks.add(new CorridorBlock(maze.getFreeCorridor(GameLogic.DARK_LEVEL)));
		corridorBlocks.add(new CorridorBlock(maze.getFreeCorridor(GameLogic.LIGHT_LEVEL)));
    }

    private void initPlayers(GameLogic logicalGame) {
		this.players = new Player[logicalGame.getNrUsers()];
		for (int i = 0; i < logicalGame.getNrUsers(); i++) {
		    GameUser logicalUser = logicalGame.getUser(i);
		    players[i] = new Player(logicalUser);
		    logicalUser.addObserver(this);
		}
    }

    private void initArrows(Maze3D maze) {
	arrows = new Vector();
	for (int z = 0; z < maze.getZDimension(); z++) {
	    for (int x = 0; x < maze.getXDimension(); x++) {
		if (!maze.isSteadyX(x)) {
		    arrows.add(new Arrow(new MazePosition3D(x, -1, z), new Orientation2D(180)));				// north edge
		    arrows.add(new Arrow(new MazePosition3D(x, maze.getYDimension(), z), new Orientation2D()));	// south edge
		}
	    }
	    for (int y = 0; y < maze.getYDimension(); y++) {
		if (!maze.isSteadyY(y)) {
		    arrows.add(new Arrow(new MazePosition3D(-1, y, z), new Orientation2D(270)));					// west edge
		    arrows.add(new Arrow(new MazePosition3D(maze.getXDimension(), y, z), new Orientation2D(90)));	// east edge
		}
	    }
	}
	setAllArrowsSelectionState(SelectionState.UNSELECTED);
    }

    private void registerPickHandler() {
	for (Enumeration e = corridorBlocks.elements(); e.hasMoreElements(); ) {
	    CorridorBlock next = (CorridorBlock)e.nextElement();
	    next.addPickListener(this);
	    if (next.getLogicalCorridor().containsTreasure()) {
		next.getTreasure().addPickListener(this);
	    }
	}
	for (Enumeration e = arrows.elements(); e.hasMoreElements(); ) {
	    ((Arrow)e.nextElement()).addPickListener(this);
	}
    }

    public GameLogic getLogicalGame() {
	return logicalGame;
    }

    public static Vector3f getGridTranslation(int x, int y, int z) {
	Vector3f result = new Vector3f();

	result.x = ((x - (GameLogic.X_DIMENSION / 2)) * CorridorBlock.BLOCK_SIZE) - (CorridorBlock.BLOCK_SIZE / 2);
	result.y = ((z - (GameLogic.Z_DIMENSION / 2)) * LEVEL_HEIGHT) + (LEVEL_HEIGHT / 2);
	result.z = ((y - (GameLogic.Y_DIMENSION / 2)) * CorridorBlock.BLOCK_SIZE) - (CorridorBlock.BLOCK_SIZE / 2);

	return result;
    }

    public static Vector3f getFreeTranslation(int z) {
	return getGridTranslation(1000, 1000, z);
    }

    private void setAllMazeCorridorBlocksSelectionState(int state) {
	for (Enumeration e = corridorBlocks.elements(); e.hasMoreElements(); ) {
	    MazeCorridor3D corridor = ((CorridorBlock)e.nextElement()).getLogicalCorridor();
	    if (corridor.getPosition().inMaze()) {
		corridor.getSelectionState().set(state);
	    }
	}
    }

    private void setAllFreeCorridorBlocksSelectionState(int state) {
	for (Enumeration e = corridorBlocks.elements(); e.hasMoreElements(); ) {
	    MazeCorridor3D corridor = ((CorridorBlock)e.nextElement()).getLogicalCorridor();
	    if (corridor.getPosition().isFree()) {
		corridor.getSelectionState().set(state);
	    }
	}
    }

    private void setAllArrowsSelectionState(int state) {
		for (Enumeration e = arrows.elements(); e.hasMoreElements(); ) {
		    ((Arrow)e.nextElement()).getLogicalArrow().getSelectionState().set(state);
		}
    }

    private void setAllReachableCorridorBlocksSelectionState(int state) {
	MazePosition3D myPos = logicalGame.getUser(logicalGame.getActiveUserRank()).getEnvironmentPosition();
	Vector reachableCorridorBlocks = logicalGame.getMaze().getReachableCorridors(myPos.getX(), myPos.getY(), myPos.getZ());
	for (Enumeration e = reachableCorridorBlocks.elements(); e.hasMoreElements(); ) {
	    MazeCorridor3D next = (MazeCorridor3D)e.nextElement();
	    next.getSelectionState().set(state);
	    if (next.containsTreasure() &&
		logicalGame.getUser(logicalGame.getActiveUserRank()).getMission().part(next.getTreasure())) {
		next.getTreasure().getSelectionState().set(state);
	    }
	}
    }

    private void setAllUnreachableCorridorBlocksSelectionState(int state) {
	MazePosition3D myPos = logicalGame.getUser(logicalGame.getActiveUserRank()).getEnvironmentPosition();
	Vector unreachableCorridorBlocks = logicalGame.getMaze().getUnreachableCorridors(myPos.getX(), myPos.getY(), myPos.getZ());
	for (Enumeration e = unreachableCorridorBlocks.elements(); e.hasMoreElements(); ) {
	    MazeCorridor3D next = (MazeCorridor3D)e.nextElement();
	    next.getSelectionState().set(state);
	    if (next.containsTreasure() &&
		logicalGame.getUser(logicalGame.getActiveUserRank()).getMission().part(next.getTreasure())) {
		next.getTreasure().getSelectionState().set(state);
	    }
	}
    }

    public Node buildModel() {
	BranchGroup model = new BranchGroup();
	TransformGroup sceneTrans = new TransformGroup();

	// add corridor blocks
	switchCorridors = new Switch(Switch.CHILD_ALL);
	switchCorridors.setCapability(Switch.ALLOW_SWITCH_WRITE);
	for (Enumeration e = corridorBlocks.elements(); e.hasMoreElements(); ) {
	    switchCorridors.addChild(((CorridorBlock)e.nextElement()).getModel());
	}
	sceneTrans.addChild(switchCorridors);

	// add arrows
	switchArrows = new Switch(Switch.CHILD_ALL);
	switchArrows.setCapability(Switch.ALLOW_SWITCH_WRITE);
	for (Enumeration e = arrows.elements(); e.hasMoreElements(); ) {
	    switchArrows.addChild(((Arrow)e.nextElement()).getModel());
	}
	sceneTrans.addChild(switchArrows);

	// add the player objects
	for (int i = 0; i < logicalGame.getNrUsers(); i++) {
	    sceneTrans.addChild(players[i].getModel());
	}

	// rotate whole scene towards the viewer
	Transform3D sceneRotation = new Transform3D();
	sceneRotation.rotX(Math.PI * 0.5);
	sceneTrans.setTransform(sceneRotation);

	model.addChild(sceneTrans);

	return model;
    }

    public void update(Observable o, Object arg) {
		if (o instanceof FiniteStateMachine) {	// game state changed
		    switch(logicalGame.getMyGameState().getCurrentState()) {
			case GameLogic.OBSERVE:
			    // fall through
			case GameLogic.WALK:
			    setArrowsVisible(false);
			    setAllArrowsSelectionState(SelectionState.UNSELECTABLE);
			    setAllUnreachableCorridorBlocksSelectionState(SelectionState.UNSELECTABLE);
			    setAllReachableCorridorBlocksSelectionState(SelectionState.UNSELECTED);
			    setAllFreeCorridorBlocksSelectionState(SelectionState.UNSELECTED);
			    break;
			case GameLogic.CHANGE_MAZE:
			    setArrowsVisible(true);
			    setAllArrowsSelectionState(SelectionState.UNSELECTED);
			    setAllUnreachableCorridorBlocksSelectionState(SelectionState.UNSELECTABLE);
			    setAllReachableCorridorBlocksSelectionState(SelectionState.UNSELECTED);
			    setAllFreeCorridorBlocksSelectionState(SelectionState.UNSELECTED);
			    break;
			case GameLogic.BLOCK_TO_INSERT_POSITIONED:
			    setArrowsVisible(true);
			    setAllArrowsSelectionState(SelectionState.UNSELECTED);
			    setAllUnreachableCorridorBlocksSelectionState(SelectionState.UNSELECTABLE);
			    setAllReachableCorridorBlocksSelectionState(SelectionState.UNSELECTED);
			    setAllFreeCorridorBlocksSelectionState(SelectionState.SELECTED);
			    break;
			case GameLogic.GAME_WON:
			    setArrowsVisible(false);
			    setAllArrowsSelectionState(SelectionState.UNSELECTABLE);
			    setAllUnreachableCorridorBlocksSelectionState(SelectionState.UNSELECTED);
			    setAllReachableCorridorBlocksSelectionState(SelectionState.UNSELECTED);
			    setAllFreeCorridorBlocksSelectionState(SelectionState.UNSELECTABLE);
			    break;
			case GameLogic.ROTATION_CHANGED:
			    // do nothing
			    break;
			case FiniteStateMachine.NOT_STARTED:
			    // another game state was changed, I'm not started yet. Do nothing.
			    break;
			default:
			    System.err.println("WARNING: unknown game state in GamesBoard.update: " + logicalGame.getMyGameState().getCurrentState());
		    }
		} else if (o instanceof GameUser && arg instanceof Integer) {
		    int userRank = ((GameUser)o).getRank();
		    Integer whatHappened = (Integer)arg;
		    if (whatHappened.equals(GameUser.NEW_ENVIRONMENT_POSITION)) {
				// a game user has a new environment position
				switch(((GameUser)o).lastMoveType()) {
				    case GameUser.WALK:				// it should walk to the new position
					players[userRank].walk(logicalGame.getMaze());
					break;
				    case GameUser.TELEPORT:			// it should teleport to the new position
					players[userRank].teleport();
					break;
				    case GameUser.PICKUP:			// it should pick up the treasure
					players[userRank].pickUp(logicalGame.getMaze());
					break;
				}
		    } else if (whatHappened.equals(GameUser.POSITION_OF_ENVIRONMENT_CHANGED) &&
				players[userRank].getLogicalUser().getEnvironmentPosition().inMaze()) {
				// position of environment changed, so move player along with the moved environment
			    players[userRank].moveAlong();
		    }
		}
    }

	public void pickPerformed(PickEvent e) {
		GameMove move = buildGameMove(e);
		if (move != null) uplink.executeGameMove(move);
    }

    private GameMove buildGameMove(PickEvent e) {
		PickableObject picked = e.getPicked();
		if (e.getModifiers() == e.BUTTON1_MASK) {			// left mouse button click
		    if (picked instanceof Arrow) {
			// position free corridor blocks on picked arrow
			Arrow pickedArrow = (Arrow)picked;
			MazePosition3D pos = pickedArrow.getPosition();
			return new GameMove.PositionFreeCorridorBlock(logicalGame.getActiveUserRank(), pos.getX(), pos.getY());
		} else if (picked instanceof CorridorBlock) {
			return handleCorridorPick(((CorridorBlock)picked).getLogicalCorridor());
		} else if (picked instanceof Teleport) {
			return handleCorridorPick(((Teleport)picked).getEnvironment());
	    } else if (picked instanceof Treasure3D) {
			// walk to treasure and pick it up
			Treasure pickedTreasure = ((Treasure3D)picked).getLogicalTreasure();
			MazeCorridor3D findSpot = pickedTreasure.getFindSpot();
			if (findSpot != null) {
			    MazePosition3D pos = findSpot.getPosition();
			    return new GameMove.PickUp(logicalGame.getActiveUserRank(), pos.getX(), pos.getY(), pos.getZ(), pickedTreasure.getColorId(), pickedTreasure.getSymbolId());
			} else {
			    System.err.println("WARNING: picked treasure hasn't got a find-spot");
			}
		    }
		} else if (e.getModifiers() == e.BUTTON3_MASK) {	// right mouse button click
		    if (picked instanceof CorridorBlock) {
				CorridorBlock corridorBlock = (CorridorBlock)picked;
				switch(logicalGame.getMyGameState().getCurrentState()) {
				    case GameLogic.BLOCK_TO_INSERT_POSITIONED: case GameLogic.ROTATION_CHANGED:
					if (corridorBlock.getLogicalCorridor().getSelectionState().getState() == SelectionState.SELECTED) {
					    MazePosition3D pos = corridorBlock.getLogicalCorridor().getPosition();
					    return new GameMove.RotateCorridorBlock(logicalGame.getActiveUserRank(), -90);
					}
					break;
					// rest: TODO
				}
		    }
		}
		// the pick event doens't trigger a move
		return null;
    }

	private GameMove handleCorridorPick(MazeCorridor3D logicalCorridor) {
		switch(logicalGame.getMyGameState().getCurrentState()) {
		    case GameLogic.BLOCK_TO_INSERT_POSITIONED: case GameLogic.ROTATION_CHANGED:
				// slide selected corridorblock into maze
				MazePosition3D oldPos = logicalCorridor.getPosition();
				MazePosition3D newPos = MazePosition3D.insertInMaze(oldPos);
				return new GameMove.InsertFreeCorridorBlock(logicalGame.getActiveUserRank(), newPos.getX(), newPos.getY());
		    case GameLogic.WALK:
				// new position for player selected
				MazePosition3D walkPos = logicalCorridor.getPosition();
				if (!walkPos.equals(logicalGame.getUser(logicalGame.getActiveUserRank()).getEnvironmentPosition())) {
					return new GameMove.WalkTo(logicalGame.getActiveUserRank(), walkPos.getX(), walkPos.getY(), walkPos.getZ());
				}
		    case GameLogic.GAME_WON:
				// ignore these pick's; since all corridor blocks are made UNSELECTED after the
				// game is won by someone, these pick can happen but should simply be ignored
				break;
		    case GameLogic.OBSERVE:
				// ignore the picks of observers, so do nothing here
				break;
		    case GameLogic.CHANGE_MAZE:
				// ignore the pick that occur while the arrows became visible
				break;
		    default:
				System.err.println("WARNING: pick on CorridorBlock occured in gamestate " + logicalGame.getMyGameState().getCurrentState());
		}
		return null;
	}
		

    public void setUpperLevelVisible(boolean visible) {
	if (upperLevelVisible != visible) {
	    if (visible) {		// show all corridors
		switchCorridors.setWhichChild(Switch.CHILD_ALL);
	    } else {			// show only free and lower corridors
		BitSet allButUpperLevelMask = buildAllButUpperLevelMask();
		switchCorridors.setChildMask(allButUpperLevelMask);
		switchCorridors.setWhichChild(Switch.CHILD_MASK);
	    }
	    upperLevelVisible = visible;
	}
    }

    public boolean getUpperLevelVisible() {
	return upperLevelVisible;
    }

    private boolean lastVisible = true;
    private void setArrowsVisible(boolean visible) {
	if (lastVisible == visible) return;
	if (visible) {
	    BitSet selectableArrowsMask = buildSelectableArrowsMask();
	    switchArrows.setChildMask(selectableArrowsMask);
	    switchArrows.setWhichChild(Switch.CHILD_MASK);
	} else {
	    switchArrows.setWhichChild(Switch.CHILD_NONE);
	}
	lastVisible = visible;
    }

    private BitSet buildAllButUpperLevelMask() {
	BitSet result = new BitSet(corridorBlocks.size());
	for (int i = 0; i < corridorBlocks.size(); i++) {
	    if (((CorridorBlock)corridorBlocks.elementAt(i)).getLogicalCorridor().getPosition().getZ() != GameLogic.LIGHT_LEVEL) {
		result.set(i);
	    } else {
		result.clear(i);
	    }
	}
	return result;
    }

    private BitSet buildSelectableArrowsMask() {
	BitSet result = new BitSet(arrows.size());
	for (int i = 0; i < arrows.size(); i++) {
	    MazePosition3D position = ((Arrow)arrows.elementAt(i)).getPosition();
	    if (position.getX() == logicalGame.getForbiddenInsertX() &&
		position.getY() == logicalGame.getForbiddenInsertY()) {
		result.clear(i);
	    } else {
		result.set(i);
	    }
	}
	return result;
    }

}
