package quest.client.game;

import quest.global.game.*;
import javax.swing.*;
import java.awt.*;

/**
 * Displays a list of events that happened in an Amazing Quest
 * game. There are 4 different
 * kind of events:
 * <UL>
 * <LI>Game moves, like a user walking or picking up a treasure.
 * <LI>User joining or leaving
 * <LI>Spectator joining or leaving
 * <LI>Meta game state changes, like another user takes his turn or
 * the game has ended.
 * </UL>
 * Each kind of event has it's own color, so it can be distinguished
 * easily in the list. Game moves that are performed by
 * yourself have a slightly different color than the moves
 * performed by other players.</P>
 * <P>
 * Events can only be added.</P>
 *
 * @author Mathijs den Burger
 */

class History extends JComponent {

	private static final String DEFAULT_FONT_NAME = "SansSerif";
	private static final int DEFAULT_FONT_STYLE = Font.PLAIN;
	private static final int DEFAULT_FONT_SIZE = 12;
	private static final int META_GAME_STATE_CHANGE_FONT_STYLE = Font.BOLD;
	private static final Color HEADER_FOREGROUND = new Color(0x33, 0x33, 0x44),
		LIST_BACKGROUND = Color.lightGray,
		GAME_MOVE_FOREGROUND = Color.darkGray,
		YOUR_GAME_MOVE_FOREGROUND = Color.black,
		SPECTATOR_PRESENCE_CHANGE_FOREGROUND = Color.blue,
		USER_PRESENCE_CHANGE_FOREGROUND = Color.red,
		META_GAME_STATE_CHANGE_FOREGROUND = Color.green.darker().darker();

	private static final String HEADER_TEXT = " History:";

	private JList list;
	private DefaultListModel history;

	/**
	 * Constructs an empty history list.
	 */
	History() {
		// build the list
		history = new DefaultListModel();
		list = new JList(history);
		list.setCellRenderer(new HistoryListCellRenderer());
		list.setEnabled(false);
		list.setBackground(LIST_BACKGROUND);

		// place the list in a scrollpane
		JScrollPane scrollpane = new JScrollPane(list);
		scrollpane.getViewport().setScrollMode(JViewport.BACKINGSTORE_SCROLL_MODE);
		scrollpane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);

		// build header to put on top of the list
	    JLabel header = new JLabel(HEADER_TEXT, SwingConstants.LEADING);
	    header.setForeground(HEADER_FOREGROUND);

	    // set layout
		GridBagLayout gridbag = new GridBagLayout();
       	GridBagConstraints c = new GridBagConstraints();
		setLayout(gridbag);

       	// add the header
       	c.gridwidth = GridBagConstraints.REMAINDER;
        c.gridheight = 1;
        c.fill = GridBagConstraints.HORIZONTAL;
		c.weightx = 1.0;
		c.weighty = 0.0;
		c.ipadx = 0;
		c.ipady = 0;
        c.anchor = GridBagConstraints.WEST;
		gridbag.setConstraints(header, c);
        add(header);

		// add scrollpane
		c.gridwidth = GridBagConstraints.REMAINDER;
        c.gridheight = GridBagConstraints.REMAINDER;
        c.gridy = GridBagConstraints.RELATIVE;
        c.fill = GridBagConstraints.BOTH;
		c.weightx = 1.0;
		c.weighty = 1.0;
		c.ipadx = 0;
		c.ipady = 0;
		c.anchor = GridBagConstraints.CENTER;
		gridbag.setConstraints(scrollpane, c);
        add(scrollpane);
	}

	private void addString(Color color, String string, Icon image, int fontStyle) {
		history.addElement(new ColoredString(color, string, image, fontStyle));
		if (list.getLastVisibleIndex() >= 0 &&
			list.getLastVisibleIndex() < history.size() - 1) {
			list.ensureIndexIsVisible(history.size() - 1);
		}
	}

	/**
	 * Adds information about the given game move to the list.
	 * You can add your own game moves this way so they have another
	 * color than the moves of other players.
	 * @param move The move to add information about
	 * @param yourName Your name
	 */
	public void addOwnGameMove(GameMove move, String yourName) {
		addGameMove(move, yourName, YOUR_GAME_MOVE_FOREGROUND);
	}

	/**
	 * Adds information about the given game move to the list.
	 * @param move The move to add information about
	 * @param yourName The name of the initiator of the move
	 */
	public void addGameMove(GameMove move, String initiatorName) {
		addGameMove(move, initiatorName, GAME_MOVE_FOREGROUND);
	}

	private void addGameMove(GameMove move, String initiatorName, Color color) {
		String message = initiatorName;
		if (move instanceof GameMove.PickUp) {
			GameMove.PickUp pickUpMove = (GameMove.PickUp)move;
			message += " picked up ";
			addString(color, message, Treasure3D.getIcon(pickUpMove.getSymbolId(), pickUpMove.getColorId()), DEFAULT_FONT_STYLE);
			return;
		}

		if (move instanceof GameMove.WalkTo) {
			message += " walked";
		} else if (move instanceof GameMove.ChangeMaze) {
			message += " is going to change the maze";
		} else if (move instanceof GameMove.PositionFreeCorridorBlock) {
			message += " positioned the free corridors";
		} else if (move instanceof GameMove.RotateCorridorBlock) {
			message += " rotated the free corridors";
		} else if (move instanceof GameMove.InsertFreeCorridorBlock) {
			message += " inserted the free corridors";
		}
		addString(color, message, null, DEFAULT_FONT_STYLE);
	}

	/**
	 * Adds information to the list about the changing presence
	 * of a spectator. This means a spectator just joined or left
	 * the game.
	 * @param spectatorName The name of the spectator
	 * @param joined true if the spectator just joined the game,
	 * false if he just left it
	 */
	public void addSpectatorPresenceChange(String spectatorName, boolean joined) {
		String message = "Spectator " + spectatorName;
		if (joined) {
			message += " joined";
		} else {
			message += " left";
		}
		addString(SPECTATOR_PRESENCE_CHANGE_FOREGROUND, message, null, DEFAULT_FONT_STYLE);
	}

	/**
	 * Adds information to the list about the changing presence
	 * of a user. This means a user just left the game.
	 * @param userName The name of the user that just left the
	 * game
	 */
	public void addUserPresenceChange(String userName) {
		String message = "User " + userName	+ " left";
		addString(USER_PRESENCE_CHANGE_FOREGROUND, message, null, DEFAULT_FONT_STYLE);
	}

	/**
	 * Adds information to the list about a meta game state
	 * change. This can be anything you want to add extra to
	 * the history, but it should be something like another
	 * user taking his turn, the game ended etc.
	 * @param message The message to display
	 */
	public void addMetaGameStateChange(String message) {
		addString(META_GAME_STATE_CHANGE_FOREGROUND, message, null, META_GAME_STATE_CHANGE_FONT_STYLE);
	}

	// inner classes

	private class ColoredString {

		private Color color;
		private String string;
		private Icon icon;
		private int fontStyle;

		public ColoredString(Color color, String string, Icon icon) {
			this(color, string, icon, Font.PLAIN);
		}

		public ColoredString(Color color, String string, Icon icon, int fontStyle) {
			this.color = color;
			this.string = string;
			this.icon = icon;
			this.fontStyle = fontStyle;
		}

		public String getString() {
			return string;
		}

		public Color getColor() {
			return color;
		}

		public Icon getIcon() {
			return icon;
		}

		public int getFontStyle() {
			return fontStyle;
		}

	}

	private class HistoryListCellRenderer extends JLabel implements ListCellRenderer {

		public Component getListCellRendererComponent(JList list, Object value, int index, boolean isSelected, boolean cellHasFocus) {
            if (value instanceof ColoredString) {
            	ColoredString coloredString = (ColoredString)value;
            	setHorizontalTextPosition(SwingConstants.LEADING);
           		setFont(new Font(DEFAULT_FONT_NAME, coloredString.getFontStyle(), DEFAULT_FONT_SIZE));
            	setForeground(coloredString.getColor());
	            setText(coloredString.getString());
            	setIcon(coloredString.getIcon());
            } else {
            	setText("");
            }
            return this;
        }

    }

}
