package quest.client.game;

import quest.global.game.*;
import javax.swing.*;
import java.awt.*;
import java.util.Observer;
import java.util.Observable;

/**
 * <P>
 * Displays information about the {@link quest.global.game.Mission}
 * of a player.
 * This means all the treasures a player still has to find
 * are displayed. The treasures themself are displayed as
 * icons that are retrieved from the {@link quest.client.game.Treasure3D#getIcon(int, int)}
 * method.</P>
 * <P>
 * The mission to display can be changed, including null, in which
 * case nothing is displayed. The current mission is always observed,
 * so the information can be updated in case the contents of the
 * mission changes.</P>
 *
 * @author Mathijs den Burger 
 */

class MissionInfo extends JComponent implements Observer {

	private static final Color HEADER_FOREGROUND = new Color(0x33, 0x33, 0x44),
		TODO_BACKGROUND = Color.lightGray;
	private static final String HEADER_TEXT = " Your mission:";
	private static final String MISSION_ACCOMPLISHED_TEXT = " Mission accomplished!",
		MISSION_FAILED_TEXT = " Mission failed...";
	private static final Font MISSION_ACCOMPLISHED_FONT = new Font("SansSerif", Font.BOLD, 16),
		MISSION_FAILED_FONT = MISSION_ACCOMPLISHED_FONT;

	private Box toDo;
	private Mission mission;

	/**
	 * Constructs an empty mission info component that doens't
	 * contain a mission.
	 */
	MissionInfo() {
		this.mission = null;

		// build the todo-list of treasures to find
		toDo = new Box(BoxLayout.X_AXIS);
		toDo.setBackground(TODO_BACKGROUND);

		// place the todo-list in a scrollpane
		JScrollPane scrollpane = new JScrollPane(toDo);
		//scrollpane.getViewport().setScrollMode(JViewport.BACKINGSTORE_SCROLL_MODE);
		scrollpane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		scrollpane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_NEVER);

		// build header to put on top of the list
	    JLabel header = new JLabel(HEADER_TEXT, SwingConstants.LEADING);
	    header.setForeground(HEADER_FOREGROUND);

	    // set layout
		GridBagLayout gridbag = new GridBagLayout();
       	GridBagConstraints c = new GridBagConstraints();
		setLayout(gridbag);

       	// add the header
       	c.gridwidth = GridBagConstraints.REMAINDER;
        c.gridheight = 1;
        c.fill = GridBagConstraints.HORIZONTAL;
		c.weightx = 1.0;
		c.weighty = 0.0;
		c.ipadx = 0;
		c.ipady = 0;
        c.anchor = GridBagConstraints.WEST;
		gridbag.setConstraints(header, c);
        add(header);

		// add scrollpane
		c.gridwidth = GridBagConstraints.REMAINDER;
        c.gridheight = GridBagConstraints.REMAINDER;
        c.gridy = GridBagConstraints.RELATIVE;
        c.fill = GridBagConstraints.BOTH;
		c.weightx = 1.0;
		c.weighty = 1.0;
		c.ipadx = 0;
		c.ipady = 0;
		c.anchor = GridBagConstraints.CENTER;
		gridbag.setConstraints(scrollpane, c);
        add(scrollpane);

        setBackground(TODO_BACKGROUND);
	}

	/**
	 * Checks if this mission info component contains a mission.
	 * @return true if this mission info component contains a mission
	 * (set by {@link quest.client.game.MissionInfo#setMission(quest.global.game.Mission)}),
	 * false otherwise
	 */
	boolean hasMission() {
		return (mission != null);
	}

	private void updateToDo() {
		toDo.removeAll();
		if (mission != null) {
			if (mission.accomplished()) {
				drawMissionAccomplished();
			} else if (mission.failed()) {
				drawMissionFailed();
			} else {
				drawTreasuresToDo();
			}
		}
		toDo.validate();
		toDo.repaint();
	}

	private void drawMissionAccomplished() {
		JLabel done = new JLabel(MISSION_ACCOMPLISHED_TEXT);
		done.setFont(MISSION_ACCOMPLISHED_FONT);
		done.setForeground(Color.black);
		toDo.add(done);
	}

	private void drawMissionFailed() {
		JLabel failed = new JLabel(MISSION_FAILED_TEXT);
		failed.setFont(MISSION_FAILED_FONT);
		failed.setForeground(Color.black);
		toDo.add(failed);
	}

	private void drawTreasuresToDo() {
		Treasure[] treasures = mission.getMissingTreasures();
		for (int i = 0; i < treasures.length; i++) {
			JLabel label = new JLabel(Treasure3D.getIcon(treasures[i].getSymbolId(), treasures[i].getColorId()));
			label.setBorder(BorderFactory.createEmptyBorder(0, 2, 0, 0));
			toDo.add(label);
		}
	}

	/**
	 * Sets a new mission to display and removes the old one.
	 * @mission The mission to display the treasures a player still
	 * has to find of. This can be null, in which case nothing is displayed
	 */
	public void setMission(Mission mission) {
		if (this.mission != null) {
			this.mission.deleteObserver(this);
		}
		this.mission = mission;
		updateToDo();
		this.mission.addObserver(this);
	}

	/**
	 * This method is called whenever the current {@link quest.global.game.Mission}
	 * changes. Applications shouldn't call this method directly.
	 */
	public void update(Observable o, Object arg) {
		if (o instanceof Mission) {
			updateToDo();
		}
	}

}
