package quest.client.game;

import javax.vecmath.*;

/**
 * <P>
 * Provides a number of functions that can be used to contruct 3D models
 * more conveniently. Most of the methods generate float arrays that can be used
 * as (texture)coordinates, normals, etc. for 'blocks': simple 3D rectangles.
 * There are also some functions for the 
 * translation of arrays of vertices and the cloning of float arrays.</P>
 * <P>
 * This class is a <A HREF="../../../overview-summary.html#patternSingleton">Singleton</A>.</P>
 *
 * @author Mathijs den Burger
 */

class ModelTool {

	/**
	 * The number of vertices that together can define a 3D block.
	 */
	public static final int NR_BLOCK_VERTICES = 24;
	
	/**
	 * The number of vertices that together can define a square plane
	 */
	public static final int NR_PLANE_SQUARE_VERTICES = 4;

	private static ModelTool uniqueInstance;
	private float[] blockNormals,
		planeSquareXZNormals;

	/**
	 * Constructs the single ModelTool object. Use {@link quest.client.game.ModelTool#instance()}
	 * to get a reference to this single instance.
	 */
	protected ModelTool() {
		blockNormals = initBlockNormals();
		planeSquareXZNormals = initPlaneSquareXZNormals();
	};

	/**
	 * Gives the single instance of this class.
	 * @return The single instance
	 */
	public static ModelTool instance() {
		if (uniqueInstance == null) {
			uniqueInstance = new ModelTool();
		}
		return uniqueInstance;
	}

	private float[] initBlockNormals() {
		float[] result = {
			// front face
			0, 0, 1,
			0, 0, 1,
			0, 0, 1,
			0, 0, 1,
			// right face
			1, 0, 0,
			1, 0, 0,
			1, 0, 0,
			1, 0, 0,
			// back face
			0, 0, -1,
			0, 0, -1,
			0, 0, -1,
			0, 0, -1,
			// left face
			-1, 0, 0,
			-1, 0, 0,
			-1, 0, 0,
			-1, 0, 0,
			// bottom face
			0, -1, 0,
			0, -1, 0,
			0, -1, 0,
			0, -1, 0,
			// top face
			0, 1, 0,
			0, 1, 0,
			0, 1, 0,
			0, 1, 0,
		};
		return result;
	}

	private float[] initPlaneSquareXZNormals() {
		float[] result = {
			0, 1, 0,
			0, 1, 0,
			0, 1, 0,
			0, 1, 0,
		};
		return result;
	}

	/**
 	 * Construct an array of floats that can be used as coordinates
 	 * in a QuadArray. The floats then form an block with it's origin
 	 * in the back-bottom-left corner and the given size in the X,
 	 * Y and Z direction.
 	 *
 	 * @param	dx	size in the X direction
 	 * @param	dy	size in the Y direction
 	 * @param	dz	size in the Z direction
 	 *
 	 * @return an array of 24 * 3 floats
 	 */
	public float[] blockCoordinates(float dx, float dy, float dz) {
		float[] result = {
			// front face
			0, 0, dz,
			dx, 0, dz,
			dx, dy, dz,
			0, dy, dz,
			// right face
			dx, 0, 0,
			dx, dy, 0,
			dx, dy, dz,
			dx, 0, dz,
			// back face
			0, 0, 0,
			0, dy, 0,
			dx, dy, 0,
			dx, 0, 0,
			// left face
			0, 0, 0,
			0, 0, dz,
			0, dy, dz,
			0, dy, 0,
			// bottom face
			0, 0, 0,
			dx, 0, 0,
			dx, 0, dz,
			0, 0, dz,
			// top face
			0, dy, 0,
			0, dy, dz,
			dx, dy, dz,
			dx, dy, 0
		};
		return result;
	}

	/**
 	 * Construct an array of floats that can be used as the normals
 	 * of the vertices in a QuadArray created with 'blockCoordinates'.
 	 *
 	 * @return An array of 24 * 3 floats that represent 24 normals.
 	 */
	public float[] blockNormals() {
		return blockNormals;
	}

	/**
 	 * Construct an array of floats that can be used as coordinates
 	 * in a QuadArray. The floats then form an square in the X-Z-plane
 	 * with it's origin in the back-left corner and the given size in the X
 	 * and Z direction.
 	 *
 	 * @param	dx	Size in the X direction
 	 * @param	dz	Size in the Z direction
 	 *
 	 * @return An array of 24 * 3 floats that represent 24 normals.
 	 */
	public float[] planeSquareXZCoordinates(float dx, float dz) {
		float[] result = {
			0, 0, 0,
			0, 0, dz,
			dx, 0, dz,
			dx, 0, 0
		};
		return result;
	}

	/**
	 * Constructs an array of floats that can be used as the normals
	 * of the vertices in a QuadArray created with 
	 * {@link quest.client.game.ModelTool#planeSquareXZCoordinates(float, float)}.
	 *
	 * @return An array of 4 * 3 floats that represent 4 normals.
	 */
	public float[] planeSquareXZNormals() {
		return planeSquareXZNormals;
	}

	/**
	 * Constructs an array of floats that can be used as texture coordinates
	 * for the vertices in a QuadArray created with 
	 * {@link quest.client.game.ModelTool#blockCoordinates(float, float, float)}.
	 * @param frontS The s-coordinate for the texture on the front plane of the block
	 * @param frontT The t-coordinate for the texture on the front plane of the block
	 * @param rightS The s-coordinate for the texture on the right plane of the block
	 * @param rightT The t-coordinate for the texture on the right plane of the block
	 * @param backS The s-coordinate for the texture on the back plane of the block
	 * @param backT The t-coordinate for the texture on the back plane of the block
	 * @param leftS The s-coordinate for the texture on the left plane of the block
	 * @param leftT The t-coordinate for the texture on the left plane of the block
	 * @param bottomS The s-coordinate for the texture on the bottom plane of the block
	 * @param bottomT The t-coordinate for the texture on the bottom plane of the block
	 * @param topS The s-coordinate for the texture on the top plane of the block
	 * @param topT The t-coordinate for the texture on the top plane of the block
	 *
	 * @return An array of 6 * 4 * 2 floats that represent texture coordinates
	 * for all the planes of a block.
	 */
	public float[] blockTexCoords2(float frontS, float frontT, float rightS, float rightT,
		float backS, float backT, float leftS, float leftT,
		float bottomS, float bottomT, float topS, float topT) {
		float[] result = {
			// front face
			0f, 0f,
			frontS, 0f,
			frontS, frontT,
			0f, frontT,
			// right face
			rightS, 0f,
			rightS, rightT,
			0f, rightT,
			0f, 0f,
			// back face
			backS, 0f,
			backS, backT,
			0f, backT,
			0f, 0f,
			// left face
			0f, 0f,
			leftS, 0f,
			leftS, leftT,
			0f, leftT,
			// bottom face
			0f, 0f,
			bottomS, 0f,
			bottomS, bottomT,
			0f, bottomT,
			// top face
			0f, topT,
			0f, 0f,
			topS, 0f,
			topS, topT,
		};
		return result;
	}

	/**
	 * Translates a list of vertices by a vector.
	 *
	 * @param vertices		An array of floats, of which each triplet is
	 *						treated as an vertex in 3D space. The total number
	 *						of floats should be an multitude of 3.
	 * @param vector		The vector by which all the vertices are translated.
	 */
	public static void translate(float[] vertices, Vector3f vector) {
		int offset;
		for (int i = 0; i < vertices.length / 3; i++) {
			offset = i * 3;
			vertices[offset] += vector.x;
			vertices[offset + 1] += vector.y;
			vertices[offset + 2] += vector.z;
		}
	}

	/**
	 * Clones a float array
	 *
	 * @param x	The float array to clone
	 *
	 * @return An exact copy of the given float array
	 */
	public static float[] clone(float[] x) {
		float[] result = new float[x.length];
		for (int i = 0; i < x.length; i++) result[i] = x[i];
		return result;
	}
	
}
