package quest.client.game;

import java.util.Vector;
import java.awt.Component;
import java.net.URL;
import javax.media.j3d.*;
import com.sun.j3d.utils.image.TextureLoader;
import javax.vecmath.*;

/**
 * <P>
 * A flyweight factory for (almost) all Java3D node components. </P>
 * <P>
 * The rendering performance of a scene in Java3D can be increased
 * by sharing as much node components as possible. By using this 
 * flyweight factory this is achieved automatically.</P>
 * <P>
 * Each node component has an associated ID, which is created from the 
 * distinguishing parts of such a component. Whenever the application request
 * a component that doens't exist yet (which can be checked by the various ID's),
 * it is constructed. When a component is requested that does exist already, 
 * a reference to the already created component is returned.</P>
 * <P>
 * This class is a <A HREF="../../../overview-summary.html#patternSingleton">Singleton</A>,
 * and also en example of the <A HREF="../../../overview-summary#patternFacade">Facade design pattern</A>.</P>
 *
 * @author Mathijs den Burger
 */

class NodeComponentFactory {

	private static NodeComponentFactory uniqueInstance = null;
	private Vector nodeComponents, nodeComponentIDs;

	/**
	 * Constructs the single NodeComponentFactory object. Use 
	 * {@link quest.client.game.NodeComponentFactory#instance()}
	 * to get a reference to this single instance.
	 */
	protected NodeComponentFactory() {
		nodeComponents = new Vector();
		nodeComponentIDs = new Vector();
	}

	/**
	 * Gives a reference to the single instance of this object.
	 * @return An reference to that instance
	 */
	public static NodeComponentFactory instance() {
		if (uniqueInstance == null) uniqueInstance = new NodeComponentFactory();
		return uniqueInstance;
	}

	/**
	 * Gives an (possibly shared) appearance with the given material and
	 * rendering attributes and the other components set to null.
	 * @param material The material of the appearance
	 * @param renderingAttributes The rendering attributes of the appearance
	 * @return The (possibly shared) appearance
	 */
	public Appearance getAppearance(Material material, RenderingAttributes renderingAttributes) {
		return getAppearance(material, renderingAttributes, null);
	}

	/**
	 * Gives an (possibly shared) appearance with the given material,
	 * rendering- and polygon attributes and the other components set to null.
	 * @param material The material of the appearance
	 * @param renderingAttributes The rendering attributes of the appearance
	 * @param polygonattributes The polygon attributes of the appearance
	 * @return The (possibly shared) appearance
	 */
	public Appearance getAppearance(Material material, RenderingAttributes renderingAttributes, PolygonAttributes polygonAttributes) {
		AppearanceID appearanceID = new AppearanceID(material, renderingAttributes, polygonAttributes);
		int index = nodeComponents.indexOf(appearanceID);
		if (index < 0) {	// appearance doens't exist
			Appearance appearance = new Appearance();
			appearance.setMaterial(material);
			appearance.setRenderingAttributes(renderingAttributes);
			appearance.setPolygonAttributes(polygonAttributes);
			nodeComponents.add(appearance);
			nodeComponentIDs.add(appearanceID);
			return appearance;
		} else {			// return shared appearance
			return (Appearance)nodeComponents.elementAt(index);
		}
	}

	/**
  	 * Gives (possible shared) rendering attributes with the given parameters.
	 * @param visible The visible parameter of a rendering attributes object
	 * @return The (possibly shared) rendering attributes
  	 */
  	public RenderingAttributes getRenderingAttributes(boolean visible) {
		RenderingAttributesID renderingAttributesID = new RenderingAttributesID(visible);
		int index = nodeComponents.indexOf(renderingAttributesID);
		if (index < 0) {	// appearance doens't exist
			RenderingAttributes renderingAttributes = new RenderingAttributes();
			renderingAttributes.setVisible(visible);
			nodeComponents.add(renderingAttributes);
			nodeComponentIDs.add(renderingAttributesID);
			return renderingAttributes;
		} else {			// return shared rendering attributes
			return (RenderingAttributes)nodeComponents.elementAt(index);
		}
  	}

  	/**
  	 * Gives (possible shared) polygon attributes with the given parameters.
  	 * @param cullFace The cullFace parameter of a polygon attributes object
  	 * @param backFaceNormalFlip The backFaceNormalFlip parameter of a polygon attributes object
  	 * @return The (possibly shared) polygon attributes
  	 */
  	public PolygonAttributes getPolygonAttributes(int cullFace, boolean backFaceNormalFlip) {
		PolygonAttributesID polygonAttributesID = new PolygonAttributesID(cullFace, backFaceNormalFlip);
		int index = nodeComponents.indexOf(polygonAttributesID);
		if (index < 0) {	// appearance doens't exist
			PolygonAttributes polygonAttributes = new PolygonAttributes();
			polygonAttributes.setCullFace(cullFace);
			polygonAttributes.setBackFaceNormalFlip(backFaceNormalFlip);
			nodeComponents.add(polygonAttributes);
			nodeComponentIDs.add(polygonAttributesID);
			return polygonAttributes;
		} else {			// return shared rendering attributes
			return (PolygonAttributes)nodeComponents.elementAt(index);
		}
  	}

  	/**
	 * Gives a (possible shared) material with the given parameters.
	 * @param ambientColor The ambientColor parameter of a material object
	 * @param emissiveColor The emissiveColor parameter of a material object
	 * @param diffuseColor The diffuseColor parameter of a material object
	 * @param specularColor The specularColor parameter of a material object
	 * @param shininess The shininess parameter of a material object
	 * @return The (possibly shared) material
	 */
	public Material getMaterial(Color3f ambientColor, Color3f emissiveColor, Color3f diffuseColor, Color3f specularColor, float shininess) {
		MaterialID materialID = new MaterialID(ambientColor, emissiveColor, diffuseColor, specularColor, shininess);
		int index = nodeComponentIDs.indexOf(materialID);
		if (index < 0) {	// material doesn't exist
			Material material = new Material(ambientColor, emissiveColor, diffuseColor, specularColor, shininess);
			nodeComponents.add(material);
			nodeComponentIDs.add(materialID);
			return material;
		} else {			// return shared material
			return (Material)nodeComponents.elementAt(index);
		}
	}

  	/**
	 * Gives a (possible shared) quad array with the given parameters and the other
	 * parameters set to null.
	 * @param vertexCount The vertexCount parameter of a quad array object
	 * @param vertexFormat The vertexFormat parameter of a quad array object
	 * @param coordinates The coordinates parameter of a quad array object
	 * @param normals The normals parameter of a quad array object
	 * @return The (possibly shared) quad array
	 */
	public QuadArray getQuadArray(int vertexCount, int vertexFormat, float[] coordinates, float[] normals) {
		return getQuadArray(vertexCount, vertexFormat, coordinates, normals, null);
	}

  	/**
	 * Gives a (possible shared) quad array with the given parameters and the other
	 * parameters set to null.
	 * @param vertexCount The vertexCount parameter of a quad array object
	 * @param vertexFormat The vertexFormat parameter of a quad array object
	 * @param coordinates The coordinates parameter of a quad array object
	 * @param normals The normals parameter of a quad array object
	 * @param texCoords The texCoords parameter of a quad array object
	 * @return The (possibly shared) quad array
	 */
	public QuadArray getQuadArray(int vertexCount, int vertexFormat, float[] coordinates, float[] normals, float[] texCoords) {
		QuadArrayID quadArrayID = new QuadArrayID(vertexCount, vertexFormat, coordinates, normals, texCoords);
		int index = nodeComponentIDs.indexOf(quadArrayID);
		if (index < 0) {	// quad array doesn't exist
			QuadArray quadArray = new QuadArray(vertexCount, vertexFormat);
			quadArray.setCoordinates(0, coordinates);
			quadArray.setNormals(0, normals);
			if (texCoords != null) quadArray.setTextureCoordinates(0, 0, texCoords, 0, texCoords.length / 2);
			nodeComponents.add(quadArray);
			nodeComponentIDs.add(quadArrayID);
			return quadArray;
		}					// return shared quad array
		return (QuadArray)nodeComponents.elementAt(index);
	}


	/**
	 * Gives a (possible shared) line strip array with the given parameters
	 * @param vertexCount The vertexCount parameter of a line strip array object
	 * @param vertexFormat The vertexFormat parameter of a line strip array object
	 * @param stripVertexCounts The stripVertexCounts parameter of a line strip array object
	 * @param coordinates The coordinates parameter of a line strip array object
	 * @param colors The colors parameter of a line strip array object
	 * @return The (possibly shared) line strip array
	 */
	public LineStripArray getLineStripArray(int vertexCount, int vertexFormat, int[] stripVertexCounts, float[] coordinates, byte[] colors) {
		LineStripArrayID lineStripArrayID = new LineStripArrayID(vertexCount, vertexFormat, stripVertexCounts, coordinates, colors);
		int index = nodeComponentIDs.indexOf(lineStripArrayID);
		if (index < 0) {	// line strip array doesn't exist
			LineStripArray lineStripArray = new LineStripArray(vertexCount, vertexFormat, stripVertexCounts);
			lineStripArray.setCoordinates(0, coordinates);
			lineStripArray.setColors(0, colors);
			nodeComponents.add(lineStripArray);
			nodeComponentIDs.add(lineStripArrayID);
			return lineStripArray;
		}					// return shared line strip array
		return (LineStripArray)nodeComponents.elementAt(index);
	}

	/**
	 * Gives a (possible shared) texture with the given parameters
	 * @param url The url of the texture
	 * @param format The format parameter of a {@link com.sun.j3d.utils.image.TextureLoader} object
	 * @param flags The flags parameter of a {@link com.sun.j3d.utils.image.TextureLoader} object
	 * @return The (possibly shared) texture. This can be null if the texture couldn't
	 * be loaded
	 */
	public Texture getTexture(URL url, String format, int flags) {
		TextureID textureID = new TextureID(url, format, flags);
		int index = nodeComponentIDs.indexOf(textureID);
		if (index < 0) {	// texture doesn't exist
			TextureLoader tl = new TextureLoader(url, format, null);
			Texture texture = tl.getTexture();
			if (texture == null) {
				System.err.println("WARNING: cannot load texture " + url.toString());
			} else {
				nodeComponents.add(texture);
				nodeComponentIDs.add(textureID);
			}
			return texture;
		}					// return shared texture
		return (Texture)nodeComponents.elementAt(index);
	}

	
	/**
	 * Gives a (possible shared) texture attributes with the given parameters
	 * @param textureMode The textureMode parameter of a texture attributes object
	 * @return The (possibly shared) texture attributes.
	 */
	public TextureAttributes getTextureAttributes(int textureMode) {
		TextureAttributesID textureAttributesID = new TextureAttributesID(textureMode);
		int index = nodeComponentIDs.indexOf(textureAttributesID);
		if (index < 0) {	// texture attributes doesn't exist
			TextureAttributes ta = new TextureAttributes();
			ta.setTextureMode(textureMode);
			return ta;
		}					// return shared texture
		return (TextureAttributes)nodeComponents.elementAt(index);
	}


	// INNER CLASSES

	private class AppearanceID {

		private MaterialID materialID;
		private RenderingAttributesID renderingAttributesID;
		private PolygonAttributesID polygonAttributesID;

		AppearanceID(Material material, RenderingAttributes renderingAttributes, PolygonAttributes polygonAttributes) {
			if (material == null) this.materialID = null;
			else this.materialID = new MaterialID(material);

			if (renderingAttributes == null) this.renderingAttributesID = null;
			else this.renderingAttributesID = new RenderingAttributesID(renderingAttributes);

			if (polygonAttributes == null) this.polygonAttributesID = null;
			else this.polygonAttributesID = new PolygonAttributesID(polygonAttributes);
		}

		public boolean equals(Object rhs) {
			if (rhs instanceof AppearanceID) {
				AppearanceID appearanceID = (AppearanceID)rhs;
				if (this.materialID == null && appearanceID.materialID != null) return false;
				if (this.renderingAttributesID == null && appearanceID.renderingAttributesID != null) return false;
				if (this.polygonAttributesID == null && appearanceID.polygonAttributesID != null) return false;
				return (materialID.equals(appearanceID.materialID) &&
					renderingAttributesID.equals(appearanceID.renderingAttributesID) &&
					polygonAttributesID.equals(appearanceID.polygonAttributesID));
			} else {
				return false;
			}
		}
	}

	private class RenderingAttributesID {

		private boolean visible;

		RenderingAttributesID(RenderingAttributes renderingAttributes) {
			this.visible = renderingAttributes.getVisible();
		}

		RenderingAttributesID(boolean visible) {
			this.visible = visible;
		}

		public boolean equals(Object rhs) {
			if (rhs instanceof RenderingAttributesID) {
				RenderingAttributesID renderingAttributesID = (RenderingAttributesID)rhs;
				return (visible == renderingAttributesID.visible);
			} else {
				return false;
			}
		}
	}

	private class PolygonAttributesID {

		private int cullFace;
		private boolean backFaceNormalFlip;

		PolygonAttributesID(PolygonAttributes polygonAttributes) {
			this.cullFace = cullFace;
			this.backFaceNormalFlip = polygonAttributes.getBackFaceNormalFlip();
		}

		PolygonAttributesID(int cullFace, boolean backFaceNormalFlip) {
			this.cullFace = cullFace;
			this.backFaceNormalFlip = backFaceNormalFlip;
		}

		public boolean equals(Object rhs) {
			if (rhs instanceof PolygonAttributesID) {
				PolygonAttributesID polygonAttributesID = (PolygonAttributesID)rhs;
				return (cullFace == polygonAttributesID.cullFace &&
					backFaceNormalFlip == polygonAttributesID.backFaceNormalFlip);
			} else {
				return false;
			}
		}
	}

	private class MaterialID {

		private Color3f ambientColor, emissiveColor, diffuseColor, specularColor;
		private float shininess;

		MaterialID(Material material) {
			material.getAmbientColor(ambientColor = new Color3f());
			material.getEmissiveColor(emissiveColor = new Color3f());
			material.getDiffuseColor(diffuseColor = new Color3f());
			material.getSpecularColor(specularColor = new Color3f());
			shininess = material.getShininess();
		}

		MaterialID(Color3f ambientColor, Color3f emissiveColor,
			Color3f diffuseColor, Color3f specularColor, float shininess) {
			this.ambientColor = ambientColor;
			this.emissiveColor = emissiveColor;
			this.diffuseColor = diffuseColor;
			this.specularColor = specularColor;
			this.shininess = shininess;
		}

		public boolean equals(Object rhs) {
			if (rhs instanceof MaterialID) {
				MaterialID materialID = (MaterialID)rhs;
				return ((ambientColor.equals(materialID.ambientColor)) &&
						(emissiveColor.equals(materialID.emissiveColor)) &&
						(diffuseColor.equals(materialID.diffuseColor)) &&
						(specularColor.equals(materialID.specularColor)));
			} else {
				return false;
			}
		}
	}

	private class QuadArrayID {

		private int vertexCount, vertexFormat;
		private float[] coordinates, normals, texCoords;

		QuadArrayID(int vertexCount, int vertexFormat, float[] coordinates, float[] normals, float[] texCoords) {
			this.vertexCount = vertexCount;
			this.vertexFormat = vertexFormat;
			this.coordinates = new float[coordinates.length];
			for (int i = 0; i < coordinates.length; i++) {
				this.coordinates[i] = coordinates[i];
			}
			if (normals == null) {
				this.normals = null;
			} else {
				this.normals = new float[normals.length];
				for (int i = 0; i < normals.length; i++) {
					this.normals[i] = normals[i];
				}
			}
			if (texCoords == null) {
				this.texCoords = null;
			} else {
				this.texCoords = new float[texCoords.length];
				for (int i = 0; i < texCoords.length; i++) {
					this.texCoords[i] = texCoords[i];
				}
			}
		}

		public boolean equals(Object rhs) {
			if (rhs instanceof QuadArrayID) {
				QuadArrayID id = (QuadArrayID)rhs;

				// check vertexCount and -format
				if (vertexCount != id.vertexCount || vertexFormat != id.vertexFormat)
					return false;

				// check coordinates
				if (coordinates.length != id.coordinates.length) return false;
				for (int i = 0; i < coordinates.length; i++) {
					if (coordinates[i] != id.coordinates[i]) return false;
				}

				// check normals
				if ((normals == null && id.normals != null) || (normals != null && id.normals == null)) {
					return false;
				} else if (normals != null && id.normals != null) {
					if (normals.length != id.normals.length) {
						return false;
					} else {
						for (int i = 0; i < normals.length; i++) {
							if (normals[i] != id.normals[i]) return false;
						}
					}
				}

				// check texture coordinates
				if ((texCoords == null && id.texCoords != null) || (texCoords != null && id.texCoords == null)) {
					return false;
				} else if (texCoords != null && id.texCoords != null) {
					if (texCoords.length != id.texCoords.length) {
						return false;
					} else {
						for (int i = 0; i < texCoords.length; i++) {
							if (texCoords[i] != id.texCoords[i]) return false;
						}
					}
				}
				return true;
			} else {
				return false;
			}
		}
	}

	private class LineStripArrayID {

		private int vertexCount, vertexFormat;
		private float[] coordinates;
		private int[] stripVertexCounts;
		private byte[] colors;

		LineStripArrayID(int vertexCount, int vertexFormat, int[] stripVertexCounts, float[] coordinates, byte[] colors) {
			this.vertexCount = vertexCount;
			this.vertexFormat = vertexFormat;
			this.stripVertexCounts = new int[stripVertexCounts.length];
			for (int i = 0; i < stripVertexCounts.length; i++) {
				this.stripVertexCounts[i] = stripVertexCounts[i];
			}
			this.coordinates = new float[coordinates.length];
			for (int i = 0; i < coordinates.length; i++) {
				this.coordinates[i] = coordinates[i];
			}
			if (colors == null) {
				this.colors = null;
			} else {
				this.colors = new byte[colors.length];
				for (int i = 0; i < colors.length; i++) {
					this.colors[i] = colors[i];
				}
			}
		}

		public boolean equals(Object rhs) {
			if (rhs instanceof LineStripArrayID) {
				LineStripArrayID id = (LineStripArrayID)rhs;
				// check vertexCount and -format
				if (vertexCount != id.vertexCount || vertexFormat != id.vertexFormat)
					return false;

				// check coordinates
				if (coordinates.length != id.coordinates.length) return false;
				for (int i = 0; i < coordinates.length; i++) {
					if (coordinates[i] != id.coordinates[i]) return false;
				}

				// check strip vertex counts
				if (stripVertexCounts.length != id.stripVertexCounts.length) return false;
				for (int i = 0; i < stripVertexCounts.length; i++) {
					if (stripVertexCounts[i] != id.stripVertexCounts[i]) return false;
				}

				// check colors
				if ((colors == null && id.colors != null) || (colors != null && id.colors == null)) {
					return false;
				} else if (colors != null && id.colors != null) {
					if (colors.length != id.colors.length) {
						return false;
					} else {
						for (int i = 0; i < colors.length; i++) {
							if (colors[i] != id.colors[i]) return false;
						}
					}
				}

				return true;
			} else {
				return false;
			}
		}

	}

	private class TextureID {

		private URL url;
		private String format;
		private int flags;

		TextureID(URL url, String format, int flags) {
			this.url = url;
			this.format = format;
			this.flags = flags;
		}

		public boolean equals(Object rhs) {
			if (rhs instanceof TextureID) {
				TextureID textureID = (TextureID)rhs;
				return (url.equals(textureID.url) &&
					format.equals(textureID.format) &&
					flags == textureID.flags);
			} else {
				return false;
			}
		}

	}

	private class TextureAttributesID {

		private int textureMode;

		TextureAttributesID(int textureMode) {
			this.textureMode = textureMode;
		}

		public boolean equals(Object rhs) {
			if (rhs instanceof TextureAttributesID) {
				TextureAttributesID taID = (TextureAttributesID)rhs;
				return (textureMode == taID.textureMode);
			} else {
				return false;
			}
		}

	}


}	
