package quest.client.game;

import java.awt.*;
import java.awt.event.*;
import java.util.Enumeration;
import javax.media.j3d.*;
import com.sun.j3d.utils.picking.*;

/**
 * <P>
 * A Java3D behavior that enables users to pick {@link quest.client.game.PickableObject}s 
 * in the 3D world with their mouse. The user always picks the nearest object
 * under his mouse pointer. This generates a {@link quest.client.game.PickEvent}
 * that is given to the {@link quest.client.game.PickableObject#dispatchEvent(quest.client.game.PickEvent)}
 * method of the picked object. This distributes it to all the {@link quest.client.game.PickListener}s 
 * that listen to that object.</P>
 * <P>
 * The actual picking takes place at the moment the user releases his left mouse 
 * button. Between pressing and releasing the mouse button, a user may move
 * the mouse pointer a little bit. This is still interpreted as a single 'click'
 * (to provide a smoother collaboration with the 
 * {@link com.sun.j3d.utils.behaviors.mouse.MouseRotate} behavior).</P>
 *
 * @author Mathijs den Burger
 */ 

class PickBehavior extends Behavior {

	/**
	 * The pick mode used for picking object in the 3D scene
	 */
	protected static final int DEFAULT_PICK_MODE = PickTool.GEOMETRY;
	
	/**
	 * The pick tolerance used
	 */
	protected static final float DEFAULT_PICK_TOLERANCE = 0f;
	
	/**
	 * The mouse tolerance used. This is the maximum number of pixels a user may
	 * move his mouse pointer horizontally and/or vertically between a
	 * PRESSED and a RELEASED mouse event and these two event still form a single
	 * mouse click
	 */
	protected static final int DEFAULT_MOUSE_TOLERANCE = 5;	// pixels
	
	private static final int UNKNOWN = -1;
	
	private int lastPressedX, lastPressedY;

	/**
	 * The wakeup criteria for this behavior
	 */
	protected WakeupCriterion[] conditions;
	
	/**
	 * The wakup condition for this behavior
	 */
	protected WakeupCondition wakeupCondition;

	/**
	 * The canvas used for picking
	 */
	protected PickCanvas pickCanvas;

	/**
	 * Constructs a pick behavior for the given canvas that monitors picks on
	 * all pickable objects under the given root within the given bounds.
	 * @param canvas The canvas to pick on
	 * @param monitorRoot The scene graph root to monitor
	 * @param schedulingBounds The bounds for this behavior
	 */
	public PickBehavior(Canvas3D canvas, BranchGroup monitorRoot, Bounds schedulingBounds) {
		super();
    	pickCanvas = new PickCanvas(canvas, monitorRoot);
    	pickCanvas.setMode(DEFAULT_PICK_MODE);
    	setSchedulingBounds(schedulingBounds);
     	pickCanvas.setTolerance(DEFAULT_PICK_TOLERANCE);
     	lastPressedX = lastPressedY = UNKNOWN;
	}

	/**
	 * Initialization function of this behavior. Applications shouldn't call
	 * this method directly
	 */
	public void initialize() {
	    conditions = new WakeupCriterion[2];
	    conditions[0] = new WakeupOnAWTEvent(Event.MOUSE_DOWN);
	    conditions[1] = new WakeupOnAWTEvent(Event.MOUSE_UP);
	    //wakeupCondition = new WakeupOnAWTEvent(MouseEvent.MOUSE_CLICKED);
		//wakeupCondition = new WakeupOnAWTEvent(MouseEvent.MOUSE_PRESSED);
	    //new WakeupOr(conditions);

	    wakeupCondition = new WakeupOr(conditions);
	    wakeupOn(wakeupCondition);
	}
	
	/**
	 * This method is called by the Java3D renderer every time all the
	 * wakup criteria are satisfied. Applications shouldn't call this method
	 * directly.
	 */
 	public void processStimulus (Enumeration criteria) {
 		WakeupCriterion wakeupCriterion;
    	AWTEvent[] events = null;
    	
		while(criteria.hasMoreElements()) {
      		wakeupCriterion = (WakeupCriterion)criteria.nextElement();
      		if (wakeupCriterion instanceof WakeupOnAWTEvent) {
				events = ((WakeupOnAWTEvent)wakeupCriterion).getAWTEvent();
			}
		}

    	if (events[0] instanceof MouseEvent && mouseEventTriggersPick((MouseEvent)events[0])) {
    		handleMouseEvent((MouseEvent)events[0]);
    	}
	    wakeupOn(wakeupCondition);
  	}
  	
  	private boolean mouseEventTriggersPick(MouseEvent e) {
		boolean result = false;

  		if (lastPressedX == UNKNOWN || lastPressedY == UNKNOWN) {
  			if (e.getID() == MouseEvent.MOUSE_PRESSED) {
	  			lastPressedX = e.getX();
  				lastPressedY = e.getY();
  			}
  		} else if (e.getID() == MouseEvent.MOUSE_RELEASED) {
			if (Math.abs(e.getX() - lastPressedX) <= DEFAULT_MOUSE_TOLERANCE &&
  				Math.abs(e.getY() - lastPressedY) <= DEFAULT_MOUSE_TOLERANCE) {
  				result = true;
  			}
	  		lastPressedX = UNKNOWN;
  			lastPressedY = UNKNOWN;
  		}
  		return result;
  	}

	/**
	 * Handles the mouse event that triggers this behavior. This is where the
	 * actual picking takes place
	 * @param e The mouse event that triggered this behavior
	 */
	protected void handleMouseEvent(MouseEvent e) {
		pickCanvas.setShapeLocation(e);
   	 	PickResult result = pickCanvas.pickClosest();

   	 	if (result != null) {	// we picked at least something
   	 		SceneGraphPath sgp = result.getSceneGraphPath();
   	 		for (int i = sgp.nodeCount() - 1; i >= 0; i--) {
   	 			Node pickedObject = sgp.getNode(i);
		   		if (pickedObject.getUserData() instanceof PickableObject) {
					PickableObject cause = (PickableObject)pickedObject.getUserData();
					boolean isPopupTrigger = (e.getModifiers() == e.BUTTON3_MASK);
					PickEvent pickEvent = new PickEvent(e.getComponent(), cause,
						e.getID(), e.getWhen(), e.getModifiers(), e.getX(),
						e.getY(), e.getClickCount(), isPopupTrigger);
					cause.dispatchEvent(pickEvent);
					return;
		   		}
		   	}
		}
    }

}