package quest.client.game;

import java.util.*;
import java.awt.event.MouseEvent;
import javax.media.j3d.*;

/**
 * <P>
 * Subclasses of this object represent objects in the 3D world
 * that can be picked. When the object is picked, all added pick
 * listeners are notified. This can be temporarily disabled
 * by using the method {@link quest.client.game.PickableObject#setReportPicks(boolean)}</P>
 * <P>
 * It is the responsibility of the subclasses to enable picking
 * for specific parts of the 3D model.</P>
 * <P>
 * A PickableObject uses the user data field of the root of the
 * 3D model to store a reference to itself in. This means users
 * shouldn't use the user data field of the root node of their
 * pickable 3D models themself, otherwise picking won't function
 * properly.<?P>
 *
 * @see quest.client.game.PickListener
 * @author Mathijs den Burger
 */

abstract class PickableObject extends VisibleObject {

	private Vector pickListeners;
	private boolean reportPicks;

	/**
	 * Constructs a pickable object without any listeners.
	 */
	public PickableObject() {
		pickListeners = new Vector();
		addModelCapability(Node.ENABLE_PICK_REPORTING);
		setModelUserData(this);
		reportPicks = true;
	}

	/**
	 * Adds a pick listener to this pickable object.
	 * @param listener The pick listener to add. Duplicates aren't
	 * added.
	 */
	public void addPickListener(PickListener listener) {
		if (!pickListeners.contains(listener)) {
			pickListeners.add(listener);
		}
	}

	/**
	 * Removes a pick listener from this pickable object
	 * @param listener The listener to remove. After this call
	 * the given listener isn't contained in the list of pick listeners
	 * of this object.
	 */
	public void removePickListener(PickListener listener) {
		pickListeners.remove(listener);
	}

	/**
	 * Distributes the given pick event across all pick listeners
	 * of this pickable object, if picks are set to be reported.
	 * @param e The pick event to distribute
	 * @see quest.client.game.PickableObject#setReportPicks(boolean)
	 */
	public void dispatchEvent(PickEvent e) {
		if (reportPicks) {
			for (Enumeration enum = pickListeners.elements(); enum.hasMoreElements(); ) {
				((PickListener)enum.nextElement()).pickPerformed(e);
			}
		}
	}
	
	/**
	 * Enables or disables the report of pick events by this pickable object.
	 * When you disable this, it means that the 3D model can still be picked 
	 * but the generated {@link quest.client.game.PickEvent} won't be passed
	 * to all the pick listeners.
	 * @param enabled Use true if you want to enable to report of pick event,
	 * 				or false if you don't
	 */
	public void setReportPicks(boolean enabled) {
		this.reportPicks = enabled;
	}
	
	/**
	 * Builds the 3D model and makes sets the root of it's scene graph to be
	 * pickable. The building of the actual 3D model is done in the method
	 * {@link quest.client.game.PickableObject#buildPickableModel()}.
	 * @return The root of the constructed subtree that forms the 3D model
	 * of this pickable object
	 * @see quest.client.game.VisibleObject#getModel()
	 */
	public final Node buildModel() {
		Node model = buildPickableModel();
		model.setPickable(true);
		return model;
	}
	
	/**
	 * Builds the 3D model of this pickable object.
	 * @return The model
	 * @see quest.client.game.PickableObject#buildModel()
	 */
	abstract Node buildPickableModel();

}