package quest.client.game;

import quest.global.game.SelectionState;
import quest.global.game.Selectable;
import java.util.Vector;
import java.util.Observer;
import java.util.Observable;
import javax.media.j3d.*;

/**
 * <P>
 * A pickable Quest object has an associated {@link quest.global.game.Selectable}
 * object, which usually represent the logical representation of this 3D object
 * in the game. Depending on the {@link quest.global.game.SelectionState} 
 * of this object, picking is enabled or disabled for the 3D model.
 * <OL>
 * <LI><B>UNSELECTABLE</B><BR>
 *     The 3D model won't report picks to pick listeners (which means
 *     effectivily that users can pick it in the 3D world)
 * <LI><B>UNSELECTED</B><BR>
 *     The 3D model reports picks to pick listeners
 * <LI><B>SELECTED</B> and <B>TARGET</B><BR>
 *	   The previous behavior isn't changed
 * </OL>
 * Each pickable Quest object observes the selection state of the associated
 * selectable object to adapt to it's state changes. Subclasses that would like
 * to respond to such a state change too can override the 
 * {@link quest.client.game.PickableQuestObject#updateSelectable(java.util.Observable, java.lang.Object arg)}.</P>
 * <P>
 * The 3D model is automatically made pickable
 *
 * @see quest.client.game.PickListener
 * @author Mathijs den Burger
 */

 abstract class PickableQuestObject extends PickableObject implements Observer {

	private SelectionState selectionState;

	/**
	 * Creates an pickable quest object that is associated with the given
	 * {@link quest.global.game.Selectable} object.
	 * @param s The associated selectable object
	 * @see quest.global.game
	 */
	public PickableQuestObject(Selectable s) {
		super();
		addModelCapability(Node.ALLOW_PICKABLE_WRITE);
		selectionState = s.getSelectionState();
		selectionState.addObserver(this);
	}

	/**
	 * This method is called whenever the selection state of the associated
	 * {@link quest.global.game.Selectable} object changes.
	 * Applications shouldn't call this method directly
	 */
	public final void update(Observable o, Object arg) {
		if (o instanceof SelectionState) {
			handleSelectionStateConsequences(getModel());
		}
		updateSelectable(o, arg);
	}

	/**
	 * Subclasses can override this method to respond to selection state changes 
	 * of the associated selectable object. This method is called after the
	 * pick behavior is changed of the model. The default implementation
	 * does nothing. If other parts than the selection state are observed too,
	 * the calls to the update method will be passed to this method too.
	 * @param Observable The observable object
	 * @param arg An argument passed to the notifyObservers method
	 */
	protected void updateSelectable(Observable o, Object arg) {}

	private void handleSelectionStateConsequences(Node model) {
		switch(selectionState.getState()) {
			case SelectionState.UNSELECTABLE:
				setReportPicks(false);
				break;
			case SelectionState.UNSELECTED:
				setReportPicks(true);
				break;
			case SelectionState.SELECTED:
				// do nothing
				break;
			case SelectionState.TARGET:
				// do nothing
				break;
		}
	}

	/**
	 * Builds the 3D model and initializes the correct behavior pick behavior 
	 * depending on the selection state of the associated selectable object.
	 * The building of the actual 3D model is done in the method
	 * {@link quest.client.game.PickableQuestObject#buildSelectableModel()}.
	 * @return The root of the constructed subtree that forms the 3D model
	 * of this pickable Quest object
	 * @see quest.client.game.PickableObject#buildModel()
	 */
	public final Node buildPickableModel() {
		Node model = buildSelectableModel();
		handleSelectionStateConsequences(model);
		return model;
	}

	/**
	 * Builds the 3D model of this pickable quest object.
	 * @return The model
	 * @see quest.client.game.PickableQuestObject#buildPickableModel()
	 */
	abstract Node buildSelectableModel();

}