package quest.client.game;

import quest.global.game.*;
import java.awt.Color;
import java.awt.event.MouseEvent;
import java.util.*;
import javax.media.j3d.*;
import javax.vecmath.*;
import com.sun.j3d.utils.geometry.*;

/**
 * <P>
 * Represents a 3D player in the maze. A player has a certain color, which 
 * corresponds with his rank. He controls his position in the maze by himself.
 * A player can e.g. be told to walk to a new location, in which case he builds 
 * an animation and executes it. If another animation was still running, it's
 * forced to completion before the other one is started.</P>
 * <P>
 * Multiple player can be in the same corridor block. Because players handle their
 * own position, tehy also have to cope with the problem of positioning themselves
 * on top of each other when multiple users are in the same corridor block.
 * To make this possible, all player's keep references to each other, so they
 * can aler each other's position if that's nessecary. All of this is done 
 * automatically.</P>
 *
 * @author Mathijs den Burger
 */

class Player extends PickableQuestObject {

	/**
	 * The time it takes for a player to walk from one corridor to a neighbour
	 * corridor (in milliseconds)
	 */
	public static final int WALK_TO_NEIGHBOUR_CORRIDOR_TIME = 300;
	
	/**
	 * Time to wait after a user arrives at a corridor block before he picks up 
	 * the treasure there (in milliseconds)
	 */
	public static final int PICKUP_TREASURE_DELAY = 100;				
	
	private static final float PLAYER_MODEL_RADIUS = Walls.CORRIDOR_WIDTH / 3;	// the radius of a player model (in metres)

	// appearance that correspond with the various colors
	private static final Appearance[] appearances = new Appearance[GameLogic.MAX_NR_USERS];
	static {
		NodeComponentFactory ncf = NodeComponentFactory.instance();
		appearances[0] = ncf.getAppearance(ncf.getMaterial(new Color3f(0.4f, 0f, 0f), new Color3f(0.8f, 0f, 0f), new Color3f(0.8f, 0.6f, 0.6f), new Color3f(0f, 0f, 0f), 10), null);
		if (GameLogic.MAX_NR_USERS > 1) appearances[1] = ncf.getAppearance(ncf.getMaterial(new Color3f(0f, 0f, 0.4f), new Color3f(0f, 0f, 0.8f), new Color3f(0.6f, 0.6f, 0.8f), new Color3f(0f, 0f, 0f), 10), null);
		if (GameLogic.MAX_NR_USERS > 2) appearances[2] = ncf.getAppearance(ncf.getMaterial(new Color3f(0f, 0.4f, 0f), new Color3f(0f, 0.7f, 0f), new Color3f(0.6f, 0.7f, 0.6f), new Color3f(0f, 0f, 0f), 10), null);
		if (GameLogic.MAX_NR_USERS > 3) appearances[3] = ncf.getAppearance(ncf.getMaterial(new Color3f(0.4f, 0.4f, 0f), new Color3f(0.8f, 0.8f, 0f), new Color3f(0.8f, 0.8f, 0.6f), new Color3f(0f, 0f, 0f), 10), null);
		if (GameLogic.MAX_NR_USERS > 4) appearances[4] = ncf.getAppearance(ncf.getMaterial(new Color3f(0f, 0.4f, 0.4f), new Color3f(0f, 0.8f, 0.8f), new Color3f(0.6f, 0.8f, 0.8f), new Color3f(0f, 0f, 0f), 10), null);
		if (GameLogic.MAX_NR_USERS > 5) appearances[5] = ncf.getAppearance(ncf.getMaterial(new Color3f(0.4f, 0f, 0.4f), new Color3f(0.8f, 0f, 0.8f), new Color3f(0.8f, 0.6f, 0.8f), new Color3f(0f, 0f, 0f), 10), null);
	}

	private static final Color[] awtColors = {
		Color.red,
		Color.blue,
		Color.green,
		Color.yellow,
		Color.cyan,
		Color.magenta
	};

	/**
	 * The animation branchgroups, one for each player.
	 * The players add their walking, teleporting and moving animation to their
	 * own branchgroup.
	 */
	private static BranchGroup[] animators = new BranchGroup[GameLogic.MAX_NR_USERS];
	static {
		for (int i = 0; i < GameLogic.MAX_NR_USERS; i++) {
			animators[i] = null;
		}
	}

	private static Vector allPlayers = new Vector(GameLogic.MAX_NR_USERS);

	private GameUser logicalUser;
	private TransformGroup positionTg;
	private BranchGroup animationBg;
	private Thread delayedPickUpTreasure;
	private long lastAnimationStartTime, lastAnimationDuration;

	/**
	 * Constructs a player that is associated with the given user
	 * @param logicalUser The user that forms the logical part of this player
	 */
	public Player(GameUser logicalUser) {
		super(logicalUser);
		this.logicalUser = logicalUser;
		allPlayers.add(this);
		delayedPickUpTreasure = null;
		lastAnimationStartTime = System.currentTimeMillis();
		lastAnimationDuration = 0;
	}

	/**
	 * Gives the logical version of this player
	 * @return The logical version of this player
	 */
	public GameUser getLogicalUser() {
		return logicalUser;
	}

	private Appearance getAppearance() {
		return appearances[logicalUser.getRank()];
	}

	/**
	 * Gives the color that is associated with to the given rank.
	 * @param rank The rank of a player
	 * @return The color that is associated with the given rank
	 */
	public static Color getAWTColor(int rank) {
		return awtColors[rank];
	}

	private Transform3D getPositionMatrix() {
		Transform3D result = new Transform3D();
		MazePosition3D pos = logicalUser.getEnvironmentPosition();
		Vector3f translation = GamesBoard.getGridTranslation(pos.getX(), pos.getY(), pos.getZ());
		translation.y += getSharingRank() * 2 * PLAYER_MODEL_RADIUS;
		result.set(translation);
		return result;
	}

	/**
	 * Builds the 3D model of this user.
	 * @return The root node of that model
	 */
	public Node buildSelectableModel() {
		// create player object
		Sphere playerObject = new Sphere(PLAYER_MODEL_RADIUS, Primitive.ENABLE_GEOMETRY_PICKING | Primitive.GENERATE_NORMALS, getAppearance());

		// position player object in the middle of an imaginary corridor block
		Transform3D internalTrans = new Transform3D();
		internalTrans.setTranslation(new Vector3f(CorridorBlock.BLOCK_SIZE / 2, Floor.HEIGHT + (Walls.HEIGHT / 2), CorridorBlock.BLOCK_SIZE / 2));
		TransformGroup internalTg = new TransformGroup(internalTrans);
		internalTg.addChild(playerObject);

		// now move the player object to the right place in the network of corridors
		positionTg = new TransformGroup(getPositionMatrix());
		positionTg.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		positionTg.addChild(internalTg);

		// add animator branch
		animationBg = new BranchGroup();
		animationBg.setCapability(TransformGroup.ALLOW_CHILDREN_EXTEND);
		animationBg.setCapability(TransformGroup.ALLOW_CHILDREN_WRITE);
		positionTg.addChild(animationBg);

		return positionTg;
	}

	/**
	 * Moves a player along with it's (moving) environment. 
	 * This is actually a small walking animation, in which the user walks
	 * at the same time his environment moves. Therefore it seems the user
	 * 'moves along' with the environment. The old and new location in the
	 * {@link quest.global.game.MazePosition3D} object of the logical user is
	 * used to determine how the player should move.
	 */
	void moveAlong() {
		Point3f tempPos = new Point3f();

		// build the standard animation
		PositionPathInterpolator animation = CorridorBlock.buildPositionAnimation(logicalUser.getEnvironmentPosition(), positionTg);

		// check if you share the moving corridor block with other players.
		// if so, adapt the start and destination position to the right sharing rank
		animation.getPosition(0, tempPos);
		tempPos.y += getSharingRank() * 2 * PLAYER_MODEL_RADIUS;
		animation.setPosition(0, tempPos);
		int lastIndex = animation.getArrayLengths() - 1;
		animation.getPosition(lastIndex, tempPos);
		tempPos.y += getSharingRank() * 2 * PLAYER_MODEL_RADIUS;
		animation.setPosition(lastIndex, tempPos);

		// execute the animation
		executeAnimation(animation);

		// store information about the start time and duration of an animation
		storeAnimationInfo(animation);
	}

	/**
	 * Lets the player walk through the maze. The old and new location in the
	 * {@link quest.global.game.MazePosition3D} object of the logical user 
	 * are used as the start and end point of the walk. The route itself
	 * is determined by the method 
	 * {@link quest.global.game.Maze3D#getPathBetween(int, int, int, int, int, int)}.
	 * @param maze The maze in which the user will walk
	 */
	void walk(Maze3D maze) {
		Point3f tempPos = new Point3f();

		// build standard animation
		PositionPathInterpolator animation = buildWalkingAnimation(maze);

		// check if you shared the starting corridor block with other players,
		// and adapt their position if so
		adaptOtherPlayersPosition();

		// check if you have to share the destination corridor block with
		// other players, and adapt your destination position if so
		int lastIndex = animation.getArrayLengths() - 1;
		animation.getPosition(lastIndex, tempPos);
		moveOnTopOfOtherPlayersInSameCorridorBlock(tempPos);
		animation.setPosition(lastIndex, tempPos);

		// execute the animation
		executeAnimation(animation);

		// store information about the start time and duration of an animation
		storeAnimationInfo(animation);
	}

	/**
	 * Lets the player pickup a treasure in the maze. The treasure to pick up
	 * is the one that lies in the new location of the 
	 * {@link quest.global.game.MazePosition3D} object of the logical user.
	 * If the player is currently not in that position, he will walk to that
	 * location first (by executing {@link quest.client.game.Player#walk(quest.global.game.Maze3D)}).
	 * @param maze The maze in which the user will walk and pick up the treasure
	 */
	void pickUp(Maze3D maze) {
		// first, walk to the find spot
		walk(maze);

		// now set a thread that will pickup the treasure after the last
		// animation is finished
		long now = System.currentTimeMillis();
		MazePosition3D destination = logicalUser.getEnvironmentPosition();
		MazeCorridor3D findSpot = maze.getCorridor(destination.getX(), destination.getY(), destination.getZ());

		if (delayedPickUpTreasure != null) {
			// force previous pickup operation to complete
			try {
				delayedPickUpTreasure.interrupt();
			} catch (SecurityException e) {
			    //e.printStackTrace();
			    //System.err.println("WARNING: couldn't force previous pickup operation to complete in Player.pickUp");
			    
			}
		}
		long waitTime = Math.max((lastAnimationStartTime + lastAnimationDuration) - now, 0);
		delayedPickUpTreasure = new Player.DelayedPickUpTreasure(logicalUser, findSpot, waitTime + PICKUP_TREASURE_DELAY);
		delayedPickUpTreasure.start();
	}

	/**
	 * Teleports the player from the old to the new location in the
	 * {@link quest.global.game.MazePosition3D} object of the logical user.
	 * Teleporting means that the user is transferred immediately to his new
	 * location (so you can see it as a infinite quick walk).
	 */
	void teleport() {
		Point3f tempPos = new Point3f();

		// build the standard animation
		PositionPathInterpolator animation = CorridorBlock.buildPositionAnimation(logicalUser.getEnvironmentPosition(), positionTg);

		// check if you share the moving corridor block with other players.
		// if so, adapt the start and destination position to the right sharing rank
		animation.getPosition(0, tempPos);
		tempPos.y += getSharingRank() * 2 * PLAYER_MODEL_RADIUS;
		animation.setPosition(0, tempPos);
		int lastIndex = animation.getArrayLengths() - 1;
		animation.getPosition(lastIndex, tempPos);
		tempPos.y += getSharingRank() * 2 * PLAYER_MODEL_RADIUS;
		animation.setPosition(lastIndex, tempPos);

		// execute the animation
		executeAnimation(animation);

		// store information about the start time and duration of an animation
		storeAnimationInfo(animation);

		// to make sure other teleporting players end in the right ranking position,
		// adapt their animations if they started earlier with a higher rank
		adaptOtherPlayersMoveAlongAnimation();
	}

	private void storeAnimationInfo(PositionPathInterpolator animation) {
		Alpha alpha = animation.getAlpha();
		lastAnimationStartTime = alpha.getStartTime();
		lastAnimationDuration = getTotalAlphaDuration(alpha);
	}

	private long getTotalAlphaDuration(Alpha alpha) {
		return (alpha.getPhaseDelayDuration() +
			 alpha.getIncreasingAlphaDuration() +
			 alpha.getIncreasingAlphaRampDuration() +
			 alpha.getAlphaAtOneDuration() +
			 alpha.getDecreasingAlphaDuration() +
			 alpha.getDecreasingAlphaRampDuration() +
			 alpha.getAlphaAtZeroDuration());
	}

	private void adaptOtherPlayersPosition() {
		for (Enumeration e = allPlayers.elements(); e.hasMoreElements(); ) {
			Player otherPlayer = (Player)e.nextElement();
			if (otherPlayer.logicalUser != logicalUser &&
				otherPlayer.logicalUser.getEnvironmentPosition() == logicalUser.getOldEnvironmentPosition()) {
				// you shared the corridor block with this other player, so
				// move the other player to it's right sharing rank position
				otherPlayer.positionTg.setTransform(otherPlayer.getPositionMatrix());
			}
		}
	}

	private void adaptOtherPlayersMoveAlongAnimation() {
		boolean rankIsHigher = false;
		for (Enumeration e = allPlayers.elements(); e.hasMoreElements(); ) {
			Player otherPlayer = (Player)e.nextElement();
			if (otherPlayer.logicalUser == logicalUser) {
				rankIsHigher = true;
			} else if (otherPlayer.logicalUser.getEnvironmentPosition() == logicalUser.getEnvironmentPosition() &&
				rankIsHigher) {
				// you share the corridor block with this other player, his animation
				// started earlier than yours and his rank is higher. This means he
				// didn't know about you when he build his animation, and therefore
				// moves to the wrong sharing rank position. Solution: let him build
				// and execute his move along animation again
				otherPlayer.moveAlong();
			}
		}
	}

	private int getSharingRank() {
		int result = 0;
		for (Enumeration e = allPlayers.elements(); e.hasMoreElements(); ) {
			GameUser otherPlayer = (GameUser)(((Player)e.nextElement()).getLogicalUser());
			if (otherPlayer == logicalUser) {
				return result;
			} else if (otherPlayer.getEnvironmentPosition() == logicalUser.getEnvironmentPosition()) {
				result += 1;
			}
		}
		// if you reach this, this player wasn't found in the vector of all players.
		// this should be impossible :)
		System.err.println("Warning: something strange happened in Player.getSharingRank");
		return result;
	}

	private void moveOnTopOfOtherPlayersInSameCorridorBlock(Tuple3f position) {
		int nrOtherPlayersInSameCorridorBlock = 0;
		for (Enumeration e = allPlayers.elements(); e.hasMoreElements(); ) {
			GameUser otherPlayer = (GameUser)(((Player)e.nextElement()).getLogicalUser());
			if (otherPlayer != logicalUser &&
				otherPlayer.getEnvironmentPosition() == logicalUser.getEnvironmentPosition()) {
				// you share the corridor block with this other player, so count it as the number of player under you
				nrOtherPlayersInSameCorridorBlock += 1;
			}
		}
		position.y += nrOtherPlayersInSameCorridorBlock * 2 * PLAYER_MODEL_RADIUS;
	}

	private void executeAnimation(Behavior animation) {
		int myRank = logicalUser.getRank();
		if (animators[myRank] != null)
			animators[myRank].detach();
		animators[myRank] = new BranchGroup();
		animators[myRank].setCapability(BranchGroup.ALLOW_DETACH);
		animators[myRank].addChild(animation);
		animators[myRank].compile();
		animationBg.addChild(animators[myRank]);
	}

	private PositionPathInterpolator buildWalkingAnimation(Maze3D maze) {
		// get maze coordinates of path
		MazePosition3D start = logicalUser.getOldEnvironmentPosition(),
			end = logicalUser.getEnvironmentPosition();
		Vector path = maze.getPathBetween(start.getX(), start.getY(), start.getZ(),
			end.getX(), end.getY(), end.getZ());

		// build positions
		Point3f[] positions = new Point3f[path.size()];
		for (int i = 0; i < path.size(); i++) {
			MazePosition3D pos = (MazePosition3D)path.elementAt(i);
			positions[i] = new Point3f(GamesBoard.getGridTranslation(pos.getX(), pos.getY(), pos.getZ()));
		}

		// build knots
		float[] knots = new float[path.size()];
		for (int i = 0; i < path.size(); i++) {
			knots[i] = i * (1.0f / (path.size() - 1));
		}

		// alpha value
		Alpha alpha = new Alpha(1, (path.size() - 1) * WALK_TO_NEIGHBOUR_CORRIDOR_TIME);
		alpha.setStartTime(System.currentTimeMillis());

		// build behavior
		PositionPathInterpolator walk = new PositionPathInterpolator(
			alpha, positionTg, new Transform3D(), knots, positions);
		walk.setSchedulingBounds(Game.EVERYTHING);

		return walk;
	}

	// INNER CLASSES

	private static class DelayedPickUpTreasure extends Thread {

		private GameUser user;
		private MazeCorridor3D findSpot;
		private long delay;

		DelayedPickUpTreasure(GameUser user, MazeCorridor3D findSpot, long delay) {
			this.user = user;
			this.findSpot = findSpot;
			this.delay = delay;
		}

		public void run() {
			try {
				sleep(delay);
			} catch (InterruptedException e) {
				// do nothing
			}
			user.pickupTreasure(findSpot, true);
		}

	    public MazeCorridor3D getFindSpot() {
		return findSpot;
	    }

	    public GameUser getUser() {
		return user;
	    }

	}

}
