package quest.client.game;

import quest.global.game.GameUser;
import quest.global.game.GameLogic;
import quest.global.game.Mission;
import java.awt.*;
import java.util.*;
import javax.swing.*;
import javax.swing.table.*;
import javax.swing.border.*;

/**
 * <P>
 * Displays a list of player's names with a colored icon in front
 * of each name. There is always one name 'active', which is
 * displayed slightly different than the other names. This active player
 * rank starts at 0 (zero). The number of treasures each player has still 
 * to find is also displayed.</P>
 * <P>
 * The names are always displayed in the same
 * order as they had in the Vector when the PlayerList is initialized.
 * These positions are taken as the ranks of the players.
 * The color of the icons depends directly on the colors the method
 * {@link quest.client.game.Player#getAWTColor(int)} returns for
 * the various ranks.</P>
 * <P>
 * The list observes the {@link quest.global.game.Mission} objects
 * of all the players in the list, so it can always display the
 * actual number of treasures a player has to find.</P>
 *
 * @author Mathijs den Burger
 */

class PlayerList extends JComponent implements Observer {

	private static final EmptyBorder PLAYER_BORDER = new EmptyBorder(0, 1, 0, 0),
		PROGRESS_BORDER = new EmptyBorder(2, 5, 1, 0);

	private static final int PROGRESS_BAR_WIDTH = 30;	// pixels

	private static final Font PASSIVE_PLAYER_FONT = new Font("SansSerif", Font.PLAIN, 12),
		ACTIVE_PLAYER_FONT = new Font("SansSerif", Font.BOLD, 12),
		LEFT_PLAYER_FONT = new Font("SansSerif", Font.ITALIC, 12);

	private static final Color DEFAULT_BACKGROUND = Color.lightGray,
		PASSIVE_PLAYER_FOREGROUND = Color.darkGray,
		ACTIVE_PLAYER_FOREGROUND = Color.black,
		LEFT_PLAYER_FOREGROUND = Color.darkGray,
		HEADER_FOREGROUND = new Color(0x33, 0x33, 0x44);

	private static final ColorIcon[] RANK_ICONS = new ColorIcon[GameLogic.MAX_NR_USERS];
	static {
		for (int i = 0; i < GameLogic.MAX_NR_USERS; i++) {
			RANK_ICONS[i] = new ColorIcon(Player.getAWTColor(i), 13, 13);
		}
	}

	private static final String HEADER_TEXT = " Players:";

  	private JLabel header;
  	private JList userList;
	private int activePlayerRank;

    /**
     * Constructs a player list that contains the given names.
     * @param players The vector of Player objects to display
     * @param activePlayerRank The rank of the player that's currently
     * active (whatever that means). His name is displayed slightly
     * different than the other names.
     */
     PlayerList(Vector players, int activePlayerRank) {
		this.activePlayerRank = activePlayerRank;

		// build the user list
		userList = new JList(players);
		userList.setCellRenderer(new GameUserListCellRenderer());
		userList.setEnabled(false);
		userList.setBackground(DEFAULT_BACKGROUND);
		userList.setVisibleRowCount(players.size());

		// register this as an observer of the missions
		for (int i = 0; i < players.size(); i++) {
			((GameUser)players.elementAt(i)).getMission().addObserver(this);
		}

		// build header to put on top of the lists
	    header = new JLabel(HEADER_TEXT, SwingConstants.LEADING);
	    header.setForeground(HEADER_FOREGROUND);

		// place the lists together in a scrollpane
		JScrollPane scrollpane = new JScrollPane(userList);

		// set layout
	    GridBagLayout gridbag = new GridBagLayout();
       	GridBagConstraints c = new GridBagConstraints();
		setLayout(gridbag);

       	// add the header
	    c.gridwidth = GridBagConstraints.REMAINDER;
        c.gridheight = 1;
        c.fill = GridBagConstraints.HORIZONTAL;
		c.weightx = 1.0;
		c.weighty = 0.0;
		c.ipadx = 0;
		c.ipady = 0;
        c.anchor = GridBagConstraints.WEST;
		gridbag.setConstraints(header, c);
        add(header);

		// add the list
		c.gridwidth = GridBagConstraints.REMAINDER;
        c.gridheight = GridBagConstraints.REMAINDER;
        c.gridy = GridBagConstraints.RELATIVE;
        c.fill = GridBagConstraints.BOTH;
		c.weightx = 1.0;
		c.weighty = 1.0;
		c.ipadx = 0;
		c.ipady = 0;
		c.anchor = GridBagConstraints.CENTER;
		gridbag.setConstraints(scrollpane, c);
        add(scrollpane);

        // set maximum size
   		setMaximumSize(new Dimension(Integer.MAX_VALUE, header.getHeight() + userList.getCellRenderer().getListCellRendererComponent(userList, players.firstElement(), 0, false, false).getHeight() * players.size()));

    	setBackground(DEFAULT_BACKGROUND);
    }

	/**
	 * Changes the active rank. This causes another player to
	 * be displayed as the 'active' one. Ranks start at 0 (zero).
	 * @param rank The new rank
	 */
    public void setActivePlayerRank(int rank) {
		if (rank != activePlayerRank) {
		    activePlayerRank = rank;
		    repaint();
		}
    }
    
	/**
	 * The method that's called whenever the Mission of a player
	 * in the list changes. Applications shouldn't call this
	 * method directly.
	 */
	public void update(Observable o, Object arg) {
		if (o instanceof Mission) {
			userList.repaint();
		}
	}

	// INNER CLASSES


	private class GameUserListCellRenderer extends DefaultListCellRenderer implements ListCellRenderer {

		GameUserListCellRenderer() {
			setOpaque(true);
		}

		public Component getListCellRendererComponent(JList list, Object value, int index, boolean isSelected, boolean cellHasFocus) {
	        if (value instanceof GameUser) {
	        	GameUser player = (GameUser)value;
            	
            	// set the label
               	setBorder(PLAYER_BORDER);
				setBackground(DEFAULT_BACKGROUND);
				if (player.isPresent()) {
					int toDo = player.getMission().getNrMissingTreasures();
					if (toDo > 0) {		// print number of treasures to go
						setText(player.getName() + " | " + player.getMission().getNrMissingTreasures() + " |");
					} else {			// print "wins"
						setText(player.getName() + " wins");
					}
				} else {				// print that the player has left the game
					setText(player.getName() + " left");
				}
				
				// set the icon
				setIcon(RANK_ICONS[player.getRank()]);
				
				// set the font
				if (player.getRank() == activePlayerRank) {   // render the active player
					setFont(ACTIVE_PLAYER_FONT);
					setForeground(ACTIVE_PLAYER_FOREGROUND);
				} else if (!player.isPresent()) {			  // render a player that left
					setFont(LEFT_PLAYER_FONT);
					setForeground(LEFT_PLAYER_FOREGROUND);										
				} else {									  // render a passive player
					setFont(PASSIVE_PLAYER_FONT);
					setForeground(PASSIVE_PLAYER_FOREGROUND);
				}
			}
			return this;
        }
	}

}


