package quest.client.game;

import quest.global.game.Treasure;
import quest.global.util.ExtraMath;
import quest.client.ClientApplet;
import java.net.URL;
import java.net.MalformedURLException;
import java.awt.*;
import java.awt.image.*;
import java.awt.geom.AffineTransform;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.media.j3d.*;
import javax.vecmath.*;
import com.sun.j3d.utils.image.TextureLoader;
import com.sun.j3d.utils.picking.PickTool;

/**
 * <P>
 * Represents a 3D treasure in the maze. Each treasure has a associated
 * logical {@link quest.global.game.Treasure} with a colorId and a symbolId.
 * These are translated to a certain color and a texture, respectively.
 * The treasures itself are small black boxes with their symbol in their color
 * on the sides.</P>
 *
 * @author Mathijs den Burger
 */

class Treasure3D extends PickableQuestObject {

 	/**
 	 * The size (width) of a treasure (in metres).
 	 */
    private static final float SIZE = Walls.CORRIDOR_WIDTH * 0.666f;
    
    private static final float HEIGHT = 1f;

    /*
     * Textures that represent the symbols
     */
    private static final String[] SYMBOL_TEXTURE_URLS = new String[Treasure.MAX_NR_SYMBOLS];
    static {
    	for (int i = 0; i < Treasure.MAX_NR_SYMBOLS; i++) {
	    if (GameController.LOCAL_TEST_MODE) {
		SYMBOL_TEXTURE_URLS[i] = "quest/client/game/images/symbol" + i + ".png";
	    } else {
		SYMBOL_TEXTURE_URLS[i] = "quest/client/game/images/symbol" + i + ".png";
	    }
    	}
    }
    private static final String SYMBOL_TEXTURE_FORMAT = "RGBA";
    private static final int SYMBOL_TEXTURE_FLAGS = TextureLoader.GENERATE_MIPMAP;

    /*
     * Icons for all types of treasures
     */
    private static Icon[][] ICONS = new Icon[Treasure.MAX_NR_SYMBOLS][Treasure.MAX_NR_COLORS];
    static {
    	for (int i = 0; i < Treasure.MAX_NR_SYMBOLS; i++) {
	    for (int j = 0; j < Treasure.MAX_NR_COLORS; j++) {
		ICONS[i][j] = null;
	    }
    	}
    }
    private static final int ICON_WIDTH = 15,	// pixels
	ICON_HEIGHT = 15;	// pixels
    private static final Color TEXTURE_SYMBOL_COLOR = Color.black,	// color of the symbol-part in the texture image
	TEXTURE_BACKGROUND_COLOR = Color.white;						// color of the background in the texture image

    // scaling factors of the texture that covers the sides of a treasure
    private static final float FRONT_S = 3f, FRONT_T = 1f,
	RIGHT_S = 3f, RIGHT_T = 1f,
	BACK_S = 3f, BACK_T = 1f,
	LEFT_S = 3f, LEFT_T = 1f,
	BOTTOM_S = 1f, BOTTOM_T = 1f,
	TOP_S = 1f, TOP_T = 1f;


    /*
     * Materials that represent the colors
     */
    private static final Material[] materials = new Material[Treasure.MAX_NR_COLORS];
    static {
	NodeComponentFactory ncf = NodeComponentFactory.instance();
		if (Treasure.MAX_NR_COLORS > 0) materials[0] = ncf.getMaterial(new Color3f(0.4f, 0f, 0f), new Color3f(0.8f, 0f, 0f), new Color3f(0.8f, 0.6f, 0.6f), new Color3f(0f, 0f, 0f), 10);
		if (Treasure.MAX_NR_COLORS > 1) materials[1] = ncf.getMaterial(new Color3f(0f, 0f, 0.4f), new Color3f(0f, 0f, 0.8f), new Color3f(0.6f, 0.6f, 0.8f), new Color3f(0f, 0f, 0f), 10);
		if (Treasure.MAX_NR_COLORS > 2) materials[2] = ncf.getMaterial(new Color3f(0f, 0.5f, 0f), new Color3f(0f, 0.8f, 0f), new Color3f(0.6f, 0.7f, 0.6f), new Color3f(0f, 0f, 0f), 10);
		if (Treasure.MAX_NR_COLORS > 3) materials[3] = ncf.getMaterial(new Color3f(0.4f, 0.4f, 0f), new Color3f(0.8f, 0.8f, 0f), new Color3f(0.8f, 0.8f, 0.6f), new Color3f(0f, 0f, 0f), 10);
    }
    private static final Color[] colors = {
		new Color(1f, 0.3f, 0.3f),   // red
		new Color(0.5f, 0.5f, 1f),   // blue
		new Color(0f, 1f, 0f), // green
		Color.yellow   // yellow
    };

    private Treasure logicalTreasure;
    private Shape3D model;

	/**
	 * Construct a treasure that is associated with the given logical treasure.
	 * @param logicalTreasure The logical treasure that defines what this 3D one
	 * will look like
	 */
    Treasure3D(Treasure logicalTreasure) {
		super(logicalTreasure);
		this.logicalTreasure = logicalTreasure;
    }

	/**
	 * Gives the logical treasure
	 * @return The logical treasure
	 */
    public Treasure getLogicalTreasure() {
    	return logicalTreasure;
    }

	/**
	 * Gives an icon that represent a treasure with the given symbol and color.
	 * @param symbolId The symbol ID of the treasure
	 * @param colorId The color ID of the treasure
	 * @return An icon that represent a treasure with the given symbol and color.
	 */
    public static Icon getIcon(int symbolId, int colorId) {
    	symbolId = ExtraMath.clamp(symbolId, 0, Treasure.MAX_NR_SYMBOLS);
    	colorId = ExtraMath.clamp(colorId, 0, Treasure.MAX_NR_COLORS);
    	if (ICONS[symbolId][colorId] == null) {
	    // load icon
	    try {
		URL iconURL = new URL(Settings.instance().getCodeBase().toString() + SYMBOL_TEXTURE_URLS[symbolId]);
		ICONS[symbolId][colorId] = createIcon(iconURL, colors[colorId]);
	    } catch (MalformedURLException e) {
		System.err.println("WARNING: unable to build an URL of codebase " + Settings.instance().getCodeBase() + " and relative URL " + SYMBOL_TEXTURE_URLS[symbolId]);
	    }
	}
	return ICONS[symbolId][colorId];
    }

    private static Icon createIcon(URL url, Color color) {
	// get texture image
	TextureLoader tl = new TextureLoader(url, null); 
	BufferedImage image = tl.getImage().getImage();
		
	// color image
	replaceColor(image, TEXTURE_SYMBOL_COLOR, color);
	replaceColor(image, TEXTURE_BACKGROUND_COLOR, Color.black);

	// scale the image
	AffineTransform at = new AffineTransform();
	at.scale((float)ICON_WIDTH / image.getWidth(), (float)ICON_HEIGHT / image.getHeight());
	AffineTransformOp ato = new AffineTransformOp(at, AffineTransformOp.TYPE_NEAREST_NEIGHBOR);
	
	// return icon
	return new ImageIcon(ato.filter(image, null));
    }
	
	
    private static void replaceColor(BufferedImage bi, Color src, Color dst) {
	DataBuffer db1 = bi.getRaster().getDataBuffer();

	int fromcolor = src.getRGB();
	int dstcolor = dst.getRGB();

	for (int i = 0; i < db1.getSize(); i++) {
	    int e = db1.getElem(i);
	    if (e == fromcolor) db1.setElem(i, dstcolor);
	}
    }

	/**
	 * Builds the 3D model of this treasure.
	 * @return The root node of the model
	 */
    public Node buildSelectableModel() {
    	NodeComponentFactory ncf = NodeComponentFactory.instance();
	ModelTool mt = ModelTool.instance();
	int vertexFormat = QuadArray.COORDINATES | QuadArray.NORMALS | QuadArray.TEXTURE_COORDINATE_2;
	model = new Shape3D();

	// build geometry
	float[] treasureCoords = mt.blockCoordinates(SIZE, HEIGHT, SIZE);
	mt.translate(treasureCoords, new Vector3f((CorridorBlock.BLOCK_SIZE - SIZE) / 2f, Floor.HEIGHT, (CorridorBlock.BLOCK_SIZE - SIZE) / 2f));

	// add geometry
	model.addGeometry(ncf.getQuadArray(mt.NR_BLOCK_VERTICES,
					   vertexFormat, treasureCoords, mt.blockNormals(),
					   mt.blockTexCoords2(FRONT_S, FRONT_T, RIGHT_S, RIGHT_T, BACK_S, BACK_T, LEFT_S, LEFT_T, BOTTOM_S, BOTTOM_T, TOP_S, TOP_T)));

	// build appearance
	Appearance appearance = new Appearance();
	appearance.setMaterial(materials[logicalTreasure.getColorId()]);
	try {
	    URL textureURL = new URL(Settings.instance().getCodeBase().toString() + SYMBOL_TEXTURE_URLS[logicalTreasure.getSymbolId()]);	 	appearance.setTexture(ncf.getTexture(textureURL, SYMBOL_TEXTURE_FORMAT, SYMBOL_TEXTURE_FLAGS));
	    appearance.setTexture(ncf.getTexture(textureURL, SYMBOL_TEXTURE_FORMAT, SYMBOL_TEXTURE_FLAGS));
	    appearance.setTextureAttributes(ncf.getTextureAttributes(TextureAttributes.BLEND));
	} catch (MalformedURLException e) {
	    System.err.println("WARNING: unable to build an URL of codebase " + Settings.instance().getCodeBase() + " and relative URL " + SYMBOL_TEXTURE_URLS[logicalTreasure.getSymbolId()]);
	}
	model.setAppearance(appearance);

	// enable picking
	PickTool.setCapabilities(model, PickTool.INTERSECT_TEST);

	return model;
    }

}
