package quest.client.game;

import quest.client.net.ClientDaemon;
import quest.global.game.*;

/**
 * The connection with the game to the outside world.
 * Over this connection are the game moves sent that the players of the game
 * create by using the interface. You can create two types of uplinks:
 * <UL>
 * <LI>An online uplink that will communicate with the game server
 * <LI>A local uplink that will route the game moves directly back to the local
 *     {@link quest.global.game.GameLogic} object. This can be used to test
 *     the game offline.
 * </UL>
 * Each variant has it's own constructor.
 *
 * @author Mathijs den Burger
 */

class Uplink {

	// used for the local test mode
	private GameController gameController;

	// used for the "real" online mode
	private ClientDaemon postman;
	private String gameName;

	/**
	 * Constructs an uplink that can be used to test the game locally.
	 * In this mode, all game messages the server would respond with when
	 * are directly given to the game controller.
	 * @param gameController The controller of the local game
	 * @param gameName The name of this game
	 */
	public Uplink(GameController gameController, String gameName) {
		this.gameController = gameController;
		this.gameName = gameName;
		postman = null;
	}

	/**
	 * Constructs an uplink that can be used in the real online mode.
	 * In this mode all game messages are sent the server, which distributes
	 * them to all the users and spectators of this game.
	 * @param postman The client deamon that will send the game messages to the server
	 * @param gameName The name of this game
	 */
	public Uplink(ClientDaemon postman, String gameName) {
		this.postman = postman;
		this.gameName = gameName;
		gameController = null;
	}

	/**
	 * Gives the name of the game this uplink belongs to.
	 * @return The name of the game
	 */
	public String getGameName() {
		return gameName;
	}

	/**
	 * Executes the given game move. The exact behavior depends one the
	 * way the uplink is constructed. See the constructors.
	 * @param move The game move to execute
	 */
	public void executeGameMove(GameMove move) {
		if (gameController == null) {	// real online mode
			GameMessage.C2S_Gamemove msg = new GameMessage.C2S_Gamemove(gameName, move);
			postman.sendMessage(msg);
		} else {						// local test mode
			GameMessage.S2C_Gamemove msg = new GameMessage.S2C_Gamemove(gameName, move);
			gameController.incomingMessage(msg);
		}
	}

	/**
	 * Logs you out of the game you're currently in
	 */
	public void logout() {
		if (gameController == null) {	// real online mode
			GameMessage.C2S_Logout msg = new GameMessage.C2S_Logout(gameName);
			postman.sendMessage(msg);
		} else {						// local test mode
			GameMessage.S2C_Leaveself msg = new GameMessage.S2C_Leaveself(gameName);
			gameController.incomingMessage(msg);
		}
	}

}
