package quest.client.game;

import java.util.Vector;
import javax.media.j3d.*;

/**
 * <P>
 * The base class for all objects that can be seen in the 3D world.
 * Each visible object has a associated 3D model. This model is
 * positioned relative to it's parent in the scene graph.</P>
 * <P>
 * The actual model is build at the moment the user requests it
 * by the method {@link quest.client.game.VisibleObject#getModel()}.
 * If a subclass needs the model to be build earlier, it should
 * call this method.</P>
 *
 * @author Mathijs den Burger
 */

abstract class VisibleObject {

	private Vector modelCapabilities = new Vector();
	private Object userData = null;

	private Node modelRoot = null;

	/**
	 * Adds a capability to the 3D model root node. This can only be
	 * done as long as the model hasn't been build. Otherwise a
	 * Runtime exception is thrown.
	 *
	 * @param capability The capability to add to the 3D model of this visible obeject.
	 */
	protected void addModelCapability(int capability) {
		if (modelRoot == null) {
			modelCapabilities.add(new Integer(capability));
		} else {
			throw new RuntimeException("Attempting to add a capability to a model that's already build");
		}
	}

	/**
	 * Sets the user data associated with the 3D model root node.
	 * This can only be done as long as the model
	 * hasn't been build. Otherwise a Runtime exception is thrown.
     *
	 * @param userData The data to associate with the 3D model.
	 */
	protected void setModelUserData(Object userData) {
		if (modelRoot == null) {
			this.userData = userData;
		} else {
			throw new RuntimeException("Attempting to set the user data of a model that's already build");
		}
	}

	/**
	 * Returns the root of the subtree of the scenegraph that
	 * constructs the 3D model that this object represents.
	 *
	 * @return The root of the subtree that forms the model
	 */
	public final Node getModel() {
		if (modelRoot == null) {
			modelRoot = buildModel();
			for (int i = 0; i < modelCapabilities.size(); i++) {
				modelRoot.setCapability(((Integer)modelCapabilities.elementAt(i)).intValue());
			}
			if (userData != null) modelRoot.setUserData(userData);
		}
		return modelRoot;
	}

	/**
	 * Builds the subtree of the scenegraph that represents
	 * the 3D model of this visible object. This method is
	 * only called once, when the 3D model is constructed.
	 * If the object needs to alter parts of
	 * the 3D model later, appropriate capabilities must be set
	 * for parts of the scene graph subtree the model consists of.
	 * Applications should not call this method directly, but
	 * use {@link quest.client.game.VisibleObject#getModel()} instead.
	 *
	 * @return The root of the constructed subtree that forms the 3D model
	 * of this visible object
	 */
	abstract Node buildModel();

}