package quest.client.game;

import quest.global.game.MazeCorridor3D;
import javax.media.j3d.*;
import javax.vecmath.*;

/**
 * A set of 3D walls that represent one of the possible 2D shapes that
 * define a maze corridor. They can furthermore have a certain visinility type:
 * <UL>
 * <LI><B>SOLID</B> walls are the normal visible ones
 * <LI><B>CONTOURS</B> walls show only the contours of the 2D shape.
 * <LI><B>PITCH_DARK</B> walls show no walls at all: they're just black blocks.
 * </UL>
 *
 * @see quest.global.game.MazeCorridor3D#getShape2D()
 * @author Mathijs den Burger
 */

class Walls extends Shape3D {

	/**
	 * The height of a wall (in metres)
	 */
	public static final float HEIGHT = 4f;
	
	/**
	 * The width of a wall (in metres)
	 */
	public static final float WIDTH = 2f;
	
	/**
	 * The width of a corridor (the space between 2 walls) (in metres)
	 */
	public static final float CORRIDOR_WIDTH = 3f;
	
	/**
	 * Visibility constant, meaning the walls should be normally visible
	 */
	public static final int SOLID = 0;
	
	/**
	 * Visibility constant, meaning you can only see the contours of the walls
	 */
	public static final int CONTOURS = 1;
	
	/**
	 * Visibility constant, meaning you can see no walls at all.
	 */
	public static final int	PITCH_DARK = 2;

	private int shape2D;
	private boolean contours;
	private int mode;

	/**
	 * Construct a new set of walls that form the given 2D shape with the
	 * given visibility.
	 * @param shape2D The 2D shape the walls should form
	 * @param yTranslation The translation along the Y-axis of these walls.
	 * @param mode The visibility mode of the walls (SOLID, CONTOURS or PITCH_DARK)
	 * @see quest.global.game.MazeCorridor3D#getShape2D()
	 */
	public Walls(int shape2D, float yTranslation, int mode) {
		this.shape2D = shape2D;
		this.mode = mode;
		switch(mode) {
			case PITCH_DARK:
				buildPitchDark(yTranslation);
				break;
			case SOLID: case CONTOURS:
				switch(shape2D) {
					case MazeCorridor3D.STRAIGHT:
						if (mode == SOLID) buildStraightSolid(yTranslation);
						else buildStraightContours(yTranslation);
						break;
					case MazeCorridor3D.CORNER:
						if (mode == SOLID) buildCornerSolid(yTranslation);
						else buildCornerContours(yTranslation);
						break;
					case MazeCorridor3D.TJUNCTION:
						if (mode == SOLID) buildTJunctionSolid(yTranslation);
						else buildTJunctionContours(yTranslation);
						break;
					case MazeCorridor3D.CROSSING:
						if (mode == SOLID) buildCrossingSolid(yTranslation);
						else buildCrossingContours(yTranslation);
						break;
					default:
						System.err.println("No model defined for MazeCorridor3D type " + shape2D + " in Walls.<init>");
				}
				break;
			default:
				throw new RuntimeException("Unknown mode in Walls.<init>: " + mode);
		}
	}

	/**
	 * Gives the 2D shape of these walls
	 * @return The 2D shape
	 * @see quest.global.game.MazeCorridor3D#getShape2D()
	 */
	public final int getShape2D() {
		return shape2D;
	}

	/**
	 * Gives the visibility mode of these walls
	 * @return The visibility mode of these walls (SOLID, CONTOURS or PITCH_DARK)
	 */
	public final int getMode() {
		return mode;
	}

	private void buildPitchDark(float yTranslation) {
		NodeComponentFactory ncf = NodeComponentFactory.instance();
		ModelTool mt = ModelTool.instance();
		int vertexFormat = QuadArray.COORDINATES | QuadArray.NORMALS | QuadArray.COLOR_3;

		// make coordinates
		float[] blockCoords = mt.blockCoordinates((2 * WIDTH) + CORRIDOR_WIDTH, HEIGHT, (2 * WIDTH) + CORRIDOR_WIDTH);

		// translate to the right place
		mt.translate(blockCoords, new Vector3f(0, yTranslation, 0));

		/*
		// make colors
		byte[] black = new byte[mt.NR_BLOCK_VERTICES * 3];
		for (int i = 0; i < mt.NR_BLOCK_VERTICES * 3; i++) {
			black[i] = 0;
		}
		*/

		// add geometry
		addGeometry(ncf.getQuadArray(mt.NR_BLOCK_VERTICES, vertexFormat, blockCoords, mt.blockNormals()));
	}

	private void buildStraightSolid(float yTranslation) {
		NodeComponentFactory ncf = NodeComponentFactory.instance();
		ModelTool mt = ModelTool.instance();
		int vertexFormat = QuadArray.COORDINATES | QuadArray.NORMALS;

		// make coordinates
		float[] westWallCoords = mt.blockCoordinates(WIDTH, HEIGHT, (2 * WIDTH) + CORRIDOR_WIDTH);
		float[] eastWallCoords = mt.blockCoordinates(WIDTH, HEIGHT, (2 * WIDTH) + CORRIDOR_WIDTH);

		// translate parts to the right place
		mt.translate(westWallCoords, new Vector3f(0, yTranslation, 0));
		mt.translate(eastWallCoords, new Vector3f(WIDTH + CORRIDOR_WIDTH, yTranslation, 0));

		// add geometry
		addGeometry(ncf.getQuadArray(mt.NR_BLOCK_VERTICES, vertexFormat, westWallCoords, mt.blockNormals()));
		addGeometry(ncf.getQuadArray(mt.NR_BLOCK_VERTICES, vertexFormat, eastWallCoords, mt.blockNormals()));
	}

	private void buildStraightContours(float yTranslation) {
		NodeComponentFactory ncf = NodeComponentFactory.instance();
		ModelTool mt = ModelTool.instance();
		int vertexFormat = QuadArray.COORDINATES | QuadArray.COLOR_3;
		int nrVertices = 8;
		int[] stripVertexCounts = { 4, 4 };

		// make coordinates
		float[] coords = {
			// first strip: back-left-edge, to top, left-top, front-left-edge
			WIDTH, yTranslation, 0,
			WIDTH, yTranslation + HEIGHT, 0,
			WIDTH, yTranslation + HEIGHT, (2 * WIDTH) + CORRIDOR_WIDTH,
			WIDTH, yTranslation, (2 * WIDTH) + CORRIDOR_WIDTH,
			// second strip: back-right-edge, to top, right-top, front-right-edge
			WIDTH + CORRIDOR_WIDTH, yTranslation, 0,
			WIDTH + CORRIDOR_WIDTH, yTranslation + HEIGHT, 0,
			WIDTH + CORRIDOR_WIDTH, yTranslation + HEIGHT, (2 * WIDTH) + CORRIDOR_WIDTH,
			WIDTH + CORRIDOR_WIDTH, yTranslation, (2 * WIDTH) + CORRIDOR_WIDTH,
		};

		// make colors
		byte[] grey = new byte[nrVertices * 3];
		for (int i = 0; i < (nrVertices * 3); i++) {
			grey[i] = (byte)128;
		}

		// add geometry
		addGeometry(ncf.getLineStripArray(nrVertices, vertexFormat, stripVertexCounts, coords, grey));
	}

	private void buildCornerSolid(float yTranslation) {
		NodeComponentFactory ncf = NodeComponentFactory.instance();
		ModelTool mt = ModelTool.instance();
		int vertexFormat = QuadArray.COORDINATES | QuadArray.NORMALS;

		// make coordinates
		float[] westWallCoords = mt.blockCoordinates(WIDTH, HEIGHT, WIDTH + CORRIDOR_WIDTH);
		float[] southBigWallCoords = mt.blockCoordinates((2 * WIDTH) + CORRIDOR_WIDTH, HEIGHT, WIDTH);
		float[] northEastWallCoords = mt.blockCoordinates(WIDTH, HEIGHT, WIDTH);

		// translate parts to the right place
		mt.translate(westWallCoords, new Vector3f(0, yTranslation, 0));
		mt.translate(southBigWallCoords, new Vector3f(0, yTranslation, WIDTH + CORRIDOR_WIDTH));
		mt.translate(northEastWallCoords, new Vector3f(WIDTH + CORRIDOR_WIDTH, yTranslation, 0));

		// add geometry
		addGeometry(ncf.getQuadArray(mt.NR_BLOCK_VERTICES, vertexFormat, westWallCoords, mt.blockNormals()));
		addGeometry(ncf.getQuadArray(mt.NR_BLOCK_VERTICES, vertexFormat, southBigWallCoords, mt.blockNormals()));
		addGeometry(ncf.getQuadArray(mt.NR_BLOCK_VERTICES, vertexFormat, northEastWallCoords, mt.blockNormals()));
	}

	private void buildCornerContours(float yTranslation) {
		NodeComponentFactory ncf = NodeComponentFactory.instance();
		ModelTool mt = ModelTool.instance();
		int vertexFormat = QuadArray.COORDINATES | QuadArray.COLOR_3;
		int nrVertices = 10;
		int[] stripVertexCounts = { 5, 5 };

		// make coordinates
		float[] coords = {
			// first strip: back-left-edge, to top, left lower corner, right-near-edge
			WIDTH, yTranslation, 0,
			WIDTH, yTranslation + HEIGHT, 0,
			WIDTH, yTranslation + HEIGHT, WIDTH + CORRIDOR_WIDTH,
			(2 * WIDTH) + CORRIDOR_WIDTH, yTranslation + HEIGHT, WIDTH + CORRIDOR_WIDTH,
			(2 * WIDTH) + CORRIDOR_WIDTH, yTranslation, WIDTH + CORRIDOR_WIDTH,
			// second strip: back-right-edge, to top, right upper corner, right-far-edge
			WIDTH + CORRIDOR_WIDTH, yTranslation, 0,
			WIDTH + CORRIDOR_WIDTH, yTranslation + HEIGHT, 0,
			WIDTH + CORRIDOR_WIDTH, yTranslation + HEIGHT, WIDTH,
			(2 * WIDTH) + CORRIDOR_WIDTH, yTranslation + HEIGHT, WIDTH,
			(2 * WIDTH) + CORRIDOR_WIDTH, yTranslation, WIDTH
		};

		// make colors
		byte[] grey = new byte[nrVertices * 3];
		for (int i = 0; i < (nrVertices * 3); i++) {
			grey[i] = (byte)128;
		}

		// add geometry
		addGeometry(ncf.getLineStripArray(nrVertices, vertexFormat, stripVertexCounts, coords, grey));
	}

	private void buildTJunctionSolid(float yTranslation) {
		NodeComponentFactory ncf = NodeComponentFactory.instance();
		ModelTool mt = ModelTool.instance();
		int vertexFormat = QuadArray.COORDINATES | QuadArray.NORMALS;

		// make coordinates
		float[] bigWallCoords = mt.blockCoordinates((2 * WIDTH) + CORRIDOR_WIDTH, HEIGHT, WIDTH);
		float[] leftSmallWallCoords = mt.blockCoordinates(WIDTH, HEIGHT, WIDTH);
		float[] rightSmallWallCoords = mt.blockCoordinates(WIDTH, HEIGHT, WIDTH);

		// translate parts to the right place
		mt.translate(bigWallCoords, new Vector3f(0, yTranslation, WIDTH + CORRIDOR_WIDTH));
		mt.translate(leftSmallWallCoords, new Vector3f(0, yTranslation, 0));
		mt.translate(rightSmallWallCoords, new Vector3f(WIDTH + CORRIDOR_WIDTH, yTranslation, 0));

		// add geometry
		addGeometry(ncf.getQuadArray(mt.NR_BLOCK_VERTICES, vertexFormat, bigWallCoords, mt.blockNormals()));
		addGeometry(ncf.getQuadArray(mt.NR_BLOCK_VERTICES, vertexFormat, leftSmallWallCoords, mt.blockNormals()));
		addGeometry(ncf.getQuadArray(mt.NR_BLOCK_VERTICES, vertexFormat, rightSmallWallCoords, mt.blockNormals()));
	}

	private void buildTJunctionContours(float yTranslation) {
		NodeComponentFactory ncf = NodeComponentFactory.instance();
		ModelTool mt = ModelTool.instance();
		int vertexFormat = QuadArray.COORDINATES | QuadArray.COLOR_3;
		int nrVertices = 14;
		int[] stripVertexCounts = { 5, 5, 4 };

		// make coordinates
		float[] coords = {
			// first strip: back-left-edge, to top, back-left-top corner, left-far-edge
			WIDTH, yTranslation, 0,
			WIDTH, yTranslation + HEIGHT, 0,
			WIDTH, yTranslation + HEIGHT, WIDTH,
			0, yTranslation + HEIGHT, WIDTH,
			0, yTranslation, WIDTH,
			// second strip: right-far-edge, to top, back-right-top corner, back-right-edge
			(2 * WIDTH) + CORRIDOR_WIDTH, yTranslation, WIDTH,
			(2 * WIDTH) + CORRIDOR_WIDTH, yTranslation + HEIGHT, WIDTH,
			WIDTH + CORRIDOR_WIDTH, yTranslation + HEIGHT, WIDTH,
			WIDTH + CORRIDOR_WIDTH, yTranslation + HEIGHT, 0,
			WIDTH + CORRIDOR_WIDTH, yTranslation, 0,
			// third strip: left-near-edge, to top, near-top-edge, right-near-edge
			0, yTranslation, WIDTH + CORRIDOR_WIDTH,
			0, yTranslation + HEIGHT, WIDTH + CORRIDOR_WIDTH,
			(2 * WIDTH) + CORRIDOR_WIDTH, yTranslation + HEIGHT, WIDTH + CORRIDOR_WIDTH,
			(2 * WIDTH) + CORRIDOR_WIDTH, yTranslation, WIDTH + CORRIDOR_WIDTH
		};

		// make colors
		byte[] grey = new byte[nrVertices * 3];
		for (int i = 0; i < (nrVertices * 3); i++) {
			grey[i] = (byte)128;
		}

		// add geometry
		addGeometry(ncf.getLineStripArray(nrVertices, vertexFormat, stripVertexCounts, coords, grey));
	}

	private void buildCrossingSolid(float yTranslation) {
		NodeComponentFactory ncf = NodeComponentFactory.instance();
		ModelTool mt = ModelTool.instance();
		int vertexFormat = QuadArray.COORDINATES | QuadArray.NORMALS;

		// make coordinates
		float[] northWestWallCoords = mt.blockCoordinates(WIDTH, HEIGHT, WIDTH);
		float[] northEastWallCoords = mt.blockCoordinates(WIDTH, HEIGHT, WIDTH);
		float[] southWestWallCoords = mt.blockCoordinates(WIDTH, HEIGHT, WIDTH);
		float[] southEastWallCoords = mt.blockCoordinates(WIDTH, HEIGHT, WIDTH);

		// translate parts to the right place
		mt.translate(northWestWallCoords, new Vector3f(0, yTranslation, 0));
		mt.translate(northEastWallCoords, new Vector3f(WIDTH + CORRIDOR_WIDTH, yTranslation, 0));
		mt.translate(southWestWallCoords, new Vector3f(0, yTranslation, WIDTH + CORRIDOR_WIDTH));
		mt.translate(southEastWallCoords, new Vector3f(WIDTH + CORRIDOR_WIDTH, yTranslation, WIDTH + CORRIDOR_WIDTH));

		// add geometry
		addGeometry(ncf.getQuadArray(mt.NR_BLOCK_VERTICES, vertexFormat, northWestWallCoords, mt.blockNormals()));
		addGeometry(ncf.getQuadArray(mt.NR_BLOCK_VERTICES, vertexFormat, northEastWallCoords, mt.blockNormals()));
		addGeometry(ncf.getQuadArray(mt.NR_BLOCK_VERTICES, vertexFormat, southWestWallCoords, mt.blockNormals()));
		addGeometry(ncf.getQuadArray(mt.NR_BLOCK_VERTICES, vertexFormat, southEastWallCoords, mt.blockNormals()));
	}

	private void buildCrossingContours(float yTranslation) {
		NodeComponentFactory ncf = NodeComponentFactory.instance();
		ModelTool mt = ModelTool.instance();
		int vertexFormat = QuadArray.COORDINATES | QuadArray.COLOR_3;
		int nrVertices = 20;
		int[] stripVertexCounts = { 5, 5, 5, 5 };

		// make coordinates
		float[] coords = {
			// first strip: back-left-edge, to top, back-left-top corner, left-far-edge
			WIDTH, yTranslation, 0,
			WIDTH, yTranslation + HEIGHT, 0,
			WIDTH, yTranslation + HEIGHT, WIDTH,
			0, yTranslation + HEIGHT, WIDTH,
			0, yTranslation, WIDTH,
			// second strip: left-near-edge, to top, near-left-top corner, front-left-edge
			0, yTranslation, WIDTH + CORRIDOR_WIDTH,
			0, yTranslation + HEIGHT, WIDTH + CORRIDOR_WIDTH,
			WIDTH, yTranslation + HEIGHT, WIDTH + CORRIDOR_WIDTH,
			WIDTH, yTranslation + HEIGHT, (2 * WIDTH) + CORRIDOR_WIDTH,
			WIDTH, yTranslation, (2 * WIDTH) + CORRIDOR_WIDTH,
			// third strip: front-right-edge, to top, front-right-top corner, right-near-edge
			WIDTH + CORRIDOR_WIDTH, yTranslation, (2 * WIDTH) + CORRIDOR_WIDTH,
			WIDTH + CORRIDOR_WIDTH, yTranslation + HEIGHT, (2 * WIDTH) + CORRIDOR_WIDTH,
			WIDTH + CORRIDOR_WIDTH, yTranslation + HEIGHT, WIDTH + CORRIDOR_WIDTH,
			(2 * WIDTH) + CORRIDOR_WIDTH, yTranslation + HEIGHT, WIDTH + CORRIDOR_WIDTH,
			(2 * WIDTH) + CORRIDOR_WIDTH, yTranslation, WIDTH + CORRIDOR_WIDTH,
			// fourth strip: right-far-edge, to top, back-right-top corner, back-right-edge
			(2 * WIDTH) + CORRIDOR_WIDTH, yTranslation, WIDTH,
			(2 * WIDTH) + CORRIDOR_WIDTH, yTranslation + HEIGHT, WIDTH,
			WIDTH + CORRIDOR_WIDTH, yTranslation + HEIGHT, WIDTH,
			WIDTH + CORRIDOR_WIDTH, yTranslation + HEIGHT, 0,
			WIDTH + CORRIDOR_WIDTH, yTranslation, 0
		};

		// make colors
		byte[] grey = new byte[nrVertices * 3];
		for (int i = 0; i < (nrVertices * 3); i++) {
			grey[i] = (byte)128;
		}

		// add geometry
		addGeometry(ncf.getLineStripArray(nrVertices, vertexFormat, stripVertexCounts, coords, grey));
	}
}