package quest.client.net;

import java.io.*;
import java.net.*;
import java.util.*;
import java.lang.reflect.*;
import quest.global.net.*;

/**
 * The client side daemon responsible for providing communication facilities to the processes.
 *
 * @author    Merijn Evertse
 */
public class ClientDaemon {
	private InetAddress serverAddress;
	private MessageClientThread receivingAgent;
	private MessageStream messageStream;
	private static ClientDaemon uniqueInstance = null;
	private Socket connection;

	/**
	 * Initializes the client daemon.
	 *
	 * @param    username    The username of the client
	 * @param    password    The password of the client
	 */
	public ClientDaemon(InetAddress serverAddress) throws IOException {
		try {
			if (uniqueInstance != null)
				throw new RuntimeException("singleton constraint violation");
	
			this.serverAddress = serverAddress;
			uniqueInstance = this;
			run();
		} catch (IOException e) {
			uniqueInstance = null;
			throw(e);
		}
	}

	/**
	 * Runs the client daemon. It connects to the server and authenticates the user.
	 * It also starts to listening proces for incoming messages.
	 */
	protected void run() throws IOException {
	  // Setup connection
	  connection = new Socket(serverAddress, ClientConfiguration.getInstance().getServerPort());
	  
	  // Start communication facilities
	  messageStream = new MessageStream(connection, MessageStream.CLIENT_SIDE);      
	  
	  // Start receiving agent
	  receivingAgent = new MessageClientThread(messageStream);
	  receivingAgent.start();
	}
	
	/**
	 * Registers a proces, so it can receive messages.
	 *
	 * @param    messageType        The type of the messages the proces uses
	 * @param    messageListener    The listener the client uses for handling incoming messages
	 */
	public void registerProces(Class messageType, MessageClientListener messageListener) {
		receivingAgent.registerProces(new Proces(messageType, messageListener));
	}

	/**
	 * Unregisters a proces, so no messages will be passes to the proces
	 * anymore.
	 *
	 * @param    MessageType    The type of the message to unregister
	 */
	public void unregisterProces(Class messageType) {
		receivingAgent.unregisterProces(new Proces(messageType, null));
	}

	/**
	 * Sends the message to the server.
	 *
	 * @param    message    The message to be send
	 */
	public void sendMessage(Message message) {
	  try {
		messageStream.sendMessage(message);
	  } catch (IOException e) {
		System.out.println("WARNING: Connection problems: " + e.getMessage());
	  }
	}
	
	/**
	 * Returns an instance of this class, making sure only one object of it exists.
	 *
	 * @return    A reference to the instance of this class
	 */
	public static ClientDaemon getInstance() {
		if (uniqueInstance == null) {
			throw new RuntimeException("singleton class has to be initialized by the constructor first");
		}

		return uniqueInstance;
	}

	/**
	 * Closes the communication with the server and the client daemon itself.
	 */
	public void close() {
	  System.out.println("ClientDaemon is being closed");
	  uniqueInstance = null;
	
		try {
		  if (messageStream != null) {
			messageStream.close();
			messageStream = null;
		  }
		  if (receivingAgent != null) {
			receivingAgent.close();
			receivingAgent = null;
		  }
		  if (connection != null) {
			connection.close();
			connection = null;
		  }
		} catch (IOException e) {}
	}
}


