package quest.global.chat;

import java.io.Serializable;
import quest.global.net.Message;

/**
 * Contains all channel messages used for communication between client and server.
 * Channel messages inform the clients of the channel currently active on the server.
 * There are two types of communication.
 * <UL>
 * <LI>A client requests a full update and gets info about all the channels.
 * <LI>A client subscribes for partial updates and gets at regular intervals
 *     info about all the channels that have changed since the last update.
 * </UL>
 * <PRE>
 * Packet          | Reply				| Description
 * ------------------------------------------------------------
 * C2S_Subscribe	 S2C_FullUpdate		  Client subscribes for auto updates
 *                   S2C_PartialUpdate    At regular times all changes are distributes to 
 *                                        subscribed clients
 * C2S_UnSubscribe   N/A                  Client unsubscribes for auto updates 
 * C2S_Request		 S2C_FullUpdate       Client requests a single full update
 *
 * NOTE: This message is not to be uses for communication, this functions only
 *       as a root in the hierarchy tree.
 *
 * @author    Merijn Evertse & Sander Brandenburg
 */ 
public class ChannelMessage extends Message implements Serializable {

	/* Server -> Client messages */
	
	/**
	 * Contains the information about all the channels currently active on the server.
	 *
	 * @author    Merijn Evertse & Sander Brandenburg
	 */  	
	public static class S2C_FullUpdate extends ChannelMessage implements Serializable {
		/**
		 * A list of ChannelInfo objects. One for each channel on the server.
		 */
		public ChannelInfo[] array;
		
		/**
		 * Initializes the message with the given list of info objects.
		 */
		public S2C_FullUpdate(ChannelInfo[] arr) {
			this.array = arr;
		}
	}
	
	/**
	 * Contains all the changes of the channels since the previous partial update.
	 *
	 * @author    Merijn Evertse & Sander Brandenburg
	 */
	public static class S2C_PartialUpdate extends ChannelMessage implements Serializable {
		/**
		 * This channel is created between the two updates.
		 */
		public final static int UPDATE_INSERT = 0;
		
		/**
		 * This channel is deleted between the two updates.
		 */
		public final static int UPDATE_DELETE = 1;
		
		/**
		 * This channel is changed between the two updates.
		 */ 
		public final static int UPDATE_MODIFY = 2;
		
		/**
		 * Container class representing an update of the info about this channel.
		 * Consists of the type of the change in the channel and the new 
		 * information of the channel.
		 *
		 * @author    Merijn Evertse & Sander Brandenburg
		 */
		public final static class Update {
			/**
			 * The type of the change.
			 */
			public int mode;
			
			/**
			 * The new information about the channel.
			 */
			public ChannelInfo channel;
			
			/**
			 * Initializes the object with the given parameters.
			 *
			 * @param    mode    The type of the change
			 * @param    ci      The ChannelInfo object
			 */
			public Update(int mode, ChannelInfo ci) {
				this.mode = mode;
				this.channel = ci;
			}
		}
		
		/**
		 * The list of all changes since the previous update.
		 */
		public Update[] updateArray;
		
		/**
		 * Initializes the oject with the given parameters.
		 *
		 * @param    updateArray    The list off all changes in the channels
		 */
		public S2C_PartialUpdate(Update[] updateArray) {
			this.updateArray = updateArray;
		}
	}
	
	/**
	 * The client wants to subscribe for auto updates on the changes in the channels.
	 *
	 * @author    Merijn Evertse & Sander Brandenburg
	 */
	public static class C2S_Subscribe extends ChannelMessage implements Serializable {
		/**
		 * Initializes the object.
		 */
		public C2S_Subscribe() {}
	}
	
	/**
	 * The client wants to unsubscribe for auto updates.
	 *
	 * @author    Merijn Evertse & Sander Brandenburg
	 */
	public static class C2S_UnSubscribe extends ChannelMessage implements Serializable {
		/**
		 * Initializes the object.
		 */
		public C2S_UnSubscribe() {}
	}
	
	/**
	 * The client requests a full update on the currently active channels.
	 *
	 * @author    Merijn Evertse & Sander Brandenburg
	 */
	public static class C2S_Request extends ChannelMessage implements Serializable {
		/**
		 * Initialized the object.
		 */
		public C2S_Request() {}
	}
}
