package quest.global.chat;

import java.io.*;
import java.lang.reflect.*;
import quest.global.net.*;

/**
 * Contains all chat messages used for communication between client and server.
 * <PRE>
 * Tabel with the different modes of a channel
 *
 * Mode(int)    | Constant             | Description
 * =============+======================+========================
 * 0            | CHANMODE_UNKNOWN     | Undefined channel mode
 * 1            | CHANMODE_BEGIN       | Begin / Chat fase
 * 2            | CHANMODE_GAME        | Game fase
 * 3            | CHANMODE_END         | Game ended
 * 
 * Tabel with the different modes of a user
 *
 * Mode(int)    | Constant             | Description
 * =============+======================+====================
 * 0            | USERMODE_UNDEFINED   | Undefined, no game can be started
 * 1            | USERMODE_PLAYER      | Player
 * 2            | USERMODE_PLAYERADMIN | Player and administrator
 * 3            | USERMODE_SPEC_PLAYER | Spectator, asking to become player
 * 4            | USERMODE_SPEC_SPEC   | Spectator
 *   
 * Tabel with the different modes of a signoff
 * 
 * Mode(int)    | Constant             | Description
 * =============+======================+====================
 * 1            | SIGNOFF_LEAVE        | Client leaves channel
 * 2            | SIGNOFF_KICK         | Administrator kicks client
 * 3            | SIGNOFF_TIMEOUT      | Client leaves due to a timeout
 *
 * Protocol overview:
 *
 * NOTE: []'s indicate a selective send of the message, otherwise
 *            all clients receive the message.
 *       ()'s indicate an optional message which is sent only special cases.
 *       ...* indicates multiple values, by means of an array. 
 *
 * Client -&gt; Server packets
 *
 * Request       | Reply          | Arguments                    | Description
 * --------------------------------------------------------------------------------------------------
 * C2S_CHAT        S2C_MSG          channelname, text              Normal textmessage for whole channel
 * C2S_WHISPER     [S2C_WHISPER]    channelname, username, text    Whisper message to target
 * C2S_JOIN        S2C_JOINSELF     channelname                    Join request (will enter as spectator)
 *                 [S2C_USERJOIN]   
 * C2S_CREATE      S2C_JOINSELF     channelname                     Request for creation of channel
 *                                                                 (client will become administrator)
 * C2S_CHUSERMODE  S2C_CHUSERMODE   channelname, mode              Request for mode change of client
 * C2S_KICK        S2C_SIGNOFF      channelname, username          Kick player out of channel (admin only)
 *                 [S2C_LEAVESELF]
 * C2S_LOCKOUT     S2C_LOCKOUT      channelname, "1"|"0"           Lock request for channel (admin only)
 * C2S_LEAVE       S2C_SIGNOFF      channelname                    Request for leaving the channel
 *                 [S2C_LEAVESELF]
 *                 (S2C_CHUSERMODE)                                If leaving client was admin, inform
 *                                                                 all clients of the new admin
 * C2S_LOGIN       S2C_JOINSELF                                    Login request for global chat channel
 *                 [S2C_USERJOIN]
 * C2S_LOGOUT      S2C_SIGNOFF                                     Logout request for global chat channel
 *                 [S2C_LEAVESELF]
 *                 (S2C_CHUSERMODE)                                If leaving client was admin, inform
 *                                                                 all clients of the new admin
 * N/A             S2C_SIGNOFF                                     Client leaves channel due to a timeout
 *                 (S2C_CHUSERMODE)                                If leaving client was admin, inform
 *                                                                 all clients of the new admin
 * C2S_STARTGAME   S2C_CHCHANMODE                                  Admin request for game to be started
 * C2S_STOPGAME    S2C_CHCHANMODE                                  Admin request for the game to be stopped
 *
 * Server -&gt; Client packets
 *
 * Request       | Reply          | Arguments                    | Description
 * --------------------------------------------------------------------------------------------------
 * S2C_BROADCAST   N/A              channelname, text              Textmessage from channel
 * S2C_MSG         N/A              channelname, username, text    Textmessage from a client in the channel
 * S2C_JOINSELF    N/A              channelname, (naam, umode)*    Join reply, with info on other clients
 * S2C_LEAVESELF   N/A              channelname                    Leave reply
 * S2C_USERJOIN    N/A              channelname, username, umode   New client join channel, with this info
 * S2C_SIGNOFF     N/A              channelname, username, umode   Client leaves channel
 * S2C_CHUSERMODE  N/A              channelname, username, umode   Usermode of client changes
 * S2C_CHCHANMODE  N/A              channelname, cmode             Mode of the channel changes
 *                                                                 (for instance when game starts)
 * S2C_WHISPER     N/A              channelname, username, text    Textmessage from a client for target
 * S2C_LOCKOUT     N/A              channelname, "1"|"0"           Channel is being (un)locked
 * </PRE>
 * This class is the standard chatmessage used as root in the hierarchy tree. 
 * This message is not to be used in normal communications. 
 * All messages to be transferred are to extend this message to ensure
 * safe transport and handling of the message.
 *
 * @author Sander Brandenburg, Merijn Evertse
 */
public abstract class ChatMessage extends Message implements Serializable {
    /**
     * The name of the channel this message originated from.
     */
    public String channel;
  
    /**
     * Constructor to create a new message with the given parameters.
     */
    public ChatMessage(String channel) {
	this.channel = channel;
    }
    
    /**
     * The name of the global channel.
     */
    public final static String GLOBALCHANNELNAME = "Global";

    /**
     * User leaves.
     */
    public final static int SIGNOFF_LEAVE        = 1;

    /**
     * User is kicked out by the administrator.
     */
    public final static int SIGNOFF_KICK        = 2;

    /**
     * User timedout and is removed by the system.
     */
    public final static int SIGNOFF_TIMEOUT        = 3;
  
    /**
     * The mode of the channel is not known at the moment.
     */
    public final static int CHANMODE_UNKNOWN    = 0;
  
    /**
     * The channel is in the pre-game state at the moment.
     */
    public final static int CHANMODE_BEGIN        = 1;

    /**
     * The channel is in the game state at the moment.
     */
    public final static int CHANMODE_GAME        = 2;

    /**
     * The channel is in the post-game state at the moment.
     */
    public final static int CHANMODE_END        = 3;
  
    /**
     * The mode of the user is not known at the moment.
     */
    public final static int USERMODE_UNDEFINED        = 0;

    /** 
     * The user is a player.
     */
    public final static int USERMODE_PLAYER        = 1;

    /**
     * The user is the administrator of the channel, and therefor also a player.
     */
    public final static int USERMODE_PLAYERADMIN    = 2;
  
    /**
     * The user is a spectator, but wants to become a player.
     */
    public final static int USERMODE_SPEC_PLAYER    = 3;

    /**
     * The user is a spectator.
     */
    public final static int USERMODE_SPEC_SPEC        = 4;
  
    /**
     * A list of the descriptions for the different usermodes.
     */
    public final static String userModes[] = {
	"Undefined",
	"Player",
	"Administrator",
	"Spectator",
	"Spectator",
	null
    };
  
    /**
     * A list of the descriptions for the different channelmodes.
     */
    public final static String channelModes[] = {
	"Unknown",
	"Pre-game",
	"Game in progress",
	"Post-game",
	null
    };

    /*
     * Client to server messages.
     */
  
    /**
     * A chatmessage containing a normal text message.
     * Is sent by a user when he wants to say something in the channel to all the users.
     * @author Sander Brandenburg, Merijn Evertse 
     */
    public static class C2S_Chat extends ChatMessage implements Serializable {
    
	/**
	 * The text the user typed in the channel.
	 */
	public String text;
    
	/**
	 * Constructor the create the message with the given parameters.
	 *
	 * @param    channel    The name of the channel the message orgininated from
	 * @param    text       The text
	 */
	public C2S_Chat(String channel, String text) {
	    super(channel);
	    this.text = text;
	}
    }
    
    /**
     * A chatmessage containing a text message whisperred to another user.
     * Is sent by a user when he wants to whisper something to another user in the channel.
     * @author Sander Brandenburg, Merijn Evertse
     */
    public static class C2S_Whisper extends ChatMessage implements Serializable {
	/**
	 * The username of the user whisperred to.
	 */
	public String username;

	/**
	 * The text whisperred to the user.
	 */
	public String text;
    
	/**
	 * Constructor to create the whisper message with the given parameters.
	 *
	 * @param    channel    The name of the channel the message originated from
	 * @param    text       The text whisperred
	 * @param    username   The username of the target user
	 */
	public C2S_Whisper(String channel, String text, String username) {
	    super(channel);
	    this.text = text;
	    this.username = username;
	}
    }
    
    /**
     * A chatmessage containing a join request for a channel.
     * Is sent by a user when he wants to join the channel.
     * @author Sander Brandenburg, Merijn Evertse
     */
    public static class C2S_Join extends ChatMessage implements Serializable {
	/**
	 * Constructor to create the join message with the given parameters.
	 *
	 * @param    channel    The name of the channel the message originated from
	 */
	public C2S_Join(String channel) {
	    super(channel);
	}
    }
  
    /**
     * A chatmessage containing a create channel request.
     * Is sent by a user when he wants to create a new channel, so he can get some users together
     * and start a game, or when he just wants to chat in a more private environment.
     * @author Sander Brandenburg, Merijn Evertse
     */
    public static class C2S_Create extends ChatMessage implements Serializable {    
	/**
	 * Constructor to create the create message with the given parameters.
	 *
	 * @param    channel    The name of the channel the message originated from
	 */
	public C2S_Create(String channel) {
	    super(channel);
	}
    }

    /**
     * A chatmessage containing a change usermode request.
     * Is sent by a user or the administrator of the channel when he requests for a change 
     * of the mode of that user.
     * 
     * NOTE: The check for the validity of the modechange is made by the server. 
     * @author Sander Brandenburg, Merijn Evertse
     */  
    public static class C2S_Chusermode extends ChatMessage implements Serializable {
	/**
	 * The username of the user from which the usermode is to be changed.
	 */
	public String username;

	/**
	 * The new mode of the user.
	 */
	public int mode;
    
	/**
	 * Constructor to create the change username message with the given parameters.
	 *
	 * @param    channel     The name of the channel the message originated from
	 * @param    username    The name of the user
	 * @param    mode        The new mode of the user
	 */
	public C2S_Chusermode(String channel, String username, int mode) {
	    super(channel);
	    this.username = username;
	    this.mode = mode;
	}
    }
  
    /**
     * A chatmessage containing a kick request for a user.
     * Is sent by the administrator of the channel when he wants to kick a user out of the channel.
     *
     * NOTE: Can only happen in main channel when an administrator is online at that moment.
     *       in other channels an administrator is always online because it is his channel.
     * @author Sander Brandenburg, Merijn Evertse
     */
    public static class C2S_Kick extends ChatMessage implements Serializable {
	/**
	 * The username of the user to be kicked.
	 */
	public String username;
    
	/**
	 * Constructor to create the kick message with the given parameters.
	 *
	 * @param    channel     The name of the channel the message originated from
	 * @param    username    The name of the user to be kicked
	 */
	public C2S_Kick(String channel, String username) {
	    super(channel);
	    this.username = username;
	}
    }
  
    /**
     * A chatmessage containing a lockout request of the channel.
     * Is sent by the administrator of the channel when he wants to lock / unlock the channel.
     * When a channel is locked no new users can join the channel.
     * @author Sander Brandenburg, Merijn Evertse
     */
    public static class C2S_Lockout extends ChatMessage implements Serializable {
	/**
	 * The type of the lockout request.
	 * Can be:    TRUE:    when the channel has to be locked
	 *            FALSE:   when the channel had to be unlocked
	 */
	public boolean lockout;
    
	/**
	 * Constructor to create the lockout message with the given parameters.
	 *
	 * @param    channel    The name of the channel the message originated from
	 * @param    lockout    The new state of the channel 
	 */
	public C2S_Lockout(String channel, boolean lockout) {
	    super(channel);
	    this.lockout = lockout;
	}
    }
  
    /**
     * A chatmessage containing a leave request from a user.
     * Is sent by a user when he wants to leave the channel he is in at the moment.
     * @author Sander Brandenburg, Merijn Evertse
     */
    public static class C2S_Leave extends ChatMessage implements Serializable {
	/**
	 * Constructor to create the leave message with the given parameters.
	 *
	 * @param    channel    The name of the channel the message originated from
	 */
	public C2S_Leave(String channel) {
	    super(channel);
	}
    }
  
    /**
     * A chatmessage containing a login request from a user.
     * Is sent by a user when he just enterred the chat system en wants to join the main channel.
     * @author Sander Brandenburg, Merijn Evertse
     */
    public static class C2S_Login extends ChatMessage implements Serializable {
	/**
	 * Constructor to create the login message with the given parameters.
	 *
	 * @param    channel    The name of the channel the message originated from
	 */
	public C2S_Login(String channel) {
	    super(channel);
	}
    }
  
    /**
     * A chatmessage containing a logout request from a user.
     * Is sent by a user when he wants to leave the chat system and therefor wants to leave 
     * the main channel.
     * @author Sander Brandenburg, Merijn Evertse
     */
    public static class C2S_Logout extends ChatMessage implements Serializable {
	/**
	 * Constructor to create the logout message with the given parameters.
	 *
	 * @param    channel    The name of the channel the message originated from
	 */
	public C2S_Logout(String channel) {
	    super(channel);
	}
    }
  
    /**
     * A chatmessage containing a startgame request from the administrator of the channel.
     * Is sent by the administrator of the channel when he thinks he has enough players for the game 
     * and wants to start playing. The number of players is now fixed and can only become less, 
     * when players leave the game. The channel is put into the game state.
     * @author Sander Brandenburg, Merijn Evertse
     */
    public static class C2S_Startgame extends ChatMessage implements Serializable {
	/**
	 * Constructor to create the startgame message with the given parameters.
	 *
	 * @param    channel    The name of the channel the message originated from
	 */
	public C2S_Startgame(String channel) {
	    super(channel);
	}
    }
  
    /**
     * A chatmessage containing an abortgame request from the administrator of the channel.
     * Is sent by the administrator of the channel when he notices that not enough players entered the 
     * game, after it was started, maybe because some players had connection problems for instance.
     * The game is then aborted and the channel is put back into the pre-game state. 
     * @author Sander Brandenburg, Merijn Evertse
     */
    public static class C2S_Abortgame extends ChatMessage implements Serializable {
	public C2S_Abortgame(String channel) {
	    super(channel);
	}
    }

    /**
     * Contains the request to stop the game by the administrator of the channel.
     * Is normally issued when administrator presses the 'stop game' button(STOP_STOP), or 
     * when a certain irrecoverable error has occurred(STOP_ABORT).
     * @author Sander Brandenburg, Merijn Evertse
     */
    public static class C2S_Stopgame extends ChatMessage implements Serializable {
	/**
	 * Represents an irrecoverable error, for example: not enough players to continue.
	 */
	public final static int STOP_ABORT = 1;
	/**
	 * Represents the administrator has pressed the 'stop game' button.
	 */
	public final static int STOP_STOP = 2;

	public int type;

	/**
	 * Contains the request to stop the game by the administrator of the channel.
	 * @param channel The channel name.
	 * @param type STOP_* fields to designate the type of stop request.
	 */
	public C2S_Stopgame(String channel, int type) {
	    super(channel);
	    this.type = type;
	}
    }
  
    /*
     * Server to client messages.
     */

    /**
     * A chatmessage containing a text message sent by the server to all users in the channel.
     * Is sent by the server to notify people of urgent messages, or special information of the server.
     * @author Sander Brandenburg, Merijn Evertse
     */
    public static class S2C_Broadcast extends ChatMessage implements Serializable {
	/**
	 * The text the server sends to all users.
	 */
	public String text;
    
	/**
	 * Constructor to create the broadcast message with the given parameters.
	 *
	 * @param    channel    The name of the channel the message originated from
	 * @param    text       The text of the message    
	 */
	public S2C_Broadcast(String channel, String text) {
	    super(channel);
	    this.text = text;
	}
    }
  
    /**
     * A chatmessage containing a normal text message for all users in the channel.
     * Is sent by the server to notify all users of what one user wanted to say in the channel.
     * @author Sander Brandenburg, Merijn Evertse
     */
    public static class S2C_Msg extends ChatMessage implements Serializable {
	/**
	 * The username of the originator of the message.
	 */
	public String username;

	/**
	 * The text of the message.
	 */
	public String text;
    
	/**
	 * Constructor to create the message with the given parameters.
	 *
	 * @param    channel    The name of the channel the message originated from
	 * @param    username   The name of the originator of the message
	 * @param    text       The text of the message
	 */
	public S2C_Msg(String channel, String username, String text) {
	    super(channel);
	    this.username = username;
	    this.text = text;
	}
    }
  
    /**
     * A chatmessage containing a join grant for the user originally requesting the join.
     * Is sent by the server as a reply to the user, telling him he is allowed to join the channel,
     * and also informing him of the other users in the channel, their usermode, and the mode
     * of the channel.
     * @author Sander Brandenburg, Merijn Evertse
     */
    public static class S2C_Joinself extends ChatMessage implements Serializable {
	/**
	 * The list of other users in the channel.
	 */
	public String[] usernames;

	/**
	 * The list of the usermodes of the users in the channel.
	 *
	 * NOTE: this list is in the same order as the usernames list.
	 */
	public int[] umodes;
    
	/**
	 * The mode of the channel.
	 */
	public int channelmode;
    
	/**
	 * Constructor to create the joinself message with the given parameters.
	 *
	 * @param    channel      The name of the channel the message originated from
	 * @param    usernames    The list of usernames
	 * @param    umodes       The list of usermodes
	 * @param    channelmode  The mode of the channel
	 */
	public S2C_Joinself(String channel, String[] usernames, int[] umodes, int channelmode) {
	    super(channel);
	    this.usernames = usernames;
	    this.umodes = umodes;
	    this.channelmode = channelmode;
	}
    }
  
    /**
     * A chatmessage containing a leave grant for the user originally request the leave.
     * Is sent by the server as a reply to the user, telling him he is allowed to leave the channel.
     * @author Sander Brandenburg, Merijn Evertse
     */
    public static class S2C_Leaveself extends ChatMessage implements Serializable {
	/**
	 * Constructor to create the leaveself message with the given parameters.
	 *
	 * @param    channel    The name of the channel the message originated from
	 */
	public S2C_Leaveself(String channel) {
	    super(channel);
	}    
    }
  
    /**
     * A chatmessage containing a userjoin notification.
     * Is sent by the server to inform all users other than the joining user, 
     * of the joining user and his usermode.
     * @author Sander Brandenburg, Merijn Evertse
     */
    public static class S2C_Userjoin extends ChatMessage implements Serializable {
	/**
	 * The username of the joining user.
	 */
	public String username;

	/**
	 * The usermode of the joining user.
	 */
	public int umode;
    
	/**
	 * Constructor to create the userjoin message with the given parameters.
	 *
	 * @param    channel    The name of the channel the message originated from
	 * @param    username   The name of the joining user
	 * @param    umode      The mode of the joining user
	 */
	public S2C_Userjoin(String channel, String username, int umode) {
	    super(channel);
	    this.username = username;
	    this.umode = umode;
	}
    }
  
    /**
     * A chatmessage containing a signoff notification.
     * Is sent by the server to notify all users except the leaving user, of the leaving user, 
     * and the reason he is leaving.
     * @author Sander Brandenburg, Merijn Evertse
     */
    public static class S2C_Signoff extends ChatMessage implements Serializable {
	/**
	 * The name of the leaving user.
	 */
	public String username;

	/**
	 * The reason the user is leaving.
	 * See the signoff identifiers for the different reasons.
	 */   
	public int mode;
    
	/**
	 * Constructor to create the signoff message with the given parameters.
	 *
	 * @param    channel    The name of the channel the message originated from
	 * @param    username   The name of the leaving user
	 * @param    mode       Identifier for the reason the user is leaving
	 */
	public S2C_Signoff(String channel, String username, int mode) {
	    super(channel);
	    this.username = username;
	    this.mode = mode;
	}
    }
  
    /**
     * A chatmessage containing a change usermode reply.
     * Is sent by the server to inform all users of the change of the usermode of a user.
     * @author Sander Brandenburg, Merijn Evertse
     */
    public static class S2C_Chusermode extends ChatMessage implements Serializable {
	/**
	 * The name of the user from which the usermode changed.
	 */
	public String username;

	/**
	 * The new usermode of the user.
	 */
	public int umode;
    
	/**
	 * Constructor to create the change usermode message with the given parameters.
	 *
	 * @param     channel    The name of the channel the message originated from
	 * @param     username   The name of the user
	 * @param     umode      The new mode of the user 
	 */
	public S2C_Chusermode(String channel, String username, int umode) {
	    super(channel);
	    this.username = username;
	    this.umode = umode;
	}
    }
  
    /**
     * A chatmessage containing a whisper message.
     * Is sent by the server to notify the target user of the text another user whisperred to him.
     * @author Sander Brandenburg, Merijn Evertse
     */ 
    public static class S2C_Whisper extends ChatMessage implements Serializable {
	/**
	 * The username of the sender of the message.
	 */
	public String username;

	/**
	 * The text being whisperred.
	 */
	public String text;
    
	/**
	 * Constructor to create the whisper message with the given parameters.
	 *
	 * @param    channel    The name of the channel the message originated from
	 * @param    username   The name of the user sending the message
	 * @param    text       The text the sender whisperred
	 */
	public S2C_Whisper(String channel, String username, String text) {
	    super(channel);
	    this.username = username;
	    this.text = text;
	}
    }
  
    /**
     * A chatmessage containing a whisper confirmation to the originator.
     * Is sent by the server to the originator of the message as a confirmation.
     * @author Sander Brandenburg, Merijn Evertse
     */
    public static class S2C_Whisperself extends ChatMessage implements Serializable {
	/**
	 * The username of the originator.
	 */
	public String username;

	/**
	 * The text of the whisper message.
	 */
	public String text;
    
	/**
	 * Constructor to create the whisperself message with the given parameters.
	 *
	 * @param    channel    The name of the channel the message originated from
	 * @param    username   The name of the sender of the message
	 * @param    text       The whisperred text
	 */
	public S2C_Whisperself(String channel, String username, String text) {
	    super(channel);
	    this.username = username;
	    this.text = text;
	}
    }
  
    /**
     * A chatmessage containing a lockout reply.
     * Is sent by the server to inform all users of the new lock state of the channel.
     * @author Sander Brandenburg, Merijn Evertse
     */
    public static class S2C_Lockout extends ChatMessage implements Serializable {
	/**
	 * The new state of the channel.
	 */
	public boolean lockout;
    
	/**
	 * Constructor to create the lockout message with the given parameters.
	 *
	 * @param    channel    The name of the channel the message originated from
	 * @param    lockout    The new mode of the channel
	 */
	public S2C_Lockout(String channel, boolean lockout) {
	    super(channel);
	    this.lockout = lockout;
	}
    }
  
    /**
     * A chatmessage containing a notification of a channelmode change.
     * Is sent by the server when the mode of the channel changes to inform all user of that change.
     * This occurs when for instance a game is started or ended.
     *
     * NOTE: For users other than the administrator this also tells for instance that they
     *       have to start a gameclient when the administrator initiated this start of the game.
     * @author Sander Brandenburg, Merijn Evertse
     */
    public static class S2C_Chchanmode extends ChatMessage implements Serializable {    
	/**
	 * The new mode of the channel.
	 */
	public int mode;
    
	/**
	 * Constructor to create the change channelmode message with the given parameters.
	 *
	 * @param    channel    The name of the channel the message originated from
	 * @param    mode       The new mode of the channel
	 */
	public S2C_Chchanmode(String channel, int mode) {
	    super(channel);
	    this.mode = mode;
	}
    }
}

