package quest.global.chat;

import java.io.Serializable;
import quest.global.net.Message;

/**
 * Contains all ranking messages used for communication between client and server.
 * Ranking messages are used to request and reply for all kinds of ranking overviews.
 *
 * <PRE>
 * Packet              | Reply				| Description
 * ----------------------------------------------------------------
 * C2S_RequestRanking    S2C_SendRanking      Clients requests for a ranking overview
 * NOTE: This message is not to be uses for communication, this functions only
 *       as a root in the hierarchy tree.
 *
 * @author    Merijn Evertse & Sander Brandenburg
 */ 
public class RankingMessage extends Message implements Serializable {
	/**
	 * Ranking overview of the players on the top of the ranking table.
	 */
	public final static int RANKING_TOP = 0;
	
	/**
	 * Ranking overview of the whole ranking table.
	 */
	public final static int RANKING_WHOLE = 1;
	
	/**
	 * Ranking overview of the players which are closely above and below in rank
	 * to the requesting client.
	 */
	public final static int RANKING_NEIGHBOUR = 2;

	/**
	 * The clients requests for a ranking overview.
	 *
	 * @author    Merijn Evertse & Sander Brandenburg
	 */
	public static class C2S_RequestRanking extends RankingMessage {
		/**
		 * The type of overview requested.
		 */
		public int mode;
		
		/**
		 * Initializes the object with the given parameters.
		 *
		 * @param    mode    The type of overview requested
		 */
		public C2S_RequestRanking (int mode) {
			this.mode = mode;
		}
	}

	/**
	 * Contains the ranking overview requested by a client.
	 *
	 * @author    Merijn Evertse & Sander Brandenburg
	 */
	public static class S2C_SendRanking extends RankingMessage {
		/**
		 * The type of overview requested.
		 */
		public int mode;
		
		/**
		 * The current ranking position of the requesting client.
		 */
		public int ownpos;
		
		/**
		 * The total number of players registerred at the system.
		 */
		public int maxpos;
		
		/**
		 * The ranking of the requesting client.
		 */
		public int ranking;
		
		/**
		 * The list of rankings of all players in the overview. These are UserRanking objects.
		 */
		public UserRanking[] rankings;
		
		/**
		 * Initializes the object with the given parameters.
		 *
		 * @param    mode     The type of overview requested
		 * @param    ownpos   The current ranking position of the requesting client
		 * @param    maxpos   The total number of players registerred at the system
		 * @param    ranking  The ranking of the requesting client
		 * @param    rankings The list of rankings of all players in the overview
		 */
		public S2C_SendRanking(int mode, int ownpos, int maxpos, int ranking, UserRanking[] rankings) {
			this.ownpos = ownpos;
			this.maxpos = maxpos;
			this.ranking = ranking;
			this.mode = mode;
			this.rankings = rankings;
		}
	}
}
