package quest.global.chat;

import java.io.Serializable;
import quest.global.net.Message;

/**
 * Contains all search messages used for communication between client and server.
 * Search messages are used to request and reply for search requests of player names
 * <PRE>
 * Packet          | Reply				| Description
 * ------------------------------------------------------------
 * C2S_Search	     S2C_Result           Clients requests a search
 *
 * NOTE: This message is not to be uses for communication, this functions only
 *       as a root in the hierarchy tree.
 *
 * @author    Merijn Evertse & Sander Brandenburg
 */ 
public class SearchMessage extends Message implements Serializable {

	/**
	 * Clients requests for a search of a playername.
	 *
	 * @author    Merijn Evertse & Sander Brandenburg
	 */	
	public static class C2S_Search extends SearchMessage {
		/**
		 * The name, or part of it, of the user to search for.
		 */
		public String username;
		
		/**
		 * Initializes the object with the given parameter.
		 *
		 * @param    username    Initializes the object with the given parameter
		 */
		public C2S_Search(String username) {
			this.username = username;
		}
	}

	/**
	 * Server sends the results from the search back to the requesting client.
	 *
	 * @author    Merijn Evertse & Sander Brandenburg
	 */
	public static class S2C_Result extends SearchMessage {
		/**
		 * Contains the information of one player who matches with the search string.
		 *
		 * @author    Merijn Evertse & Sander Brandenburg
		 */
		public static class SearchResult implements Serializable {
			/**
			 * The name of the player.
			 */
			public String username;
			
			/**
			 * The channel the player is in at the moment, if any.
			 */
			public String currentchannel;
			
			/**
			 * The state of the player.
			 */
			public int userstate;
			
			/**
			 * The rating of the player.
			 */
			public int rating;
			
			/**
			 * The number of games the player has played.
			 */
			public int nrgames;
			
			/**
			 * If the player is in a channel, this indicated the state of that channel.
			 */ 
			public int gamestate;

			/**
	 		 * Initializes the object with the given parameters.
	 		 *
	 		 * @param    username        The name of the player
	 		 * @param    userstate       The state of the player
	 		 * @param    rating          The rating of the player
	 		 * @param    nrgames         The number of games the player has played
	 		 * @param    gamestate       The state of the channel, if the player is online
	 		 */
			public SearchResult(String username, int userstate, int rating, int nrgames, 
								String currentchannel, int gamestate) {
				this.username = username;
				this.currentchannel = currentchannel;
				this.userstate = userstate;
				this.rating = rating;
				this.nrgames = nrgames;
				this.gamestate = gamestate;
			}
			
			/**
			 * Overrides the toString in Message.
			 *
			 * @return    A description of the object
			 */ 
			public String toString() {
				return getClass().getName() + ": \"" + username + "\"";
			}
		}
		
		/**
		 * The list of players who match with the search request.
		 */
		public SearchResult[] result;
		
		/**
		 * Initializes the object with the given parameter.
		 *
		 * @param    result    The list of players who match with the search request
		 */
		public S2C_Result(SearchResult[] result) {
			this.result = result;
		}
	}
}
