package quest.global.game;

import java.util.Observable;
import java.util.Vector;
import java.util.Enumeration;
import quest.global.util.SerializableObservable;


/**
 * A backpack belongs to a user and contains all the treasures
 * he or she has picked up so far. Currently, treasures can only be
 * added to it. Observers are notified when this happens.
 *
 * @author Mathijs den Burger
 */

public class Backpack extends SerializableObservable implements Cloneable {

    private Vector treasures;
    private Treasure lastAddedTreasure;

    /**
     * Creates an empty backpack.
     */
    public Backpack() {
		treasures = new Vector();
		lastAddedTreasure = null;
    }

 	/**
 	 * Copy constructor used by clone()
 	 */
    protected Backpack(Backpack original) {
		treasures = new Vector();
		for (Enumeration e = original.treasures.elements(); e.hasMoreElements(); ) {
		    treasures.add(((Treasure)e.nextElement()).clone());
		}
    }

	/**
	 * Adds a treasure to the backpack. All observers are notified
	 * when this happens.
	 * @param treasure The treasure to add. If it's null, nothing is
	 *	added and no observers are notified.
	 */
    public void addTreasure(Treasure treasure) {
		if (treasure != null) {
			treasures.add(treasure);
			lastAddedTreasure = treasure;
			//System.out.println("DEBUG: treasure added to backpack, notifying " + countObservers() + " observers");
			setChanged();
			notifyObservers();
		}
    }

	/**
	 * Gives the treasure that was added last.
	 * @return The last added treasure, or null if no treasures are
	 *	added yet
	 */
	public Treasure getLastAddedTreasure() {
		return lastAddedTreasure;
	}

	/**
	 * Gives the number of treasures this backpack currently contains
	 * @return The number of treasures
	 */
    public int getNrTreasures() {
		return treasures.size();
    }

	/**
	 * Removes all treasures from this backpack
	 * @return All the treasures that were contained in this backpack
	 */
    public Treasure[] removeAllTreasures() {
		Treasure[] result = new Treasure[treasures.size()];
		for (int i = 0; i < treasures.size(); i++) {
		    result[i] = (Treasure)treasures.elementAt(i);
		}
		treasures.clear();
		setChanged();
		notifyObservers();
		return result;
    }


	/**
	 * Builds a copy of this object and all it's parts.
	 * @return The copy
	 */
	public Object clone() {
		return new Backpack(this);
    }

}
