package quest.global.game;

import quest.global.util.*;
import quest.client.game.GameController;	// only for the LOCAL_TEST_MODE variable
import java.util.Vector;
import java.util.Enumeration;
import java.util.Observer;
import java.util.Observable;
import javax.vecmath.Point3d;
import java.io.Serializable;

/**
 * <P>
 * This object contains the complete logical representation of a game.
 * It also keeps track of the game state of each user, which is modeled
 * as a FiniteStateMachine. By 'feeding' this object GameMove objects,
 * the game states change. By observing the logical game's parts,
 * objects that e.g. implement the visualisation of a game can
 * monitor various parts of the game and take action or animate
 * something as the game changes (a user walks, the game ends etc.)</P>
 * <P>
 * A game has two types of people that use it: users and spectators.
 * Users are the ones that are activily playing the game, and take turns to play.
 * Spectators are only observing the game, and will never take turns.
 * The number of users is limited, but the number of spectators can be as large
 * as you want.</P>
 * <P>
 * Each user gets a rank according to it's position in the list the game gets when
 * it is constructed. These ranks start at 0.</P>
 *
 * @see quest.global.game.GameMove
 * @see quest.global.util.FiniteStateMachine
 * @see quest.global.game.MazeCorridor3D
 *
 * @author Mathijs den Burger
 */
public class GameLogic implements Observer, Serializable, Cloneable {

    /**
     * The size of the games' board along the x-axis, measured in maze corridors.
     * This corresponds to the 'width' of a level of the games' board.
     */
    public static final int X_DIMENSION = 7;

    /**
     * The size of the games' board along the y-axis, measured in maze corridors.
     * This corresponds to the 'height' of a level of the games' board.
     */
    public static final int Y_DIMENSION = 7;

    /**
     * The size of the games' board along the z-axis, measured in maze corridors.
     * This corresponds to the number of levels the games' board consists of.
     */
    public static final int Z_DIMENSION = 2;

    /**
     * The index of the level which is dark and therefore not all corridors
     * in it can be seen.
     */
    public static final int DARK_LEVEL = 0;

    /**
     * The index of the level which is light and all corridors in it can be seen.
     */
    public static final int LIGHT_LEVEL = 1;

    /**
     * The maximum number of users that can participate in a game.
     */
    public static final int MAX_NR_USERS = 6;

    /**
     * A constant for user ranks, indicating that you're not an active player
     * (you're nobody in the game)
     */
    public static final int NOBODY = -1;

    /**
     * A constant for user ranks, indicating that it's currently unknown what
     * that rank should be.
     */
    public static final int UNKNOWN = -2;

    /**
     * A constant for user rank, indicating you are the game server
     */
    public static final int SERVER = -3;

    /**
     * Game state: indicates the game isn't started yet.
     */
    public static final int NOT_STARTED = 0;

    /**
     * Game state: indicates you can only observe the actions of other players.
     */
    public static final int OBSERVE = 1;

    /**
     * Game state: indicates that you can walk around in the maze an perform actions.
     */
    public static final int WALK = 2;

    /**
     * Game state: indicates that you can change the maze by inserting the free corridor shaft
     * into it.
     */
    public static final int CHANGE_MAZE = 3;

    /**
     * Game state: indicates that the rotation of the free corridors outside the maze has changed.
     */
    public static final int ROTATION_CHANGED = 4;

    /**
     * Game state: indicates that the free maze corridors are positioned for insertion
     */
    public static final int BLOCK_TO_INSERT_POSITIONED = 5;

    /**
     * Game state: indicates someone has won the game and the game is
     * finished.
     */
    public static final int GAME_WON = 6;

    /**
     * The number of game states.
     */
    private static final int NR_STATES = 7;

    /**
     * The rank of the user that may start
     */
    private static final int STARTING_RANK = 0;

    private static final int[][] GAME_STATE_TRANSITIONS = {
		{ OBSERVE, WALK },
		{ WALK, WALK },
		{ WALK, CHANGE_MAZE },
		{ CHANGE_MAZE, BLOCK_TO_INSERT_POSITIONED },
		{ BLOCK_TO_INSERT_POSITIONED, BLOCK_TO_INSERT_POSITIONED },
		{ BLOCK_TO_INSERT_POSITIONED, ROTATION_CHANGED },
		{ ROTATION_CHANGED, ROTATION_CHANGED },
		{ ROTATION_CHANGED, BLOCK_TO_INSERT_POSITIONED },
		{ BLOCK_TO_INSERT_POSITIONED, OBSERVE },
		{ ROTATION_CHANGED, OBSERVE },
		{ WALK, GAME_WON },
		{ OBSERVE, GAME_WON },
		{ GAME_WON, GAME_WON },
		// THE FOLLOWING TWO TRANSITIONS ARE ADDED TO ENABLE THE HACK TO PLAY WITH MULTIPLE USERS ON ONE MACHINE
		{ BLOCK_TO_INSERT_POSITIONED, WALK },
		{ ROTATION_CHANGED, WALK },
    };


    /**
     * Initial positions of the players in the maze, as factors of the maze size.
     */
    private static final Point3d[] RANK_POSITIONS = {
		new Point3d(0.0, 0.0, 1.0),		// rank 0
		new Point3d(1.0, 1.0, 1.0),		// rank 1
		new Point3d(1.0, 0.0, 1.0),		// rank 2
		new Point3d(0.0, 1.0, 1.0),		// rank 3
		new Point3d(0.5, 0.0, 1.0),		// rank 4
		new Point3d(0.5, 1.0, 1.0),		// rank 5
    };

    private FiniteStateMachine[] gameStates;
    private FiniteStateMachine myGameState;
    private GameUserList users;
    private SpectatorSet spectators;
    private int activeUserRank, myRank, winnerRank;
    private Maze3D maze;
    private GameMove lastGameMove;
    private int forbiddenInsertX, forbiddenInsertY;

    /**
     * Creates a new logical representation of an Amazing Quest game.
     * @param userNames The names of the people that will be playing (known as users).
     * The first user in this list may start.
     * @param spectatorNames The names of the people that will only observe the game (known as spectators)
     */
    public GameLogic(String[] userNames, String[] spectatorNames) {
    	// initialize the maze
    	this.maze = new Maze3D(X_DIMENSION, Y_DIMENSION, Z_DIMENSION);

	// generate the treasures
	Treasure[] treasures = generateTreasures(ExtraMath.clamp(userNames.length, 0, MAX_NR_USERS));

    	// initialize the users
    	if (userNames.length > MAX_NR_USERS) {
	    users = initUsers(userNames, MAX_NR_USERS, maze, treasures);
    	} else {
	    users = initUsers(userNames, userNames.length, maze, treasures);
	}
	users.addObserver(this);

	// add the treasures to the maze
	addTreasuresToMaze(treasures, maze);

	// initialize the spectators
	spectators = initSpectators(spectatorNames);

	// initialize the game states
	gameStates = new FiniteStateMachine[users.size()];
	for (int i = 0; i < users.size(); i++) {
	    gameStates[i] =  new FiniteStateMachine(NR_STATES, GAME_STATE_TRANSITIONS);
	}
	myGameState = new FiniteStateMachine(NR_STATES, GAME_STATE_TRANSITIONS);

	// initialize the various ranks
	this.myRank = UNKNOWN;
	activeUserRank = NOBODY;
	winnerRank = UNKNOWN;

	// initialize the last known game move
	lastGameMove = null;

	// initialze the forbidden insert coordinates
	forbiddenInsertX = forbiddenInsertY = MazePosition3D.UNKNOWN;
    }

	/**
	 * Copy constructor, used by clone()
	 */
    protected GameLogic(GameLogic original) {
	// copy the maze
	maze = (Maze3D)original.maze.clone();

	// copy the users. Note that this should be done after cloning the maze
	users = (GameUserList)(original.users.clone(maze));
	users.addObserver(this);
	for (Enumeration e = users.elements().elements(); e.hasMoreElements(); ) {
		((GameUser)e.nextElement()).addObserver(this);
	}

	// copy the spectators
	spectators = (SpectatorSet)original.spectators.clone();

	// copy the game states
	gameStates = new FiniteStateMachine[users.size()];
	for (int i = 0; i < users.size(); i++) {
	    gameStates[i] = (FiniteStateMachine)original.gameStates[i].clone();
	}
	myGameState = (FiniteStateMachine)original.myGameState.clone();

	// copy the various ranks
	activeUserRank = original.activeUserRank;
	myRank = original.myRank;
	winnerRank = original.winnerRank;

	// copy the last known game move
	if (original.lastGameMove != null) {
	    lastGameMove = (GameMove)original.lastGameMove.clone();
	} else {
	    lastGameMove = null;
	}

	// copy the forbidden insert coordinates
	forbiddenInsertX = original.forbiddenInsertX;
	forbiddenInsertY = original.forbiddenInsertY;
    }

    private GameUserList initUsers(String[] userNames, int number, Maze3D maze, Treasure[] treasures) {
	Vector v = new Vector(number);
	int nrTreasuresPerUser = (int)(treasures.length / number);
	for (int i = 0; i < number; i++) {
	    MazePosition3D userPosition = initUserPosition(i, maze);

	    // set treasures to find
	    Treasure[] treasuresToFind = new Treasure[nrTreasuresPerUser];
	    for (int j = 0; j < nrTreasuresPerUser; j++) {
		treasuresToFind[j] = treasures[(i * nrTreasuresPerUser) + j];
	    }

	    GameUser user = new GameUser(userNames[i], i, userPosition, new Mission(treasuresToFind));
	    user.addObserver(this);
	    v.add(user);
	}
	return new GameUserList(v);
    }

    private MazePosition3D initUserPosition(int rank, Maze3D maze) {
	int xPos = (int)(RANK_POSITIONS[rank].x * (X_DIMENSION - 1)),
	    yPos = (int)(RANK_POSITIONS[rank].y * (Y_DIMENSION - 1)),
	    zPos = (int)(RANK_POSITIONS[rank].z * (Z_DIMENSION - 1));
	return maze.getCorridor(xPos, yPos, zPos).getPosition();
    }

    private SpectatorSet initSpectators(String[] spectatorNames) {
    	Vector v = new Vector(spectatorNames.length);
    	for (int i = 0; i < spectatorNames.length; i++) {
	    v.add(new GameSpectator(spectatorNames[i]));
    	}
    	return new SpectatorSet(v);
    }

    private Treasure[] generateTreasures(int nrUsers) {
	int nrTreasures = (Treasure.MAX_NR_TREASURES / nrUsers) * nrUsers;
	Treasure[] result = Treasure.generateTreasures(nrTreasures);

	// shuffle treasures
	Treasure temp;
	int switchIndex;
	for (int i = 0; i < result.length; i++) {
	    temp = result[i];
	    switchIndex = (int)(Math.random() * result.length);
	    result[i] = result[switchIndex];
	    result[switchIndex] = temp;
	}

	return result;
    }

    private void addTreasuresToMaze(Treasure[] treasures, Maze3D maze) {
		Vector corridors = maze.getCorridorsThatCanContainTreasures();
		for (int i = 0; i < treasures.length; i++) {
		    int index = (int)(Math.random() * corridors.size());
		    MazeCorridor3D chosen = (MazeCorridor3D)corridors.get(index);
		    chosen.setTreasure(treasures[i], false);
		    treasures[i].setFindSpot(chosen);
		    corridors.remove(chosen);
		}
    }

    /**
     * Gives the current logical representation of the 3D maze.
     * @return The current logical representation of the 3D maze.
     */
    public Maze3D getMaze() {
    	return maze;
    }

    /**
     * Gives the number users of this game.
     * @return The number of users of this game.
     */
    public int getNrUsers() {
    	return users.size();
    }

    /**
     * Gives the number of spectators of this game.
     * @return The number of spectators of this game.
     */
    public int getNrSpectators() {
	return spectators.size();
    }

    /**
     * Gives the rank of the active user (the user that's currently
     * performing game moves).
     * @return The rank of the active user. This can be NOBODY, which indicates that
     * the game isn't started yet and therefore there is no active user.
     */
    public int getActiveUserRank() {
    	return activeUserRank;
    }

    /**
     * Gives your own rank.
     * @return Your own rank. This can be GameLogic.NOBODY, in
     * which case you are a spectator. It can also be set to GameLogic.UNKNOWN,
     * in which case your rank is not known yet because the game isn't started.
     */
    public int getMyRank() {
    	return myRank;
    }

    /**
     * Gives the rank of the game participant with the given name. This can be a
     * user or a spectator.
     * @param name The name of the participant.
     * @return The rank of the participant. If the name was of a user, this will
     * be a rank between 0 and the number of users - 1. If the name is that of a
     * spectator, the rank will be NOBODY. If the name is unknown, the rank will be
     * UNKNOWN.
     */
    public int getRank(String name) {
	// check if it is a user
	for (int i = 0; i < users.size(); i++) {
	    if (users.get(i).getName().equals(name)) return i;
	}
	// check if it is a spectator
	for (Enumeration e = spectators.elements().elements(); e.hasMoreElements(); ) {
	    if (((GameSpectator)e.nextElement()).getName().equals(name)) return NOBODY;
	}
	// name is unknown
	return UNKNOWN;
    }

    /**
     * Checks if the given name belongs to a user in this game. This only covers
     * the current users, not the ones that have left the game.
     * @return true if the name belongs to a current user in this game, false otherwise
     */
    public boolean isKnownUser(String name) {
	for (int i = 0; i < users.size(); i++) {
	    if (users.get(i).getName().equals(name)) return true;
	}
	return false;
    }

    /**
     * Checks if the given name belongs to a spectator in this game. This only covers
     * the current spectators, not the ones the have left the game.
     * @return true if the name belongs to a current spectator in this game, false otherwise.
     */
    public boolean isKnownSpectator(String name) {
	for (Enumeration e = spectators.elements().elements(); e.hasMoreElements(); ) {
	    if (((GameSpectator)e.nextElement()).getName().equals(name)) return true;
	}
	return false;
    }

    private boolean validRank(int rank) {
	return ((rank >= 0 && rank < users.size()) || rank == NOBODY || rank == SERVER);
    }

    /**
     * Gives the user with the given rank.
     * @return If the rank is valid, the user with the given rank, else null.
     */
    public GameUser getUser(int rank) {
    	if (!validRank(rank)) return null;
    	else return (GameUser)users.get(rank);
    }

    /**
     * Gives the list of spectators that are currently playing this game.
     * Modifying this list is immediately reflected in the game.
     * @return A vector of GameUser objects.
     */
    public GameUserList getUsers() {
    	return users;
    }

    /**
     * Gives the rank of the user that has won this game.
     * @return If someone has won this game: his rank<BR>
     *         If no-one has won this game: GameLogic.NOBODY<BR>
     *         else: GameLogic.UNKNOWN
     */
    public int getWinnerRank() {
		if (gameFinished()) {
		    winnerRank = activeUserRank;
		}
		return winnerRank;
    }

    /**
     * Gives the set of spectators that are currently observing this game.
     * Modifying this set is immediately reflected in the game.
     * @return A set of spectators that are currently observing this game.
     */
    public SpectatorSet getSpectators() {
    	return spectators;
    }

    /**
     * Gives the finite state machine that reflects the current game state
     * of a user with a given rank.
     * @param rank The rank of the user
     * @return The finite state machine that reflects the current game state.
     *		If the rank wasn't valid, null is returned.
     */
    public FiniteStateMachine getUserGameState(int rank) {
		if (rank < 0 || rank > users.size() - 1) {
		    System.err.println("WARNING: rank " + rank + " is not a valid player rank");
		    return null;
		} else {
		    return gameStates[rank];
		}
    }

    /**
     * Gives the finite state machine that reflects your current game
     * state.
     * @return The finite state machine.
     */
    public FiniteStateMachine getMyGameState() {
		return myGameState;
    }


    /**
     * Gives the last legal game move that was performed in this game.
     * This can be null, in which case no game moves have been performed yet.
     * @return The last game move that was performed in this game.
     */
    public GameMove getLastGameMove() {
    	return lastGameMove;
    }

    /**
     * Gives the maze position along the X-axis of the game in which no
     * corridor block may be inserted during the current turn. This is the
     * X-position next to the games board.
     * @return The position along the X-axis. This can be MazePosition3D.UNKNOWN,
     *			in which case there is no forbidden position (yet).
     * @see quest.global.game.MazePosition3D
     */
    public int getForbiddenInsertX() {
		return forbiddenInsertX;
    }

    /**
     * Gives the maze position along the Y-axis of the game in which no
     * corridor block may be inserted during the current turn. This is the
     * Y-position next to the games board.
     * @return The position along the Y-axis. This can be MazePosition3D.UNKNOWN,
     *			in which case there is no forbidden position (yet).
     * @see quest.global.game.MazePosition3D
     */
    public int getForbiddenInsertY() {
		return forbiddenInsertY;
    }

    /**
     * Performs the given move in this logical game, if possible.
     * @throws IllegalMoveException When the given move is not permitted.
     * @return true if the game is finished after this move<BR>
     *         false otherwise
     */
    public boolean doMove(GameMove move) throws IllegalMoveException {
    	boolean result;
		if (legalMove(move)) {
		    lastGameMove = move;
		    if (move instanceof GameMove.WalkTo) {
			doWalkTo((GameMove.WalkTo)move);
			//} else if (move instanceof GameMove.TeleportTo) {
			//	doTeleportTo((GameMove.TeleportTo)move);
		    } else if (move instanceof GameMove.PickUp) {
			doPickUp((GameMove.PickUp)move);
		    } else if (move instanceof GameMove.ChangeMaze) {
			doChangeMaze((GameMove.ChangeMaze)move);
		    } else if (move instanceof GameMove.PositionFreeCorridorBlock) {
			doPositionFreeCorridorBlock((GameMove.PositionFreeCorridorBlock)move);
		    } else if (move instanceof GameMove.RotateCorridorBlock) {
			doRotateCorridorBlock((GameMove.RotateCorridorBlock)move);
		    } else if (move instanceof GameMove.InsertFreeCorridorBlock) {
			doInsertFreeCorridorBlock((GameMove.InsertFreeCorridorBlock)move);
		    }
		} else {
		    throw new IllegalMoveException();
		}
		// check if the game is won
		result = gameFinished();
		if (result) handleGameWon();
		return result;
    }

    private boolean legalMove(GameMove move) {
    	if (move.getInitiatorRank() != activeUserRank) {
	    return false;
	} else if (move instanceof GameMove.WalkTo) {
	    return legalWalkTo((GameMove.WalkTo)move);
	} else if (move instanceof GameMove.PickUp) {
	    return legalPickUp((GameMove.PickUp)move);
	} else if (move instanceof GameMove.ChangeMaze) {
	    return legalChangeMaze((GameMove.ChangeMaze)move);
	} else if (move instanceof GameMove.PositionFreeCorridorBlock) {
	    return legalPositionFreeCorridorBlock((GameMove.PositionFreeCorridorBlock)move);
	} else if (move instanceof GameMove.RotateCorridorBlock) {
	    return legalRotateCorridorBlock((GameMove.RotateCorridorBlock)move);
	} else if (move instanceof GameMove.InsertFreeCorridorBlock) {
	    return legalInsertFreeCorridorBlock((GameMove.InsertFreeCorridorBlock)move);
	} else {	// unknown move type
	    return false;
	}
    }

    private boolean legalWalkTo(GameMove.WalkTo move) {
	if (gameStates[move.getInitiatorRank()].getCurrentState() != WALK) {
	    System.err.println("GameLogic.legalWalkTo: user " + move.getInitiatorRank() + " is not in WALK state but in state " + gameStates[move.getInitiatorRank()].getCurrentState());
	    return false;
	} else {
	    MazePosition3D activePos = ((GameUser)users.get(activeUserRank)).getEnvironmentPosition();
	    if (!maze.pathExistsBetween(activePos.getX(), activePos.getY(), activePos.getZ(),
					move.getX(), move.getY(), move.getZ())) {
		System.err.println("GameLogic.legalWalkTo: there exists no path between (" + activePos.getX() + ", " + activePos.getY() + ", " + activePos.getZ() + ") and (" + move.getX() + ", " + move.getY() + ", " + move.getZ() + ")");
		return false;
	    }
	}
	return true;
    }

    /*
      private boolean legalTeleportTo(GameMove.TeleportTo move) {
      if (gameState.getCurrentState() != OBSERVE &&
      gameState.getCurrentState() != WALK) {
      return false;
      } else {
      MazePosition3D activePos = ((GameUser)users.get(activeUserRank)).getEnvironmentPosition();
      MazePosition3D newPos = new MazePosition3D(move.getX(), move.getY(), move.getZ());
      if (activePos.edgeOfMaze() && newPos.edgeOfMaze() &&
      (activePos.getX() == newPos.getX() || activePos.getY() == newPos.getY())) {
      return true;
      } else if (maze.containsTeleport(activePos.getX(), activePos.getY()) &&
      activePos.getX() == newPos.getX() && activePos.getY() == newPos.getY()) {
      return true;
      } else {
      return false;
      }
      }
      }
    */

    private boolean legalPickUp(GameMove.PickUp move) {
	if (gameStates[move.getInitiatorRank()].getCurrentState() != WALK) {
	    return false;
	} else {
	    MazePosition3D activePos = ((GameUser)users.get(activeUserRank)).getEnvironmentPosition();
	    if (!maze.pathExistsBetween(activePos.getX(), activePos.getY(), activePos.getZ(),
					move.getX(), move.getY(), move.getZ())) {
		return false;
	    } else {
		MazeCorridor3D corridor = maze.getCorridor(move.getX(), move.getY(), move.getZ());
		if (!corridor.containsTreasure() ||
		    !users.get(activeUserRank).getMission().part(corridor.getTreasure()) ||
		    corridor.getTreasure().getColorId() != move.getColorId() ||
		    corridor.getTreasure().getSymbolId() != move.getSymbolId()) {
		    return false;
		}
	    }
	}
	return true;
    }

    private boolean legalChangeMaze(GameMove.ChangeMaze move) {
	return (gameStates[move.getInitiatorRank()].validNextState(CHANGE_MAZE));
    }

    private boolean legalPositionFreeCorridorBlock(GameMove.PositionFreeCorridorBlock move) {
	MazePosition3D pos = new MazePosition3D(move.getX(), move.getY(), 0);
	if (!gameStates[move.getInitiatorRank()].validNextState(BLOCK_TO_INSERT_POSITIONED)) {
	    return false;
	} else if (forbiddenInsertX != MazePosition3D.UNKNOWN && forbiddenInsertY != MazePosition3D.UNKNOWN &&
		   forbiddenInsertX == move.getX() && forbiddenInsertY == move.getY()) {
	    // trying to position the free corridors at the forbidden insert position
	    return false;
	} else if (pos.nextToMaze() &&
		   !maze.isSteadyX(move.getX()) && !maze.isSteadyY(move.getY())) {
	    return true;
	} else {
	    return false;
	}
    }

    private boolean legalRotateCorridorBlock(GameMove.RotateCorridorBlock move) {
		if (!ExtraMath.dividableBy(move.getRotation(), 90)) {
		    System.err.println("ERROR: legalRotateCorridorBlock, rotation of " + move.getRotation() + " is not a multiple of 90");
		    return false;
		} else if (gameStates[move.getInitiatorRank()].getCurrentState() != BLOCK_TO_INSERT_POSITIONED &&
			   gameStates[move.getInitiatorRank()].getCurrentState() != ROTATION_CHANGED) {
		    System.err.println("ERROR: legalRotateCorridorBlock, game state of initiator not valid (" + gameStates[move.getInitiatorRank()].getCurrentState() + ")");
		    return false;
		}
		return true;
    }

    private boolean legalInsertFreeCorridorBlock(GameMove.InsertFreeCorridorBlock move) {
	if (!gameStates[move.getInitiatorRank()].validNextState(WALK) &&
	    !gameStates[move.getInitiatorRank()].validNextState(OBSERVE)) {
	    return false;
	} else if (ExtraMath.clamp(forbiddenInsertX, 0, maze.getXDimension() - 1) == move.getX() &&
		   ExtraMath.clamp(forbiddenInsertY, 0, maze.getYDimension() - 1) == move.getY()) {
	    // trying to insert a corridor from the forbidden insertion position
	    return false;
	} else if (move.getX() == 0 || move.getX() == maze.getXDimension() - 1 ||
		   move.getY() == 0 || move.getY() == maze.getYDimension() - 1) {
	    // trying to insert a corridor in an edge
	    if (((move.getX() == 0 || move.getX() == maze.getXDimension() - 1)
		 && maze.isSteadyY(move.getY()))
		||
		((move.getY() == 0 || move.getY() == maze.getYDimension() - 1)
		 && maze.isSteadyX(move.getX()))
		) {
		// cannot insert corridor into a steady edge block
		return false;
	    } else {
		return true;
	    }
	} else {
	    return false;
	}
    }

    private boolean gameCanBeStarted(int myRank) {
	if(!validRank(myRank)){
	    System.err.println("ERROR: Not a valid rank: " + myRank);
	    return false;
	}
	if (myRank >= 0 && gameStates[myRank].getCurrentState() != FiniteStateMachine.NOT_STARTED) {
	    System.err.println("ERROR: my state machine is already started");
	    return false;
	}
	return true;
    }

    /**
     * Starts the game. If you are the user that may begin, you will be able to
     * do some moves now.
     * @param myRank Your rank. Use NOBODY if you're a spectator. Use
     * SERVER if you are the game server.
     */
    public void start(int myRank) {
		if (gameCanBeStarted(myRank)) {
		    this.myRank = myRank;
		    activeUserRank = STARTING_RANK;

		    // if you're NOBODY (a spectator) or the server,
		    // myGameState will be kept seperate from the users'
		    // game states. In these cases, start it in OBSERVE mode
		    if (myRank == NOBODY || myRank == SERVER) {
			myGameState.start(OBSERVE);
		    }

		    for (int i = 0; i < users.size(); i++) {
			// I'm a user, so my game state was seperate and should be in the correct
			// place in the game state array now
			if (i == myRank) myGameState = gameStates[i];

			// start the game state
			if (i == activeUserRank) {
			    gameStates[i].start(WALK);
			
			} else {
			    gameStates[i].start(OBSERVE);
			}
		    }
		} else {
		    System.out.println("WARNING: GameLogic.start: Cannot start the game in state " + gameStates[myRank].getCurrentState() + " with my rank " + myRank);
		}
    }

    private void doWalkTo(GameMove.WalkTo move) {
	MazePosition3D destination = maze.getCorridor(move.getX(), move.getY(), move.getZ()).getPosition();
	((GameUser)users.get(activeUserRank)).setEnvironmentPosition(destination, GameUser.WALK);
	gameStates[move.getInitiatorRank()].setState(WALK);
    }

    /*
      private void doTeleportTo(GameMove.TeleportTo move) {
      MazePosition3D destination = maze.getCorridor(move.getX(), move.getY(), move.getZ()).getPosition();
      ((GameUser)users.get(activeUserRank)).setEnvironmentPosition(destination, GameUser.TELEPORT);
      gameState.setState(WALK);
      }
    */

    private void doPickUp(GameMove.PickUp move) {
		MazeCorridor3D findSpot = maze.getCorridor(move.getX(), move.getY(), move.getZ());
		MazePosition3D destination = findSpot.getPosition();
		GameUser activeUser = (GameUser)(users.get(activeUserRank));
		if (!activeUser.getEnvironmentPosition().equals(destination)) {
		    // walk to the destination, pickup the treasure but don't notify the find spot about thiscorridor
		    activeUser.setEnvironmentPosition(destination, GameUser.PICKUP);
		    ((GameUser)users.get(activeUserRank)).pickupTreasure(findSpot, false);
		} else {
		    // you're already in the same corridor block as the treasure,
		    // so pick it up immediately and notify the find spot about this
		    ((GameUser)users.get(activeUserRank)).pickupTreasure(findSpot, true);
		}
		gameStates[move.getInitiatorRank()].setState(WALK);
    }

    private void doChangeMaze(GameMove.ChangeMaze move) {
	gameStates[move.getInitiatorRank()].setState(CHANGE_MAZE);
    }

    private void doPositionFreeCorridorBlock(GameMove.PositionFreeCorridorBlock move) {
	for (int z = 0; z < maze.getZDimension(); z++) {
	    maze.getFreeCorridor(z).getPosition().set(move.getX(), move.getY(), z);
	}
	gameStates[move.getInitiatorRank()].setState(BLOCK_TO_INSERT_POSITIONED);
    }

    private void doRotateCorridorBlock(GameMove.RotateCorridorBlock move) {
		for (int z = 0; z < maze.getZDimension(); z++) {
		    maze.getFreeCorridor(z).getOrientation().rotate(move.getRotation());
		}
		gameStates[move.getInitiatorRank()].setState(ROTATION_CHANGED);
    }

    private void doInsertFreeCorridorBlock(GameMove.InsertFreeCorridorBlock move) {
		// insert corridor block
		for (int z = 0; z < maze.getZDimension(); z++) {
		    maze.getFreeCorridor(z).getPosition().set(move.getX(), move.getY(), z);
		}

		// calculate the new forbidden insert position
		if (move.getX() == 0) {
		    forbiddenInsertX = maze.getXDimension();
		    forbiddenInsertY = move.getY();
		} else if (move.getX() == maze.getXDimension() - 1) {
		    forbiddenInsertX = -1;
		    forbiddenInsertY = move.getY();
		} else if (move.getY() == 0) {
		    forbiddenInsertX = move.getX();
		    forbiddenInsertY = maze.getYDimension();
		} else if (move.getY() == maze.getYDimension() - 1) {
		    forbiddenInsertX = move.getX();
		    forbiddenInsertY = -1;
		} else {
		    System.out.println("WARNING: illegal insert free corridor block position in GameLogic.doInsertFreeCorridorBlock: (" + move.getX() + ", " + move.getY());
		}

		// select the next user
		selectNextUser();
    }

    private void selectNextUser() {
		// let the stopped user become an observer
		gameStates[activeUserRank].setState(OBSERVE);

		// choose the next user
		do {
			activeUserRank = (activeUserRank + 1) % getNrUsers();
		} while (!users.get(activeUserRank).isPresent());

		// HACK TO ENABLE TESTING OF MULTIPLE PLAYERS ON ONE CLIENT:
		if (GameController.LOCAL_TEST_MODE) {
		    myRank = activeUserRank;
		    myGameState = gameStates[myRank];
		}
		// END HACK

		// enable the new user
		gameStates[activeUserRank].setState(WALK);
    }

    /**
     * Checks whether this game is finished
     * @return true if the game is finished, false otherwise
     */
    private boolean gameFinished() {
	if (gameStates[activeUserRank].getCurrentState() == GAME_WON) {
	    return true;
	} else {
	    for (Enumeration e = users.elements().elements(); e.hasMoreElements(); ) {
		if (((GameUser)e.nextElement()).getMission().accomplished()) {
		    // someone has completed it's mission, so the game has ended
		    return true;
		}
	    }
	    return false;
	}
    }

    private void handleGameWon() {
    	// set all game states to "GAME_WON",
	// set all missions of the losers to "failed"
	for (int i = 0; i < users.size(); i++) {
	    gameStates[i].setState(GAME_WON);
	    if (i != winnerRank) users.get(i).getMission().setFailed();
    	}
	myGameState.setState(GAME_WON);
    }

    /**
     * The update routine that is called when some Observable part
     * that is observed by this GameLogic changes. You should not call
     * this method directly.
     */
    public void update(Observable o, Object arg) {
		if (o instanceof GameUserList) {
			handleGameUserListUpdate();
		} else if (o instanceof GameUser && arg instanceof Integer) {
			handleGameUserUpdate((GameUser)o, (Integer)arg);
		}
	}
	
	private void handleGameUserListUpdate() {
		GameUser leftUser = users.getLastChangedPresentStatusUser();
		
    	// set his mission to be failed
    	leftUser.getMission().setFailed();	
		
    	// remove all his treasures that are still in the maze by picking
    	// them up manually
    	Treasure[] treasures = leftUser.getMission().getMissingTreasures();
    	for (int i = 0; i < treasures.length; i++) {
			leftUser.pickupTreasure(treasures[i].getFindSpot(), true);
    	}
	}
	
	private void handleGameUserUpdate(GameUser gameUser, Integer arg) {
	    if (arg.equals(GameUser.POSITION_OF_ENVIRONMENT_CHANGED)) {
		    // position of environment changed
			if (!gameUser.getEnvironmentPosition().inMaze()) {
			    // player is thrown out of the maze, teleport him to the other side
			    MazePosition3D newPosition = maze.getFreeCorridor(gameUser.getEnvironmentPosition().getZ()).getPosition();
			    gameUser.setEnvironmentPosition(newPosition, GameUser.TELEPORT);
			}
	    }
	}
	
	/**
	 * Builds a copy of this object and all it's parts.
	 * @return The copy.
	 */
    public Object clone() {
		return new GameLogic(this);
    }

}
