package quest.global.game;

import java.io.*;
import quest.global.net.*;

/**
 * Contains all game messages used for communication between client and server.
 * <PRE>
 * Protocol overview:
 *
 * NOTE: []'s indicate a selective send of the message, otherwise
 *            all clients receive the message.
 *       ()'s indicate an optional message which is sent only special cases.
 *       ...* indicates multiple values, by means of an array. 
 *
 * Client -&gt; Server packets
 *
 * Request       | Reply          | Arguments                    | Description
 * --------------------------------------------------------------------------------------------------
 * C2S_Login       S2C_Joinself     gamename                       Client logs into the game when game is not started
 *                 [S2C_Userjoin]                                  Userjoins only when client is a player
 *  or             S2C_InitGame                                    Spectator logs into a running game
 *                 [S2C_Userjoin]
 * C2S_logout      S2C_Leaveself    gamename                       Client logs out of the game
 *                 [S2C_Userleave]                                 Userleave only when game isn't ended yet
 * C2S_Gamemove    S2C_Gamemove     gamename, gamemove             Current player does a move
 * N/A             S2C_InitGame
 * N/A             S2C_AbortGame
 * N/A             [S2C_Userleave]                                 Client timed out, or logged out at chat
 *
 * Server -&gt; Client packets
 *
 * Request       | Reply          | Arguments                    | Description
 * --------------------------------------------------------------------------------------------------
 * S2C_Joinself    N/A              gamename, uname*, umode*       Client joins game, give info of other users
 * S2C_Userjoin    N/A              gamename, uname, umode         Client joins game, give users his info
 * S2C_Leaveself   N/A              gamename                       Client leaves game
 * S2C_Userleave   N/A              gamename, uname                Client leaves game, notify other users
 * S2C_Gamemove    N/A              gamename, gamemove             Propagate move to other users
 * S2C_InitGame    N/A              gamename, gamelogic            All players are online, start game with given bord
 *                                                                 or give bord to joining spectator on running game
 * S2C_AbortGame   N/A              gamename                       Game is aborted by admin
 *
 * @author    Merijn Evertse & Mathijs den Burger
 */
public class GameMessage extends Message implements Serializable {
  /**
   * Game client is player in this game.
   */
  public static final int USERMODE_PLAYER    = 1;

  /**
   * Game client is spectator in this game.
   */
  public static final int USERMODE_SPECTATOR = 2;

  /**
   * The name of the game.
   */
  public String gamename;
  
  /**
   * Initializes this object with the given parameter.
   *
   * @param    gamename    The name of the game
   */
  public GameMessage(String gamename) {
	this.gamename = gamename;
  }
  
  /* Client to server messages */

  /**
   * Client wants to login to the gamesystem.
   *
   * @author    Merijn Evertse
   */
  public static class C2S_Login extends GameMessage implements Serializable {
	/**
	 * Initializes the object with the given parameter.
	 *
	 * @param    gamename    The name of the game
	 */
	public C2S_Login(String gamename) {
	  super(gamename);
	}
  }

  /**
   * Client wants to logout of the gamesystem.
   *
   * @author    Merijn Evertse
   */
  public static class C2S_Logout extends GameMessage implements Serializable {
	/**
	 * Initializes the object with the given parameter.
	 *
	 * @param    gamename    The name of the game
	 */
	public C2S_Logout(String gamename) {
	  super(gamename);
	}
  }

  /**
   * Client does a move in the game.
   *
   * @author    Merijn Evertse
   */
  public static class C2S_Gamemove extends GameMessage implements Serializable {
	/**
	 * The game move.
	 */
	public GameMove gamemove;
	
	/**
	 * Initializes the object with the given parameter.
	 *
	 * @param    gamename    The name of the game
	 * @param    gamemove    The game move
	 */
	public C2S_Gamemove(String gamename, GameMove gamemove) {
	  super(gamename);
	  this.gamemove = gamemove;
	}
  }

  /* Server to client messages */
  
  /**
   * The server sents the joining client a join grant. The message also includes
   * the usernames and corresponding usermodes of all other users in the game.
   *
   * @author    Merijn Evertse
   */
  public static class S2C_Joinself extends GameMessage implements Serializable {
	/**
	 * The list of usernames.
	 */
	public String[] unames;
	
	/**
	 * The list of usermodes.
	 */
	public int[] umodes;
	
	/**
	 * Initializes the object with the given parameter.
	 *
	 * @param    gamename    The name of the game
	 * @param    unames      The list of usernames
	 * @param    umodes      The list of usermodes
	 */
	public S2C_Joinself(String gamename, String[] unames, int[] umodes) {
	  super(gamename);
	  this.unames = unames;
	  this.umodes = umodes;
	}
  }

  /**
   * The server sents all clients except the leaving client in the game a messages informing them
   * of the name and usermode of the joining client.
   *
   * @author    Merijn Evertse
   */
  public static class S2C_Userjoin extends GameMessage implements Serializable {
	/**
	 * The username of the joining client.
	 */
	public String uname;
	
	/**
	 * The usermode of the joining client.
	 */
	public int umode;
	
	/**
	 * Initializes the object with the given parameter.
	 *
	 * @param    gamename    The name of the game
	 * @param    uname       The username of the joining client
	 * @param    umode       The usermode of the joining client
	 */
	public S2C_Userjoin(String gamename, String uname, int umode) {
	  super(gamename);
	  this.uname = uname;
	  this.umode = umode;
	}
  }

  /**
   * The server sends the leaving client a leave grant.
   *
   * @author    Merijn Evertse
   */
  public static class S2C_Leaveself extends GameMessage implements Serializable {
	/**
	 * Initializes the object with the given parameter.
	 *
	 * @param    gamename    The name of the game
	 */
	public S2C_Leaveself(String gamename) {
	  super(gamename);
	}
  }

  /**
   * The server sends all clients except the leaving client in the game a message
   * informing them of the leaving client.
   *
   * @author    Merijn Evertse
   */
  public static class S2C_Userleave extends GameMessage implements Serializable {
	/**
	 * The username of the leaving client.
	 */
	public String uname;
	
	/**
	 * Initializes the object with the given parameter.
	 *
	 * @param    gamename    The name of the game
	 * @param    uname       The username of the leaving client
	 */
	public S2C_Userleave(String gamename, String uname) {
	  super(gamename);
	  this.uname = uname;
	}
  }

  /**
   * The server sends all clients the game move done by the currently playing client.
   * The client who did the move will also receive the message, as a check that the move
   * was valid and to keep all the clients as synchronous as possible.
   *
   * @author    Merijn Evertse
   */
  public static class S2C_Gamemove extends GameMessage implements Serializable {
	/**
	 * The game move.
	 */
	public GameMove gamemove;
	
	/**
	 * Initializes the object with the given parameter.
	 *
	 * @param    gamename    The name of the game
	 * @param    gamemove    The game move
	 */
	public S2C_Gamemove(String gamename, GameMove gamemove) {
	  super(gamename);
	  this.gamemove = gamemove;
	}
  }

	/**
	 * The servers sends all clients the message that the game can begin. All players
	 * are online so the game can be started. The gamelogic the server created is
	 * distributed with this message to all clients, ensuring that everyone has
	 * the same gamesituation.
	 *
	 * @author    Merijn Evertse
	 */
	public static class S2C_InitGame extends GameMessage implements Serializable {
		/**
		 * The gamelogic.
		 */
		public GameLogic gamelogic;
	
		/**
		 * Initializes the object with the given parameter.
		 *
		 * @param    gamename    The name of the game
		 * @param    gamelogic   The gamelogic
		 */
		public S2C_InitGame(String gamename, GameLogic gamelogic) {
	  		super(gamename);
	  		this.gamelogic = gamelogic;
		}
  	}

	/**
	 * The server sends all clients the message that the game is being aborted by the
	 * administrator of the game. 
	 * NOTE: All clients are required to do a normal logout procedure.
	 *
	 * @author    Merijn Evertse
	 */
	public static class S2C_AbortGame extends GameMessage implements Serializable {
		/**
		 * Initializes the object with the given parameter.
		 *
		 * @param    gamename    The name of the game
		 */
		public S2C_AbortGame(String gamename) {
			super(gamename);
		}
  	}
} 

