package quest.global.game;

import java.io.Serializable;

/**
 * Represents a move of a player in the games' board.
 * These objects are sent across the network and can be feeded
 * to a GameLogic object. Subclasses specify the exact move.
 *
 * @see quest.global.game.GameLogic
 * @author Mathijs den Burger
 */
public class GameMove implements Serializable, Cloneable {

  private int initiatorRank;

  /**
   * Constructs a game move.
   * @param initiatorRank Rank of the user that initiates this move
   */
  protected GameMove(int initiatorRank) {
	this.initiatorRank = initiatorRank;
  }

  /**
   * @return The initiator of this move
   */
  public int getInitiatorRank() {
	return initiatorRank;
  }

  /**
   * Builds a complete copy of this object.
   * @return The copy
   */
  public Object clone() {
	return new GameMove(initiatorRank);
  }

  // INNER CLASSES

	/**
	 * The move in which a user walks to a new position in the maze.
	 */
  public static class WalkTo extends GameMove implements Serializable, Cloneable {

	private int x, y, z;

	/**
	 * Creates a walk to move.
	 * @param initiatorRank The rank of the user that initiated this move
	 *	(which is the user that walks)
	 * @param x The X-coordinate of the destination of the walk
	 * @param y The Y-coordinate of the destination of the walk
	 * @param z The Z-coordinate of the destination of the walk
	 * @see quest.global.game.MazePosition3D
	 */
	public WalkTo(int initiatorRank, int x, int y, int z) {
	  super(initiatorRank);
	  this.x = x;
	  this.y = y;
	  this.z = z;
	}

	/**
	 * Gives the X-coordinate of the destination of the walk
	 * @return The X-coordinate
	 */
	public int getX() {
	  return x;
	}

	/**
	 * Gives the Y-coordinate of the destination of the walk
	 * @return The Y-coordinate
	 */
	public int getY() {
	  return y;
	}

	/**
	 * Gives the Z-coordinate of the destination of the walk
	 * @return The Z-coordinate
	 */
	public int getZ() {
	  return z;
	}

	/**
	 * Builds a copy of this move
	 * @return The copy
	 */
	public Object clone() {
		return new GameMove.WalkTo(super.initiatorRank, x, y, z);
	}

  }

	/**
	 * The move in which a user picks up a treasure in a certain maze corridor
	 */
	public static class PickUp extends GameMove implements Serializable, Cloneable {

		private int x, y, z, colorId, symbolId;

		/**
		 * Creates a pick up move.
		 * @param initiatorRank The rank of the user that initiated this move
		 *	(which is the user that picks up the treasure)
		 * @param x The X-coordinate of the corridor in which the treasure to pick up resides
		 * @param y The Y-coordinate of the corridor in which the treasure to pick up resides
		 * @param z The Z-coordinate of the corridor in which the treasure to pick up resides
		 * @param colorId The color ID of the treasure the user picks up
		 * @param symbolId The symbol ID of the treasure the users picks up
		 * @see quest.global.game.MazePosition3D
		 * @see quest.global.game.Treasure
		 */
		public PickUp(int initiatorRank, int x, int y, int z, int colorId, int symbolId) {
	  		super(initiatorRank);
	  		this.x = x;
	  		this.y = y;
	  		this.z = z;
	  		this.colorId = colorId;
	  		this.symbolId = symbolId;
		}

		/**
		 * Gives the X-coordinate of the corridor in which the treasure to pick up resides
		 * @return The X-coordinate
		 */
		public int getX() {
	  		return x;
		}

		/**
		 * Gives the Y-coordinate of the corridor in which the treasure to pick up resides
		 * @return The Y-coordinate
		 */
		public int getY() {
	  		return y;
		}

		/**
		 * Gives the Z-coordinate of the corridor in which the treasure to pick up resides
		 * @return The Z-coordinate
		 */
		public int getZ() {
	  		return z;
		}

		/**
		 * Gives the color ID of the treasure the user picks up
		 * @return The color ID
		 */
		public int getColorId() {
			return colorId;
		}

		/**
		 * Gives the symbol ID of the treasure the user picks up
		 * @return The symbol ID
		 */
		public int getSymbolId() {
			return symbolId;
		}

	/**
	 * Builds a copy of this move
	 * @return The copy
	 */
		public Object clone() {
	  		return new GameMove.PickUp(super.initiatorRank, x, y, z, colorId, symbolId);
		}

  	}


	/**
	 * The move in which a user teleports to a new position in the maze.
	 */
  public static class TeleportTo extends GameMove implements Serializable, Cloneable {

	private int x, y, z;

	/**
	 * Creates a teleport to move.
	 * @param initiatorRank The rank of the user that initiated this move
	 *	(which is the user that teleports)
	 * @param x The X-coordinate of the destination of the teleport
	 * @param y The Y-coordinate of the destination of the teleport
	 * @param z The Z-coordinate of the destination of the teleport
	 * @see quest.global.game.MazePosition3D
	*/
	public TeleportTo(int initiatorRank, int x, int y, int z) {
	  super(initiatorRank);
	  this.x = x;
	  this.y = y;
	  this.z = z;
	}

	/**
	 * Gives the X-coordinate of the destination of the teleport
	 * @return The X-coordinate
	 */
	public int getX() {
	  return x;
	}

	/**
	 * Gives the Y-coordinate of the destination of the teleport
	 * @return The Y-coordinate
	 */
	public int getY() {
	  return y;
	}

	/**
	 * Gives the Z-coordinate of the destination of the teleport
	 * @return The Z-coordinate
	 */
	public int getZ() {
	  return z;
	}

	/**
	 * Builds a copy of this move
	 * @return The copy
	 */
	public Object clone() {
	  return new GameMove.TeleportTo(super.initiatorRank, x, y, z);
	}

  }

	/**
	 * The move in which a user has stopped walking around and
	 * is going to change the maze now.
	 */
  public static class ChangeMaze extends GameMove implements Serializable, Cloneable {

	/**
	 * Creates an change maze move.
	 * @param initiatorRank The rank of the user that initiated this move
	 *	(which is the user that's going to the change the maze)
	 */
	public ChangeMaze(int initiatorRank) {
	  super(initiatorRank);
	}

	/**
	 * Builds a copy of this move
	 * @return The copy
	 */
	public Object clone() {
	  return new GameMove.ChangeMaze(super.initiatorRank);
	}
  }

	/**
	 * The move in which a user positions the free corridor blocks
	 * somewhere next to the maze.
	 */
  public static class PositionFreeCorridorBlock extends GameMove implements Serializable, Cloneable {

	private int x, y;

	/**
	 * Creates a position free corridor block move.
	 * @param initiatorRank The rank of the user that initiated this move
	 *	(which is the user that positiones the free corridor blocks)
	 * @param x The X-coordinate of the position of the positioned corridor blocks
	 * @param y The Y-coordinate of the position of the positioned corridor blocks
	 */
	public PositionFreeCorridorBlock(int initiatorRank, int x, int y) {
	  super(initiatorRank);
	  this.x = x;
	  this.y = y;
	}

	/**
	 * Gives the X-coordinate of the position the user positions the
	 * free corridor blocks at.
	 * @return The X-coordinate
	 */
	public int getX() {
	  return x;
	}

	/**
	 * Gives the Y-coordinate of the position the user positions the
	 * free corridor blocks at.
	 * @return The Y-coordinate
	 */
	public int getY() {
	  return y;
	}

	/**
	 * Builds a copy of this move
	 * @return The copy
	 */
	public Object clone() {
	  return new GameMove.PositionFreeCorridorBlock(super.initiatorRank, x, y);
	}

  }


	/**
	 * The move in which a user inserts the free corridor blocks
	 * he has positioned next to the maze into the maze.
	 */
  public static class InsertFreeCorridorBlock extends GameMove implements Serializable, Cloneable {

	private int x, y;

	/**
	 * Creates an insert free corridor block move.
	 * @param initiatorRank The rank of the user that initiated this move
	 *	(which is the user that inserts the free corridor blocks)
	 * @param x The X-coordinate of the position the inserted free corridor blocks will have.
	 * @param y The Y-coordinate of the position the inserted free corridor blocks will have.
	 */
	public InsertFreeCorridorBlock(int initiatorRank, int x, int y) {
	  super(initiatorRank);
	  this.x = x;
	  this.y = y;
	}

	/**
	 * Gives the X-coordinate of the position the inserted free corridor blocks will have.
	 * @return The X-coordinate
	 */
	public int getX() {
	  return x;
	}

	/**
	 * Gives the Y-coordinate of the position the inserted free corridor blocks will have.
	 * @return The Y-coordinate
	 */
	public int getY() {
	  return y;
	}

	/**
	 * Builds a copy of this move
	 * @return The copy
	 */
	public Object clone() {
	  return new GameMove.InsertFreeCorridorBlock(super.initiatorRank, x, y);
	}

  }

	/**
	 * The move in which the free corridor blocks next to the maze
	 * are rotated.
	 */
  public static class RotateCorridorBlock extends GameMove implements Serializable, Cloneable {

	private int rotation;

	/**
	 * Creates a rotate corridor block move
	 * @param initiatorRank The rank of the user that initiated this move
	 *	(which is the user that rotates the free corridor blocks)
	 * @param rotation The number of degrees to rotate the free corridor blocks by
	 */
	public RotateCorridorBlock(int initiatorRank, int rotation) {
	  super(initiatorRank);
	  this.rotation = rotation;
	}

	/**
	 * Gives the number of degrees to rotate the free corridor blocks by
	 * @return The number of degrees
	 */
	public int getRotation() {
	  return rotation;
	}

	/**
	 * Builds a copy of this move
	 * @return The copy
	 */
	public Object clone() {
	  return new GameMove.RotateCorridorBlock(super.initiatorRank, rotation);
	}

  }


}
