package quest.global.game;

import java.util.Observer;
import java.util.Observable;
import quest.global.util.SerializableObservable;

/**
 * <P>
 * The logical representation of a user in an Amazing Quest game.
 * A user is someone who actually plays the game and takes turns
 * with other players.</P>
 * <P>
 * Users have a name, a rank (for taking turns), a position in the
 * maze, a mission and a backpack with the treasures they've picked up.
 * Each user user also has an 'present' status, which indicates if he's
 * still present in the game (playing or observing) that he has left the game.</P>
 * <P>
 * A user has to be able to move along with a moving environment (the
 * corridor he's currently in). Therefore he maintains a pointer to
 * the MazePosition3D object of the corridor block he's currently in.
 * He observes this position, and if it changes e notifies it's observers
 * with the argument POSITION_OF_ENVIRONMENT_CHANGED. This gives observers
 * the change to e.g. move to representation of the player along with that
 * of the environment.</P>
 * <P>
 * A game user can walk trough the maze by setting a new environment position.
 * He notifies observers of this change with the argument NEW_ENVIRONMENT_POSITION.</P>
 * <P>
 * When the 'active' status of a game user changes, he notifies all observers
 * with the argument ACTIVE_STATUS_CHANGED.</P>
 *
 * @author Mathijs den Burger
 * @version 1.0
 * @see quest.global.game.MazePosition3D
 */

public class GameUser extends SerializableObservable implements Selectable, Observer, Cloneable {

    /**
     * Argument given to observers when the position of the environment
     * has changed. This means the user moved because the environment did.
     */
    public static final Integer POSITION_OF_ENVIRONMENT_CHANGED = new Integer(0);

    /**
     * Argument given to observers when a new environment position has been
     * set for this user. This means the user has moved himself to that new position.
     */
    public static final Integer NEW_ENVIRONMENT_POSITION = new Integer(1);

    /**
     * Argument given to observers when the 'present' status of this user has changed.
     * This means the user has either left the game while he was playing, or just joined
     * the game.
     */
    public static final Integer PRESENT_STATUS_CHANGED = new Integer(2);

    /**
     * Move type argument for setEnvironmentPosition, indicating that
     * the user has walked to the new position n the maze.
     */
    public static final int WALK = 0;

    /**
     * Move type argument for setEnvironmentPosition, indicating that
     * the user has teleported to the new position in the maze.
     */
    public static final int TELEPORT = 1;

    /**
     * Move type argument for setEnvironmentPosition, indicating that
     * the user has walked to the new position in the maze and picked
     * up the treasure that lies in the corridor he walked to.
     */
    public static final int PICKUP = 2;

    private static final int MOVE_TYPE_IF_ENVIRONMENT_POSITION_CHANGES = WALK;

    private String name;			// name of this user
    private int rank;				// rank of this user (for taking turns) starting at 0
    private int lastMoveType;
    private Mission mission;
    private boolean isPresent;

    private MazePosition3D environmentPosition,
		oldEnvironmentPosition;

    private SelectionState selectionState;
    private Backpack backpack;

    /**
     * Creates a new game user. This user is marked as present.
     * @param name The name of the user
     * @param rank The rank of the user
     * @param environmentPosition The position of the user in the maze.
     *	This should be the same object as used by the MazeCorridor3D objects
     *	the maze consists of.
     * @param mission The mission of this user
     */
    public GameUser(String name, int rank, MazePosition3D environmentPosition, Mission mission) {
    	this.name = name;
    	this.rank = rank;
    	this.environmentPosition = oldEnvironmentPosition = environmentPosition;
    	this.mission = mission;
    	this.isPresent = true;
    	environmentPosition.addObserver(this);
    	selectionState = new SelectionState();
    	lastMoveType = TELEPORT;
    	this.backpack = new Backpack();
    	backpack.addObserver(this);
    }

	/**
	 * Copy constructor used by clone()
	 */
    protected GameUser(GameUser original, MazePosition3D environmentPosition, MazePosition3D oldEnvironmentPosition) {
		name = new String(original.name);
		rank = original.rank;
		this.environmentPosition = environmentPosition;
		environmentPosition.addObserver(this);
		this.oldEnvironmentPosition = oldEnvironmentPosition;
		mission = (Mission)original.mission.clone();
		isPresent = original.isPresent;
		selectionState = (SelectionState)original.selectionState.clone();
		lastMoveType = original.lastMoveType;
		backpack = (Backpack)original.backpack.clone();
		backpack.addObserver(this);
    }

	/**
	 * Gives the name of this user
	 * @return The name of this user
	 */
    public String getName() {
		return name;
    }

	/**
	 * Gives the rank of this user
	 * @return The rank of this user
	 */
    public int getRank() {
    	return rank;
    }
    
    /**
     * Checks if this user is still present (playing the game or observing), 
     * or that he has  left the game.
     * @return true if this user is still present, false if he has left the game.
     */
    public boolean isPresent() {
    	return isPresent;
    }
    
    /**
     * Sets the 'present' status of this user.
     * @param active Use true to indicate that this user is still present in the
     * game (activily playing it or observing ), or false to indicate that he 
     * has left the game.
     */
    public void setPresent(boolean isPresent) {
		if (this.isPresent != isPresent) {
    		this.isPresent = isPresent;
			setChanged();
		    notifyObservers(PRESENT_STATUS_CHANGED);
		}
    }

	/**
	 * Gives the current position of the environment (the maze
	 * 	corridor) the user is currently in.
	 * @return The current position of the environment
	 */
    public MazePosition3D getEnvironmentPosition() {
    	return environmentPosition;
    }

	/**
	 * Gives the previous position of the environment (the maze
	 *	corridor) the user was in.
	 * @return The previous position
	 */
    public MazePosition3D getOldEnvironmentPosition() {
    	return oldEnvironmentPosition;
    }

	/**
	 * Gives the mission of this user
	 * @return The mission of this user
	 */
    public Mission getMission() {
    	return mission;
    }

	/**
	 * Gives the type of the last move this user performed. This can
	 *  be WALK, TELEPORT of PICKUP, which indicates how the user
	 *  moves from the old environment position to the current one.
	 * @return The type of the last move.
	 */
    public int lastMoveType() {
    	return lastMoveType;
    }

	/**
	 * Gives the backpack of this user
	 * @return The backpack of this user
	 */
    public Backpack getBackpack() {
		return backpack;
    }

	/**
	 * Sets the new environment position of this user. This should be
	 * the same MazePosition3D object as that of the corridor the user
	 * is in. After setting the new position all observers are notified
	 * with the argument NEW_ENVIRONMENT_POSITION. If the given
	 * environment position is the same as the current one, nothing
	 * happens.
	 * @param position The new environment position
	 * @param moveType The type of movement from the current position
	 *	to the new (given) one. This can be WALK, TELEPORT of PICKUP.
	 */
    public void setEnvironmentPosition(MazePosition3D position, int moveType) {
    	if (!environmentPosition.equals(position)) {
		    oldEnvironmentPosition = environmentPosition;
		    oldEnvironmentPosition.deleteObserver(this);
		    environmentPosition = position;
		    environmentPosition.addObserver(this);
		    lastMoveType = moveType;
		    setChanged();
		    notifyObservers(NEW_ENVIRONMENT_POSITION);
		}
    }

	/**
	 * Let the user pick up the treasure at the given find spot, and
	 * add that treasure to it's backpack. Note that users can pick up
	 * treasures anyware in the games' board, so it's the game's responsibility
	 * to move the user to the find spot first if that the desired behavior.
	 * @param findSpot The corridor that contains the treasure to pick up
	 * @param notifyCorridorObservers If true, the observers of the find spot
	 * 	will be notified that the treasure has been removed from it. If false,
	 * 	this won't happen.
	 */
    public void pickupTreasure(MazeCorridor3D findSpot, boolean notifyCorridorObservers) {
		backpack.addTreasure(findSpot.setTreasure(null, notifyCorridorObservers));
    }

    private void checkMoveType(int type) {
		if (type != WALK && type != TELEPORT)
		    throw new RuntimeException("Unknown move type in GameUser.checkMoveType: " + type);
    }

    /**
     * Gives the selection state of this user
     * @return The selection state of this user
     */
    public SelectionState getSelectionState() {
		return selectionState;
    }

    /**
     * Checks if this user is equal to the given one.
     * Two users are the same if they have the same name.
     * @param rhs The other game user
     * @return true if this user is equal to the given one, false otherwise.
     */
    public boolean equals(GameUser rhs) {
		return (name == rhs.name);
    }

	/**
     * The update routine that is called when some Observable part
     * that is observed by this GameUser changes. You should not call
     * this method directly.
     */
    public void update(Observable o, Object arg) {
		if (o instanceof MazePosition3D) {
		    // environment position changed
		    lastMoveType = MOVE_TYPE_IF_ENVIRONMENT_POSITION_CHANGES;
		    setChanged();
		    notifyObservers(POSITION_OF_ENVIRONMENT_CHANGED);
		} else if (o instanceof Backpack) {
		    // treasure added to the backpack, add it to your mission
		    //System.out.println("DEBUG: gameuser is notified of added treasure to backpack and changes it mission");
		    mission.found(backpack.getLastAddedTreasure());
		}
    }

	/**
	 * Dummy method, you should use clone(MazePosition3D, MazePosition3D) instead.
	 * Invoking this method will cause a RuntimeException.
	 */
    public Object clone() {
		throw new RuntimeException("WARNING: you're using GameUser.clone(). Use GameUser.clone(MazePosition3D, MazePosition3D) instead.");
    }

	/**
	 * Builds a copy of this object and all it's parts. Since cloning
	 * a game user can be part of an clone action that clones the
	 * maze corridors too, the old en current environment position
	 * have to be supplied to ensure the cloned user will be living in
	 * the cloned corridors.
	 * @param environmentPosition The position of the environment the clone will be in.
	 * @param oldEnvironmentPosition The position of the environment the clone was in.
	 * @return The copy.
	 */
    public Object clone(MazePosition3D environmentPosition, MazePosition3D oldEnvironmentPosition) {
		return new GameUser(this, environmentPosition, oldEnvironmentPosition);
    }

}
