package quest.global.game;

import java.util.*;
import quest.global.util.SerializableObservable;
import quest.global.util.ExtraMath;

/**
 * Maintains a list of game users, which get all the elements at
 * construction time. Once the list is created, it can't be changed.
 * When the 'present' status of one of the contained game users changes,
 * all observers are notified.
 *
 * @see quest.global.game.GameUser
 * @author Mathijs den Burger
 */

public class GameUserList extends SerializableObservable implements Observer, Cloneable {

    private LinkedList gameUsers;
	private GameUser lastChangedPresentStatusUser = null;

   	/**
     * Constructs a list that contains the game users
     * @param gameUsers The users the list will contain
     */
    public GameUserList(Vector gameUsers) {
		this.gameUsers = new LinkedList(gameUsers);
		for (Enumeration e = gameUsers.elements(); e.hasMoreElements(); ) {
			((GameUser)e.nextElement()).addObserver(this);
		}
    }

	/**
	 * Copy constructor used by clone()
	 */
    protected GameUserList(GameUserList original, Maze3D environment) {
    	MazePosition3D userPos, userPosOld, mazePos, mazePosOld;
    	GameUser user;

		gameUsers = new LinkedList();
		for (ListIterator li = original.gameUsers.listIterator(0); li.hasNext(); ) {
		    user = (GameUser)li.next();
		    userPos = user.getEnvironmentPosition();
		    mazePos = environment.getCorridor(userPos.getX(), userPos.getY(), userPos.getZ()).getPosition();
		    userPosOld = user.getOldEnvironmentPosition();
		    mazePosOld = environment.getCorridor(userPosOld.getX(), userPosOld.getY(), userPosOld.getZ()).getPosition();
			GameUser clonedUser = (GameUser)user.clone(mazePos, mazePosOld);
		    this.gameUsers.add((GameUser)clonedUser);
		    clonedUser.addObserver(this);
		}
    }

	/**
	 * Gives the last user of which the 'present' status changed.
	 * @return The last user of which the 'present' status changed, or null
	 * if no 'present' status has changed yet.
	 */
	public GameUser getLastChangedPresentStatusUser() {
		return lastChangedPresentStatusUser;
	}

	/**
	 * Gives the number of game users in the list
	 * @return The number of game users in the list
	 */
    public int size() {
		return gameUsers.size();
    }

	/**
	 * Gives the game user at the given index in the list
	 * If the given index is out of range it is clamped.
	 * @param The index of the game user, 0 <= index < size()
	 * @return The game user at the given index
	 * @see quest.global.util.ExtraMath#clamp
	 */
    public GameUser get(int index) {
		return (GameUser)gameUsers.get(ExtraMath.clamp(index, 0, size()));
    }

	/**
	 * Gives all the elements in the list as a vector.
	 * @return All the elements of the list
	 */
    public Vector elements() {
		Vector result = new Vector(gameUsers.size());
		for (ListIterator i = gameUsers.listIterator(0); i.hasNext(); ) {
		    result.add(i.next());
		}
		return result;
    }

	/**
	 * This method is called whenever the 'active' field of one of the contained
	 * game users changed. Applications shouldn't call this method directly.
	 */
	public void update(Observable o, Object arg) {
		if (o instanceof GameUser && arg.equals(GameUser.PRESENT_STATUS_CHANGED)) {
			lastChangedPresentStatusUser = (GameUser)o;
			setChanged();
			notifyObservers();
		}
	}

	/**
	 * Dummy method, you should use clone(Maze3D) instead.
	 * Invoking this method will cause a RuntimeException.
	 */
    public Object clone() {
		throw new RuntimeException("WARNING: you're using GameUserList.clone(). Use GameUserList.clone(Maze3D) instead");
    }

	/**
	 * Builds a copy of this object. Because the game users need to
	 * have the maze position of the corridors there in, and that
	 * maze may be cloned to, the maze the cloned game users this
	 * list will live in has to be supplied too.
	 * @param maze The maze the cloned game users in this list will
	 * 	live in
	 * @return The copy
	 * @see quest.global.game.GameUser#clone(quest.global.game.MazePosition3D, quest.global.game.MazePosition3D)
	 */
    public Object clone(Maze3D maze) {
		return new GameUserList(this, maze);
    }

}
