package quest.global.game;

import quest.global.util.*;
import java.util.*;

/**
 * <P>
 * The logical representation of the 3D maze that forms the games' board.
 * An random maze of is generated with the given sizes when this class
 * is instantiated. It maintains a 3D grid of MazeCorridor3D objects that
 * together form the maze.</P>
 * <P>
 * The structure of the maze can be seen as several levels (in the
 * Z-direction) of which each level has the same size in X and
 * Y direction. Each level has two types of corridors: steady and
 * nonsteady corridors. If a corridor at a certain (x, y) position
 * is steady can be checked with the isSteadyX() en isSteadyY() methods.
 * Only the corridors of which both the X and Y coordinate are steady
 * are steady themself. The idea is that of steady corridors the position
 * can't be changed, but it is the responsibility of the classes that
 * use a Maze3D to control this.
 * The layout of the steady and nonsteady corridors is the same
 * at each level. </P>
 * <P>
 * The type of the corridors that are fixed is also fixed. See the
 * manual of Amazing Quest for a picture of this fixed layout.</P>
 * <P>
 * This class also contain some methods that deal with "maze logic":
 * there are methods that check if a certain corridor
 * is reachable from some other point in the maze, and ones that
 * return all reachable corridors. The information about the current layout of
 * the corridors is stored for as long as the maze structure isn't changed,
 * so multiple calls to these methods will only require one 'expensive' search
 * to determine the layout of the corridors.</P>
 * <P>
 * The maze observes the positions of all the MazeCorridor3D objects
 * the maze consists of, and adapts it's view of the maze layout
 * whenever the structure of the maze corridors changes.</P>
 *
 * @author Mathijs den Burger
 * @version 1.0
 */

public class Maze3D extends SerializableObservable implements Observer, Cloneable {

    private static final int X = 0, Y = 1, Z = 2;
    private static final int[][] TELEPORT_SHAFTS;				// (x, y) combinations of staircase positions
    private static final int[][] STEADY_TJUNCTION_ROTATIONS;	// (x, y, rotation) combinations of steady t-junctions in the middle
	private static final int[] SHRINE = new int[3];				// (x, y, z) coordinate of the shrine (in maze corridors)

	// initialize maze layout, depending on it's size
	static {
		if (GameLogic.X_DIMENSION == 7 && GameLogic.Y_DIMENSION == 7) {
			// 7x7
			TELEPORT_SHAFTS = new int[4][2];
			initTeleportShaft(0, 2, 2);
			initTeleportShaft(1, 4, 2);
			initTeleportShaft(2, 2, 4);
			initTeleportShaft(3, 4, 4);
			STEADY_TJUNCTION_ROTATIONS = new int[4][3];
			initSteadyTJunctionRotation(0, 2, 2, 270);
			initSteadyTJunctionRotation(1, 4, 2, 180);
			initSteadyTJunctionRotation(2, 2, 4, 0);
			initSteadyTJunctionRotation(3, 4, 4, 90);
			initShrine(3, 3, 0);
		} else if (GameLogic.X_DIMENSION == 9 && GameLogic.Y_DIMENSION == 9) {
			// 9x9
			TELEPORT_SHAFTS = new int[4][2];
			initTeleportShaft(0, 2, 2);
			initTeleportShaft(1, 2, 6);
			initTeleportShaft(2, 6, 2);
			initTeleportShaft(3, 6, 6);
			STEADY_TJUNCTION_ROTATIONS = new int[4][3];
			initSteadyTJunctionRotation(0, 4, 2, 90);
			initSteadyTJunctionRotation(1, 6, 4, 0);
			initSteadyTJunctionRotation(2, 4, 6, 270);
			initSteadyTJunctionRotation(3, 2, 4, 180);
			initShrine(4, 4, 0);
		} else {
			throw new RuntimeException("Maze3D: games' board size not supported: " + GameLogic.X_DIMENSION + "x" + GameLogic.Y_DIMENSION);
		}
	}

    private static final double[] RANDOM_CORRIDOR_CHANGES = {
	0,			// change to select a CLOSED
	0,			// change to select a DEAD_END
	1.0 / 3.0,	// change to select a STRAIGHT
	1.0 / 3.0,	// change to select a CORNER
	1.0 / 3.0,	// change to select a TJUNCTION
	0			// change to select a CROSSING
    };

    private static Random randomGenerator = new Random();
    private int xDimension, yDimension, zDimension;
    private MazeCorridor3D[][][] maze;
    private MazeCorridor3D[] freeCorridors;

    // variables for path-calculation in the maze
    private int[][][] corridorIndexMatrix;
    private boolean corridorIndexMatrixValid;

	/**
	 * Constructs a 3D maze of the given size, and initializes
	 * it with random corridors.
	 * @param xDimension The size of the maze along the X-axis.
	 *	This corresponds to one of the dimensions of a level
	 * @param yDimension The size of the maze along the Y-axis.
	 *  This corresponds to the other dimension of a level
	 * @param zDimension The size of the maze along the Z-axis
	 * 	This corresponds to the number of levels of the maze
	 */
    Maze3D(int xDimension, int yDimension, int zDimension) {
		this.xDimension = xDimension;
		this.yDimension = yDimension;
		this.zDimension = zDimension;
		initMaze();
		initFreeCorridors();
		corridorIndexMatrix = new int[xDimension][yDimension][zDimension];
		corridorIndexMatrixValid = false;
    }

	/**
	 * Copy constructor used by clone()
	 * @see quest.global.game.Maze3D#clone()
	 */
    protected Maze3D(Maze3D original) {
	// copy maze dimensions
	xDimension = original.xDimension;
	yDimension = original.yDimension;
	zDimension = original.zDimension;
	// copy the maze
	maze = new MazeCorridor3D[xDimension][yDimension][zDimension];
	for (int x = 0; x < xDimension; x++) {
	    for (int y = 0; y < yDimension; y++) {
		for (int z = 0; z < zDimension; z++) {
		    maze[x][y][z] = (MazeCorridor3D)(original.maze[x][y][z].clone());
		    maze[x][y][z].getPosition().addObserver(this);
		    //maze[x][y][z].getOrientation().addObserver(this);
		}
	    }
	}
	// copy the free corridors
	freeCorridors = new MazeCorridor3D[zDimension];
	for (int z = 0; z < zDimension; z++) {
	    freeCorridors[z] = (MazeCorridor3D)(original.freeCorridors[z].clone());
	    freeCorridors[z].getPosition().addObserver(this);
	}
	// copy the corridor index matrix
	corridorIndexMatrix = new int[xDimension][yDimension][zDimension];
	for (int x = 0; x < xDimension; x++) {
	    for (int y = 0; y < yDimension; y++) {
		for (int z = 0; z < zDimension; z++) {
		    corridorIndexMatrix[x][y][z] = original.corridorIndexMatrix[x][y][z];
		}
	    }
	}
	corridorIndexMatrixValid = original.corridorIndexMatrixValid;
    }

	private static void initTeleportShaft(int index, int x, int y) {
		TELEPORT_SHAFTS[index][0] = x;
		TELEPORT_SHAFTS[index][1] = y;
	}

	private static void initSteadyTJunctionRotation(int index, int x, int y, int rotation) {
		STEADY_TJUNCTION_ROTATIONS[index][0] = x;
		STEADY_TJUNCTION_ROTATIONS[index][1] = y;
		STEADY_TJUNCTION_ROTATIONS[index][2] = rotation;
	}

	private static void initShrine(int x, int y, int z) {
		SHRINE[0] = x;
		SHRINE[1] = y;
		SHRINE[2] = z;
	}

    private void initMaze() {
    	maze = new MazeCorridor3D[xDimension][yDimension][zDimension];
	for (int x = 0; x < xDimension; x++) {
	    for (int y = 0; y < yDimension; y++) {
		for (int z = 0; z < zDimension; z++) {
		    maze[x][y][z] = getInitCorridor(x, y, z);
		}
	    }
	}
    }

    private void initFreeCorridors() {
	freeCorridors = new MazeCorridor3D[zDimension];
	for (int z = 0; z < zDimension; z++) {
	    MazePosition3D position = new MazePosition3D(z);
	    position.addObserver(this);
	    freeCorridors[z] = getRandomCorridor(position);
	}
    }

    private MazeCorridor3D getInitCorridor(int x, int y, int z) {
		MazeCorridor3D result;
		MazePosition3D position = new MazePosition3D(x, y, z);
		position.addObserver(this);
		if ((x == 0) && (y == 0)) {
		    // north-west corner
		    result = new MazeCorridor3D(position, MazeCorridor3D.CORNER, new Orientation2D(270));
		} else if ((x == 0) && (y == (yDimension - 1))) {
		    // south-west corner
		    result = new MazeCorridor3D(position, MazeCorridor3D.CORNER, new Orientation2D(0));
		} else if ((x == xDimension - 1) && (y == yDimension - 1)) {
		    // south-east corner
		    result = new MazeCorridor3D(position, MazeCorridor3D.CORNER, new Orientation2D(90));
		} else if ((x == xDimension - 1) && (y == 0)) {
		    // north-east corner
		    result = new MazeCorridor3D(position, MazeCorridor3D.CORNER, new Orientation2D(180));
		} else if ((x == 0) && isSteadyY(y)) {
		    // west edge
		    result = new MazeCorridor3D(position, MazeCorridor3D.TJUNCTION, new Orientation2D(270));
		} else if (isSteadyX(x) && (y == yDimension - 1)) {
		    // south edge
		    result = new MazeCorridor3D(position, MazeCorridor3D.TJUNCTION, new Orientation2D(0));
		} else if ((x == xDimension - 1) && isSteadyY(y)) {
		    // east edge
		    result = new MazeCorridor3D(position, MazeCorridor3D.TJUNCTION, new Orientation2D(90));
		} else if (isSteadyX(x) && (y == 0)) {
		    // north edge
		    result = new MazeCorridor3D(position, MazeCorridor3D.TJUNCTION, new Orientation2D(180));
		} else if (isSteadyX(x) && isSteadyY(y)) {
		    // steady corridors in the middle of the board
		    if (containsTeleport(x, y)) {
				// staircase corridor
				boolean ceilingOpen = (z != 0);
				boolean floorOpen = (z != (zDimension - 1));
				int rotation = getSteadyTJunctionRotation(x, y);
				if (rotation == -1) {	// at (x, y) isn't a T-junction, so it should be a crossing
					result = new MazeCorridor3D(position, MazeCorridor3D.CROSSING, ceilingOpen, floorOpen, new Orientation2D(0));
				} else {				// there's a rotated T-junction with a staircase at (x, y)
					result = new MazeCorridor3D(position, MazeCorridor3D.TJUNCTION, ceilingOpen, floorOpen, new Orientation2D(rotation));
				}
		    } else if (shrineShaft(x, y)) {
				// a corridor in the same shaft as the shrine. This is always a crossing
				result = new MazeCorridor3D(position, MazeCorridor3D.CROSSING, new Orientation2D(0));
		    } else {
				// a steady t-junction with a certain rotation, but without a staircase
				result = new MazeCorridor3D(position, MazeCorridor3D.TJUNCTION, false, false, new Orientation2D(getSteadyTJunctionRotation(x, y)));
		    }
		} else {	// movable corridor block
		    result = getRandomCorridor(position);
		}
		//result.getOrientation().addObserver(this);
		return result;
    }

	/**
	 * Checks if a maze corridor at the given (x, y) position contains a teleport.
	 * @param x The x position of the maze corridor
	 * @param y The y position of the maze corridor
	 * @return true if a maze corridor at the given (x, y) position contains a teleport,
	 *	false otherwise
	 */
    boolean containsTeleport(int x, int y) {
		for (int i = 0; i < TELEPORT_SHAFTS.length; i++) {
		    if ((TELEPORT_SHAFTS[i][X] == x) && (TELEPORT_SHAFTS[i][Y] == y)) {
				return true;
		    }
		}
		return false;
    }

    private boolean containsShrine(int x, int y, int z) {
		return ((x == SHRINE[X]) && (y == SHRINE[Y]) && (z == SHRINE[Z]));
    }

    private boolean shrineShaft(int x, int y) {
		return ((x == SHRINE[X]) && (y == SHRINE[Y]));
    }

    private int getSteadyTJunctionRotation(int x, int y) {
		for (int i = 0; i < STEADY_TJUNCTION_ROTATIONS.length; i++) {
		    if ((STEADY_TJUNCTION_ROTATIONS[i][X] == x) && (STEADY_TJUNCTION_ROTATIONS[i][Y] == y)) {
			return STEADY_TJUNCTION_ROTATIONS[i][2];
		    }
		}
		//System.err.println("GameLogic.getSteadyTJunctionRotation: no steady t-junction at (" + x + ", " + y + ")");
		return -1;
    }

    private MazeCorridor3D getRandomCorridor(MazePosition3D position) {
		double rand = randomGenerator.nextDouble();
		int selectType = 0;
		double treshold = RANDOM_CORRIDOR_CHANGES[selectType];
		while (rand > treshold) {
		    selectType++;
		    treshold += RANDOM_CORRIDOR_CHANGES[selectType];
		}
		return new MazeCorridor3D(position, selectType, new Orientation2D(getRandomRotation()));
    }

    private int getRandomRotation() {
		return randomGenerator.nextInt(4) * 90;
    }

	/**
	 * Gives the dimension of the maze along the X-axis. This corresponds with
	 * one dimension of a level.
	 * @return The dimension of the maze along the X-axis (in maze corridors)
	 */
    public int getXDimension() {
		return xDimension;
    }

	/**
	 * Gives the dimension of the maze along the Y-axis. This corresponds with
	 * one dimension of a level.
	 * @return The dimension of the maze along the Y-axis (in maze corridors)
	 */
    public int getYDimension() {
		return yDimension;
    }

    /**
	 * Gives the dimension of the maze along the Z-axis. This corresponds with
	 * the number of levels in the maze.
	 * @return The dimension of the maze along the Z-axis (in maze corridors)
	 */
	public int getZDimension() {
		return zDimension;
    }

	/**
	 * Checks if an corridor with the given x-coordinate could be a steady corridor,
	 * that is: not moveable.
	 * @param x The x-coordinate to check (in maze corridors)
	 * @return true if a corridor with the given x-coordinate could be a steady corridor,
	 * 	false otherwise.
	 */
    public static boolean isSteadyX(int x) {
		return ExtraMath.even(x);
    }

	/**
	 * Checks if an corridor with the given y-coordinate could be a steady corridor,
	 * that is: not moveable.
	 * @param y The y-coordinate to check (in maze corridors)
	 * @return true if a corridor with the given y-coordinate could be a steady corridor,
	 * 	false otherwise.
	 */
    public static boolean isSteadyY(int y) {
		return ExtraMath.even(y);
    }

	/**
	 * Gives the maze corridor that's currently at the given position in the maze.
	 * @param x The x-coordinate of the corridor (in maze corridors)
	 * @param y The y-coordinate of the corridor (in maze corridors)
	 * @param z The z-coordinate of the corridor (in maze corridors)
	 * @return The corridor at the given position
	 */
    public MazeCorridor3D getCorridor(int x, int y, int z) {
		checkPosition(x, y, z);
		return maze[x][y][z];
    }

    /**
     * Gives the free corridor that's currently at the given level (z-coordinate).
     * @param z The z-coordinate of the corridor (in maze corridors)
     * @return The free corridor at the given level
     */
    public MazeCorridor3D getFreeCorridor(int z) {
		checkPosition(0, 0, z);
		return freeCorridors[z];
    }

    private void checkPosition(int x, int y, int z) {
		if (!validPosition(x, y, z)) {
		    throw new IndexOutOfBoundsException("maze index out of bounds (" + x + ", " + y + ", " + z + ")");
		}
    }

	/**
	 * Checks if the maze position (x, y, z) is a valid one, that is: if it
	 * specifies a position inside the maze.
	 * @param x The x-coordinate (in maze corridors)
	 * @param y The y-coordinate (in maze corridors)
	 * @param z The z-coordinate (in maze corridors)
	 * @return true if the position (x, y, z) lies inside the maze, false otherwise
	 */
    public boolean validPosition(int x, int y, int z) {
	if ((x >= 0) && (x < xDimension) &&
	    (y >= 0) && (y < yDimension) &&
	    (z >= 0) && (z < zDimension)) {		// correct maze position
	    return true;
	} else {								// incorrect position
	    return false;
	}
    }

	/**
	 * Gives all the corridors the maze consists of. Note that this doesn't include the
	 * free corridors.
	 * @return All the corridors the maze consists of
	 */
    public Vector getAllMazeCorridors() {
		Vector result = new Vector(xDimension * yDimension * zDimension);
		for (int x = 0; x < xDimension; x++) {
		    for (int y = 0; y < yDimension; y++) {
			for (int z = 0; z < zDimension; z++) {
			    result.add(maze[x][y][z]);
			}
		    }
		}
		return result;
    }

	/**
	 * Gives all the corridor in the maze that are reachable from a given position.
	 * A reachable corridor is a corridor to which there's a path from the corridor
	 * at the given position. Such a path can cross various levels. The position must
	 * lie inside the maze. The free corridors can never be reached from inside the maze
	 * and are therefore not included in the search.
	 * @param fromX The x-coordinate of the starting corridor (in maze corridors)
	 * @param fromY The y-coordinate of the starting corridor (in maze corridors)
	 * @param fromZ The z-coordinate of the starting corridor (in maze corridors)
	 * @return All the corridors that can be reached from the corridor at the (fromX, fromY, fromZ)
	 * position in the maze.
	 */
    public Vector getReachableCorridors(int fromX, int fromY, int fromZ) {
		checkPosition(fromX, fromY, fromZ);
		Vector result = new Vector(1);
		calculateCorridorIndexMatrix();
		for (int x = 0; x < xDimension; x++) {
		    for (int y = 0; y < yDimension; y++) {
			for (int z = 0; z < zDimension; z++) {
			    if (corridorIndexMatrix[x][y][z] == corridorIndexMatrix[fromX][fromY][fromZ]) {
				result.add(maze[x][y][z]);
			    }
			}
		    }
		}
		return result;
    }

	/**
	 * Gives all the corridor in the maze that are not reachable from a given position.
	 * A Unreachable corridor is a corridor to which there exists no path from the corridor
	 * at the given position. Such a path could cross various levels. The position must
	 * lie inside the maze. Note that this method only search in the corridor inside the maze,
	 * so the free corridors are not included in teh returned Vector.
	 * @param fromX The x-coordinate of the starting corridor (in maze corridors)
	 * @param fromY The y-coordinate of the starting corridor (in maze corridors)
	 * @param fromZ The z-coordinate of the starting corridor (in maze corridors)
	 * @return All the corridors that cannot be reached from the corridor at the
	 *	(fromX, fromY, fromZ) position in the maze.
	 */
    public Vector getUnreachableCorridors(int fromX, int fromY, int fromZ) {
		Vector reachableCorridors = getReachableCorridors(fromX, fromY, fromZ);
		Vector result = new Vector(getAllMazeCorridors());
		result.removeAll(reachableCorridors);
		return result;
    }

	/**
	 * Gives all the corridors in the maze that can contain treasures. This doens't
	 * include the free corridors.
	 * @return All the corridor in the maze that can contain treasures.
	 */
	public Vector getCorridorsThatCanContainTreasures() {
		Vector result = getAllMazeCorridors();
		Object[] checkCopies = result.toArray();
		for (int i = 0; i < checkCopies.length; i++) {
		    MazeCorridor3D corridor = (MazeCorridor3D)checkCopies[i];
		    MazePosition3D pos = corridor.getPosition();
		    if (containsTeleport(pos.getX(), pos.getY())) {
				result.remove(corridor);
		    }
		}
		return result;
    }

    private void printCorridorIndexMatrix() {
	System.out.println("-------------------------------");
	for (int y = 0; y < yDimension; y++) {
	    for (int x = 0; x < xDimension; x++) {
		System.out.print("" + corridorIndexMatrix[x][y][1] + " ");
	    }
	    System.out.print("  |  ");
	    for (int x = 0; x < xDimension; x++) {
		System.out.print("" + corridorIndexMatrix[x][y][0] + " ");
	    }
	    System.out.println();
	}
    }

    /**
     * Check if there exists a path between two corridors A and B.
     * @param xA	x-coordinate of corridor A (in maze corridors)
     * @param yA	y-coordinate of corridor A (in maze corridors)
     * @param zA	z-coordinate of corridor A (in maze corridors)
     * @param xA	x-coordinate of corridor B (in maze corridors)
     * @param yA	y-coordinate of corridor B (in maze corridors)
     * @param zA	z-coordinate of corridor B (in maze corridors)
     * @return	true if there exists a path between A and B, false otherwise
     */
    public final boolean pathExistsBetween(int xA, int yA, int zA, int xB, int yB, int zB) {
		checkPosition(xA, yA, zA);
		checkPosition(xB, yB, zB);
		calculateCorridorIndexMatrix();
		return (corridorIndexMatrix[xA][yA][zA] == corridorIndexMatrix[xB][yB][zB]);
    }

    /**
     * Gives the maze coordinates of the corridors that form a path from corridor
     * A to B. If there doens't exist such a path, null is returned.
     * @return A vector of MazePosition3D's that form the path from A to B.
     *         The vector is sorted, such that the first element is the starting
     *         position, and the last element is the destination position.
     */
    public Vector getPathBetween(int xA, int yA, int zA, int xB, int yB, int zB) {
		if (!pathExistsBetween(xA, yA, zA, xB, yB, zB)) {
		    return null;
		} else {
		    Vector foundPath = new Vector(2);
		    foundPath.add(maze[xA][yA][zA].getPosition());
		    getPathCoordinates(foundPath, xB, yB, zB);
		    return foundPath;
		}
    }

    private boolean getPathCoordinates(Vector foundPath, int xB, int yB, int zB) {
	MazePosition3D lastPos = (MazePosition3D)foundPath.lastElement();
	if (lastPos.getX() == xB && lastPos.getY() == yB && lastPos.getZ() == zB) {
	    // path complete, don't search further
	    return true;
	} else {
	    // search destination
	    MazeCorridor3D current = maze[lastPos.getX()][lastPos.getY()][lastPos.getZ()];
	    for (int i = 0; i < MazeCorridor3D.NR_DIMENSIONS; i++) {
		if (current.passageOpen(i)) {
		    int xNeighbour = lastPos.getX() + MazeCorridor3D.NEIGHBOUR_OFFSET[i][0];
		    int yNeighbour = lastPos.getY() + MazeCorridor3D.NEIGHBOUR_OFFSET[i][1];
		    int zNeighbour = lastPos.getZ() + MazeCorridor3D.NEIGHBOUR_OFFSET[i][2];
		    if (validPosition(xNeighbour, yNeighbour, zNeighbour)) {
			MazePosition3D nextPosition = maze[xNeighbour][yNeighbour][zNeighbour].getPosition();
			if (maze[xNeighbour][yNeighbour][zNeighbour].passageOpen(MazeCorridor3D.oppositePassage(i)) &&
			    !foundPath.contains(nextPosition)) {
			    foundPath.add(nextPosition);						// add next position to path
			    if (getPathCoordinates(foundPath, xB, yB, zB)) {// path found
				return true;
			    } else {										// recursive search found no path
				foundPath.removeElementAt(foundPath.size() - 1);
			    }
			}
		    }
		}
	    }
	    return false;
	}
    }

    private void calculateCorridorIndexMatrix() {
	if (!corridorIndexMatrixValid) {
	    //System.out.println("Recalculating corridor index matrix");
	    int corridorIndex = 0;
	    resetCorridorIndexMatrix();
	    for (int x = 0; x < xDimension; x++) {
		for (int y = 0; y < yDimension; y++) {
		    for (int z = 0; z < zDimension; z++) {
			if (corridorIndexMatrix[x][y][z] == -1) {
			    fillCorridor(x, y, z, corridorIndex);
			    corridorIndex++;
			}
		    }
		}
	    }
	    //printCorridorIndexMatrix();		// DEBUG
	    corridorIndexMatrixValid = true;
	}
    }

    private void resetCorridorIndexMatrix() {
	for (int x = 0; x < xDimension; x++) {
	    for (int y = 0; y < yDimension; y++) {
		for (int z = 0; z < zDimension; z++) {
		    corridorIndexMatrix[x][y][z] = -1;
		}
	    }
	}
    }

    private void fillCorridor(int x, int y, int z, int index) {
	corridorIndexMatrix[x][y][z] = index;
	MazeCorridor3D current = maze[x][y][z];
	for (int i = 0; i < MazeCorridor3D.NR_DIMENSIONS; i++) {
	    if (current.passageOpen(i)) {
		int xNeighbour = x + MazeCorridor3D.NEIGHBOUR_OFFSET[i][0];
		int yNeighbour = y + MazeCorridor3D.NEIGHBOUR_OFFSET[i][1];
		int zNeighbour = z + MazeCorridor3D.NEIGHBOUR_OFFSET[i][2];
		if (validPosition(xNeighbour, yNeighbour, zNeighbour) &&
		    maze[xNeighbour][yNeighbour][zNeighbour].passageOpen(MazeCorridor3D.oppositePassage(i)) &&
		    corridorIndexMatrix[xNeighbour][yNeighbour][zNeighbour] != index) {
		    fillCorridor(xNeighbour, yNeighbour, zNeighbour, index);
		}
	    }
	}
    }

    /**
     * The update routine that is called when some Observable part
     * that is observed by this GameLogic changes. You should not call
     * this method directly.
     */
    public void update(Observable o, Object arg) {
		if (o instanceof MazePosition3D) {	// position of one of the maze corridors changed
		    MazePosition3D pos = (MazePosition3D)o;
		    if (pos.isFree() && pos.wasFree()) {
				// free corridor block is positioned at an edge of the maze, don't slide anything
		    } else if (!pos.isFree() && pos.wasFree()) {
				// inserted free corridor block
				MazeCorridor3D freeBlock = freeCorridors[pos.getZ()];
				if (pos.getX() == 0) {
				    // block inserted in west edge, shift east neighbour recursively
				    maze[pos.getX()][pos.getY()][pos.getZ()].getPosition().set(pos.getX() + 1, pos.getY(), pos.getZ());
				} else if (pos.getX() == xDimension - 1) {
				    // block inserted in east edge, shift west neighbour recursively
				    maze[pos.getX()][pos.getY()][pos.getZ()].getPosition().set(pos.getX() - 1, pos.getY(), pos.getZ());
				} else if (pos.getY() == 0) {
				    // block inserted in north edge, shift south neighbour recursively
				    maze[pos.getX()][pos.getY()][pos.getZ()].getPosition().set(pos.getX(), pos.getY() + 1, pos.getZ());
				} else if (pos.getY() == yDimension - 1) {
				    // block inserted in south edge, shift north neighbour recursively
				    maze[pos.getX()][pos.getY()][pos.getZ()].getPosition().set(pos.getX(), pos.getY() - 1, pos.getZ());
				}
				// after recursion, the whole row or column is slided and the
				// free corridor block is the one thrown out. Now insert the initiator.
				maze[pos.getX()][pos.getY()][pos.getZ()] = freeBlock;
				corridorIndexMatrixValid = false;
			    } else if (pos.isFree() && !pos.wasFree()) {
				// this block just became free, so add it to the free corridors
				freeCorridors[pos.getZ()] = maze[pos.getOldX()][pos.getOldY()][pos.getOldZ()];
			    } else {	// !pos.isFree() && !pos.wasFree()
					// block is moved inside the maze
					int dx = pos.getX() - pos.getOldX();
					int dy = pos.getY() - pos.getOldY();
					if (Math.abs(dx) != 1 && Math.abs(dy) != 1) {
					    // illegal position change
					    throw new RuntimeException("Illegal position change of corridorblock: from (" +
								       pos.getOldX() + ", " + pos.getOldY() + ", " + pos.getOldZ() + ") to (" +
								       pos.getX() + ", " + pos.getY() + ", " + pos.getZ() + ")");
					} else {
				    // slide this block recursively to neighbour position
				    maze[pos.getX()][pos.getY()][pos.getZ()].getPosition().set(pos.getX() + dx, pos.getY() + dy, pos.getZ());
				    maze[pos.getX()][pos.getY()][pos.getZ()] = maze[pos.getOldX()][pos.getOldY()][pos.getOldZ()];
				}
		    }
		}
    }

	/**
	 * Builds a copy of this object.
	 * @return The copy
	 */
    public Object clone() {
		return new Maze3D(this);
    }

}
