package quest.global.game;

import java.util.Observer;
import java.util.Observable;
import quest.global.util.SerializableObservable;

/**
 * <P>
 * A maze corridor is a building block of a 3D maze. It can be visualised a
 * a square room with six possible passages: one in each wall, one in the floor and
 * one in the ceiling. Each of these passages can be open or closed.</P>
 * <P>
 * Since each maze block has to be referred to by a 3D-coordinate, there
 * must be an agreement upon the meaning of the six names that refer to the
 * passages. The intuition is that the NORTH-EAST-SOUTH-WEST plane of the maze
 * lies in the X-Y plane in 3D space, and the ceiling has a greater z-coordinate
 * than the floor. So the meaning of the names are:
 * <UL>
 * <LI>NORTH lies in the negative Y direction
 * <LI>EAST lies in the positive X direction
 * <LI>SOUTH lies in the positive Y direction
 * <LI>WEST lies in the negative X direction
 * <LI>FLOOR lies in the negative Z direction
 * <LI>CEILING lies in the positive Z direction
 * </UL></P>
 * <P>
 * There are six possible shapes you can create in a 2D corridor block with four
 * possible exists that can be open or closed, ignoring rotation changes.
 * These '2D shapes' don't look at the
 * passages at the floor or the ceiling. You can get this shape2D type of a maze corridor,
 * and even create a maze corridor by specifying the shape2D type.</P>
 * <P>
 * A maze corridor can contain one treasure. This treasure can be changed or removed,
 * in which case all observers are notified. Each maze corridor observes it own
 * orientation, and adapts it internal representation of the passages if this orientation
 * changes.</P>
 *
 * @author Mathijs den Burger
 * @version 1.0
 */

public class MazeCorridor3D extends SerializableObservable implements Observer, Selectable, Cloneable {

 	/**
 	 * Number of dimensions (possible passages) in a maze corridor
 	 */
    public static final int NR_DIMENSIONS = 6;

    /**
     * The index used in arrays for the north passage data of a maze corridor
     */
    public static final int NORTH = 0;

    /**
     * The index used in arrays for the east passage data of a maze corridor
     */
    public static final int EAST = 1;

    /**
     * The index used in arrays for the south passage data of a maze corridor
     */
    public static final int SOUTH = 2;

    /**
     * The index used in arrays for the west passage data of a maze corridor
     */
	public static final int WEST = 3;

    /**
     * The index used in arrays for the floor passage data of a maze corridor
     */
    public static final int FLOOR = 4;

    /**
     * The index used in arrays for the ceiling passage data of a maze corridor
     */
    public static final int CEILING = 5;

    /**
     * The relative offsets per coordinate of a neighbour corridor block.
     * The first index indicates the passage (NORTH, EAST, SOUTH, WEST, FLOOR, CEILING),
     * the second one the coordinate (0 for X, 1 for Y, 2 for Z).
     */
    public static final int NEIGHBOUR_OFFSET[][] = {
		{0, -1, 0}, {1, 0, 0}, {0, 1, 0}, {-1, 0, 0}, {0, 0, -1}, {0, 0, 1}
    };

    /**
     * A 2D corridor shape, with all four passages closed
     */
    public static final int CLOSED = 0;

    /**
     * A 2D corridor shape, with only one passage open (the default is
     * the NORTH passage open)
     */
	public static final int DEAD_END = 1;

	/**
     * A 2D corridor shape, with two passages across each other open (the default is
     * the NORTH and SOUTH passage open)
     */
	public static final int STRAIGHT = 2;

	/**
     * A 2D corridor shape, with two passages next to each other (the default is
     * the NORTH and EAST passage open)
     */
	public static final int CORNER = 3;

	/**
     * A 2D corridor shape, with only one passages closed (the default is
     * the NORTH, EAST and WEST passage open and the SOUTH passage closed)
     */
	public static final int TJUNCTION = 4;

	/**
     * A 2D corridor shape, with all four passages open
     */
	public static final int CROSSING = 5;

	/**
     * The number of possible 2D corridor shapes that are different under rotation
     */
	public static final int NR_SHAPE2D = 6;

    protected boolean[] passageOpen;
    protected int shape2D;
    protected MazePosition3D position;
    protected Orientation2D orientation;
    protected SelectionState selectionState;
    protected Treasure treasure;

    /**
     * Creates a maze corridor with all passages closed and a standard orientation.
     * @param position Position of this corridor in the maze
     */
    public MazeCorridor3D(MazePosition3D position) {
		this(position, false, false, false, false, false, false);
    }

    /**
     * Creates a maze corridor with the given passages open and a standard orientation.
     * @param position		Position of this corridor in the maze
     * @param northPassage	Indicates whether the north passage is open or closed
     * @param eastPassage	Indicates whether the east passage is open or closed
     * @param southPassage	Indicates whether the south passage is open or closed
     * @param westPassage	Indicates whether the west passage is open or closed
     * @param floorPassage	Indicates whether the floor passage is open or closed
     * @param ceilingPassage	Indicates whether the ceiling passage is open or closed
     */
    public MazeCorridor3D(MazePosition3D position, boolean northPassageOpen, boolean eastPassageOpen, boolean southPassageOpen,
			  boolean westPassageOpen, boolean floorPassageOpen, boolean ceilingPassageOpen) {
	this(position, northPassageOpen, eastPassageOpen, southPassageOpen, westPassageOpen,
	     floorPassageOpen, ceilingPassageOpen, new Orientation2D());
    }

    /**
     * Creates a maze corridor with the given passages open and a given initial orientation.
     * @param position		Position of this corridor in the maze
     * @param northPassage	Indicates whether the north passage is open or closed
     * @param eastPassage	Indicates whether the east passage is open or closed
     * @param southPassage	Indicates whether the south passage is open or closed
     * @param westPassage	Indicates whether the west passage is open or closed
     * @param floorPassage	Indicates whether the floor passage is open or closed
     * @param ceilingPassage	Indicates whether the ceiling passage is open or closed
     * @param orientation	The 2D orientation of the corridor
     */
    public MazeCorridor3D(MazePosition3D position, boolean northPassageOpen, boolean eastPassageOpen, boolean southPassageOpen,
			  boolean westPassageOpen, boolean floorPassageOpen, boolean ceilingPassageOpen, Orientation2D orientation) {
	this.position = position;
	passageOpen = new boolean[NR_DIMENSIONS];
	passageOpen[NORTH] = northPassageOpen;
	passageOpen[EAST] = eastPassageOpen;
	passageOpen[SOUTH] = southPassageOpen;
	passageOpen[WEST] = westPassageOpen;
	passageOpen[FLOOR] = floorPassageOpen;
	passageOpen[CEILING] = ceilingPassageOpen;
	this.orientation = orientation;
	orientation.addObserver(this);
	selectionState = new SelectionState();
	this.shape2D = calculateShape2D(northPassageOpen, eastPassageOpen, southPassageOpen, westPassageOpen);
	rotate(orientation.getRotation());
	treasure = null;
    }

    /**
     * Creates a maze corridor of the given 2D shape and initial rotation. The floor
     * and ceiling are closed.
     *
     * @param position		Position of this corridor in the maze
     * @param shape2D		The 2D shape
     * @param orientation	The 2D orientation of the corridor
     */
    public MazeCorridor3D(MazePosition3D position, int shape2D, Orientation2D orientation) {
	this(position, shape2D, false, false, orientation);
    }

    /**
     * Creates a maze corridor of the given 2D shape, floor, ceiling and initial rotation.
     * @param position		Position of this corridor in the maze
     * @param shape2D		The 2D shape
     * @param floorPassage	Indicates whether the floor passage is open or closed
     * @param ceilingPassage	Indicates whether the ceiling passage is open or closed
     * @param orientation	The 2D orientation of the corridor
     */
    public MazeCorridor3D(MazePosition3D position, int shape2D, boolean floorPassageOpen, boolean ceilingPassageOpen, Orientation2D orientation) {
	this.position = position;
	passageOpen = new boolean[NR_DIMENSIONS];
	setPassages2D(shape2D);
	passageOpen[FLOOR] = floorPassageOpen;
	passageOpen[CEILING] = ceilingPassageOpen;
	this.orientation = orientation;
	orientation.addObserver(this);
	selectionState = new SelectionState();
	this.shape2D = shape2D;
	rotate(orientation.getRotation());
	treasure = null;
    }

	/**
	 * Copy constructor used by clone()
	 * @see quest.global.game.MazeCorridor3D#clone()
	 */
    protected MazeCorridor3D(MazeCorridor3D original) {
	// copy passage open value
	passageOpen = new boolean[original.passageOpen.length];
	for (int i = 0; i < original.passageOpen.length; i++) {
	    passageOpen[i] = original.passageOpen[i];
	}
	shape2D = original.shape2D;
	position = (MazePosition3D)original.position.clone();
	orientation = (Orientation2D)original.orientation.clone();
	orientation.addObserver(this);
	selectionState = (SelectionState)original.selectionState.clone();
	if (original.treasure != null) {
	    treasure = (Treasure)original.treasure.clone();
	    treasure.setFindSpot(this);
	} else {
	    treasure = null;
	}
    }

    private void setPassages2D(int shape2D) {
	switch(shape2D) {
	    case CLOSED:
		passageOpen[NORTH] = passageOpen[EAST] = passageOpen[SOUTH] = passageOpen[WEST] = false;
		break;
	    case DEAD_END:
		passageOpen[NORTH] = true; passageOpen[EAST] = passageOpen[SOUTH] = passageOpen[WEST] = false;
		break;
	    case STRAIGHT:
		passageOpen[NORTH] = true; passageOpen[EAST] = false; passageOpen[SOUTH] = true; passageOpen[WEST] = false;
		break;
	    case CORNER:
		passageOpen[NORTH] = true; passageOpen[EAST] = true; passageOpen[SOUTH] = false; passageOpen[WEST] = false;
		break;
	    case TJUNCTION:
		passageOpen[NORTH] = true; passageOpen[EAST] = true; passageOpen[SOUTH] = false; passageOpen[WEST] = true;
		break;
	    case CROSSING:
		passageOpen[NORTH] = passageOpen[EAST] = passageOpen[SOUTH] = passageOpen[WEST] = true;
		break;
	    default:
		System.err.println("Unknown shape2D in MazeCorridor3D.setPassages2D: " + shape2D);
	}
    }

    /**
     * Examines the values of the north, east, south and west gate, and returns
     * the 2D shape this combination resembles. The rotation is assumed to be 0.
     */
    private int calculateShape2D(boolean north, boolean east, boolean south, boolean west) {
	int result = -1;
	if (!north && !east && !south && !west) result = CLOSED;
	else if (north && !east && !south && !west) result = DEAD_END;
	else if (north && east && !south && !west) result = CORNER;
	else if (north && !east && south && !west) result = STRAIGHT;
	else if (north && east && !south && west) result = TJUNCTION;
	else if (north && east && south && west) result = CROSSING;
	else System.err.println("Unknown shape2D in MazeCorridor3D.calculateShape2D");
	return result;
    }

    /**
     * Checks if the given passage is open.
     * @param passage	The passage to check (NORTH, EAST, SOUTH, WEST, FLOOR, CEILING)
     * @return 	true if the passage is open, false otherwise
     */
    public final boolean passageOpen(int passage) {
	if ((passage < 0) || (passage > NR_DIMENSIONS - 1)) {
	    throw new IndexOutOfBoundsException("passage value of " + passage + " is outside range of (0, " + (NR_DIMENSIONS - 1));
	} else {
	    return passageOpen[passage];
	}
    }

    /**
     * Gives the opposite passage of the given one.
     * @param passage	The given passage (NORTH, EAST, SOUTH, WEST, FLOOR or CEILING)
     * @return	The opposite passage
     */
    public static final int oppositePassage(int passage) {
	if (passage == NORTH) return SOUTH;
	else if (passage == EAST) return WEST;
	else if (passage == SOUTH) return NORTH;
	else if (passage == WEST) return EAST;
	else if (passage == FLOOR) return CEILING;
	else if (passage == CEILING) return FLOOR;
	else throw new IndexOutOfBoundsException("passage value of " + passage + " is outside range of (0, " + (NR_DIMENSIONS - 1));
    }

    /**
     * Rotates this corridor counter clockwise around the z-axis. This means that the
     * floor and ceiling stay where they are, but the north wall turns to the east,
     * the east one to the south etc.
     */
    protected void turnCounterClockwise() {
	boolean tmp = passageOpen[WEST];
	passageOpen[WEST] = passageOpen[NORTH];
	passageOpen[NORTH] = passageOpen[EAST];
	passageOpen[EAST] = passageOpen[SOUTH];
	passageOpen[SOUTH] = tmp;
    }

	/**
	 * Rotates this corridor for the given number of degrees counter clockwise.
	 * If this is not a multiple
	 * of 90, the value is floored to the nearest multiple of 90.
	 */
    protected void rotate(int degrees) {
		int nrCounterClockwiseRotations = (int)Math.floor(degrees / 90);
		for (int i = 0; i < nrCounterClockwiseRotations; i++) turnCounterClockwise();
    }

	/**
	 * Gives the orientation of this corridor
	 * @return The orientation of this corridor
	 */
    public Orientation2D getOrientation() {
		return orientation;
    }

	/**
	 * Gives the position of this corridor in the maze
	 * @return The position of this corridor
	 */
    public MazePosition3D getPosition() {
		return position;
    }

    /**
	 * Gives the 2D shape of this corridor. This can be
	 * CLOSED, DEAD_END, STRAIGHT, CORNER, TJUNCTION or CROSSING.
	 * @return The 2D shape of this corridor
	 */
    public final int getShape2D() {
		return shape2D;
    }

	/**
	 * Gives the selection state of this corridor
	 * @return The selection state of this corridor
	 */
    public SelectionState getSelectionState() {
		return selectionState;
    }

    /*
     * Gives the treasure that lies in this corridor block.
     * @return If this corridor block contains a treasure: that treasure,
     *         else null.
     */
    public Treasure getTreasure() {
    	return treasure;
    }

    /*
     * Puts the given treasure in this corridor block. If there was
     * a treasure in this block already, that treasure is returned.
     * Use null as an argument to only remove the current treasure.
     * This method notifies all observers.
     * @param treasure The treasure to add. Use null to add no treasure,
     *                 which effectivily removes the current treasure.
     * @return The treasure that was already in this corridor block.
     *         If there wasn't such a treasure, null is returned.
     */
    public Treasure setTreasure(Treasure newTreasure, boolean notifyObservers) {
    	Treasure oldTreasure = treasure;
    	treasure = newTreasure;
    	setChanged();
    	if (notifyObservers) notifyObservers();
    	return oldTreasure;
    }

    /**
     * Checks if this corridor contains a treasure
     * @return true if this corridor contains a treasure, false otherwise
     */
    public boolean containsTreasure() {
    	return (treasure != null);
    }

    /**
     * The update routine that is called when some Observable part
     * that is observed by this MazeCorridor changes. You should not call
     * this method directly.
     */
    public void update(Observable o, Object arg) {
		if (o instanceof Orientation2D) {			// orientation changed
		    Orientation2D orientation = (Orientation2D)o;
		    int rotatedDegrees = orientation.getRotation() - orientation.getOldRotation();
		    if (rotatedDegrees < 0) rotatedDegrees += 360;
		    rotate(rotatedDegrees);
		}
    }

    /**
     * Builds a copy of this object
     * @return The copy
     */
    public Object clone() {
		return new MazeCorridor3D(this);
    }

}
