package quest.global.game;

import java.util.Observable;
import quest.global.util.*;

/**
 * <P>
 * Specifies a position in or next to the 3D maze that forms the games' board.
 * Positions are measured in MazeCorridors. All the maze corridors together form
 * a 3D grid, and a maze position points to a specific corridor in that grid.</P>
 * <P>
 * Since that grid is 3-dimensional, a maze position has three dimensions too (x, y and z).
 * In each dimension a maze corridor has a coordinate ranging from 0 to the size
 * of the grid in that dimension minus one. A position next to the maze therefore always
 * has coordinates with a value of -1 or the size of a dimension of the grid.
 * A maze position next to the maze in the X-Y-plane is said to be 'free', which can be checked with
 * the method isFree(). The previous value of the free-ness is also rembered, which can
 * be checked with the method wasFree(). Note that the value of the z-coordinate
 * doens't influence the 'free-ness' of a corridor.</P>
 * <P>
 * A maze position can be changed, in which case all observers are notified of 
 * the change. The previous position is always remembered, and can be read with
 * the getOld..() methods.</P>
 *
 * @author Mathijs den Burger
 * @version 1.0
 */

public class MazePosition3D extends SerializableObservable implements Cloneable {

	private static final int MIN_X = -1,
		MIN_Y = -1,
		MIN_Z = 0,
		MAX_X = GameLogic.X_DIMENSION,
		MAX_Y = GameLogic.Y_DIMENSION,
		MAX_Z = GameLogic.Z_DIMENSION;

	/**
	 * Value of a coordinate when the exact value isn't known
	 */
	public static final int UNKNOWN = MAX_X + MAX_Y + MAX_Z;

	private int x, y, z;
	private int oldX, oldY, oldZ;
	private boolean isFree, wasFree;

	/**
	 * Constructs a new maze position with the given coordinates. The old coordinates
	 * get the initial values too. Values that lie outside the permitted range
	 * are clamped. 
	 * @param x The x-coordinate (in maze corridors)
	 * @param y The y-coordinate (in maze corridors)
	 * @param z The z-coordinate (in maze corridors)
	 * @see quest.global.util.ExtraMath#clamp
	 */
	public MazePosition3D(int x, int y, int z) {
		this.x = oldX = ExtraMath.clamp(x, MIN_X, MAX_X);
		this.y = oldY = ExtraMath.clamp(y, MIN_Y, MAX_Y);
		this.z = oldZ = ExtraMath.clamp(z, MIN_Z, MAX_Z);
		calculateFree();
		wasFree = isFree;
	}

	/**
	 * Constructs a new maze position with the given z-coordinates and the other
	 * coordinates set to UNKNOWN. The old coordinates get these initial values 
	 * too. Values that lie outside the permitted range are clamped. 
	 * A maze position constructed this way is initially free.
	 * @param z The z-coordinate (in maze corridors)
	 * @see quest.global.util.ExtraMath#clamp
	 */
	public MazePosition3D(int z) {
		this.x = oldX = this.y = oldY = UNKNOWN;
		this.z = oldZ = z;
		this.z = ExtraMath.clamp(z, MIN_Z, MAX_Z);
		wasFree = isFree = true;
	}

    /**
     * Copy constructor used by clone()
     * @see quest.global.game.MazePosition3D#clone()
     */
    protected MazePosition3D(MazePosition3D original) {
		this.x = original.x;
		this.y = original.y;
		this.z = original.z;
		this.oldX = original.oldX;
		this.oldY = original.oldY;
		this.oldZ = original.oldZ;
		this.isFree = original.isFree;
		this.wasFree = original.wasFree;
    }

	/**
	 * Reports whether the current coordinates point to a position next to the maze
	 * in the X-Y-plane. The z-coordinate doens't influence the free-ness of a corridor.
	 * @return true if the current coordinates point to a position next to the maze in
	 * the X-Y-plane, false otherwise
	 */
	public boolean isFree() {
		return isFree;
	}

	/**
	 * Reports whether the previous coordinates pointed to a position next to the maze
	 * @return true if the previous coordinates pointed to a position next to the maze, false otherwise
	 */
	public boolean wasFree() {
		return wasFree;
	}

	/**
	 * Gives the current x-coordinate
	 * @return The current x-coordinate
	 */
	public int getX() {
		return x;
	}

	/**
	 * Gives the current y-coordinate
	 * @return The current y-coordinate
	 */
	public int getY() {
		return y;
	}

	/**
	 * Gives the current z-coordinate
	 * @return The current z-coordinate
	 */
	public int getZ() {
		return z;
	}

	/**
	 * Gives the previous x-coordinate
	 * @return The previous x-coordinate
	 */
	public int getOldX() {
		return oldX;
	}

	/**
	 * Gives the previous y-coordinate
	 * @return The previous y-coordinate
	 */
	public int getOldY() {
		return oldY;
	}

	/**
	 * Gives the previous z-coordinate
	 * @return The previous z-coordinate
	 */
	public int getOldZ() {
		return oldZ;
	}

	/**
	 * Sets the new coordinates of the maze position. Values that lie outside the
	 * valid range are clamped. The given coordinates are only
	 * set if they point to different point in the maze than the previous ones, 
	 * or the current position is free. If this is the case, the current 
	 * coordinates become the previous ones, and the previous ones are lost.
	 * Then the free-ness is recalculated and all observers are notified.
	 * @param x The new x-coordinate (in maze corridors)
	 * @param y The new y-coordinate (in maze corridors)
	 * @param z The new z-coordinate (in maze corridors)
	 * @see quest.global.util.ExtraMath#clamp
	 */
	public void set(int x, int y, int z) {
		x = ExtraMath.clamp(x, MIN_X, MAX_X);
		y = ExtraMath.clamp(y, MIN_Y, MAX_Y);
		z = ExtraMath.clamp(z, MIN_Z, MAX_Z);
		if (this.x != x || this.y != y || this.z != z ||	// something changed
			isFree) {										// or visibilty of free item may have changed
			oldX = this.x;
			oldY = this.y;
			oldZ = this.z;
			this.x = x;
			this.y = y;
			this.z = z;
			wasFree = isFree;
			calculateFree();
			setChanged();
			notifyObservers();
		}
	}
	
	private void calculateFree() {
		if (x == MIN_X || x == MAX_X || y == MIN_Y || y == MAX_Y) {
			isFree = true;
		} else {
			isFree = false;
		}
	}		

	/**
	 * Checks whether the current position lies inside the maze. Note that this
	 * isn't the complement of isFree(), since the latter doens't take the z-coordinate
	 * into account, and this method does.
	 * @return true if the current position lies inside the maze, false otherwise
	 */
	public boolean inMaze() {
		return ((x >= 0) && (x < GameLogic.X_DIMENSION) &&
			(y >= 0) && (y < GameLogic.Y_DIMENSION) &&
			(z >= 0) && (z < GameLogic.Z_DIMENSION));
	}

	/**
	 * Checks whether the current position lies at the edge of the maze. 'The edge'
	 * is defined as 
	 * <CODE>
	 * (x == 0) || (x == GameLogic.X_DIMENSION - 1) || (y == 0) || (y == GameLogic.Y_DIMENSION - 1)
	 * </CODE>
	 * and can intuitively be seen as the corridors at the edge of a X-Y-plane.
	 * @return true if the current position lies at the edge of the maze, false otherwise
	 */
	public boolean edgeOfMaze() {
		return ((x == 0) || (x == GameLogic.X_DIMENSION - 1) ||
			(y == 0) || (y == GameLogic.Y_DIMENSION - 1));
	}

	/**
	 * Checks whether the current position lies next to the maze. 'next to' is defined as
	 * <CODE>
	 * x == -1 || x == GameLogic.X_DIMENSION ||	y == -1 || y == GameLogic.Y_DIMENSION
	 * </CODE>
	 */	 
	public boolean nextToMaze() {
		return (x == -1 || x == GameLogic.X_DIMENSION ||
			y == -1 || y == GameLogic.Y_DIMENSION);
	}

	/**
	 * Calculates the position if the given position would be 'inserted' into
	 * the maze. Insertion only applies to the x and y-coordinate.
	 * Effectively, the nearest position inside the maze to the given position is returned.
	 * @return A new MazePosition3D object
	 */
	public static MazePosition3D insertInMaze(MazePosition3D outsideMaze) {
		return new MazePosition3D(ExtraMath.clamp(outsideMaze.getX(), 0, GameLogic.X_DIMENSION - 1),
			ExtraMath.clamp(outsideMaze.getY(), 0, GameLogic.Y_DIMENSION - 1),
			outsideMaze.getZ());
	}
	
	/**
	 * Checks if this maze position is equals to a given one. Two maze positions
	 * are equals if there current position is the same.
	 * @return true if this maze position is equals to the given one, false otherwise
	 */
	public boolean equals(MazePosition3D rhs) {
		return (x == rhs.x && y == rhs.y && z == rhs.z);
	}

	/**
	 * Builds a copy of this object
	 * @return The copy
	 */
    public Object clone() {
		return new MazePosition3D(this);
    }

}
