package quest.global.game;

import java.util.*;
import quest.global.util.SerializableObservable;

/**
 * <P>
 * Defines the goal of a game user. The current mission is to find
 * a number of treasures in any order. When a user finds a treasure it can 
 * be reported as found
 * to the mission. When all the treasures to find are reported as found, 
 * the mission is accomplished. You can also mark the mission as failed,
 * in which case the mission will never be able to be accomplished.</P>
 * <P>
 * When a treasure is added to the mission, all observers are notified.</P>
 * 
 * @see quest.global.game.GameUser
 * @author Mathijs den Burger
 * @version 1.0
 */

public class Mission extends SerializableObservable implements Cloneable {

    private Vector targets;
    private int totalNrTreasuresToFind;
	private boolean missionFailed;

    /**
     * Constructs a new mission.
     * @param treasuresToFind All the treasures a user must find to complete this mission
     */
    public Mission(Treasure[] treasuresToFind) {
		Arrays.sort(treasuresToFind, Mission.TreasureComparator.instance());
		this.targets = new Vector(treasuresToFind.length);
		totalNrTreasuresToFind = targets.size();
		for (int i = 0; i < treasuresToFind.length; i++) {
		    targets.add(treasuresToFind[i]);
		}
		missionFailed = false;
    }

	/**
	 * Copy constructor used by clone()
	 * @see quest.global.game.Mission#clone()
	 */
    protected Mission(Mission original) {
		targets = new Vector(original.targets.size());
		for(Enumeration e = original.targets.elements(); e.hasMoreElements(); ) {
		    targets.add((Treasure)(((Treasure)e.nextElement()).clone()));
		}
		missionFailed = original.missionFailed;
    }

    /**
     * Reports that the given treasure is found. The treasure is
     * removed from the to-do-list, and all observers are notified.
     * When a treasure is found that isn't part of this mission, or this mission
     * is already marked as failed, nothing happens.
     * @param treasure The found treasure
     */
    public void found(Treasure treasure) {
    	if (missionFailed) return;
		if (targets.remove(treasure)) {
		    setChanged();
		    notifyObservers();
		} else {
	    	System.err.println("WARNING in Mission.found: a treasure was found (symbolId " + treasure.getSymbolId() + ", colorId " + treasure.getColorId() + "), but it was already removed from the mission");
	    }
    }

	/**
	 * Checks you still have to find the given treasure to complete this mission.
	 * @param treasure The treasure to check
	 * @return true if you still have to find the given treasure to complete this mission,
	 * false if the treasure was already found or it wasn't part of this mission at all.
	 */
    public boolean part(Treasure treasure) {
		return targets.contains(treasure);
    }

	/**
	 * Checks if this mission is accomplished. That is: all the treasures to find
	 * are found and the mission hasn't been marked as 'failed'.
	 * @return true if this mission is accomplished, false if not
	 */
    public boolean accomplished() {
		return (targets.size() == 0 && !missionFailed);
    }
    
    /**
     * Marks this mission as failed. This cannot be undone. All observers are
     * notified of this change.
     */
    public void setFailed() {
    	missionFailed = true;
	    setChanged();
	    notifyObservers();
	}
	
	/**
	 * Checks if this mission is marked as failed.
	 * @return true if this mission is marked as failed, false otherwise.
	 */
    public boolean failed() {
    	return missionFailed;
    }

	/**
	 * Gives the number of treasures still to find to complete this mission.
	 * return The number of treasures still to find to complete this mission.
	 */
    public int getNrMissingTreasures() {
		return targets.size();
    }
    
    /**
	 * Gives the total number of treasures to find to complete this mission.
	 * This never changes after the mission is constructed, of course.
	 * return The number of treasures to find to complete this mission.
	 */
    public int getTotalNrTreasuresToFind() {
    	return totalNrTreasuresToFind;
    }

    /**
	 * Gives the number of treasures already found.
	 * @return The number of treasures already found.
	 */
	public int getNrFoundTreasures() {
		return totalNrTreasuresToFind - targets.size();
	}

	/**
	 * Gives all the treasures still to find to complete the mission.
	 * @return All the treasures still to find to complete the mission.
	 */
    public Treasure[] getMissingTreasures() {
		Treasure[] result = new Treasure[targets.size()];
		for (int i = 0; i < targets.size(); i++) {
		    result[i] = (Treasure)targets.elementAt(i);
		}
		return result;
    }

	/**
	 * Builds a copy of this object.
	 * @return The copy
	 */
    public Object clone() {
		return new Mission(this);
    }

    // INNER CLASSES

    private static class TreasureComparator implements Comparator {

		private static TreasureComparator uniqueInstance = null;
	
		protected TreasureComparator() {}
	
		/**
		 * Constructs a comparator that compares Treasure objects.
		 */
		public static TreasureComparator instance() {
		    if (uniqueInstance == null) uniqueInstance = new TreasureComparator();
		    return uniqueInstance;
		}
	
		/**
		 * Compares two treasures.
		 * Treasures are first sorted according to color ID, then to symbol ID.
		 * Objects other than Treasure are always 'smaller' then Treasures.
		 * @see quest.global.game.Treasure
		 */
		public int compare(Object o1, Object o2) {
		    if (o1 instanceof Treasure && o2 instanceof Treasure) {
			Treasure t1 = (Treasure)o1,
			    t2 = (Treasure)o2;
			if (t1.getColorId() < t2.getColorId()) return -1;
			else if (t1.getColorId() > t2.getColorId()) return 1;
			else if (t1.getSymbolId() < t2.getSymbolId()) return -1;
			else return (t1.getSymbolId() > t2.getSymbolId() ? 1 : 0);
		    } else {
			return Integer.MIN_VALUE;
		    }
		}
	
		/**
		 * Checks if this comparator is equal to the given object.
		 * This is only the case when the given object is also an TreasureComparator.
		 */
		public boolean equals(Object obj) {
		    return (obj instanceof Mission.TreasureComparator);
		}

    }

}
