package quest.global.game;

import java.util.Observable;
import quest.global.util.SerializableObservable;

/**
 * <P>
 * Defines a certain orientation in 2D space. This orientation consist of
 * the number of degrees something is rotated around an imaginary axis.</P>
 * <P>
 * The orientation can only be changed by rotating a certain amount of degrees.
 * When this happens all observers are notified. The previous rotation is always 
 * remembered.</P>
 * <P>
 * The rotation lies always between 0 and 359 degrees.</P>
 * 
 * @author Mathijs den Burger
 * @version 1.0
 */
 
public class Orientation2D extends SerializableObservable implements Cloneable {

	/**
	 * The current rotation (in degrees)
	 */
	protected int rotation;						// in degrees
	
	/**
	 * The previous rotation (in degrees)
	 */
	protected int oldRotation;					// in degrees

	/**
	 * Constructs a 2D orientation with a rotation component of 0.
	 */
	public Orientation2D() {
		this(0);
	}

	/**
	 * Constructs a 2D orientation with a given rotation component.
	 * The old rotation will be set to the same value as the given rotation.
	 * @param rotation The 2D rotation (in degrees)
	 */
	public Orientation2D(int rotation) {
		this.rotation = rotation;
		this.rotation = this.rotation % 360;
		oldRotation = 0;
	}

	/**
	 * Copy constructor used by clone()
	 * @see quest.global.game.Orientation2D#clone()
	 */
    protected Orientation2D(Orientation2D original) {
		this.rotation = original.rotation;
		this.oldRotation = original.oldRotation;
    }

	/**
	 * Rotates this 2D orientation for the given number of degrees.
	 * Negative rotatin values are valid are rotate the orientation
	 * in the opposite direction. The current rotation becomes the previous one,
	 * and the old previous value is lost. All observers are notified of this change.
	 * @param degrees The number of degrees to rotate
	 */
	public void rotate(int degrees) {
		degrees = degrees % 360;
		oldRotation = rotation;
		rotation = (rotation + degrees) % 360;
		setChanged();
		notifyObservers();
	}

	/**
	 * Gives the current number of degrees the 2D orientation is rotated
	 * @return The current number of degrees the 2D orientation is rotated
	 */
	public int getRotation() {
		return rotation;
	}

	/**
	 * Gives the current number of degrees the 2D orientation is rotated
	 * @return The current number of degrees the 2D orientation is rotated
	 */
	public int getOldRotation() {
		return oldRotation;
	}

	/**
	 * Builds a copy of this object
	 * @return The copy
	 */
    public Object clone() {
		return new Orientation2D(this);
    }

}
