package quest.global.game;

import java.util.Observable;
import quest.global.util.SerializableObservable;

/**
 * Defines a simple state machine with four states: UNSELECTABLE, UNSELECTED,
 * SELECTED and TARGET. All state transitions are possible. The previous state
 * is always rembered. Observers are notified when the state has changed.
 * 
 * @author Mathijs den Burger
 * @version 1.0
 */

public class SelectionState extends SerializableObservable implements Cloneable {

	/**
	 * A selection state, meaning that something cannot be selected.
	 */
	public static final int UNSELECTABLE = 0;
	
	/**
	 * A selection state, meaning that something can be selected but currently isn't
	 */
	public static final int UNSELECTED = 1;
	
	/**
	 * A selection state, meaning that something is selected
	 */
	public static final int SELECTED = 2;
	
	/**
	 * A selection state, meaning that something is a target now, which is the selection
	 * of something that was already selected.
	 */
	public static final int TARGET = 3;

	/**
	 * The number of possible selection states.
	 */
	public static final int NR_STATES = 4;

	private int state, oldState;

	/**
	 * Constructs a selection state with the initial value UNSELECTABLE.
	 */
	SelectionState() {
		this(UNSELECTABLE);
	}

	/**
	 * Constructs a selection state with the given initial state. The old state
	 * is also set to this given state.
	 * @param state The initial selection state
	 */
	SelectionState(int state) {
		this.state = this.oldState = state;
	}

	/**
	 * Copy constructor used by clone()
	 * @see quest.global.game.SelectionState#clone()
	 */
    protected SelectionState(SelectionState original) {
		state = original.state;
		oldState = original.oldState;
    }

	/**
	 * Gives the current selection state.
	 */
	public int getState() {
		return state;
	}

	/**
	 * Gives the previous selection state.
	 */
	public int getOldState() {
		return oldState;
	}

	/**
	 * Sets the current selection state to a new one. This only succeeds if
	 * the given state is a valid one and it's different than the current one. 
	 * If this is the case, the current state
	 * becomes the previous one and the old previous state is lost. All observers
	 * are then notified of the changed state.
	 * @param state The new selection state
	 */
	public void set(int state) {
		if (this.state != state && validState(state)) {
			oldState = this.state;
			this.state = state;
			setChanged();
			notifyObservers();
		}
	}

	/**
	 * Checks if the given state is a valid one (UNSELECTABLE, UNSELECTED,
	 * SELECTED or TARGET)
	 * @param state The state to check
	 * @return true if <state> was valid, false otherwise
	 */
	public boolean validState(int state) {
		if (state < 0 || state > NR_STATES - 1)
			return false;
		else
			return true;
	}

	/**
	 * Checks if this selection state is equal to the given one. Two
	 * selection states are equals if their current state and their old state
	 * are equal.
	 * @param rhs The other selection state
	 * @return true if this selection state is equal to the given one, false otherwise.
	 */
	public boolean equals(SelectionState rhs) {
		return (this.state == rhs.state && this.oldState == rhs.oldState);
	}

	/**
	 * Builds a copy of this object.
	 * @return The copy
	 */
    public Object clone() {
		return new SelectionState(this);
    }

}
