package quest.global.game;

import java.util.*;
import quest.global.util.SerializableObservable;

/**
 * Maintains a set of game spectators. Spectators can be added to and removed from
 * the set at any time. The elements are sorted according to the natural sorting
 * order of GameSpectator objects.</P>
 * <P>
 * Observers of a spectator set are notified whenever a spectator is added to or
 * removed from the set. They can ask if the last spectator was joining or leaving
 * the set with the method lastPresenceChangeWasJoining(), and the name of this spectator
 * with the method getLastPresenceChangingSpectatorName().</P>
 *
 * @author Mathijs den Burger
 * @version 1.0
 */
 
public class SpectatorSet extends SerializableObservable implements Cloneable {

    private TreeSet spectators;
    private String lastPresenceChangingSpectatorName;
    private boolean lastPresenceChangeWasJoining;

	/**
	 * Constructs an empty spectator set
	 */
    public SpectatorSet() {
		spectators = new TreeSet();
		lastPresenceChangingSpectatorName = null;
		lastPresenceChangeWasJoining = false;
    }

	/**
	 * Constructs a spectator set that initially contains all the given spectators
	 * @param spectators All the spectators the are initially contained in this set
	 */
    public SpectatorSet(Vector spectators) {
		this.spectators = new TreeSet(spectators);
    }

	/**
	 * Copy constructor used by clone()
	 * @see quest.global.game.SpectatorSet#clone()
	 */
    protected SpectatorSet(SpectatorSet original) {
		spectators = new TreeSet();
		for (Iterator it = original.spectators.iterator(); it.hasNext(); ) {
		    spectators.add((GameSpectator)(((GameSpectator)it.next()).clone()));
		}
		this.lastPresenceChangingSpectatorName = original.lastPresenceChangingSpectatorName;
		this.lastPresenceChangeWasJoining = original.lastPresenceChangeWasJoining;
    }

	/**
	 * Adds a spectator to the set if he wasn't already present.
	 * All observers are notified if he's added. 
	 * @param spectator The spectator to add
	 */
    public void add(GameSpectator spectator) {
		if (spectators.add(spectator)) {
			lastPresenceChangingSpectatorName = spectator.getName();
			lastPresenceChangeWasJoining = true;
			setChanged();
			notifyObservers();
		}
    }

	/**
	 * Removes a spectator from the set if he's present.
	 * All observers are notified if he's removed. 
	 * @param spectator The spectator to remove
	 */
    public void remove(GameSpectator spectator) {
		if (spectators.remove(spectator)) {
			lastPresenceChangingSpectatorName = spectator.getName();
			lastPresenceChangeWasJoining = false;
			setChanged();
			notifyObservers();
		}
    }

	/**
	 * Removes a spectator with the given name from the set if he's present.
	 * All observers are notified if he's removed. 
	 * @param spectatorName The name of the spectator to remove
	 */
    public void remove(String spectatorName) {
		for (Iterator i = spectators.iterator(); i.hasNext(); ) {
		    if (((GameSpectator)i.next()).getName() == spectatorName) {
				i.remove();
				lastPresenceChangingSpectatorName = spectatorName;
				lastPresenceChangeWasJoining = false;
				setChanged();
				notifyObservers();
				return;
		    }
		}
    }

	/**
	 * Gives the name of the last spectator that changed it's presence (was added
	 * or removed).
	 * @return The name of the last spectator that changed it's presence
	 */
    public String getLastPresenceChangingSpectatorName() {
		return lastPresenceChangingSpectatorName;
    }

	/**
	 * Checks whether the last spectator was added to or removed from the set.
	 * @return true if the last spectator was added to the set, false if he was removed
	 * from the list.
	 */

    public boolean lastPresenceChangeWasJoining() {
		return lastPresenceChangeWasJoining;
    }

	/**
	 * Gives the number of spectators that are currenty contained in the set
	 */
    public int size() {
		return spectators.size();
    }

	/**
	 * Gives all the spectators that are currenty contained in the set, sorted
	 * by their natural order.
	 * @return All the spectators
	 * @see quest.global.game.GameSpectator#compareTo(Object)
	 */
    public Vector elements() {
		Vector result = new Vector(spectators.size());
		for (Iterator i = spectators.iterator(); i.hasNext(); ) {
		    result.add(i.next());
		}
		return result;
    }

	/**
	 * Builds a copy of this object
	 * @return The copy
	 */
    public Object clone() {
		return new SpectatorSet(this);
    }

}
