package quest.global.game;

import quest.global.util.ExtraMath;
import java.awt.Color;
import java.io.Serializable;

/**
 * <P>
 * The objects a game user has to find in the 3D maze.
 * A treasure is characterized by it's color and it's symbol. There are a number
 * of colors and symbols, and every possible combination of these two defines
 * one unique treasure. Both are supplied in the form of ID's, so the e.g.
 * a visualisation component may choose it's own real-world representation of
 * each color and symbol. ID's range from 0 to MAX_NR_... - 1.</P>
 * <P>
 * Besides this color and symbol, every treasure has a find spot too. This is
 * the MazeCorridor3D object that represents the corridor in which this treasure
 * can be found in the 3D maze.</P>
 *
 * @author Mathijs den Burger
 * @version 1.0
 */
 
public class Treasure implements Selectable, Cloneable, Serializable {

    /**
     * The number of colors a treasure can have
     */
    public static final int MAX_NR_COLORS = 4;
    
    /**
     * The number of symbols a treasure can have
     */
    public static final int MAX_NR_SYMBOLS = 8;
    
   	/**
   	 * The maximum possible number of treasures
   	 */ 
    public static final int MAX_NR_TREASURES = MAX_NR_COLORS * MAX_NR_SYMBOLS;

    private int colorId;
    private int symbolId;
    private MazeCorridor3D findSpot;
    private SelectionState selectionState;

	/**
	 * Constructs a treasure with the given color and symbol. Values that
	 * lie outside the permitted range are clamped. The find spot is set to null,
	 * the selection state to the default selection state.
	 * @param colorId The color of the treasure, 0 <= colorId <= MAX_NR_COLORS
	 * @param symbolId The symbol of the treasure, 0 <= symbolId <= MAX_NR_SYMBOLS
	 * @see quest.global.util.ExtraMath#clamp(int, int, int)
	 */
    public Treasure(int colorId, int symbolId) {
		this.colorId = ExtraMath.clamp(colorId, 0, MAX_NR_COLORS - 1);
		this.symbolId = ExtraMath.clamp(symbolId, 0, MAX_NR_SYMBOLS - 1);
		findSpot = null;
		this.selectionState = new SelectionState();
    }

	/**
	 * Copy constructor used by clone()
	 * @see quest.global.game.Treasure#clone()
	 */
    protected Treasure(Treasure original) {
		colorId = original.colorId;
		symbolId = original.symbolId;
		findSpot = original.findSpot; // WARNING: find spot should be changed manually after calling clone() with setFindSpot()
		selectionState = (SelectionState)original.selectionState.clone();
    }

	/**
	 * Gives the color of the treasure
	 * @return The color of the treasure
	 */
    public int getColorId() {
	return colorId;
    }

	/**
	 * Gives the symbol of the treasure
	 * @return The symbol of the treasure
	 */
    public int getSymbolId() {
	return symbolId;
    }

    /**
     * Gives the maze corridor in which this treasure can be found.
     * @return The maze corridor in which this treasure can be found, or null
     * if the find spot isn't set yet
     * @see quest.global.game.Treasure#setFindSpot(MazeCorridor3D)
     */
    public MazeCorridor3D getFindSpot() {
    	return findSpot;
    }

	/**
	 * Sets the maze corridor in which this treasure can be found
	 * @param corridor The maze corridor in which this treasure can be found
	 */
    public void setFindSpot(MazeCorridor3D corridor) {
    	findSpot = corridor;
    }

	/**
	 * Gives the selection state of this treasure
	 * @return The selection state
	 */
    public SelectionState getSelectionState() {
		return selectionState;
    }

	/**
	 * Generates the given amount of different treasures. If nrTreasures is bigger
	 * than the possible number of treasures, you ony get the maximum number of
	 * possible treasures.
	 * @param nrTreasures The number of treasures to generate
	 * @return The generated treasures
	 */
    public static Treasure[] generateTreasures(int nrTreasures) {
		Treasure[] result = new Treasure[nrTreasures];
		int nrGenerated = 0;
	
		for (int symbol = 0; symbol < MAX_NR_SYMBOLS; symbol++) {
		    for (int color = 0; color < MAX_NR_COLORS; color++) {
				result[nrGenerated] = new Treasure(color, symbol);
				nrGenerated += 1;
				if (nrGenerated == nrTreasures) return result;
		    }
		}
		return result;
    }

	/**
	 * Checks if this treasure is equal to the given one. Two treasures are equal
	 * if they have the same color and symbol
	 */
    public boolean equals(Object rhs) {
		if (rhs instanceof Treasure) {
		    Treasure treasure = (Treasure)rhs;
		    if (treasure.colorId == colorId && treasure.symbolId == symbolId) return true;
		}
		return false;
    }

    /**
     * Builds a copy of this object
     * @return The copy
     */
    public Object clone() {
		return new Treasure(this);
    }

}
