package quest.global.login;

import java.io.Serializable;
import quest.global.net.Message;

/**
 * Contains all login messages used for communication between client and server.
 * Login messages are used for logging in/out and for creating a new user.
 * <PRE>
 * NOTE: A / B means that either A or B is sent.
 *
 * Packet          | Reply				| Description
 * ------------------------------------------------------------
 * C2S_Login	     S2C_AllowLogin /     Client wants to login
 *                   S2C_DenyLogin
 * C2S_NewUser       S2C_AllowLogin /     Client wants to create a new user
 *                   S2C_DenyLogin        and wants to login with this new username
 * C2S_Logout        S2C_Exit             Client wants to logout
 *
 * NOTE: This message is not to be uses for communication, this functions only
 *       as a root in the hierarchy tree.
 *
 * @author    Merijn Evertse & Sander Brandenburg
 */ 
public class LoginMessage extends Message implements Serializable {
	/**
	 * The client wants to login to the system. If this is succesfull the username of the client
	 * is registerred at the server daemon and the client is now granted permission to
	 * communicate with the chatserver.
	 *
	 * @author    Merijn Evertse & Sander Brandenburg
	 */
	public static class C2S_Login extends LoginMessage {
		/**
		 * The username of the client.
		 */
		public String username;
		
		/**
		 * The password of the client.
		 */
		public String password;
		
		/**
		 * Initializes the object with the given parameters.
		 *
		 * @param    username    The username of the client
		 * @param    password    The password of the client
		 */
		public C2S_Login(String username, String password) {
			this.username = username;
			this.password = password;
		}
	}

	/**
	 * The client wants to create a new user account and to login with the new information.
	 * If this is succesfull the username of the client
	 * is registerred at the server daemon and the client is now granted permission to
	 * communicate with the chatserver.
	 *
	 * @author    Merijn Evertse & Sander Brandenburg
	 */
	public static class C2S_NewUser extends LoginMessage {
		/**
		 * The username of the client.
		 */
		public String username;
		/**
		 * The password of the client.
		 */
		public String password;
		
		/**
		 * Initializes the object with the given parameters.
		 *
		 * @param    username    The username of the client
		 * @param    password    The password of the client
		 */
		public C2S_NewUser(String username, String password) {
			this.username = username;
			this.password = password;
		}
	}

	/**
	 * The client wants to logout of the system. He is unregisterred at the server daemon.
	 *
	 * @author    Merijn Evertse & Sander Brandenburg
	 */
	public static class C2S_Logout extends LoginMessage {	
		/**
		 * Initializes the object.
		 */
		 public C2S_Logout() {}
	}
	
	/**
	 * The client is allowed to login. A welcome message is sent to the client as well.
	 *
	 * @author    Merijn Evertse & Sander Brandenburg
	 */
	public static class S2C_AllowLogin extends LoginMessage {
		/**
		 * The welcome message for the client.
		 */
		public String welcomeMsg;
		
		/**
		 * Initializes the object with the given parameter.
		 *
		 * @param    welcomeMsg    The welcome message for the client
		 */
		public S2C_AllowLogin(String welcomeMsg) {
			this.welcomeMsg = welcomeMsg;
		}
	}
	
	/**
	 * The client is denied access to the system. A explanation is sent to the client as well.
	 *
	 * @author    Merijn Evertse & Sander Brandenburg
	 */
	public static class S2C_DenyLogin extends LoginMessage {
		/**
		 * The explanantion of the denial.
		 */
		public String explanation;
		
		/**
		 * Initializes the object with the given parameter.
		 *
		 * @param    explanation    The explanantion of the denial
		 */
		public S2C_DenyLogin(String explanation) {
			this.explanation = explanation;
		}
	}
	
	/**
	 * The client is informed that he can safely exit the program now.
	 *
	 * @author    Merijn Evertse & Sander Brandenburg
	 */
	public static class S2C_Exit extends LoginMessage {
		/**
		 * A goodbye message for the client.
		 */
		public String exitMessage;
		
		/**
		 * Initializes the object with the given parameter.
		 *
		 * @param    exitMessage    A goodbye message for the client
		 */
		public S2C_Exit(String exitMessage) {
			this.exitMessage = exitMessage;
		}
	}
}
