package quest.global.net;

import java.io.*;
import java.net.*;
import javax.vecmath.Tuple3d;
import quest.global.chat.*;

/**
 * Defines an creates a message stream between a client and a server.
 *
 * @author    Merijn Evertse
 */
public class MessageStream {
  
  /**
   * Defines the side as the client side of the communication channel.
   */
  public static final int CLIENT_SIDE = 1;
  
  /**
   * Defines the side as the server side of the communication channel.
   */
  public static final int SERVER_SIDE = 2;

  private Socket connection;
  private InputStream input;
  private ObjectInputStream inputStream;
  private OutputStream output;
  private ObjectOutputStream outputStream;

  /**
   * Creates a message stream between a client and a server, based on the given communication
   * channel and the given side of the channel.
   *
   * @param    connection    A connection between two parties
   * @param    side	         The side of the connection of this instance
   */
  public MessageStream(Socket connection, int side) {
    this.connection = connection;
    switch(side) {
    case CLIENT_SIDE:
      try {
		output = connection.getOutputStream();
		outputStream = new ObjectOutputStream(output);
		input = connection.getInputStream();
		inputStream = new ObjectInputStream(input);
      } catch (IOException e) {
		System.out.println(e.getMessage());
      }
      break;
    case SERVER_SIDE:
      try {
		input = connection.getInputStream();
		inputStream = new ObjectInputStream(input);
		output = connection.getOutputStream();
		outputStream = new ObjectOutputStream(output);
      } catch (IOException e) {
		System.out.println(e.getMessage());
      }
      break;
    default:
      System.out.println("Something went wrong in MessageStream");
      System.exit(1);
    }
  }

  /**
   * Sends a message over the given connection
   *
   * @param     message        The message to be send
   * @param     connection     The connection for the communication
   */
  public void sendMessage(Message message) throws IOException {

	System.out.println("MessageStream: sending " + message.toString());
	outputStream.writeObject(message);
	outputStream.flush();
  }

  /**
   * Receives a message from the given connection
   *
   * @param     connection     The connection for the communication
   * @return    The received message
   */
  public Message receiveMessage() {
    Object object = new Object();
    try {
      object = inputStream.readObject();
    } catch (InterruptedIOException e) {
	return null;
    } catch (IOException e) {
	e.printStackTrace();
      System.out.println("MessageStream IOException: " + e.getMessage());
    } catch (ClassNotFoundException e) {
      System.out.println("MessageStream ClassNotFoundException: " + e.getMessage());
    }

    Message message = null;
    try{
	  message = (Message) object;
    } catch (ClassCastException e) {
	  System.out.println(e.getMessage());
    }
    System.out.println("MessageStream: received " + message.toString());
    return message;
  }

    public void setTimeout(int millisecs) {
	try {
	    connection.setSoTimeout(millisecs);
	} catch (SocketException e) {
	    System.out.println(e.getMessage());
	}
    }

  /**
   * Returns whether a message is available on the message stream or not.
   *
   * @return    TRUE:    If a message is available
   *            FALSE:   If not
   */
  public boolean messageAvailable() {
    boolean result = false;
    try {
      result = (input.available() > 0);
      if (input.available() != 0) System.out.println("Available " + input.available());
    } catch (IOException e) {
      System.out.println(e.getMessage());
    }
    return result;
  }

	/**
	 * Closes the message stream.
	 *
	 * @exception IOException    When an error occurs during the closing of the streams 
	 */
  	public void close() throws IOException {
		inputStream.close();
		outputStream.close();
  	}
}
    
