package quest.global.util;

import java.util.Observable;
import java.security.InvalidParameterException;
import quest.global.util.SerializableObservable;

/**
 * Implements a simple finite state machine.
 * A finite state machine consists of a finite number of states and a number 
 * of possible transitions between those states. 
 * This corresponds to an directed graph in which the nodes are
 * states and the edges are transitions.</P>
 * <P>
 * The states are always represented as integers, from 0 up to the total number
 * of states - 1. So a finite state machine of 4 states will contain the states 
 * 0, 1, 2 and 3.</P>
 * <P>
 * Each finite state machine must be started before it can be used
 * (with the start method). The state of a finite state machine that isn't started
 * yet is always NOT_STARTED. Users can retrieve the current state, check if
 * a transition to another given state is possible and go to a new state.</P>
 * <P>
 * All observers are notified when the finite state machine enters a new state.
 * This includes starting it.</P>
 *
 * @author Mathijs den Burger
 */

public class FiniteStateMachine extends SerializableObservable implements Cloneable {

	/**
	 * The number of states that specify a transition together.
	 */
	public static final int TRANSITION_LENGTH = 2;
	
	/**
	 * The value of the current state when the finite state machine isn't 
	 * started yet.
	 */
	public static final int NOT_STARTED = -1;

	private int[][] transitions;
	private int nrStates, currentState;

	/**
	 * Constructs a finite state machine with the given number of states and 
	 * the given transitions.
	 * @param nrStates The number of states this finite state machine will contain
	 * @param trasitions The possible transitions between those states. This is
	 * a 2-dimensional array, of which each element should consist of two integers
	 * that indicate a possible transition from the first to the second state.<BR>
	 * Example:<BR>
	 * <CODE>
	 * int[][] transitions = { {0, 0}, {0, 1}, {1, 2}, {2, 0} };
	 * FiniteStateMachine fsm = new FiniteStateMachine(3, transitions);
	 * </CODE>
	 * This will create a finite state machine with 3 different states (0, 1, and 2)
	 * and these possible transitions:
	 * <PRE>
	 *  +---+
	 *  +-> 0 --> 1 --> 2
	 *      ^           |
	 *      +-----------+
	 * </PRE>
	 */
	public FiniteStateMachine(int nrStates, int[][] transitions) {
		checkParameters(nrStates, transitions);
		this.nrStates = nrStates;
		this.transitions = transitions;
		currentState = NOT_STARTED;
	}
	
	/**
	 * Copy constructor used by clone()
	 * @see quest.global.util.FiniteStateMachine#clone()
	 */
    protected FiniteStateMachine(FiniteStateMachine original) {
		// copy transitions
		transitions = new int[original.transitions.length][TRANSITION_LENGTH];
		for (int i = 0; i < original.transitions.length; i++) {
		    for (int j = 0; j < original.transitions[i].length; j++) {
			transitions[i][j] = original.transitions[i][j];
		    }
		}
		// copy other variables
		nrStates = original.nrStates;
		currentState = original.currentState;
    }

    private void checkParameters(int nrStates, int[][] transitions) {
	    for (int state = 0; state < transitions.length; state++) {
			if (transitions[state].length != TRANSITION_LENGTH)
				throw new InvalidParameterException("Illegal transition length: " + transitions[state].length + ", allowed value: " + TRANSITION_LENGTH);
			for (int i = 0; i < TRANSITION_LENGTH; i++) {
			    checkState(transitions[state][i], nrStates);
			}	
	    }		
    }
	
	private void checkState(int state, int nrStates) {
		if (state < 0 || state > nrStates - 1) {
			throw new InvalidParameterException("Illegal state: " + state + ", allowed values: [0, " + (nrStates - 1) + "]");
		}
	}
	
	/**
	 * Gives the current state of the finite state machine.
	 * @return The current state. If the machine isn't started yet, NOT_STARTED is returned.
	 */
	public int getCurrentState() {
		return currentState;
	}
	
	/**
	 * Starts the finite state machine in the given state.
	 * @param initialState The state the machine will be in after it is started.
	 */
	public void start(int initialState) {
		checkState(initialState, nrStates);
		currentState = initialState;
		setChanged();
		notifyObservers();
	}

    /**
     * Checks if this finite state machine has already been started
     * @return true if this finite state machine is already started, false if not
     */
    public boolean isStarted() {
	return (currentState != NOT_STARTED);
    }
	
	/**
	 * Puts the finite state machine into a new state, if the transition from
	 * the current state to the given one is valid (that is: specified when the
	 * machine was constructed). All observers are notified when of the new
	 * state. If the transition isn't valid, nothing happens.
	 * @param state The new state to put the finite state machine into.
	 * @return true if the machine is in the specified state now, false if not
	 * (which means the state wasn't a valid next state)
	 */
	public boolean setState(int state) {
		if (currentState != NOT_STARTED && validNextState(state)) {
			currentState = state;
			setChanged();
			notifyObservers();
			return true;
		} else {
			System.out.println("Illegal state transition: " + currentState + " -> " + state);
			return false;
		}
	}
	
	/**
	 * Checks if the given state if a valid next state from the current state.
	 * That is: the transition <current state> --> <nextState> is specified when
	 * the machine was constructed.
	 * @param nextState The state to check
	 * @return true if the given state is a valid next state, false otherwise.
	 */
	public boolean validNextState(int nextState) {
		for (int i = 0; i < transitions.length; i++) {
			if (transitions[i][0] == currentState && transitions[i][1] == nextState)
				return true;
		}
		return false;
	}

	/**
	 * Builds a copy of this object
	 * @return The copy
	 */
    public Object clone() {
		return new FiniteStateMachine(this);
    }
			
}	
