package quest.server.chat;

import java.util.*;

import quest.global.chat.*;

/**
 * Represents a chat channel on the server, with users in it.
 *
 * @author    Merijn Evertse
 */
class Channel {

  private static final int DEFAULT_ADMINNR = 0;

  /**
   * The string to show when the user isn't in a channel.
   */
  public static final String NOCHANNEL = "No Channel";
  
  private int adminnr;
  private String channelname;
  private Vector playernames;
  private Vector spectatornames;
  private boolean locked;
  private int status;
  
  /**
   * Creates a complete channel.
   *
   * @param    channelname    The name of the new channel
   * @param    adminname      The name of the administrator of the channel
   */
  Channel(String channelname, String adminname) {
	this.channelname = channelname;
	playernames      = new Vector();
	spectatornames   = new Vector();
	addPlayer(adminname);
	locked           = false;
	status           = ChatMessage.CHANMODE_BEGIN;
  }
  
  /**
   * Adds a player to the channel, if the person isn't already in it.
   *
   * @param    username    The name of the client
   */
  public void addPlayer(String username) {
	if (!userExists(username)) {
	  System.out.println("Player toegevoegd " + username + " " + channelname);
	  playernames.addElement(username);
	}
  }
  
  /**
   * Adds a spectator to the channel, if the person isn't already in it.
   *
   * @param    username    The name of the client
   */
  public void addSpectator(String username) {
	if (!userExists(username)) {
	  System.out.println("Spectator toegevoegd " + username + channelname);
	  spectatornames.addElement(username);
	}
  }
  
  /**
   * Removes a client from the channel.
   *
   * @param    username    The name of the client
   */
  public void removeUser(String username) {
	System.out.println("User verwijderd " + username);
	playernames.removeElement(username);
	spectatornames.removeElement(username);
  }
  
  /**
   * Returns the name of the channel.
   *
   * @return    The name of the channel
   */
  public String getChannelname() {
	return channelname;
  }

  /**
   * Returns a list of the playernames.
   *
   * @return    The playernames
   */
  public Vector getPlayernames() {
	return playernames;
  }

  /**
   * Returns a list of the spectatornames.
   *
   * @return    The spectatornames
   */
  public Vector getSpectatornames() {
	return spectatornames;
  }

  /**
   * Returns the name of the administrator of the channel.
   *
   * @return    The name of the administrator
   */
  public String getAdminname() {
	if (playernames.size() != 0) {
	  return (String) playernames.elementAt(adminnr);
	}
	return null;
  }
  
  /**
   * Checks whether a the user is a member of this channel.
   *
   * @param    username    The name of the user
   * @return   TRUE:  if the user is a member of the channel
   *           FALSE: if not
   */
  public boolean userExists(String username) {
	return (playernames.contains(username) || spectatornames.contains(username));
  }
  
  /**
   * Chooses a new admin from the players and returns the name of this new admin.
   *
   * @return    The name of the newly chosen admin of the channel
   */
  public String chooseAdmin() {
	adminnr = DEFAULT_ADMINNR;
	return getAdminname();
  }
  
  /**
   * Sets the admin of the channel to the given username.
   *
   * @param    username    The name of the user
   */
  public void setAdmin(String username) {
	for (int i = 0; i < playernames.size(); i++) {
	  if (((String) playernames.elementAt(i)).equals(username)) {
		adminnr = i;
	  }
	}
  }
  
  /**
   * Locks or unlocks the channel.
   *
   * @param    locked    TRUE:  if the channel is to be locked
   *					   FALSE: if the channel is to be unlocked
   */
  public void setLock(boolean locked) {
	this.locked = locked;
  }
  
  /**
   * Sets the status of the channel.
   *
   * @param    status    The new status of the channel
   */
  public void setStatus(int status) {
	this.status = status;
  }
  
  /**
   * Returns the status of the channel.
   *
   * @return    The status of the channel
   */
  public int getStatus() {
	return status;
  }
  
  /**
   * Returns whether the channel is empty or not.
   *
   * @return    TRUE:  if the channel is empty
   *            FALSE: if not
   */
  public boolean isEmpty() {
	return playernames.size() == 0 && spectatornames.size() == 0;
  }
  
  /** Returns the list of users in the channel.
   *
   * @return    The list of users
   */
  public Vector getUsernames() {
	Vector result = new Vector();
	result.addAll(getPlayernames());
	result.addAll(getSpectatornames());
	System.out.println(result.size());
	if (channelname.equals(ChatMessage.GLOBALCHANNELNAME)) result.removeElementAt(adminnr);
	System.out.println(result.size());
	return result;
  }
  
  /**
   * Returns all info of the channel, structured in a ChannelInfo object.
   *
   * @return    A ChannelInfo object containing all info of this channel
   */
  public ChannelInfo getChannelInfo() {
	return new ChannelInfo(channelname, getAdminname(), status, 
						   locked, playernames.size(), spectatornames.size());
  }
}

