package quest.server.chat;

import quest.server.net.*;
import quest.global.chat.*;
import quest.global.net.*;

import java.util.Vector;

/**
 * The channel server handles incoming channel messages for the chat server.
 * All client request for info about channels, or update requests for changes in
 * the channel info are handled here. This means this server collects all info about
 * channels for full updates, and a defined intervals collects all info about
 * changed channels and propagates this to subscribed users.
 *
 * @author    Merijn Evertse
 */
public class ChannelServer implements MessageServerListener {

  private Channels channels;
  private ServerDaemon sd;
  private NetUsers netUsers;
  private static ChannelServer uniqueInstance;

  /**
   * The constructor initializes the server and registers the server at the 
   * server daemon, to make itself available for packet processing.
   */ 
  public ChannelServer() {
	if (uniqueInstance != null)
	  throw new RuntimeException("singleton constraint violated");

	uniqueInstance = this;
	
	channels = Channels.getInstance();
    sd = ServerDaemon.getInstance();
  	netUsers = NetUsers.getInstance();
		
	sd.registerProces(ChannelMessage.class, this, ServerDaemon.USER_BASED);

	// start thread for handling auto updates of clients
	startAutoUpdates();
  }

  /**
   * Returns a reference to the server.
   *
   * @return    A reference to this class
   */
  public static ChannelServer getInstance() {
	if (uniqueInstance == null) {
	  throw new RuntimeException("singleton class has to be initialized by the constructor first");
	}
	return uniqueInstance;
  }

  /**
   * Handled all incoming channel messages. First the message is checked
   * and processed and then the reply messages are constructed and
   * given to the server daemon to be send.
   *
   * @param    username   The sender of the message
   * @param    msg        The message
   */    	
  public void incomingMessage(final String username, final Message msg) {		
	final Runnable ra = new Runnable() {
	  public void run() {
		ChannelMessage cm = (ChannelMessage) msg;
		
		if (cm instanceof ChannelMessage.C2S_Request) {
		  request(username, cm);
		} else if (cm instanceof ChannelMessage.C2S_Subscribe) {
		  subscribe(username, cm);
		} else if (cm instanceof ChannelMessage.C2S_UnSubscribe) {
		  unsubscribe(username, cm);
		} else {
		  System.out.println("Received unsupported message: " + cm);
		}
	  }
	};
	(new Thread(ra)).start();
  }
  
  private synchronized void request(String username, ChannelMessage cm) {
	// send full update
	sendFullUpdate(username);
  }

  private synchronized void subscribe(String username, ChannelMessage cm) {
	NetUser nu = netUsers.getUser(username);
	// set subscribed flag
	nu.subscribe(ChannelMessage.S2C_PartialUpdate.class);
	
	// he was previously not subscribed, so send full update immediately
	sendFullUpdate(username);
  }
  
  private synchronized void unsubscribe(String username, ChannelMessage cm) {
	NetUser nu = netUsers.getUser(username);
	// unset subscribed flag
	nu.unsubscribe(ChannelMessage.S2C_PartialUpdate.class);
  }

  private void sendFullUpdate(String username) {
  	ChannelMessage.S2C_FullUpdate fu = new ChannelMessage.S2C_FullUpdate(channels.getAllChannelInfos());
    sd.sendMessageToUser(username, fu);
  }
  
  private void startAutoUpdates() {
	final Runnable ra = new Runnable() {
	  public void run() {
		while(true) {
		  Vector subscribedUsers = (Vector) netUsers.getSubscribedUsers(ChannelMessage.S2C_PartialUpdate.class).clone();
		  ChannelMessage.S2C_PartialUpdate pu = new ChannelMessage.S2C_PartialUpdate(channels.getAllUpdates());
		  sd.sendMessage(subscribedUsers, pu);
  	
		  // wait some time before trying again
		  try {
			Thread.currentThread().sleep(ServerConfiguration.getInstance().getSleeptime());
		  } catch (InterruptedException e) {
			System.out.println(e.getMessage());
		  }   
		}
	  }
	};
	//(new Thread(ra)).start();
  }
}


