package quest.server.chat;

import java.util.*;
import quest.global.chat.*;

/**
 * This class contains als the channels of the chatserver. 
 * For using these channel an instance should be taken from this class and all 
 * operations on channels should go through this class.
 * This is done for administration. All operations on the channels are 
 * administrated so changes in the channels can be collected as information for the 
 * clients.
 *
 * @author    Merijn Evertse
 */
public class Channels {
  private static Hashtable channels;
  private static Vector channelsChanged;
  
  protected static Channels uniqueInstance = null;
  
  /**
   * Returns a reference to this class.
   *
   * @return    A reference
   */
  public static Channels getInstance() {
	if (uniqueInstance == null)
	  return new Channels();
	
	return uniqueInstance;
  }

  /**
   * Initialized the channels for the chatServer.
   */
  protected Channels() {
	if (uniqueInstance != null)
	  throw new RuntimeException("there's already an instance present");
	
	uniqueInstance = this;
	channels = new Hashtable();
	channelsChanged = new Vector();
  }
  
  /**
   * Add a channel to the list of channels.
   *
   * @param    channelname    The name of the new channel
   * @param    adminname      The name of the creator of the channel
   */
  public void addChannel(String channelname, String adminname) {
	if (channelname != null) {
	  System.out.println("Channel " + channelname + " added to chatserver");
	  Channel channel = new Channel(channelname, adminname);
	  channels.put(channelname, channel);
	  channelsChanged.addElement(new ChannelMessage.S2C_PartialUpdate.Update(
									 ChannelMessage.S2C_PartialUpdate.UPDATE_INSERT,
									 channel.getChannelInfo()));
	}
  }
  
  /**
   * Removes a channel from the list of channels.
   *
   * @param    channelname    The name of the channel to remove
   */
  public void removeChannel(String channelname) {
	if (channelname != null) {
	  System.out.println("Channel " + channelname + " removed from chatserver");
	  Channel channel = getChannel(channelname);
	  channels.remove(channelname);
	  channelsChanged.addElement(new ChannelMessage.S2C_PartialUpdate.Update(
									 ChannelMessage.S2C_PartialUpdate.UPDATE_DELETE,
									 new ChannelInfo(channel.getChannelname(), null, 
													 ChatMessage.CHANMODE_UNKNOWN, false, 0, 0)));
	}
  }
  
  /**
   * Chooses a new admin for the channel and return the name of the newly chosen
   * admin.
   *
   * @param    channelname    The name of the channel
   * @return   The name of the newly chosen admin
   */
  public String chooseAdmin(String channelname) {
	Channel channel = getChannel(channelname);
	String admin = channel.chooseAdmin();
	channelsChanged.addElement(new ChannelMessage.S2C_PartialUpdate.Update(
								   ChannelMessage.S2C_PartialUpdate.UPDATE_MODIFY,
								   channel.getChannelInfo()));
	return admin;
  }
  
  /**
   * Adds a user to the channel with the given name. If the user is the first
   * in the channel the person is added as a player and given admin rights.
   * If an administrator already is a member of the channel the new user
   * is added as a spectator and should ask the admin for becoming a player.
   *
   * @param    channelname    The name of the channel
   * @param    username       The name of the user
   */
  public void addUser(String channelname, String username) {
	Channel channel = getChannel(channelname);
	if (channel.getAdminname() == null && !channelname.equals(ChatMessage.GLOBALCHANNELNAME)) {
	  channel.addPlayer(username);
	  channel.setAdmin(username);
	} else {
	  channel.addSpectator(username);
	}
	channelsChanged.addElement(new ChannelMessage.S2C_PartialUpdate.Update(
								   ChannelMessage.S2C_PartialUpdate.UPDATE_MODIFY,
								   channel.getChannelInfo()));
  }
  
  /**
   * Removes a user from the channel. If the user was the last user of the channel the channel is
   * removed as well, and this is indicated by returning true.
   *
   * @param channelname    The name of the channel the user is in
   * @param username       The name of the user to delete
   * @return    TRUE:      when the channel is deleted as well
   *            FALSE:     when the channel is only modified
   */
  public boolean removeUser(String channelname, String username) {
	Channel channel = getChannel(channelname);
	channel.removeUser(username);
	// if this was the last user of the channel, do UPDATE_DELETE else UPDATE_MODIFY
	if (channel.isEmpty()) {
	  removeChannel(channelname);
	  return true;
	} else {
	  channelsChanged.addElement(new ChannelMessage.S2C_PartialUpdate.Update(
									 ChannelMessage.S2C_PartialUpdate.UPDATE_MODIFY,
									 channel.getChannelInfo()));
	}
	return false;
  }
  
  /**
   * Sets the usermode of a user with the given name, in the channel with the
   * given name, to the given usermode.
   *
   * @param    channelname     The name of the channel the user is in
   * @param    username        The name of the user
   * @param    umode           The new usermode of the user
   */
  public void setUsermode(String channelname, String username, int umode) {
	Channel channel = getChannel(channelname);
	if (umode == ChatUser.USERMODE_PLAYER) {
	  // user becomes player, instead of spectator
	  channel.removeUser(username);
	  channel.addPlayer(username);	
	} else {
	  // other mode changes are not important for number of players and spectators
	  // TODO MOET WEL
	}
	channelsChanged.addElement(new ChannelMessage.S2C_PartialUpdate.Update(
								   ChannelMessage.S2C_PartialUpdate.UPDATE_MODIFY,
								   channel.getChannelInfo()));
  }
  
  /**
   * (un)locks the channel with the given name.
   *
   * @param    channelname    The name of the channel
   * @param    locked         TRUE:  if the channel should be locked
   *                          FALSE: if the channel should be unlocked
   */ 
  public void setLock(String channelname, boolean locked) {
	Channel channel = getChannel(channelname);
	channel.setLock(locked);
	channelsChanged.addElement(new ChannelMessage.S2C_PartialUpdate.Update(
								   ChannelMessage.S2C_PartialUpdate.UPDATE_MODIFY,
								   channel.getChannelInfo()));
  }
  
  /**
   * Sets the status of the channel to the given value.
   *
   * @param    channelname    The name of the channel
   * @param    status         The new status of the channel
   */
  public void setStatus(String channelname, int status) {
	Channel channel = getChannel(channelname);
	channel.setStatus(status);
	channelsChanged.addElement(new ChannelMessage.S2C_PartialUpdate.Update(
								   ChannelMessage.S2C_PartialUpdate.UPDATE_MODIFY,
								   channel.getChannelInfo()));
  }
  
  /**
   * Returns the channel with the given name.
   *
   * @param    channelname    The name of the channel to return
   * @return   The channel with the given name, or null if it
   *	       can't be found.
   */ 
  public Channel getChannel(String channelname) {
	if (channelname != null) {
  	  return (Channel) channels.get(channelname);
  	}
  	return null;
  }
  
  /**
   * Returns the complete list of usernames of the users in the channel.
   *
   * @param    channelname    The name of the channel
   * @return   The list of usernames
   */
  public Vector getUsernames(String channelname) {
	return (getChannel(channelname)).getUsernames();
  }

  /**
   * Returns a list of playernames of the users in the channel
   *
   * @param    channelname    The name of the channel
   * @return   The list of playernames
   */
  public Vector getPlayernames(String channelname) {
	return (getChannel(channelname)).getPlayernames();
  }

  /**
   * Returns a list of spectatornames of the users in the channel
   *
   * @param    channelname    The name of the channel
   * @return   The list of spectatornames
   */
  public Vector getSpectatornames(String channelname) {
	return (getChannel(channelname)).getSpectatornames();
  }

  /**
   * Returns the name of the administrator of the channel.
   *
   * @param    channelname    The name of the channel
   * @return   The name of the admin of the channel
   */
  public String getAdminname(String channelname) {
	return (getChannel(channelname)).getAdminname();
  }
  
  /**
   * Returns the status of the channel.
   *
   * @param    channelname    The name of the channel
   * @return   The status of the channel or CHANMODE_UNKNOWN
   */
  public int getStatus(String channelname) {
	if (channelname != null && !(channelname.equals(Channel.NOCHANNEL))) 
	  return (getChannel(channelname)).getStatus();
	return ChatMessage.CHANMODE_UNKNOWN;
	
  }
  
  /**
   * Returns whether a channel with the given name exists.
   *
   * @param    channelname    The name of the channel
   * @return   TRUE:  if the channel exists
   *           FALSE: if not
   */
  public boolean channelExists(String channelname) {
	return (getChannel(channelname) != null);
  }

  /**
   * Returns the name of all the channels currently active.
   *
   * @return    The name of all the channels
   */
  public Vector getAllChannelnames() {
	Vector channelnames = new Vector();
	for (Enumeration e = channels.keys(); e.hasMoreElements(); ) {
	  channelnames.addElement(((String) e.nextElement()));
	}
	return channelnames;
  }
  
  /**
   * Returns an array of channelInfo messages, which describe the current state of the
   * channel database.
   *
   * @return    The current state of the database
   */
  public ChannelInfo[] getAllChannelInfos() {
	ChannelInfo[] result = new ChannelInfo[channels.size() - 1];  // one less because of global channel
	int count = 0;
	
	for (Enumeration e = channels.elements(); e.hasMoreElements(); ) {
	  Channel channel = (Channel) e.nextElement();
	  if (!channel.getChannelname().equals(ChatMessage.GLOBALCHANNELNAME)) {
		ChannelInfo ci = channel.getChannelInfo();
		result[count++] = ci;
	  }
	}
	return result;
  }
  
  /**
   * Returns all updates about the changed channels since last time.
   * Note that this means that after this request all updates will be made final.
   *
   * @return    The complete pile of updates
   */
  public ChannelMessage.S2C_PartialUpdate.Update[] getAllUpdates() {
	ChannelMessage.S2C_PartialUpdate.Update[] result = new ChannelMessage.S2C_PartialUpdate.Update[channelsChanged.size()];
	
	for (int i = 0; i < channelsChanged.size(); i++) {
	  result[i] = (ChannelMessage.S2C_PartialUpdate.Update) channelsChanged.elementAt(i);
	}
	return result;
  }
}



