package quest.server.chat;

import java.util.*;
import quest.global.net.*;
import quest.global.chat.*;

/**
 * The message listener for the chat server.
 * Handles all incoming messages for the chat server, changes the
 * states of the chat environment en sends messages back to the clients.
 *
 * @author    Merijn Evertse
 */
public class ChatMessageServerListener implements MessageServerListener {

  private ChatServer chatServer;
  private Channels channels;

	/**
	 * Initializes the listener.
	 */
	public ChatMessageServerListener() {
	  chatServer = ChatServer.getInstance();
	  channels = channels.getInstance();
	}

	/**
	 * Handles a message received by the communication layer.
	 *
	 * @param    username          The name of the user who sent the message
	 * @param    receivedMessage   The message received from the user
	 */
	public void incomingMessage(final String username, Message receivedMessage) {
		final ChatMessage reqmessage = (ChatMessage) receivedMessage;
		if (reqmessage.channel != null) {
		
		Runnable ra = new Runnable() {
		  public void run() {
			Vector usernames;
			String[] users;
			int[] umodes;
			String adminname;
			Channel channel;
			
			System.out.println("Handler: message received");
			
			try {
			// handle message based on message type
			if (reqmessage instanceof ChatMessage.C2S_Chat) {
			  chat(username, reqmessage);
			}
			else if(reqmessage instanceof ChatMessage.C2S_Whisper) {
			  whisper(username, reqmessage);
			}
			else if (reqmessage instanceof ChatMessage.C2S_Join) {
			  join(username, reqmessage);
			}
			else if(reqmessage instanceof ChatMessage.C2S_Create) {
			  create(username, reqmessage);
			}
			else if(reqmessage instanceof ChatMessage.C2S_Chusermode) {
			  chusermode(username, reqmessage);
			}
			else if(reqmessage instanceof ChatMessage.C2S_Kick) {
			  kick(username, reqmessage);
			}
			else if(reqmessage instanceof ChatMessage.C2S_Lockout) {
			  lockout(username, reqmessage);
			}
			else if(reqmessage instanceof ChatMessage.C2S_Leave) {
			  leave(username, reqmessage);
			}
			else if(reqmessage instanceof ChatMessage.C2S_Login) {
			  login(username, reqmessage);
			}
			else if(reqmessage instanceof ChatMessage.C2S_Logout) {
			  logout(username, reqmessage);
			}
			else if(reqmessage instanceof ChatMessage.C2S_Startgame) {
			  startgame(username, reqmessage);
			}
			else if (reqmessage instanceof ChatMessage.C2S_Stopgame) {
			  stopgame(username, (ChatMessage.C2S_Stopgame) reqmessage);
			}
			else {
			  System.out.println("Unknown chat message received");
			}
			} catch (ClassCastException cce) {
			  System.out.println("WARNING: Malformed packet received " + cce.getMessage());
			  cce.printStackTrace();
			}
		  }
		};
		(new Thread(ra)).start();

		} else {
		  System.out.println("WARNING: Received packet with bad channelname, ignoring packet!");
		}
	}

  private synchronized void chat(String username, ChatMessage reqmessage)
	throws ClassCastException {
	Vector usernames;

	ChatMessage.C2S_Chat message = (ChatMessage.C2S_Chat) reqmessage;
	// client sent text so send all clients in channel this text
	usernames = channels.getUsernames(message.channel);
	
	chatServer.sendMessage(
		new ChatMessage.S2C_Msg(message.channel, username, message.text), 
		usernames);
  }

  private synchronized void whisper(String username, ChatMessage reqmessage) 
	throws ClassCastException {
	Vector usernames;
	
	ChatMessage.C2S_Whisper message = (ChatMessage.C2S_Whisper) reqmessage;
	// client whispers so send single other client in channel this text
	
	// send to source
	usernames = new Vector();
	usernames.addElement(username);
	chatServer.sendMessage(
		new ChatMessage.S2C_Whisperself(message.channel, message.username, message.text),
		usernames);
	
	// send to target
	usernames = new Vector();
	usernames.addElement(message.username);
	chatServer.sendMessage(
		new ChatMessage.S2C_Whisper(message.channel, username, 
									message.text),
		usernames);
  }  

  private synchronized void join(String username, ChatMessage reqmessage)
	throws ClassCastException {
	Vector usernames;
	
	ChatMessage.C2S_Join message = (ChatMessage.C2S_Join) reqmessage;
	// client joins a channel, inform other clients of this as well
	chatServer.joinChannel(username, message.channel);
	
	// notify users of user joining
	usernames = (Vector) (channels.getUsernames(message.channel)).clone();
	usernames.removeElement(username);
	
	chatServer.sendMessage(
		new ChatMessage.S2C_Userjoin(message.channel, username, 
									 chatServer.getUsermode(username, message.channel)), 
		usernames);
	
	// notify user joining of other users
	usernames = (Vector) (channels.getUsernames(message.channel)).clone();
	String[] users = new String[usernames.size()];
	for (int i = 0; i < usernames.size(); i++) {
	  users[i] = (String) usernames.elementAt(i);
	}
	int[] umodes = new int[usernames.size()];
	for (int j = 0; j < usernames.size(); j++) {
	  umodes[j] = chatServer.getUsermode((String) usernames.elementAt(j), 
										 message.channel);
	}			 
	usernames = new Vector();
	usernames.addElement(username);
	
	chatServer.sendMessage(
		new ChatMessage.S2C_Joinself(message.channel, users, umodes, 
									 channels.getStatus(message.channel)),
		usernames);
  }

  private synchronized void create(String username, ChatMessage reqmessage)
	throws ClassCastException {
	Vector usernames;
	
	ChatMessage.C2S_Create message = (ChatMessage.C2S_Create) reqmessage;
	// client creates channel and becomes admin of the channel
	chatServer.addChannel(message.channel, username);
	
	usernames = new Vector();
	usernames.addElement(username);
	String[] users = new String[1];
	users[0] = username;
	int[] umodes = new int[1];
	umodes[0] = chatServer.getUsermode(username, message.channel);
	
	chatServer.sendMessage(
		new ChatMessage.S2C_Joinself(message.channel, users, umodes, ChatMessage.CHANMODE_BEGIN),
		usernames);
  }

  private synchronized void chusermode(String username, ChatMessage reqmessage)
	throws ClassCastException {
	Vector usernames;
	
	ChatMessage.C2S_Chusermode message = (ChatMessage.C2S_Chusermode) reqmessage;
	// client asks for mode change, notify users if succesfull
	usernames = channels.getUsernames(message.channel);
	
	boolean succesfull = chatServer.changeUsermode(message.username, 
												   message.channel, message.mode);
	if (succesfull) {
	  chatServer.sendMessage(
		  new ChatMessage.S2C_Chusermode(message.channel, 
										 message.username, 
										 chatServer.getUsermode(message.username, message.channel)),
		  usernames);
	}
  }

  private synchronized void kick(String username, ChatMessage reqmessage)
	throws ClassCastException {
	Vector usernames;

	ChatMessage.C2S_Kick message = (ChatMessage.C2S_Kick) reqmessage;
	// admin kicks a user
	chatServer.leaveChannel(username, message.channel, ChatMessage.SIGNOFF_KICK);
	
	// notify all users of kick
	usernames = channels.getUsernames(message.channel);
	
	chatServer.sendMessage(
		new ChatMessage.S2C_Signoff(message.channel, username, 
									ChatMessage.SIGNOFF_KICK),
		usernames);
	
	// ask user to leave
	usernames = new Vector();
	usernames.addElement(username);
	
	chatServer.sendMessage(
		new ChatMessage.S2C_Leaveself(message.channel),
		usernames);
  }

  private synchronized void lockout(String username, ChatMessage reqmessage)
	throws ClassCastException {
	Vector usernames;
	
	ChatMessage.C2S_Lockout message = (ChatMessage.C2S_Lockout) reqmessage;
	// admin request for channel to be locked/unlocked, 
	//notify others of this event
	
	if (username.equals(channels.getAdminname(message.channel))) {
	  chatServer.channelLockout(message.channel, message.lockout);
	  
	  usernames = channels.getUsernames(message.channel);
	  
	  chatServer.sendMessage(
		  new ChatMessage.S2C_Lockout(message.channel, message.lockout),
		  usernames);
	}
  }

  private synchronized void leave(String username, ChatMessage reqmessage)
	throws ClassCastException {
	Vector usernames;

	ChatMessage.C2S_Leave message = (ChatMessage.C2S_Leave) reqmessage;
	// client wants to leave the channel
	String adminname = channels.getAdminname(message.channel);
	boolean emptyChannel = chatServer.leaveChannel(username, message.channel, 
												   ChatMessage.SIGNOFF_LEAVE);
	
	if (!emptyChannel) {
	  // notify all users of leaving client
	  usernames = channels.getUsernames(message.channel);
	  
	  chatServer.sendMessage(
		  new ChatMessage.S2C_Signoff(message.channel, username, 
									  ChatMessage.SIGNOFF_LEAVE),
		  usernames);
	}
	// tell client that he can leave
	usernames = new Vector();
	usernames.addElement(username);
	
	chatServer.sendMessage(
		new ChatMessage.S2C_Leaveself(message.channel),
		usernames);
	
	if (!emptyChannel) {
	  // now use old adminname to decide about new admin or not
	  if (adminname.equals(username)) {
		// choose new admin because old one just left
		adminname = channels.chooseAdmin(message.channel);
		if (adminname != null) {
		  // notify everyone in channel of admin change
		  usernames = channels.getUsernames(message.channel);
		  
		  chatServer.sendMessage(
			  new ChatMessage.S2C_Chusermode(message.channel, adminname, 
											 chatServer.getUsermode(adminname, 
																	message.channel)),
			  usernames);
		} else {
		  System.out.println("TODO: tjah wie wordt er dan admin als er geen spelers meer zijn?");
		  // sander zegt kill kanaal
		}
		
	  }
	}
  }

  private synchronized void login(String username, ChatMessage reqmessage)
	throws ClassCastException {
	Vector usernames;

	ChatMessage.C2S_Login message = (ChatMessage.C2S_Login) reqmessage;
	// client logs in to global chat
	chatServer.registerUser(username);
	
	// notify users of new user
	usernames = (Vector) (channels.getUsernames(ChatMessage.GLOBALCHANNELNAME)).clone();
	usernames.removeElement(username);
	
	chatServer.sendMessage(
		new ChatMessage.S2C_Userjoin(ChatMessage.GLOBALCHANNELNAME, 
									 username, 
									 chatServer.getUsermode(username, 
															ChatMessage.GLOBALCHANNELNAME)),
		usernames);
	
	// notify user joining of other users (start with user 1 due to root user)
	usernames = channels.getUsernames(ChatMessage.GLOBALCHANNELNAME);
	String[] users = new String[usernames.size()];
	for (int i = 0; i < usernames.size(); i++) {
	  users[i] = (String) usernames.elementAt(i);
	}
	int[] umodes = new int[usernames.size()];
	for (int j = 0; j < usernames.size(); j++) {
	  umodes[j] = chatServer.getUsermode((String) usernames.elementAt(j), 
										 ChatMessage.GLOBALCHANNELNAME);
	  System.out.println("Result: " + umodes[j]);
	}			 
	usernames = new Vector();
	usernames.addElement(username);
	
	chatServer.sendMessage(
		new ChatMessage.S2C_Joinself(ChatMessage.GLOBALCHANNELNAME,
									 users, umodes, 
									 channels.getStatus(ChatMessage.GLOBALCHANNELNAME)), 
		usernames);
  }

  private synchronized void logout(String username, ChatMessage reqmessage) 
	throws ClassCastException {
	Vector usernames;

	ChatMessage.C2S_Logout message = (ChatMessage.C2S_Logout) reqmessage;
	// client leaves global chat
	chatServer.unregisterUser(username, ChatMessage.SIGNOFF_LEAVE);
	
	// notify all users of leaving client
	usernames = channels.getUsernames(ChatMessage.GLOBALCHANNELNAME);
	
	chatServer.sendMessage(
		new ChatMessage.S2C_Signoff(ChatMessage.GLOBALCHANNELNAME, username, 
									ChatMessage.SIGNOFF_LEAVE),
		usernames);
	
	// tell client that he can leave
	usernames = new Vector();
	usernames.addElement(username);
	
	chatServer.sendMessage(
		new ChatMessage.S2C_Leaveself(ChatMessage.GLOBALCHANNELNAME),
		usernames);
  }

  private synchronized void startgame(String username, ChatMessage reqmessage)
	throws ClassCastException {
	Vector usernames;

	ChatMessage.C2S_Startgame message = (ChatMessage.C2S_Startgame) reqmessage;
	// request to start the game
	
	// first check if source is the admin of the channel
	if (channels.getAdminname(message.channel).equals(username)) {
				// ok so go ahead
	  chatServer.startGame(message.channel);
	  
	  usernames = channels.getUsernames(message.channel); 
	  // notify all users of leaving client
	  chatServer.sendMessage(
		  new ChatMessage.S2C_Chchanmode(message.channel, ChatMessage.CHANMODE_GAME),
		  usernames);
	}
  }

  private synchronized void stopgame(String username, ChatMessage.C2S_Stopgame sg) {
	if (!channels.getAdminname(sg.channel).equals(username)) {
	  Vector vec = new Vector();
	  vec.add(username);
	  chatServer.sendMessage(new ChatMessage.S2C_Broadcast(sg.channel, "You're not admin!"), vec);
	  return;
	}
	chatServer.stopGame(sg.channel);
  }
}





