package quest.server.chat;

import java.util.*;
import quest.global.net.*;
import quest.global.chat.*;
import quest.global.game.*;
import quest.server.net.*;
import quest.server.log.*;
import quest.server.game.*;

/**
 * This is the chatserver, of the chatsystem. It is responsible for handling all
 * chat messages and doing all the administration of the chatsystem.
 *
 * @author    Merijn Evertse
 */
public class ChatServer {
  private static final int NROFSHUFFLES = 10;
  
  private LogManager log;
  private ServerDaemon serverDaemon;
  private GameServer gameServer;
  private Channels channels;
  private ChatUsers users;
  private static ChatServer uniqueInstance = null;
  
  /**
   * Initializes the server and registers it at the server daemon so it can handle
   * all chat messages. Also the user and channel database are initialized.
   */
  public ChatServer() {
	if (uniqueInstance != null) {
	  throw new RuntimeException("singleton constraint violated");
	} 
	
	uniqueInstance = this;
	log = LogManager.getInstance();
	serverDaemon = ServerDaemon.getInstance();
	gameServer = GameServer.getInstance();
	
	// create user database
	users = new ChatUsers();

		// create channel database and start global channel
		channels = new Channels();

		channels.addChannel(ChatMessage.GLOBALCHANNELNAME, null);
		
		// create listeners
		new ChatMessageServerListener();
		new ChannelServer();
		new RankingServer();
		new SearchServer();
		
		// register at server daemon
		serverDaemon.registerProces(ChatMessage.class, new ChatMessageServerListener(), ServerDaemon.USER_BASED);

		log.writeEntry("ChatServer: Initiated and ready for action");
	}
	
	/**
	 * Asks the server daemon to send the given message.
	 *
	 * @param    message      The message to send
	 * @param    usernames    The usernames of the clients to send to
	 */
	public synchronized void sendMessage(ChatMessage message, Vector usernames) {
	  Vector fails = serverDaemon.sendMessage(usernames, message);
	  for (int i = 0; i < fails.size(); i++) {
		unregisterUser((String) fails.elementAt(i), ChatMessage.SIGNOFF_TIMEOUT);
	  }
	}
	
	/**
	 * Registers a user on the chat server.
	 *
	 * @param    username    The name of the user
	 */
	public void registerUser(String username) {
		users.addUser(new ChatUser(username));
		joinChannel(username, ChatMessage.GLOBALCHANNELNAME);
		log.writeEntry("ChatServer: User " + username + " registered for main chat channel");
	}

	/**
	 * Unregisters a user on the chat server.
	 *
	 * @param    username    The name of the user
	 */
	public void unregisterUser(String username, int signoffType) {
		ChatUser user = users.getUser(username);
		Vector channelnames = user.getChannelnames();
		for (int i = 0; i < channelnames.size(); i++) {
		  System.out.println((String) channelnames.elementAt(i));
		  leaveChannel(username, (String) channelnames.elementAt(i), signoffType);
		}
		users.removeUser(user);
	}
	
  /**
   * Returns the usermode of the user in the channel with the given name.
   *
   * @param    username    The name of the user
   * @param    channelname The name of the channel the user is in
   * @return   The usermode of the user in that channel
   */
	public int getUsermode(String username, String channelname) {
		return ((ChatUser) users.getUser(username)).getUsermode(channelname);
	}
	
  /**
   * Changes the usermode of the user, and returns whether the change was succesfull.
   * It could be that the change wanted is actually illegal and then it is denied.
   *
   * @param    username    The name of the user
   * @param    channelname The name of the channel the user is in
   * @param    umode       The new usermode of the user
   * @return   TRUE:  if the change was succesfull
   *           FALSE: if not
   */
	public boolean changeUsermode(String username, String channelname, int umode) {
		int oldmode = getUsermode(username, channelname);
		boolean succesfull = ((ChatUser) users.getUser(username)).
		  setUsermode(channelname, umode);
		
		if (succesfull && oldmode == ChatUser.USERMODE_PLAYERADMIN && 
			umode != ChatUser.USERMODE_PLAYERADMIN) {
			// user stopt being the admin, so choose new admin
		  System.out.println("NYI!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
		}
		
		// change channel info
		channels.setUsermode(channelname, username, umode);
		
		return succesfull;
	}

  /**
   * Adds a user to the channel with the given name. If the channel
   * in already "in-game" then the user also joins the game itself.
   *
   * @param    username    The name of the user
   * @param    channelname The name of the channel
   */
	public void joinChannel(String username, String channelname) {
		// add user to channel
		channels.addUser(channelname, username);

		// subscribe user to channel
		ChatUser user = users.getUser(username);
		if (!channelname.equals(ChatMessage.GLOBALCHANNELNAME) && 
			channels.getAdminname(channelname).equals(username)) {
			user.subscribe(channelname, ChatUser.USERMODE_PLAYERADMIN);
		} else {
			user.subscribe(channelname, ChatUser.USERMODE_SPEC_SPEC);
		}
		log.writeEntry("ChatServer: User " + username + 
					   " joins channel " + channelname);

		// if channel is "in game" player should join game also
		if (channels.getStatus(channelname) == ChatMessage.CHANMODE_GAME)
		  joinGame(channelname, username);
	}

  /**
   * Removes the user from the channel with the given name. If the channel
   * is still "in-game" then the user is also removed from the game itself.
   *
   * @param    username    The name of the user
   * @param    channelname The name of the channel
   * @param    signoffType The reason the user has to be removed,
   *                       for instance due to a timeout, or just because he wants it
   */ 
  public boolean leaveChannel(String username, String channelname, int signoffType) {
	// if channel is "in game" player should leave game also
	if (channels.getStatus(channelname) == ChatMessage.CHANMODE_GAME)
	  leaveGame(channelname, username);
	
	users.removeChannel(username, channelname);
	boolean emptyChannel = channels.removeUser(channelname, username);
	log.writeEntry("ChatServer: User " + username + " leaves channel " + channelname);
	
	// if signoff type is SIGNOFF_TIMEOUT, a userleave message should be send
	if (signoffType == ChatMessage.SIGNOFF_TIMEOUT) {
	  sendMessage(
		  new ChatMessage.S2C_Signoff(channelname, username, 
									  ChatMessage.SIGNOFF_TIMEOUT),
		  channels.getUsernames(channelname));
	}
	return emptyChannel;
  }
  
  /**
   * Adds a channel to the chatsystem, with the given name. The user
   * with the given name will be the administrator of the channel.
   *
   * @param    channelname    The name of the channel
   * @param    adminname      The name of the administrator
   */
  public void addChannel(String channelname, String adminname) {
		channels.addChannel(channelname, adminname);
		ChatUser user = users.getUser(adminname);
		user.subscribe(channelname, ChatUser.USERMODE_PLAYERADMIN);
		log.writeEntry("ChatServer: Channel " + channelname + " added");
	}
	
  /**
   * Removes a channel from the chatsystem.
   *
   * @param    channelname    The name of the channel
   */
  public void removeChannel(String channelname) {
	channels.removeChannel(channelname);
	LogManager.getInstance().writeEntry("ChatServer: Channel " + channelname + " removed");
  }
  
  /**
   * Starts the game with the current players and spectators in the channel.
   * This means the status of the channel is changed and then the clients
   * recognize that the game is started so they can start their gameclients.
   * NOTE: the playernames are shuffled to ensure that he player rankings are
   * as random as possible.
   *
   * @param    channelname    The name of the channel
   */
  public void startGame(String channelname) {
	Vector playernames = channels.getPlayernames(channelname);
	Vector spectatornames = channels.getSpectatornames(channelname);

	// enable users and spectators to send and receive game messages
	for (int i = 0; i < playernames.size(); i++) {
	  serverDaemon.subscribeUser((String) playernames.elementAt(i),
								 GameMessage.class);
	}
	for (int j = 0; j < spectatornames.size(); j++) {
	  serverDaemon.subscribeUser((String) spectatornames.elementAt(j),
								 GameMessage.class);
	}
	channels.setStatus(channelname,ChatMessage.CHANMODE_GAME);

	// shuffle playernames
	Vector randomizedPlayers = randomize((Vector) playernames.clone());

	// create game and run on game server
	gameServer.startGame(channelname, randomizedPlayers, (Vector) spectatornames.clone());
  }

  /**
   * Stops the current game in the channel. All clients get a warning of this
   * through the normal game protocol so they can logout their clients.
   * The channel is put in "after game" state for evaluation purposes.
   *
   * @param    channelname    The name of the channel
   */
  public void stopGame(String channelname) {
	Channel channel = channels.getChannel(channelname);
	channels.setStatus(channelname,ChatMessage.CHANMODE_END);
	gameServer.stopGame(channelname);
	sendMessage(
		new ChatMessage.S2C_Chchanmode(channelname, 
									   ChatMessage.CHANMODE_END), 
		channel.getUsernames());
  }

  /**
   * Aborts the current game in the channel. All clients get a warning of this 
   * through the normal game protocol. 
   * The channel is put in "before game" state to startover again.
   *
   * @param    channelname    The name of the channel
   */ 
  public void abortGame(String channelname) {
	Channel channel = channels.getChannel(channelname);
	channels.setStatus(channelname,ChatMessage.CHANMODE_BEGIN);
	gameServer.stopGame(channelname);
	sendMessage(
		new ChatMessage.S2C_Chchanmode(channelname, 
									   ChatMessage.CHANMODE_BEGIN), 
		channel.getUsernames());
  }

  /**
   * Indicates that the game is finished so the chat server can change
   * the channelmode and inform the client of the change.
   *
   * @param    gamename    The name of the finished game
   */
  public void gameFinished(String gamename) {
	Channel channel = channels.getChannel(gamename);
	channels.setStatus(gamename,ChatMessage.CHANMODE_END);
	sendMessage(
		new ChatMessage.S2C_Chchanmode(gamename, 
									   ChatMessage.CHANMODE_END), 
		channel.getUsernames());	
  }

  /**
   * Adds the user as a spectator to the running game in the channel.
   *
   * @param    channelname    The name of the channel
   * @param    spectatorname  The name of the new user
   */
  public void joinGame(String channelname, String spectatorname) {
   	gameServer.joinGame(channelname, spectatorname, GameServer.CHAT_JOINGAME);
  }
  
  /**
   * Removes a spectator from the running game in the channel.
   *
   * @param    channelname    The name of the channel
   * @param    spectatorname  The name of the user
   */
  public void leaveGame(String channelname, String spectatorname) {
	gameServer.leaveGame(channelname, spectatorname, GameServer.CHAT_LEAVEGAME);
  }

  private Vector randomize(Vector vector) {
	for (int i = 0; i < (NROFSHUFFLES + ((int) (NROFSHUFFLES * Math.random()))); i++) {
	  int one = (int) (vector.size() * Math.random());
	  int two = (int) (vector.size() * Math.random());
	  Object object = vector.elementAt(one);
	  vector.setElementAt(vector.elementAt(two), one);
	  vector.setElementAt(object, two);
	  System.out.println("SHUFFLE: suffle number " + i + " with elements " + one + " and " + two);
	}
	return vector;
  }

	/**
	 * Returns the channel with the given channelname.
	 *
	 * @param    channelname    The name of the channel
	 * @return   The channel with the given name
	 */
	public Channel getChannel(String channelname) {
		return channels.getChannel(channelname);
	}

	/**
	 * Changes the locked/unlocked status of the channel.
	 *
	 * @param    channelname    The name of the channel
	 * @param    lockout   	The new status of the channel
	 */
	public void channelLockout(String channelname, boolean lockout) {
		channels.setLock(channelname, lockout);
		if (lockout) { 
			System.out.println("Channel " + channelname + " is locked");
		} else {
			System.out.println("Channel " + channelname + " is unlocked");
		}
	}
	
	/**
	 * Changes the status of the channel.
	 *
	 * @param    channelname    The name of the channel
	 * @param    status         The new status of the channel
	 */
	public void setChannelStatus(String channelname, int status) {
		channels.setStatus(channelname, status);
		System.out.println("Channel " + channelname + " now has status " + status);
	} 

  /**
   * Returns the first channel the user is in, next to the main chat channel.
   * NOTE: at the moment the user can only be in one extra channel.
   *
   * @param    username    The name of the user
   * @return   The name of the first channel the user is in
   */
  public String getChannelname(String username) {
	System.out.println("get channelname for user: " + username);
	return users.getUser(username).getChannelname();
  }

  /**
   * Returns the current channelmode of the channel.
   *
   * @param    channelname    The name of the channel
   * @return   The channelmode
   */
  public int getChannelmode(String channelname) {
	return channels.getStatus(channelname);
  }

  /**
   * Returns whether the user is online or not.
   *
   * @param     username    The name of the user
   * @return    TRUE:  if the user is online
   *            FALSE: if not
   */
  public boolean userOnline(String username) {
	return users.userExists(username);
  }

	/**
	 * Returns an instance of this class, making sure only one object of it exists.
	 *
	 * @return    A reference to the instance of this class
	 */
	public static ChatServer getInstance() {
		if (uniqueInstance == null) {
		  throw new RuntimeException("singleton class has to be initialized by the constructor first");
		}
		return uniqueInstance;
	}

}







