package quest.server.chat;

import quest.server.net.*;
import quest.global.chat.*;
import quest.global.net.*;
import quest.server.util.QuestDatabase;

import java.util.Vector;

/**
 * <p>This is the server responsible for the functionality of providing ranking information about the players in the database. It also provides
 * an interface to to support some functionality of the chatsystem for the chatserver.
 * The messages include for instance a request for the whole ranking list, or
 * a list of the 20 people above and below a user in ranking etc.
 * </p>
 * This class is an example of a <a href="../../../overview-summary.html#patternSingleton">Singleton design pattern</a>.
 * 
 * @author    Merijn Evertse
 */
public class RankingServer implements MessageServerListener {
	private static RankingServer uniqueInstance = null;
	
	private QuestDatabase qdb = QuestDatabase.getInstance();
	private ServerDaemon sd = ServerDaemon.getInstance();	

  /**
   * Initializes the rankingserver and registers itself with the 
   * server daemon so it will receive the ranking messages for handling.
   */
  public RankingServer() {
	if (uniqueInstance != null) {
	  throw new RuntimeException("singleton constraint violated");
	} 
	
	uniqueInstance = this;
	sd.registerProces(RankingMessage.class, this, ServerDaemon.USER_BASED);
  }
  
  /**
   * Returns a reference to this object, making sure only one object exists.
   *
   * @return    The reference
   */
  public static RankingServer getInstance() {
	if (uniqueInstance == null) {
	  throw new RuntimeException("singleton class has to be initialized by the constructor first");
	}
	
	return uniqueInstance;
  }
	
  /**
   * Handles the incoming ranking messages. It checks them, processes it
   * and constructs the reply messages, so the server daemon can send these
   * to the clients.
   *
   * @param    username    The name of the user who sent the message
   * @param    msg         The message
   */
  public void incomingMessage(final String username, final Message msg) {	
	final Runnable run = new Runnable() {
	  public void run() {
		RankingMessage rm = (RankingMessage) msg;
		
		if (rm instanceof RankingMessage.C2S_RequestRanking) {
		  requestranking(username, rm);
		} else {
		  System.out.println("Received unsupported message: " + rm);
				}
	  }
	};
	
	(new Thread(run)).start();
  }

  private synchronized void requestranking(String username, RankingMessage rm) {
	RankingMessage.C2S_RequestRanking rr = (RankingMessage.C2S_RequestRanking) rm;
	int nr[] = qdb.getUserPosition(username);
	if (nr == null) {
	  return;
	}
	
	if (rr.mode == RankingMessage.RANKING_WHOLE) {
	  RankingMessage.S2C_SendRanking sr = new RankingMessage.S2C_SendRanking(
			 RankingMessage.RANKING_WHOLE, nr[0], nr[1], nr[2], qdb.getAllUserRankings());
	  System.out.println("rankings length: " + sr.rankings.length);
	  sd.sendMessageToUser(username, sr);
	} else if (rr.mode == RankingMessage.RANKING_TOP) {
	  RankingMessage.S2C_SendRanking sr = new RankingMessage.S2C_SendRanking(
			 RankingMessage.RANKING_TOP, nr[0], nr[1], nr[2], qdb.getTopUserRankings());
	  sd.sendMessageToUser(username, sr);
	} else if (rr.mode == RankingMessage.RANKING_NEIGHBOUR) {
	  RankingMessage.S2C_SendRanking sr = new RankingMessage.S2C_SendRanking(
			 RankingMessage.RANKING_NEIGHBOUR, nr[0], nr[1], nr[2], qdb.getNeighbours(username));
	  sd.sendMessageToUser(username, sr);
	} else {
	  System.out.println("invalid packet received: " + rm);
	}
  }
}
