package quest.server.chat;

import quest.server.net.ServerDaemon;
import quest.global.chat.SearchMessage;
import quest.server.util.QuestDatabase;
import quest.global.net.MessageServerListener;
import quest.global.net.Message;

import java.util.Vector;

/**
 * <p>This is the server responsible for handling the search messages, as part of the
 * chatsystem. It handles messages like a request for all usernames with "aa" as the
 * first couple of letters.</p>
 * This class is an example of a <a href="../../../overview-summary.html#patternSingleton">Singleton design pattern</a>.
 * 
 * @author    Merijn Evertse
 */
public class SearchServer implements MessageServerListener {
	private static SearchServer uniqueInstance = null;
	
	private QuestDatabase qdb = QuestDatabase.getInstance();
	private ServerDaemon sd = ServerDaemon.getInstance();	

  /**
   * Initializes the server and registers itself at the server daemon, so it will
   * receive the search messages for handling.
   */
  public SearchServer() {
	if (uniqueInstance != null) {
	  throw new RuntimeException("singleton constraint violated");
	} 
	
	uniqueInstance = this;
	sd.registerProces(SearchMessage.class, this, ServerDaemon.GLOBAL_BASED);
  }

  /**
   * Returns a reference to this object, making sure only one object exists.
   *
   * @return    The reference
   */
  public static SearchServer getInstance() {
	if (uniqueInstance == null) {
	  throw new RuntimeException("singleton class has to be initialized by the constructor first");
	}
	
	return uniqueInstance;
  }
  
  /**
   * Handles the incoming search messages. It checks them, processes it and
   * constructs the reply messages, which are given to the server daemon.
   * The server daemon then sends it to the clients.
   *
   * @param    username    The name of the user who sent the message
   * @param    msg         The message
   */
  public void incomingMessage(final String username, final Message msg) {	
	final Runnable run = new Runnable() {
	  public void run() {
		SearchMessage sm = (SearchMessage) msg;
		if (sm instanceof SearchMessage.C2S_Search) {
		  search(username, sm);
		} else {
		  System.out.println("Received unsupported message: " + sm);
		}
	  }
	};
	
	(new Thread(run)).start();
  }
  
  private synchronized void search(String username, SearchMessage sm) {
	SearchMessage.C2S_Search sch = (SearchMessage.C2S_Search) sm;
	SearchMessage.S2C_Result.SearchResult[] result = qdb.searchUsername(sch.username);
	SearchMessage.S2C_Result reply = new SearchMessage.S2C_Result(result);
	sd.sendMessageToUser(username, reply);
  }
}

