package quest.server.game;

import java.util.*;
import quest.server.chat.*;
import quest.global.game.*;

/**
 * This respresents one game currently active in the gamesystem.
 *
 * @author    Merijn Evertse
 */
public class Game {
  /**
   * Indicates the game is initializing.
   */
  public static final int INITIALIZING = 1;
  
  /**
   * Indicates the game is running.
   */
  public static final int RUNNING      = 2;

  /**
   * Indicates the game is finished.
   */
  public static final int FINISHED     = 3;
  
  /**
   * The name of the game.
   */
  public String gamename;

  /**
   * The GameLogic object, which contains all information about the current
   * game situation.
   */
  public GameLogic gamelogic;

  /**
   * The list of playernames.
   */
  public Vector playernames;

  /**
   * The list of spectatornames.
   */
  public Vector spectatornames;
 
  private Vector joiningplay;
  private Vector leavingplay;
  private Vector joiningspec;
  private Vector leavingspec;
  private int status;
  private boolean ready;

  /**
   * Initializes the game with the given parameters.
   *
   * @param    gamename       The name of the game
   * @param    playernames    The list of playernames
   * @param    spectatornames The list of spectatornames
   */
  public Game(String gamename, Vector playernames, Vector spectatornames) {
	this.gamename = gamename;
	this.playernames = new Vector();
	this.spectatornames = new Vector();
	joiningplay = playernames;
	joiningspec = spectatornames;
	leavingplay = new Vector();
	leavingspec = new Vector();
	ready = false;
  }

  /**
   * Initializes the game by constructing a gamelogic object with the
   * playernames and spectatornames. This object is then returned,
   * so it can be distributed to all clients.
   * @return    The gamelogic object
   */
  public GameLogic initializeGame() {
	String[] players = new String[playernames.size()];
	for (int i =0; i < playernames.size(); i++) {
	  players[i] = (String) playernames.elementAt(i);
	}
	String[] spectators = new String[spectatornames.size()];
	for (int i =0; i < spectatornames.size(); i++) {
	  players[i] = (String) spectatornames.elementAt(i);
	}
	for(int i=0; i < players.length; i++) {
	  	System.out.println(players[i] + "PLAYERS");
	}
	for (int i=0; i < spectators.length; i++ ) {
	  System.out.println(spectators[i] + "SPECTATORS");
	}
	gamelogic = new GameLogic(players, spectators);
	return gamelogic;
  }

  /**
   * Starts the game, by starting the gamelogic with server access.
   * This means the gamelogic object maintains a complete view of the 
   * gamesituation but the server isn't a participating user in the game.
   * This is used to validate all moves made by the clients.
   */
  public void startGame() {
	gamelogic.start(GameLogic.SERVER);
  }

  /**
   * Validates a move with the current gamesituation.
   *
   * @param     gamemove    The gamemove to validate
   * @return    TRUE:  if the move is valid
   *            FALSE: if not
   */
  public boolean doMove(GameMove gamemove) throws IllegalMoveException {
	return gamelogic.doMove(gamemove);
  }
  
  /**
   * Adds a user to the game.
   * Different types of joins are possible:
   * CHAT_JOINGAME: The user is added to the game at request of the chatserver
   *                Therefor it is a user in the proces of joining the game.
   * GAME_JOINGAME: The user joins by logging in in the gamesystem, so can
   *                be transferred from joining to active user.
   * The reason for this is that when the user joins the game on request of the
   * chatserver, he still had to login to the gamesystem to notify the gameserver
   * that he is now talking on the game protocol also.
   * If the user was in the list of players at the start of the game the user is 
   * now transferred from joining player to active player
   * and can participate in the game. Otherwise the user is added as a spectator
   * and can only look at the game.
   * 
   * @param    username    The name of the user
   * @param    type        The join type, this can be CHAT_JOINGAME or GAME_JOINGAME
   */
  public void addUser(String username, int type) {
	if (joiningplay.contains(username)) {
	  System.out.println("!!!!!!!!!!!! " + username + " transferred from joiningplay to active player");
      ready = addPlayer(username, type);
	} else {  // user is joining spectator or new user who becomes spectator
	  addSpectator(username, type);
	}
  }

  /**
   * Removes a user from the game.
   * Different types of removes are possible:
   * CHAT_LEAVEGAME: The user is removed from the game at request of the chatserver
   *                 Therefor the user leaves the chatchannel and should also
   *                 immidiately leave the game.
   * GAME_LEAVEGAME: The is logsout of the gamesystem, but still remains in the
   *                 chat channel of the game. This means that he remains in the 
   *                 proces of leaving the game until he also leaves the chat channel.
   * The reason for this is that when logging out of the game the user will remain in
   * the chat channel and therefor should remain visible there. Only his game window
   * is closed. Only when he logsout of the chat channel too, is he removed from the
   * game completely.
   * 
   * @param    username    The name of the user
   * @param    type        The leave type, this can be CHAT_LEAVEGAME or GAME_LEAVEGAME
   */                
  public void removeUser(String username, int type) {
	if (playernames.contains(username) || leavingplay.contains(username)) {
	  removePlayer(username, type);
	} else {  // user is leaving spectator
	  removeSpectator(username, type);
	}
  }

  /**
   * Adds a player to the game.
   *
   * @param    playname    The name of the player
   * @param    type        The join type
   * @return  TRUE:  if all players are online, and the game is ready to start
   *          FALSE: if not all players are online yet, so everyone should continue 
   *                 waiting for the missing players
   */
  public boolean addPlayer(String playname, int type) {
	switch (type) {
	case GameServer.CHAT_JOINGAME:
	  joiningplay.addElement(playname);
	  break;
	case GameServer.GAME_JOINGAME:
	  boolean found = joiningplay.removeElement(playname);
	  if (found) {
		playernames.addElement(playname);
		if (joiningplay.isEmpty()) return true;
	  }
	  break;
	default:
	  System.out.println("Something has gone wrong in game");
	}
	return false;
  }

  /**
   * Removes a player from the game.
   *
   * @param    playname    The name of the player
   * @param    type        The leave type
   */
  public void removePlayer(String playname, int type) {
	switch (type) {
	case GameServer.CHAT_LEAVEGAME:
	  leavingplay.removeElement(playname);
	  playernames.removeElement(playname); 
	  break;
	case GameServer.GAME_LEAVEGAME:
	  boolean found = playernames.removeElement(playname);
	  if (found) leavingplay.addElement(playname);
	  break;
	default:
	  System.out.println("Something has gone wrong in game");
	}
  }

  /**
   * Adds a spectator to the game.
   *
   * @param    specname    The name of the spectator
   * @param    type        The join type
   */
  public void addSpectator(String specname, int type) {
	switch (type) {
	case GameServer.CHAT_JOINGAME:
	  joiningspec.addElement(specname);
	  break;
	case GameServer.GAME_JOINGAME:
	  boolean found = joiningspec.removeElement(specname);
	  if (found) spectatornames.addElement(specname);
	  break;
	default:
	  System.out.println("Something has gone wrong in game");
	}
  }

  /**
   * Removes a spectator from the game.
   *
   * @param    specname    The name of the spectator
   * @param    type        The leave type
   */
  public void removeSpectator(String specname, int type) {
	switch (type) {
	case GameServer.CHAT_LEAVEGAME:
	  leavingspec.removeElement(specname);
	  spectatornames.removeElement(specname); 
	  break;
	case GameServer.GAME_LEAVEGAME:
	  boolean found = spectatornames.removeElement(specname);
	  if (found) leavingspec.addElement(specname);
	  break;
	default:
	  System.out.println("Something has gone wrong in game");
	}
  }

  /**
   * Returns whether there are players left in the game.
   *
   * @return    TRUE:  if players are still in the game
   *            FALSE: if not
   */
  public boolean empty() {
	return (playernames.size() == 0);
  }
  
  private boolean playerExists(String username) {
	for (int i = 0; i < playernames.size(); i++) {
	  if (((String) playernames.elementAt(i)).equals(username)) return true;
	}
	return false;
  }

  private boolean spectatorExists(String username) {
	for (int i = 0; i < spectatornames.size(); i++) {
	  if (((String) spectatornames.elementAt(i)).equals(username)) return true;
	}
	return false;
  }

  /**
   * Returns whether a user exists in the game.
   *
   * @param    username    The name of the user
   * @return   TRUE:  if the user exists
   *           FALSE: if not
   */
  public boolean userExists(String username) {
	return (playerExists(username) || spectatorExists(username));
  }
  
  /**
   * Returns the list of all usernames in the game.
   *
   * @return    The list of usernames
   */
  public Vector getUsernames() {
	Vector result = new Vector();
	result.addAll(playernames);
	result.addAll(spectatornames);
	return result;
  }

  /**
   * Returns the usermode of the user with the given name.
   *
   * @param    username    The name of the user
   * @return   The usermode of the user
   */
  public int getUsermode(String username) {
    if (playerExists(username)) return GameMessage.USERMODE_PLAYER;
	else return GameMessage.USERMODE_SPECTATOR;
  }

  /**
   * Returns the status of the game.
   *
   * @return    The status of the game, can be INITIALIZING, RUNNING and FINISHED
   */
  public int getStatus() {
	return status;
  }

  /**
   * Sets the status of the game to the given status.
   *
   * @param    status    The new status of the game
   */
  public void setStatus(int status) {
	this.status = status;
  }

  /**
   * Returns whether the game is ready to start.
   *
   * @return    TRUE:  if the game is ready to start
   *            FALSE: if not
   */
  public boolean gameReady() {
    return ready;
  }
}
