package quest.server.game;

import java.util.*;
import quest.global.net.*;
import quest.global.game.*;
import quest.server.chat.*;
import quest.global.chat.*;

/**
 * The message listener for the game server.
 * Handles all incoming messages for the game server, changes the
 * states of the game environment en sends messages back to the clients.
 *
 * @author    Merijn Evertse
 */
public class GameMessageServerListener implements MessageServerListener {

  private GameServer gameServer;
  private Games games;

	/**
	 * Initializes the listener.
	 */
	public GameMessageServerListener() {
	  gameServer = GameServer.getInstance();
	  games = games.getInstance();
	}

	/**
	 * Handles a message received by the communication layer. Checks the messages,
	 * processes them and constructs the reply messages, so the server daemon
	 * can send these to the clients.
	 *
	 * @param    username          The name of the user who sent the message
	 * @param    receivedMessage   The message received from the user
	 */
	public void incomingMessage(final String username, Message receivedMessage) {
		final GameMessage reqmessage = (GameMessage) receivedMessage;
		final ChatServer chatServer = ChatServer.getInstance();
		
		if (reqmessage.gamename != null && 
			games.gameExists(reqmessage.gamename)) {

		final Runnable ra = new Runnable() {
		  public void run() {
			try {
			  // handle message based on message type
			  if (reqmessage instanceof GameMessage.C2S_Login) {
				login(username, reqmessage);
			  }
			  else if (reqmessage instanceof GameMessage.C2S_Logout) {
				logout(username, reqmessage);
			  }
			  else if (reqmessage instanceof GameMessage.C2S_Gamemove) {
				gamemove(username, reqmessage);
			  }
			  else {
				System.out.println("Unknown message received in GameMessageServerListener");
			  }
			} catch (ClassCastException cce) {
			  System.out.println("WARNING: Malformed packet received " + cce.getMessage() + " from " + username);
			}
		  }
		};
		(new Thread(ra)).start();
		
		} else {
		  System.out.println("WARNING: Received packet with bad gamename, ignoring packet!");
		}
	}

  private synchronized void login(String username, GameMessage reqmessage) 
	throws ClassCastException {
	Vector usernames;

	GameMessage.C2S_Login message = (GameMessage.C2S_Login) reqmessage;
	// client requests to login to the game
	gameServer.joinGame(message.gamename, username, GameServer.GAME_JOINGAME);
	Game game = gameServer.getGame(message.gamename);
	
	if (game.gamelogic == null) {
	  System.out.println("No gamelogic yet");
	  // game isn't started yet
	  
	  // notify other clients of joining client if user is player
	  int umode = gameServer.getUsermode(message.gamename, username);
	  if (umode == GameMessage.USERMODE_PLAYER) {			  
		usernames = (Vector) (game.getUsernames()).clone();
		usernames.removeElement(username);
		gameServer.sendMessage(
			new GameMessage.S2C_Userjoin(message.gamename, username, umode),
			usernames);
	  }
	  // notify user joining of other users
	  usernames = game.getUsernames();
	  
	  String[] users = new String[usernames.size()];
	  for (int i = 0; i < usernames.size(); i++) {
		users[i] = (String) usernames.elementAt(i);
	  }
	  int[] umodes = new int[usernames.size()];
	  for (int j = 0; j < usernames.size(); j++) {
		umodes[j] = gameServer.getUsermode(message.gamename, (String) usernames.elementAt(j));
		System.out.println("Result: " + umodes[j]);
	  }	
	  
	  usernames = new Vector();
	  usernames.addElement(username);
	  
	  gameServer.sendMessage(
		  new GameMessage.S2C_Joinself(message.gamename, users, umodes),
		  usernames);
	} else {
	  // game is started so joining user is spectator
	  
	  // notify other clients of joining client if user is player
	  usernames = (Vector) (game.getUsernames()).clone();
	  usernames.removeElement(username);
	  gameServer.sendMessage(
		  new GameMessage.S2C_Userjoin(message.gamename, username, 0),
		  usernames);
	  
	  // notify user joining of game situation
	  usernames = new Vector();
	  usernames.addElement(username);
	  
	  gameServer.sendMessage(
		  new GameMessage.S2C_InitGame(message.gamename, (GameLogic) game.gamelogic.clone()),
		  usernames);
	}
	// now check if game is ready to start
	gameServer.checkGame(message.gamename);
  }	

  private synchronized void logout(String username, GameMessage reqmessage) 
	throws ClassCastException {
	Vector usernames;

	GameMessage.C2S_Logout message = (GameMessage.C2S_Logout) reqmessage;
	// clients requests to logout of the game
	Game game = gameServer.getGame(message.gamename);
	gameServer.leaveGame(message.gamename, username, GameServer.GAME_LEAVEGAME);
	
	// notify all users of leaving client, if game is still running
System.out.println("STATUS at leave" + game.getStatus());
	if (game.getStatus() == Game.RUNNING) {
	  System.out.println("go send userleave messages");
	  usernames = game.getUsernames();
	  
	  gameServer.sendMessage(
		  new GameMessage.S2C_Userleave(message.gamename, username),
		  usernames);
	}
	// tell client that he can leave
	usernames = new Vector();
	usernames.addElement(username);
	
	gameServer.sendMessage(
		new GameMessage.S2C_Leaveself(message.gamename),
		usernames);			  
  }

  private synchronized void gamemove(String username, GameMessage reqmessage) 
	throws ClassCastException {
	Vector usernames;

	GameMessage.C2S_Gamemove message = (GameMessage.C2S_Gamemove) reqmessage;
	// client does move, if valid propagate to other clients
	Game game = gameServer.getGame(message.gamename);
	
	boolean finished = false;
	boolean validmove = true;
	try {
	  finished = game.doMove(message.gamemove);
	} catch (IllegalMoveException ime) {
	  validmove = false;
	}
	if (validmove) {
	  // valid move so propagate to all clients
	  usernames = (Vector) (game.getUsernames()).clone();
	  gameServer.sendMessage(
		  new GameMessage.S2C_Gamemove(message.gamename, message.gamemove),
		  usernames);
	  if (finished){
		// game is finished but clients know it so just wait for them to logout
		// the chatserver has to be told that the game is finished
		ChatServer.getInstance().gameFinished(message.gamename);
	  }
	} else {
	  System.out.println("Illegal move by player " + username);
	}
  }
}
