package quest.server.game;

import java.util.*;
import quest.server.log.*;
import quest.server.net.*;
import quest.server.chat.*;
import quest.global.game.*;

/**
 * This is the main game server, responsible for keeping all games running
 * and administrating all activities and users.
 *
 * @author    Merijn Evertse
 */
public class GameServer {
	
  /**
   * Indicates a join game request by the chatserver for a user.
   */
  public static final int CHAT_JOINGAME  = 0;

  /**
   * Indicates a leave game request by the chatserver for a user.
   */
  public static final int CHAT_LEAVEGAME = 1;

  /**
   * Indicates a join game request by a user logging in to the gamesystem.
   */
  public static final int GAME_JOINGAME  = 2;

  /**
   * Indicates a leave game request by a user logging out of the gamesystem.
   */
  public static final int GAME_LEAVEGAME = 3;

  private LogManager log;
  private ServerDaemon serverDaemon;
  private Games games;
  private static GameServer uniqueInstance = null;
  
  /**
   * Initializes the gameserver and registers itself at the server daemon so it
   * can handle the game messages.
   */
  public GameServer() {
	if (uniqueInstance != null) {
	  throw new RuntimeException("singleton constraint violated");
	} 
	
	games = new Games();
	
	uniqueInstance = this;
	log = LogManager.getInstance();
	serverDaemon = ServerDaemon.getInstance();
	
	// create listeners
	new GameMessageServerListener();
	
	// register at server daemon
	serverDaemon.registerProces(GameMessage.class, new GameMessageServerListener(), ServerDaemon.USER_BASED);
	
	log.writeEntry("GameServer: Initiated and waiting for games");
  }
  
  /**
   * Asks the server daemon to send the given message.
   *
   * @param    message      The message to send
   * @param    usernames    The usernames of the clients to send to
   */
  public synchronized void sendMessage(GameMessage message, Vector usernames) {
	Vector fails = serverDaemon.sendMessage(usernames, message);
	for (int i = 0; i < fails.size(); i++) {
	  leaveGame(null, ((String) fails.elementAt(i)), GAME_LEAVEGAME);
	}
  }
  
  /**
   * Starts the game by adding it to the gamesystem.
   *
   * @param    gamename        The name of the game
   * @param    playernames     The list of players in the game
   * @param    spectatornames  The list of spectators in the game
   */
  public void startGame(String gamename, Vector playernames, Vector spectatornames) {
	games.addGame(gamename, playernames, spectatornames);
  }
  
  /**
   * Stops the game by aborting it.
   *
   * @param    gamename    The name of the game to abort
   */
  public void stopGame(String gamename) {
	// game will be ended so tell all users to logout (last logout will end game)
	Game game = games.getGame(gamename);
	Vector usernames = (Vector) (game.getUsernames()).clone();
	sendMessage(new GameMessage.S2C_AbortGame(gamename), usernames);
  }

  /**
   * Adds a user to the game, with the given join type.
   *
   * @param    gamename    The name of the game
   * @param    username    The name of the new user
   * @param    type        The join type (see Game for more info)
   */
  public void joinGame(String gamename, String username, int type) {
	System.out.println("User " + username + " joining game " + gamename + " with type " + type);
	games.joinGame(gamename, username, type);
  }

  /**
   * Checks whether the game can be started already, and if so tells all clients
   * the game is now started.
   *
   * @param    gamename    The name of the game
   */
  public void checkGame(String gamename) {
	boolean ready = games.checkGame(gamename);
	if (ready) {
	  // all players are now online, start the game
	  Game game = games.getGame(gamename);
	  GameLogic gamelogic = game.initializeGame();
	  Vector usernames = (Vector) (game.getUsernames()).clone();
	  sendMessage(new GameMessage.S2C_InitGame(gamename, (GameLogic) gamelogic.clone()), usernames);
	  game.startGame();
	}
	
  }

  /**
   * Removes a user from the game with the given type.
   *
   * @param    gamename    The name of the game
   * @param    username    The name of the user
   * @param    type        The leave type (see Game for more info)
   */
  public void leaveGame(String gamename, String username, int type) {
	System.out.println("User " + username + " leaving game " + gamename + " with type " + type);
	games.leaveGame(gamename, username, type);

	// disable user logging out of game to send and receive game messages
	serverDaemon.unsubscribeUser(username, GameMessage.class);

	// inform users of leaving client when type is CHAT_GAMELEAVE
	if (type == CHAT_LEAVEGAME) {
	  Vector usernames = getGame(gamename).getUsernames();
	  sendMessage(
		  new GameMessage.S2C_Userleave(gamename, username),
		  usernames);
	  
	  // tell client that he can leave
	  usernames = new Vector();
	  usernames.addElement(username);
	
	  sendMessage(
		  new GameMessage.S2C_Leaveself(gamename),
		  usernames);			  
	}
  }

  /**
   * Validates a move by a user with the current gamesituation.
   *
   * @param    gamename    The name of the game
   * @param    username    The name of the user who made the move
   * @param    gamemove    The move made by the user
   * @throws   quest.global.game.IlligalMoveException
   *           If the move wasn't valid
   * @return   TRUE:  if the move ended the game
   *           FALSE: if not
   */
  public boolean doMove(String gamename, String username, GameMove gamemove) throws IllegalMoveException {
	Game game = games.getGame(gamename);
	boolean result = game.doMove(gamemove);
	return result;
  }

  /**
   * Returns the game with the given name.
   *
   * @param    gamename    The name of the game
   * @return   The game
   */
  public Game getGame(String gamename) {
	return games.getGame(gamename);
  }

  /**
   * Returns the usermode of the user.
   *
   * @param    gamename    The name of the game the user is in
   * @param    username    The name of the user
   * @return   The usermode of the user
   */
  public int getUsermode(String gamename, String username) {
	Game game = getGame(gamename);
	if (game == null) System.out.println("WARNING: game nullpointer in GameServer.getUsermode()");
	return game.getUsermode(username);
  }

  /**
   * Returns an instance of this class, making sure only one object of it exists.
   *
   * @return    A reference to the instance of this class
   */
  public static GameServer getInstance() {
	if (uniqueInstance == null) {
	  throw new RuntimeException("singleton class has to be initialized by the constructor first");
	}
	return uniqueInstance;
  }
}

