package quest.server.game;

import java.util.*;

/**
 * This is the database of all games currently in the gamesystem.
 *
 * @author    Merijn Evertse
 */
public class Games {

  private Hashtable games;
  private static Games uniqueInstance;

  /**
   * Returns a reference to this object, making sure only one object exists.
   *
   * @return   The reference
   */
  public static Games getInstance() {
	if (uniqueInstance == null)
	  return new Games();
	
	return uniqueInstance;
  }

  /**
   * Creates an empty channellist
   */
  protected Games() {
	if (uniqueInstance != null)
	  throw new RuntimeException("there's already an instance present");
	
	uniqueInstance = this;
	games = new Hashtable();
  }
  
  /**
   * Add a game to the list of games.
   *
   * @param    gamename       The name of the new game
   * @param    playernames    The list of all player names
   * @param    spectatornames The list of all spectator names
   */
  public void addGame(String gamename, Vector playernames, Vector spectatornames) {
	Game game = new Game(gamename, playernames, spectatornames);
	games.put(gamename, game);
  }
  
  /**
   * Removes a game from the list of games.
   * 
   * @param    gamename    The name of the game to remove
   */
  public void removeGame(String gamename) {
	games.remove(gamename);
  }
  
  /**
   * Adds a user to the game with the given type.
   * 
   * @param    gamename    The name of the game
   * @param    username    The name of the user
   * @param    type        The joining type (see Game for more info)
   */
  public void joinGame(String gamename, String username, int type) {
	Game game = (Game) games.get(gamename);
	game.addUser(username, type);
  }
  
  /**
   * Checks whether the game is already finished.
   *
   * @param    gamename    The name of the game to check
   * @return   TRUE:  if the game is finished
   *           FALSE: if not
   */
  public boolean checkGame(String gamename) {
	Game game = (Game) games.get(gamename);
	return game.gameReady();
  }

  /**
   * Removes a user from the game with the given type.
   *
   * @param    gamename    The name of the game
   * @param    username    The name of the user
   * @param    type        The leave type (see Game for more info)
   */
  public void leaveGame(String gamename, String username, int type) {
	Game game;
	if (gamename != null) {
	  game = (Game) games.get(gamename);
	  //TODO game kan ook null zijn, fix dit
	  if (game == null) {
		return;
	  }
	  game.removeUser(username, type);
	  // check if game is now empty, if so remove it
	  if (game.empty()) removeGame(game.gamename);
	} else {  // game of user unknown, find game and remove user
	  for (Enumeration enum = games.elements(); enum.hasMoreElements();) {
		game = (Game) enum.nextElement();
		if (game.userExists(username)) game.removeUser(username, type);
		// check if game is now empty, if so remove it
		if (game.empty()) removeGame(game.gamename);
	  }
	}
  }
  
  /**
   * Returns whether a game exists with the given name.
   *
   * @param    gamename    The name of the game to look for.
   * @return   TRUE:  if the game exists
   *           FALSE: if not
   */
  public boolean gameExists(String gamename) {
	return (getGame(gamename) != null);
  }

  /**
   * Returns the game with the given name.
   *
   * @param    gamename    The name of the game
   * @return   The game
   */
  public Game getGame(String gamename) {
	return (Game) games.get(gamename);
  }
}

