package quest.server.login;

import quest.server.net.*;
import quest.global.chat.*;
import quest.global.net.*;
import quest.global.login.LoginMessage;
import quest.server.util.QuestDatabase;

/**
 * Server for handling the login requests and authentication of users.
 * if a user is authenticated he is then permitted to use the chat protocol.
 *
 * @see quest.global.net.MessageServerListener
 *
 * @author    Merijn Evertse
 */
public class LoginController implements MessageServerListener {
	
	private QuestDatabase qdb = QuestDatabase.getInstance();
	private ServerDaemon sd = ServerDaemon.getInstance();
	
  /**
   * Initializes the login server, and registers itself at the server daemon so it can
   * handle the login messages.
   */									 
  public LoginController() {
	sd.registerProces(LoginMessage.class, this, ServerDaemon.GLOBAL_BASED);
  }
  
  /**
   * Handles the incoming login messages. Checks the messages, processes them and
   * constructs the reply messages. These and then given to the server daemon
   * who sends them to the clients.
   */
  public void incomingMessage(final String username, final Message msg) {		
	final Runnable ra = new Runnable() {
	  public void run() {
		LoginMessage lm = (LoginMessage) msg;
		
		if (lm instanceof  LoginMessage.C2S_Login) {
		  login(username, lm);
		} else if (lm instanceof LoginMessage.C2S_NewUser) {
		  newuser(username, lm);
		} else if (lm instanceof LoginMessage.C2S_Logout) {
		  logout(username, lm);
		} else {
		  System.out.println("Received unsupported message: " + lm);
		}
	  }
	};
	(new Thread(ra)).start();
  }
  
  private synchronized void login(String username, LoginMessage lm) {
	LoginMessage.C2S_Login log = (LoginMessage.C2S_Login) lm;
	if (qdb.authenticateUser(log.username, log.password)) {
	  // passed authentication
	  LoginMessage.S2C_AllowLogin allow = new LoginMessage.S2C_AllowLogin("Welcome and stuff");
	  sd.sendMessageToUser(username, allow);
	  sd.updateUser(username, log.username); 
	  sd.subscribeUser(log.username, ChatMessage.class);
	  sd.subscribeUser(log.username, ChannelMessage.class);
	  sd.subscribeUser(log.username, RankingMessage.class);
	} else {
	  // failed authentication
	  LoginMessage.S2C_DenyLogin deny = new LoginMessage.S2C_DenyLogin("No soup for you");
	  
	  sd.sendMessageToUser(username, deny);
	  sd.unregisterUser(username);
	}
  }

  private synchronized void newuser(String username, LoginMessage lm) {
	LoginMessage.C2S_NewUser nu = (LoginMessage.C2S_NewUser) lm;
	if (qdb.insertUser(nu.username, nu.password)) {
	  // passed authentication
	  LoginMessage.S2C_AllowLogin allow = new LoginMessage.S2C_AllowLogin("Welcome to a new and allpowerfull quest exploration and stuff");
	  sd.sendMessageToUser(username, allow);
	  sd.updateUser(username, nu.username); 
	  sd.subscribeUser(nu.username, ChatMessage.class);
	  sd.subscribeUser(nu.username, ChannelMessage.class);
	  sd.subscribeUser(nu.username, RankingMessage.class);
	} else {
	  // failed authentication
	  LoginMessage.S2C_DenyLogin deny = new LoginMessage.S2C_DenyLogin("User " + nu.username + " already exists. try again.");
	  sd.sendMessageToUser(username, deny);
	  sd.unregisterUser(username);
	}					
  }

  private synchronized void logout(String username, LoginMessage lm) {
	LoginMessage.S2C_Exit exit = new LoginMessage.S2C_Exit("See you soon, you've been playing for <x> minutes");
	sd.sendMessageToUser(username, exit);
	sd.unregisterUser(username);
  }
}
