package quest.server.net;

import java.io.*;
import java.net.*;
import java.util.*;
import java.lang.reflect.*;
import quest.server.log.*;
import quest.global.net.*;

/**
 * Sever side message receiving thread for handling incoming messages and transporting it
 * to the right registered proces.
 * For a proces to be able to use the message stream, it has to register.
 *
 * @author    Merijn Evertse
 */
public class MessageServerThread extends Thread {

	private Vector processes;
	private ServerDaemon serverDaemon;
    private LogManager log;
	private ServerConfiguration configuration;
	
	/**
	 * Initializes the thread by making a proces and user database for registration.
	 */
	public MessageServerThread() {
		processes = new Vector();
		log = LogManager.getInstance();
		serverDaemon = ServerDaemon.getInstance();
		configuration = ServerConfiguration.getInstance();
	}
	
	/**
	 * Registers a proces consisting of a message type and a method for handling
	 * the message. After the registration of the proces all received messages
	 * intended for the registerd proces are supplied to the registerd message
	 * handler.
	 *
	 * @param    proces        The proces to register
	 */
	public void registerProces(Proces proces) {
		processes.addElement(proces);
		
		log.writeEntry("MessageServerThread: Registerred proces with " + 
					   proces.messageType.getName() +
					   ", " + proces.messageListener.getClass());
	}

	/**
	 * Unregisters a proces. The proces will now no longer receive incoming messages.
	 * NOTE: If incoming messages are of a type only this proces could handle, they will
	 * from now on be discarded.
	 *
	 * @param    proces    The proces to unregister
	 */
	public void unregisterProces(Proces proces) {
		processes.removeElement(proces);
	}

	private void handleMessage(Message message, NetUser user) {
		boolean messageHandled = false;
		
		for (int i = 0; i < processes.size(); i++) {
		  Proces proces = (Proces) processes.elementAt(i);
		  if (proces.messageType.isInstance(message)) {
			 switch (proces.registrationType) {
			 case ServerDaemon.USER_BASED:
			   // Only handle if protocol is allowed for this user
			   if (!(user.subscribedTo(message.getClass()))) {
				 log.writeEntry("MessageServerThread: Protocol " + 
									message.getClass() + 
									" not allowed for user " + user.username);
				 messageHandled = true;
				  break;
			   }
			   // fall through
			 case ServerDaemon.GLOBAL_BASED:
			   // Always handle message
				MessageServerListener messageListener = (MessageServerListener) proces.messageListener;
				messageListener.incomingMessage(user.username, message);
				messageHandled = true;			   
			   break;
			 default:
			   System.out.println("Uhm... something went wrond in handling a message");
			 }
		  }
		}
		if (!messageHandled) {
			log.writeEntry("Couldn't find an appropriate proces to handle the message " + message.getClass());
			// message is discarded
		}
	}	
	
	/**
	 * Runs this thread.
	 */
	public void run() {
	  setPriority(Thread.MIN_PRIORITY);
		while (true) {
			Enumeration enum = serverDaemon.getUsers();
			for (; enum.hasMoreElements();) {
				NetUser user = (NetUser) enum.nextElement();
				if (user.messageStream.messageAvailable()) {
				  // data is available so receive and handle it
				  Message message = user.messageStream.receiveMessage();
				  if (message != null) handleMessage(message, user);
				}
			}

			// wait some time before trying again
			try {
				Thread.currentThread().sleep(configuration.getSleeptime());
			} catch (InterruptedException e) {
				System.out.println(e.getMessage());
			}   
		}
	}
}
