package quest.server.net;

import java.io.*;
import java.net.*;
import java.util.*;
import java.lang.reflect.*;
import quest.global.net.*;
import quest.server.log.*;
import quest.server.chat.*;

/**
 * The server side daemon responsible for coordinating the message 
 * communication.
 * This means that it listens to incoming messages and handles them. Server processes will
 * need to register them at the server daemon to be able to receive messages for handling.
 *
 * @author    Merijn Evertse
 */
public class ServerDaemon {  
  
  	/**
  	 * The registering proces works on a user based accounting basis. This means that for
  	 * every user who wants to be able to communicate with the registering proces there
  	 * should be an entry in the server grant tables.
  	 */
    public static final int USER_BASED = 1;
	
	/**
	 * The registering proces works on a global bases accounting basis. This means that once
	 * registerred every client can communicate with this proces and request information.
	 */
	public static final int GLOBAL_BASED = 2;

    private ListenDaemon listenDaemon;
	private MessageServerThread receivingAgent;
	private NetUsers users;
	private static ServerDaemon uniqueInstance;
    private LogManager log;
	
	/**
	 * Creates a server side for the communication facilities.
	 */
	protected ServerDaemon() {
		if (uniqueInstance != null) {
			throw new RuntimeException("singleton constraint violated");
		} 
		
		uniqueInstance = this;
		log = LogManager.getInstance();
		users = NetUsers.getInstance();
	}
	
	/**
	 * Registers a proces consisiting of a message type and a method for handling
	 * the message. After the registration of the proces all received messages	
	 * intended for the registerd proces are supplied to the registerd message
	 * handler.
	 *
	 * @param    messageType        The type of the message
	 * @param    messageHandler     The handler method for the message
	 * @param    registrationType   The type of registration          
	 * 
	 * Registrationtype:
	 * quest.server.net.ServerDaemon.USER_BASED     For packet handling per user
	 * quest.server.net.ServerDaemon.GLOBAL_BASED   For packet handling for all messages
	 */
	public void registerProces(Class messageType, MessageServerListener messageListener, int registrationType) {
		receivingAgent.registerProces(new Proces(messageType, messageListener, registrationType));
	}
	
	/**
	 * Unregisters a proces, so no messages will be passes to the proces
	 * anymore.
	 *
	 * @param    MessageType    The type of the message to unregister
	 */
	public void unregisterProces(Class messageType) {
		receivingAgent.unregisterProces(new Proces(messageType,null,0));
	}

	/**
	 * Sends the message over the communication channel to the receiver.
	 *
	 * @param    usernames  The usernames
	 * @param    message    The message to send
	 * @return   A list of all users for which sending failed
	 */
	public Vector sendMessage(Vector usernames, Message message) {
		Vector streams = users.getStreams(usernames);
		Vector fails   = new Vector();
		for (int i = 0; i < streams.size(); i++) {
		  try {
			((MessageStream) streams.elementAt(i)).sendMessage(message);
		  } catch (IOException e) {
			// problems on connection with user
			String username = (String) usernames.elementAt(i);

			System.out.println("WARNING Connection problems with user " + username);
			unregisterUser(username);
			fails.addElement(username);
		  }
		}
		return fails;
	}	

  /**
   * Sends a message to a single user.
   *
   * @param    username     The name of the user
   * @param    message      The message to send
   */
  public void sendMessageToUser(String username, Message message) {
	Vector user = new Vector();
	user.addElement(username);
	sendMessage(user, message);
  }

	/**
	 * Returns an enumeration of all users currently online.
	 *
	 * @return    The enumeration of all online users
	 */
	public Enumeration getUsers() {
		return users.getUsers();
	}

	/**
	 * Registers a new user, who just came online, at the proper authorities.
	 *
	 * @param    username         The new user to register
	 * @param    messageStream    The connection stream with the user
	 */
	public void registerUser(String username, MessageStream messageStream) {
	  log.writeEntry("ServerDaemon: User " + username + " logged in");
	  users.addUser(new NetUser(username, messageStream));
	}

	/**
	 * Unregisters a user, because he goes offline.
	 *
	 * @param    username          The user to unregister
	 */
	public void unregisterUser(String username) {
	  log.writeEntry("ServerDaemon: User " + username + " logged out");
	  users.removeUser(new NetUser(username, null));
	}

	/**
	 * A new user is at the beginning registerred under a unique communication number.
	 * Only after the login procedure was succesfull, will the username of the client
	 * be registered at the server daemon.
	 * So it updates the clients information.
	 *
	 * @param    username    The current username (number) of the client
	 * @param    newUsername The new username (unique login- and nickname) of the client
	 */
  	public void updateUser(String username, String newUsername) {
		NetUser user = users.getUser(username);
		users.removeUser(user);
		user.username = newUsername;
		users.addUser(user);
  	}

  /**
   * Subscribes a user to a message protocol.
   *
   * @param    username       The name of the user
   * @param    messageType    The message protocol
   */
  public void subscribeUser(String username, Class messageType) {
	log.writeEntry("ServerDaemon: User " + username + " subscribed for " + messageType);
	users.getUser(username).subscribe(messageType);
  }

  /**
   * Unsubscribes a user to a message protocol.
   *
   * @param    username       The name of the user
   * @param    messageType    The message protocol
   */
  public void unsubscribeUser(String username, Class messageType) {
	log.writeEntry("ServerDaemon: User " + username + " unsubscribed for " + messageType);
	NetUser user = users.getUser(username);
	if (user != null) user.unsubscribe(messageType);
  }

	/**
	 * Runs the server by starting the neccesary daemons
	 */	
	protected void run() {
		// start receiving agent
		receivingAgent = new MessageServerThread();
		receivingAgent.start();
		// start listening daemon for receiving connections
		listenDaemon = new ListenDaemon();
		listenDaemon.start();
		
		log.writeEntry("ServerDaemon: Running");
	}
	
	/**
	 * Returns an instance of this class, making sure only one object of it exists.
	 *
	 * @return    A reference to the instance of this class
	 */
	public static ServerDaemon getInstance() {
		if (uniqueInstance == null) {
			uniqueInstance = new ServerDaemon();
			uniqueInstance.run();
		}
		return uniqueInstance;
	}
}

