package quest.server.util;

import java.net.*;
import java.sql.*;
import quest.server.log.*;


/**
 * Offers a simple API to communicate with the MYSQL server, still in SQL though.
 * Needs the org.gjt.mm.mysql.* package in order to work, else it throws a ClassNotFoundException and crashes.
 * <p>
 * It is an example of the <a href="../../../overview-summary.html#patternFacade">Facade design pattern</a> to offer a simplified higher-level interface to QuestDatabase.<br>
 * This class is also an example of a <a href="../../../overview-summary.html#patternSingleton">Singleton design pattern</a>.</p>
 *
 * @author Sander Brandenburg
 */
public class MYSQLCommunicator {
	
    private LogManager log;
    private static MYSQLCommunicator uniqueInstance = null;
    private Connection MYSQLConn;
	
    /**
     * Sole constructor. 
     * Throws a ClassNotFoundException when the org.gjt.mm.mysql.Driver class was not found. Crashes to the prompt afterwards.
     * Throws an SQLException when logging into the mysql server failed. Crashes directly afterwards.
     * @param hostname The ip address of the mysql server.
     * @param databse The name of the database to connect to.
     * @param username The username to use to log in the mysql server.
     * @param password The password to use to log in the mysql server.
     * @exception RuntimeException When initializing the class more than once in a context.
     */
    public MYSQLCommunicator(String hostname, String database, String username, String password) {
	if (uniqueInstance != null)
	    throw new RuntimeException("this is supposed to be an unique instance");
		
	uniqueInstance = this;
	log = LogManager.getInstance();
	log.writeEntry("MySQLCommunicator: Starting");
	try {
	    Class.forName("org.gjt.mm.mysql.Driver");
	} catch (ClassNotFoundException cnfe) {
	    cnfe.printStackTrace();
	    System.exit(-1);
	}

	final String url = "jdbc:mysql://" + hostname + "/" + database + "?user=" + username + "&password=" + password + "&autoReconnect=true";
		
	try {
	    MYSQLConn = DriverManager.getConnection(url);
	    SQLWarning warn = MYSQLConn.getWarnings();
	    if (warn != null) {
		log.writeEntry("*** MYSQL Server Connection Warnings ***");
		do {
		    log.writeEntry("SQLState: " + warn.getSQLState());
		    log.writeEntry("Message:  " + warn.getMessage());
		    log.writeEntry("Vendor:   " + warn.getErrorCode());
		    warn = warn.getNextWarning ();
		} while (warn != null);
		log.writeEntry("****************************************");
	    }
	} catch (SQLException se) {
	    System.out.println("SocketException in MYSQLCommunicator " + se.getMessage() + "\nQuiting...");
	    System.exit(-1);
	}
	log.writeEntry("MySQLCommunicator: Initialized");

    }

    /**
     * Closes the Mysql connection.
     */
    public void finalize() {
	try {
	    MYSQLConn.close();
	} catch (SQLException se) {
	}
    }
	
    /**
     * Returns a MYSQLCommunicator instance.
     * @return The already initialized MYSQLCommunicator instance.
     * @exception RuntimeException When the MYSQLCommunicator hasn't been initialized beforehand.
     */
    public static MYSQLCommunicator getInstance() {
	if (uniqueInstance == null) {
	    throw new RuntimeException("MYSQLCommunicator not initialized on getInstance()");
	}

	return uniqueInstance;
    }
	
    /**
     * Executes the specified query on the mysql server.
     * @param query The query in SQL syntax.
     * @return The ResultSet as result of the query.
     * @exception SQLException When the processing of the query resulted in a SQLException.
     */
    public ResultSet executeQuery(String query) throws SQLException {
	Statement stmt;
	ResultSet RS;
	synchronized (this) {
	    stmt = MYSQLConn.createStatement();
	    RS = stmt.executeQuery(query);
	    stmt.close();
	}
	return RS;
    }
    /**
     * Executes the specified update query on the mysql server.
     * @param query The update query in SQL syntax.
     * @return The ResultSet as result of the update query.
     * @exception SQLException When the processing of the query resulted in a SQLException.
     */
    public int executeUpdate(String update) throws SQLException {
	int changed;
	Statement stmt;
	synchronized (this) {
	    stmt = MYSQLConn.createStatement();
	    changed = stmt.executeUpdate(update);
	    stmt.close();
	}
	return changed;
    }
	
    /**
     * Executes the specified general query on the mysql server. Generally used when it's only interesting whether the query really got executed or not (for example: INSERT, DELETE).
     * @param query The query in SQL syntax.
     * @return Whether the execution has failed or not.
     * @exception SQLException When the processing of the query resulted in a SQLException.
     */
    public boolean execute(String exec) throws SQLException {
	Statement stmt;
	boolean changed;
	synchronized (this) {
	    stmt = MYSQLConn.createStatement();
	    changed = stmt.execute(exec);
	    stmt.close();
	}
	return changed;
    }
}
