package quest.server.util;

import java.sql.*;
import java.util.Vector;
import java.util.List;
import quest.global.chat.UserRanking;
import quest.global.chat.RankingMessage;
import quest.global.chat.SearchMessage;
import quest.global.chat.ChatMessage;
import quest.server.chat.ChatServer;

/**
 * Represents the database that maintains usernames, passwords and all kinds of score info.
 * Uses the MYSQLCommunicator as communication layer.
 * <p>
 * Because we cannot use the feature of a temporary table (the production machine runs an
 * ancient mysql version) this class does NOT support multiple instances which 
 * operate on the same mysql database.</p>
 * 
 * This class is an example of a <a href="../../../overview-summary.html#patternSingleton">Singleton design pattern</a>.
 * @author Sander Brandenburg
 * @see quest.server.util.MYSQLCommunicator
 */
public class QuestDatabase {
    private static QuestDatabase uniqueInstance = null;
    private MYSQLCommunicator mc = MYSQLCommunicator.getInstance();
	
    private int userCount;
    private boolean invalidate; // invalidate the result in the temp table
	
    /**
     * Sole constructor.
     * @exception RuntimeException When initializing this class more than once in the same context.
     */
    public QuestDatabase() {
	if (uniqueInstance != null)
	    throw new RuntimeException("QuestDatabase is supposed to be unique");
		
	invalidate = true;
	initTempTable();
	updateTempTable();

	uniqueInstance = this;
    }

    /**
     * Returns the QuestDatabase instance that was previously instantiated.
     * @return QuestDatabase instance that was previously instantiated.
     * @exception RuntimeException When QuestDatabase hasn't been initialized before.
     */
    public static QuestDatabase getInstance() {
	if (uniqueInstance == null)
	    throw new RuntimeException("QuestDatabase should've been initialized first");
		
	return uniqueInstance;
    }
	
    /**
     * Authenticates user according to the usertable in the database.
     * @return Whether user authentication has passed.
     */
    public boolean authenticateUser(String username, String password) {
	ResultSet RS;
	try {
	    RS = mc.executeQuery("SELECT count(*) FROM users WHERE username='" + escapeSQL(username) + "' and password=password('" + escapeSQL(password) + "')");
	    if (!RS.next()) {
		System.out.println("resultset empty in authenticateuser");
		return false;
	    }
	    int i = RS.getInt(1);
	    return i >= 1;
	} catch (SQLException se) {
	    se.printStackTrace();
	    return false;
	}
    }
	
    public boolean insertUser(String username, String password) {
	ResultSet RS;
	try {
	    int i = mc.executeUpdate("INSERT INTO users (username, password) values ('" + escapeSQL(username) + "', password('" + escapeSQL(password) + "'))");
	    if (i >= 1) {
				// succeeded
		invalidate = true;
		return true;
	    }
	} catch (SQLException se) {
	    // oppress for duplicate keys etc
	    return false;
	}
	return false;
    }
	
    /**
     * Searches for a certain partial username in the usernames table of the database.
     * @param username The (partial) username to be searched for.
     * @returns A SearchMessage.S2C_Result.SearchResult[] as result (empty if username was not found) or null when query failed.
     * @see quest.global.chat.SearchMessage.S2C_Result.SearchResult
     */
    public SearchMessage.S2C_Result.SearchResult[] searchUsername(String username) {
	updateTempTable();
	ResultSet RS;
	Vector result = new Vector(userCount);
	final String query = "select username, ranking, gamesplayed from ranking_temp where username like '%" + escapeSQL(username) + "%' order by username";
	try {
	    RS = mc.executeQuery(query);
	    while (RS.next()) {
		  ChatServer cs = ChatServer.getInstance();
		  String uname = RS.getString(1);
		  if (cs.userOnline(uname)) {
			// only when the user is online the info should be used
			String channel = cs.getChannelname(uname);
			SearchMessage.S2C_Result.SearchResult sr = 
			  new SearchMessage.S2C_Result.SearchResult(uname, 
						cs.getUsermode(uname, channel), RS.getInt(2), 
						RS.getInt(3), channel, cs.getChannelmode(channel));
			result.add(sr);
		  }
	    }
	    return (SearchMessage.S2C_Result.SearchResult[]) result.toArray(new SearchMessage.S2C_Result.SearchResult[result.size()]);
	} catch (SQLException se) {
	    se.printStackTrace();
	    return null;
	}
    }
   
    private UserRanking[] getUserRankings(int offset, int limit) {
	updateTempTable();
		
	ResultSet RS;
	Vector result = new Vector(userCount);
	final String query = "select pos, username, ranking, gamesplayed, wins, lost from ranking_temp limit " + offset + ", " + limit;
		
	try {
	    RS = mc.executeQuery(query);
	    while (RS.next()) {
		UserRanking ur = new UserRanking(RS.getInt(1), RS.getString(2), RS.getInt(3), RS.getInt(4), RS.getInt(5), RS.getInt(6));
		result.add(ur);
	    }
	    return (UserRanking[]) result.toArray(new UserRanking[result.size()]);
	} catch (SQLException se) {
	    se.printStackTrace();
	    return null;
	}
    }
    
    /**
     * Returns the rankings of all the players in the database.
     * @return Rankings of all players in the database, array empty if result is empty or null when internal error occurred.
     * @see quest.global.chat.UserRanking
     */
    public UserRanking[] getAllUserRankings() {
	return getUserRankings(0, -1);
    }
    /**
     * Returns the rankings of the top 50 players in the database.
     * @return Rankings of the top 50 players in the database, array empty if result is empty or null when internal error occurred.
     * @see quest.global.chat.UserRanking
     */
    public UserRanking[] getTopUserRankings() {
	final int topnr = 50;
		
	return getUserRankings(0, topnr);
    }
    /**
     * Returns  the rankings of the surrounding 10 neighbours of the specified username.
     * @param username The username of the player to get the UserRanking for.
     * @return Rankings of the surrounding 10 neighbours of the specified player according to the positions in the database. Returns null if playername wasn't found or an internal error occurred.
     * @see quest.global.chat.UserRanking
     */

    public UserRanking[] getNeighbours(String username) {
	// neighbours above and under seperate
	final int nrNeighbours = 10;
		
	int pos = (getUserPosition(username))[0];
	if (pos == -1)
	    return null;
		
	pos = Math.max(0, pos - nrNeighbours);
	return getUserRankings(pos, nrNeighbours*2);
    }

    /**
     * Updates
     */
    public void updateUser(String username, boolean won, int changeRanking) {
	String str = "UPDATE users set gamesplayed=gamesplayed+1, " + (won ? "wins=wins+1" : "lost=lost+1") + ", ranking = ranking+" + changeRanking + " where username='" + escapeSQL(username) + "'";
	try {
	    int res = mc.executeUpdate(str);
	    if (res <= 0) {
		System.out.println("warning: query failed: " + str);
		return;
	    }
	    invalidate = true;
	} catch (SQLException se) {
	    se.printStackTrace();
	}
    }

    public int getMaxPosition() {
	return userCount;
    }
	
    // ownpos, maxpos, ranking
    public int[] getUserPosition(String username) {
	try {
	    ResultSet RS;
	    RS = mc.executeQuery("select pos, ranking from ranking_temp where username='" + escapeSQL(username) + "'");
	    if (!RS.next()) {
		System.out.println("couldn't retrieve user position");
		return null;
	    }
			
	    return new int[] {RS.getInt(1), getMaxPosition(), RS.getInt(2)};
			
	} catch (SQLException se) {
	    se.printStackTrace();
	    return null;
	}
    }
    protected final String escapeSQL(String str) {
	int i;
	String dst = "";
	i = 0;
	for (i = 0; i < str.length(); i++) {			
	    if (str.charAt(i) == '\'') {
		dst = dst + "\\'";
	    } else if (str.charAt(i) == '\\') {
		dst = dst + "\\\\";
	    } else {
		dst = dst + str.charAt(i);
	    }
	}
		
	return dst;
    }

    /*
      unfortunately we cannot use temporary stuff because our production machine runs an ancient mysql server
    */

    protected final String queryCreateTableString = 
	"create table ranking_temp (" +
	"pos int(4) not null auto_increment primary key," +
	"username char(32) not null," +
	"ranking int(4) not null," +
	"gamesplayed int(4) not null," +
	"wins int(4) not null," +
	"lost int(4) not null" +
	")";
	
    protected final String queryDeleteTableString =
	"drop table ranking_temp";
	
    protected final String queryUpdateTable = "insert into ranking_temp (pos, username, ranking, gamesplayed, wins, lost) select NULL, username, ranking, gamesplayed, wins, lost from users order by ranking desc";
    protected final String queryDeleteTableContents = "delete from ranking_temp";

    protected void initTempTable() {
	try {
	    mc.execute(queryDeleteTableString);
	} catch(SQLException e) {
	    // ignore it, it'll generate an exception when table doesn't exist
	}

	try {
	    mc.execute(queryCreateTableString);
	} catch (SQLException e) {
	    e.printStackTrace();
	}
    }
	
    protected void updateTempTable() {
	if (!invalidate)
	    return;

	try {
	    mc.execute(queryDeleteTableContents);
	    mc.executeUpdate(queryUpdateTable);
	    ResultSet RS = mc.executeQuery("select count(*) from ranking_temp");
	    if (!RS.next())
		return;
	    userCount = RS.getInt(1);
	    invalidate = false;
	} catch (SQLException e) {
	    e.printStackTrace();
	}
    }
}
